#ifndef _RXVECDSPLINE_H_WRAP_
#define _RXVECDSPLINE_H_WRAP_

#include <RxVecdN.h>
#include <RxDlist.h>

class RxVecdNSknot : public RxDlistNode
{
	friend class RxVecdNSpline;

	RxVecdN *v;
	double u;
	double w;
	int m;
	int idx;
	RX_DLIST_NODE (RxVecdNSknot);
};

RX_DLIST_CLASS (RxVecdNSknotList, RxVecdNSknot);

class RxVecdNSpline
{

    private:
	int degree;
	int rational;

	RxVecdNSknotList knots;
	int vlen;
	int headIdx;

	void deBoor (RxVecdN *v, RxVecdN *dv, RxVecdN *ddv, 
		     RxVecdNSknot *k0, RxVecdNSknot *k1, double u, int k);
	RxVecdNSknot *findLeftKnot (double u, RxVecdNSknot *start=NULL);
	RxVecdNSknot *findRightKnot (double u, RxVecdNSknot *start=NULL);
	RxVecdNSknot *lastLeftKnot;
	RxVecdNSknot *curPnt;

    public:
	RxVecdNSpline (int len, int deg, int rat);

	void reset ();
	void reset (int deg, int rat);

	int addPnt (RxVecdN *v, double u);
	int addPnt (RxVecdN *v, double w, double u);
	int removePnts (double u);
#define RX_SPLINE_CALC_ERROR	-1
#define RX_SPLINE_CALC_OK	0x1
#define RX_SPLINE_CALC_RIGHT	0x2
#define RX_SPLINE_CALC_LEFT	0x4
	int calcPnt (RxVecdN *v, double u);
	int calcPnt (RxVecdN *v, RxVecdN *dv, 
		     RxVecdN *ddv, double u, int twoSidedDerivs=0);

	int getDegree()
	 { return (degree);
	 }
	int getVlength()
	 { return (vlen);
	 }
	int isRational()
	 { return (rational ? 1 : 0);
	 }
	int numKnots();
	RxVecdN *firstPnt (double *w=NULL, double *u=NULL, int *mp=NULL);
	RxVecdN *nextPnt (double *w=NULL, double *u=NULL, int *mp=NULL);
};

#endif
