#ifndef _RXCLIST_H_WRAP_
#define _RXCLIST_H_WRAP_

#include <RxFastAlloc.h>

#ifndef NULL
#define NULL 0
#endif

// Usage:
// 
// class XXXNode : public RxCnodeBase
// {
//  	... constructors, destructors, whatever ...
//
//	RX_NAMED_CNODE_DECLS (XXXNode, name);	// name is a private char *
//						// that names the node.
//	 *OR*
//
//	RX_CNODE_DECLS (XXXNode);
//
//	int compare (const void *ptr) const 	// optional ...
//	 { ... see if ptr references this node ...
//	 };
// }
//
// RxClist<NodeType>
// 

class RxCnodeBase : public RxFastAlloc
{
	friend class RxClistBase;

    protected:
	RxCnodeBase *next_;
	RxCnodeBase *prev_;
    public:
	RxCnodeBase()
	 { next_ = NULL;
	   prev_ = NULL;
	 };
#if 0
	RxCnodeBase *getNext()
	 { return next_;
	 };
	const RxCnodeBase *getNext() const
	 { return next_;
	 };
#endif
	RxCnodeBase *getNext(int n);
	const RxCnodeBase *getNext(int n) const
	 { return getNext(n);
	 }
#if 0
	RxCnodeBase *getPrev()
	 { return prev_;
	 };
	const RxCnodeBase *getPrev() const
	 { return prev_;
	 };
#endif
	RxCnodeBase *getPrev(int n);
	const RxCnodeBase *getPrev(int n) const
	 { return (getPrev(n));
	 }
	virtual int compare (const void *ptr) const
	 { return ((void*)this == ptr);
	 };
	virtual const char *getName () const
	 { return (NULL);
	 };
};

class RxClistBase
{
    protected:
	RxCnodeBase *head_;
	RxCnodeBase *tail_;
	RxCnodeBase *findNodeByName (const char *) const;
	RxCnodeBase *findNodeByPtr (const void *ptr) const;
	int autoDelete;
	virtual void freeNode (void *)
	 {
	 };
    public:
	RxClistBase()
	 { head_ = NULL; 
	   tail_ = NULL;
	   autoDelete = 1;
	 };
//	~RxClistBase();
	void init()
	 { head_ = NULL; 
	   tail_ = NULL;
	 };
	RxCnodeBase *addHead(RxCnodeBase* ptr);
	RxCnodeBase *addTail(RxCnodeBase* ptr);
	RxCnodeBase *addAfter(RxCnodeBase* ref, RxCnodeBase* ptr);
	RxCnodeBase *addBefore(RxCnodeBase* ref, RxCnodeBase* ptr);
	RxCnodeBase *remove(RxCnodeBase* ptr);
	int remove(RxCnodeBase* ptr1, RxCnodeBase* ptr2);
	void removeAndDelete(RxCnodeBase* ptr)
	 { remove (ptr);
	   freeNode (ptr);
	 }
	int removeAndDelete(RxCnodeBase* ptr1, RxCnodeBase* ptr2);
	int removeAll();
	int removeAndDeleteAll();

	void shift (int n);

	void setAutoDelete(int val)
	 { autoDelete = val;
	 }
	int getAutoDelete() const
	 { return autoDelete;
	 }

	int length() const;
	int empty() const
	 { return (head_==NULL);
	 }
};

#define RX_NAMED_CNODE_DECLS(N,name) \
    private: \
	char *name; \
    public: \
	char const *getName() const \
	 { return (name); \
	 } \
	void setName(const char *str) \
	 { free ((void*)name); \
	   name = strdup(str); \
	 } \
	N *next() \
	 { return ((N*)next_); \
	 } \
	N *next(int n) \
	 { return ((N*)getNext(n)); \
	 } \
	N *prev() \
	 { return ((N*)prev_); \
	 } \
	N *prev(int n) \
	 { return ((N*)getPrev(n)); \
	 } \
	const N *next() const \
	 { return ((const N*)next_); \
	 } \
	const N *next(int n) const \
	 { return ((const N*)getNext(n)); \
	 } \
	const N *prev()	const \
	 { return ((const N*)prev_); \
	 } \
	const N *prev(int n) const \
	 { return ((const N*)getPrev(n)); \
	 }
	
#define RX_CNODE_DECLS(N) \
	N *next() \
	 { return ((N*)next_); \
	 } \
	N *next(int n) \
	 { return ((N*)getNext(n)); \
	 } \
	N *prev() \
	 { return ((N*)prev_); \
	 } \
	N *prev(int n) \
	 { return ((N*)getPrev(n)); \
	 } \
	const N *next() const \
	 { return ((const N*)next_); \
	 } \
	const N *next(int n) const \
	 { return ((const N*)getNext(n)); \
	 } \
	const N *prev()	const \
	 { return ((const N*)prev_); \
	 } \
	const N *prev(int n) const \
	 { return ((const N*)getPrev(n)); \
	 }

template<class T>
class RxCnode : public RxCnodeBase
{
public:
	T val;
	RX_CNODE_DECLS(RxCnode);	
	RX_FAST_ALLOC_DECLS_L(RxCnode,next_);
};

template<class N>
class RxClist : public RxClistBase
{
	void freeNode(void *ptr)
	 { delete (N*)ptr;
	 };
	int testNode(const RxCnodeBase *node, const void *ptr)
	 { return (((N*)node)->compare(ptr));
	 };
    public:
	RxClist() : RxClistBase()
	 {
	 }
#if 1
	~RxClist()
	 { if (autoDelete) removeAndDeleteAll();
	 };
#endif
	N* head()
	 { return (N*)head_;
	 };
	N* tail()
	 { return (N*)tail_;
	 };
	N* findByPtr(const void *ptr)
	 { return (N*)findNodeByPtr(ptr);
	 };
	N* findByName(const char *ptr)
	 { return (N*)findNodeByName(ptr);
	 };

	const N* head() const
	 { return (const N*)head_;
	 };
	const N* tail() const
	 { return (const N*)tail_;
	 };
	const N* findByPtr(const void *ptr) const
	 { return (const N*)findNodeByPtr(ptr);
	 };
	const N* findByName(const char *ptr) const
	 { return (const N*)findNodeByName(ptr);
	 };
};
	
#endif
