#ifndef _RXDLIST_H_WRAP_
#define _RXDLIST_H_WRAP_

#include <RxFastAlloc.h>

#ifndef NULL
#define NULL 0
#endif

// Usage:
// 
// class XXXNode : public RxDnodeBase
// {
//  	... constructors, destructors, whatever ...
//
//	RX_NAMED_DNODE_DECLS (XXXNode, name);	// name is a private char *
//						// that names the node.
//	 *OR*
//
//	RX_DNODE_DECLS (XXXNode);
//
//	int compare (const void *ptr) const 	// optional ...
//	 { ... see if ptr references this node ...
//	 };
// }
//
// RxDlist<NodeType>
// 

class RxDnodeBase : public RxFastAlloc
{
	friend class RxDlistBase;

    protected:
	RxDnodeBase *next_;
	RxDnodeBase *prev_;
    public:
	RxDnodeBase()
	 { next_ = NULL;
	   prev_ = NULL;
	 };
#if 0
	RxDnodeBase *getNext() 
	 { return next_;
	 };
	const RxDnodeBase *getNext() const
	 { return next_;
	 };
#endif
	RxDnodeBase *getNext(int n) ;
	const RxDnodeBase *getNext(int n) const
	 { return getNext(n);
	 }
#if 0
	RxDnodeBase *getPrev() 
	 { return prev_;
	 };
	const RxDnodeBase *getPrev() const
	 { return prev_;
	 };
#endif
	RxDnodeBase *getPrev(int n) ;
	const RxDnodeBase *getPrev(int n) const
	 { return (getPrev(n));
	 }
	virtual int compare (const void *ptr) const
	 { return ((void*)this == ptr);
	 };
	virtual const char *getName () const
	 { return (NULL);
	 };
};

class RxDlistBase
{
    protected:
	RxDnodeBase *head_;
	RxDnodeBase *tail_;
	RxDnodeBase *findNodeByName (const char *) const;
	RxDnodeBase *findNodeByPtr (const void *ptr) const;
	int autoDelete;
	virtual void freeNode (void *)
	 {
	 };
    public:
	void init()
	 { head_ = NULL; 
	   tail_ = NULL;
	 };
	void set(RxDlistBase &ref)
	 { head_ = ref.head_;
	   tail_ = ref.tail_;
	 }
	RxDlistBase()
	 { head_ = NULL; 
	   tail_ = NULL;
	   autoDelete = 1;
	 }
//	~RxDlistBase();
	RxDnodeBase *addHead(RxDnodeBase* ptr);
	RxDnodeBase *addTail(RxDnodeBase* ptr);
	RxDnodeBase *addAfter(RxDnodeBase* ref, RxDnodeBase* ptr);
	RxDnodeBase *addBefore(RxDnodeBase* ref, RxDnodeBase* ptr);
	RxDnodeBase *remove(RxDnodeBase* ptr);
	int remove(RxDnodeBase* ptr1, RxDnodeBase* ptr2);
	void removeAndDelete(RxDnodeBase* ptr)
	 { remove (ptr);
	   freeNode (ptr);
	 }
	int removeAndDelete(RxDnodeBase* ptr1, RxDnodeBase* ptr2);
	int removeAll();
	int removeAndDeleteAll();

	void setAutoDelete(int val)
	 { autoDelete = val;
	 }
	int getAutoDelete() const
	 { return autoDelete;
	 }

	int length() const;
	int empty() const
	 { return (head_==NULL);
	 }
};

#define RX_NAMED_DNODE_DECLS(N,name) \
    private: \
	char *name; \
    public: \
	const char *getName() const \
	 { return (name); \
	 } \
	void setName(const char *str) \
	 { free ((void*)name); \
	   name = strdup(str); \
	 } \
	N *next() \
	 { return ((N*)next_); \
	 } \
	N *next(int n) \
	 { return ((N*)getNext(n)); \
	 } \
	N *prev() \
	 { return ((N*)prev_); \
	 } \
	N *prev(int n) \
	 { return ((N*)getPrev(n)); \
	 } \
	const N *next() const \
	 { return ((N*)next_); \
	 } \
	const N *next(int n) const \
	 { return ((N*)getNext(n)); \
	 } \
	const N *prev() const \
	 { return ((N*)prev_); \
	 } \
	const N *prev(int n) const \
	 { return ((N*)getPrev(n)); \
	 }
	

#define RX_DNODE_DECLS(N) \
	N *next() \
	 { return ((N*)next_); \
	 } \
	N *next(int n) \
	 { return ((N*)getNext(n)); \
	 } \
	N *prev() \
	 { return ((N*)prev_); \
	 } \
	N *prev(int n) \
	 { return ((N*)getPrev(n)); \
	 } \
	const N *next() const \
	 { return ((N*)next_); \
	 } \
	const N *next(int n) const \
	 { return ((N*)getNext(n)); \
	 } \
	const N *prev() const \
	 { return ((N*)prev_); \
	 } \
	const N *prev(int n) const \
	 { return ((N*)getPrev(n)); \
	 }

template<class T>
class RxDnode : public RxDnodeBase
{
public:
	T val;
	RX_DNODE_DECLS(RxDnode);	
	RX_FAST_ALLOC_DECLS_L(RxDnode,next_);
};

template<class N>
class RxDlist : public RxDlistBase
{
	void freeNode(void *ptr)
	 { delete (N*)ptr;
	 };
	int testNode(const RxDnodeBase *node, const void *ptr)
	 { return (((N*)node)->compare(ptr));
	 };
    public:
	RxDlist() : RxDlistBase()
	 {
	 }
#if 1
	~RxDlist()
	 { if (autoDelete) removeAndDeleteAll();
	 };
#endif
	N* head()
	 { return (N*)head_;
	 };
	N* tail()
	 { return (N*)tail_;
	 };
	N* findByPtr(const void *ptr)
	 { return (N*)findNodeByPtr(ptr);
	 };
	N* findByName(const char *ptr)
	 { return (N*)findNodeByName(ptr);
	 };
	const N* head() const
	 { return (N*)head_;
	 };
	const N* tail()	const
	 { return (N*)tail_;
	 };
	const N* findByPtr(const void *ptr) const
	 { return (N*)findNodeByPtr(ptr);
	 };
	const N* findByName(const char *ptr) const
	 { return (N*)findNodeByName(ptr);
	 };
};
	
#endif
