#ifndef _RXJOINT_H_WRAP_
#define _RXJOINT_H_WRAP_

#include <RxPrintable.h>
#include <RxVecd.h>
#include <RxVecd6.h>
#include <RxTrsf.h>
#include <RxMatd.h>
#include <RxTwist.h>
#include <RxWrench.h>
#include <RxSpatialAccel.h>

class RxJoint : public RxPrintable
{
private:
	
protected:
	int numdof_;
	int numc_;
	int type_;
	double *limbuf;	// buffer for limit information
	RxVecd jvels_;
	RxVecd maxJlims;
	RxVecd minJlims;
	RxVecd maxJvels;
	RxVecd maxJaccs;
	RxVecd maxJfors;
	RxVecd outConv;	// output conversion factor
	void setCopy (RxVecd &vr, const RxVecd &v1, int j) const
	 { int i;
	   for (i=0; i<numc_; i++)
	    { vr(i) = v1(i+j);
	    }
	 }
	void getCopy (RxVecd &vr, const RxVecd &v1, int j) const
	 { int i;
	   for (i=0; i<numc_; i++)
	    { vr(i+j) = v1(i);
	    }
	 }

	virtual int oscanField (class RxParseObj *obj);
	virtual int checkFields (class RxParseObj *obj);
	virtual int cprintFields (class RxCoutput *co,class RxPrintSpec *ps,
				  int init=1) const;
	void maybeOutputNewline (int &init, class RxCoutput *co,
				 int lspc) const;
	void maybePrintLimVec (RxCoutput *co, RxPrintSpec *ps,
		 	       const char *pfx, const RxVecd &vec,
			       double defaultValue, int &init) const;
public:
//	static const int RxJoint::ID;
	virtual const char *typeName() const
	 { return "Joint";
	 }

//	virtual RxJoint *clone() const = 0;
#if 0
	void copy (const RxJoint &link)
	 { numdof_ = link.dof_;
	   type_ = link.type_;
	   A_ = link.A_;
	 }
#endif
	RxJoint (int numc, int numdof);
	~RxJoint();
	
	virtual void setJnt (double val) = 0;
	virtual double getJnt () const = 0;
	void setJvel (double vel)
	 { jvels_(0) = vel;
	 }
	double getJvel() const
	 { return jvels_(0);
	 }

	virtual void setJnt (const RxVecd &vec, int j=0) = 0;
	virtual void getJnt (RxVecd &vec, int j=0) const = 0;
	void setJvel (const RxVecd &vec, int j=0)
	 { setCopy (jvels_, vec, j);
	 }
	void getJvel (RxVecd &vec, int j=0) const
	 { getCopy (vec, jvels_, j);
	 }
	
	virtual void getTrsf (RxTrsf &T) const = 0;
	virtual void postMulTrsf (RxTrsf &TR, const RxTrsf &T) const = 0;
	virtual void postMulTrsf (RxTrsf &T) const = 0;
	virtual void preMulTrsf (RxTrsf &TR, const RxTrsf &T) const = 0;
	virtual void preMulTrsf (RxTrsf &T) const = 0;
	
	virtual void addVel (RxTwist &vel) const = 0;
	virtual void addAcc (RxSpatialAccel &acc, const RxTwist &v0,
			     const RxVecd &jacc, int j=0) const = 0;

	virtual double getJfor (const RxWrench &wr) const = 0;
	virtual void getJfor (RxVecd &jfor, const RxWrench &wr, 
			      int j=0) const = 0;

	void getVel (RxTwist &vel) const
	 { vel.setZero();
	   addVel (vel);
	 }
	int getAcc (RxSpatialAccel &acc, const RxTwist &v0,
		     const RxVecd &jacc, int j=0) const
	 { acc.setZero();
	   addAcc (acc, v0, jacc, j);
	 }

	virtual void getJcol (RxMatd *cols, const RxTrsf &U, 
			      int j=0) const = 0;

	double getMaxJlim() const
	 { return maxJlims(0);
	 }
	double getMinJlim() const
	 { return minJlims(0);
	 }
	double getMaxJvel() const
	 { return maxJvels(0);
	 }
	double getMaxJacc() const
	 { return maxJaccs(0);
	 }
	double getMaxJfor() const
	 { return maxJfors(0);
	 }
	double getOutputConv() const
	 { return outConv(0);
	 }

	void setMaxJlim(double lim)
	 { maxJlims(0) = lim;
	 }
	void setMinJlim(double lim)
	 { minJlims(0) = lim;
	 }
	void setMaxJvel(double lim)
	 { maxJvels(0) = lim;
	 }
	void setMaxJacc(double lim)
	 { maxJaccs(0) = lim;
	 }
	void setMaxJfor(double lim)
	 { maxJfors(0) = lim;
	 }

	void getMaxJlim(RxVecd &vec, int j=0) const
	 { getCopy (vec, maxJlims, j);
	 }
	void getMinJlim(RxVecd &vec, int j=0) const
	 { getCopy (vec, minJlims, j);
	 }
	void getMaxJvel(RxVecd &vec, int j=0) const
	 { getCopy (vec, maxJvels, j);
	 }
	void getMaxJacc(RxVecd &vec, int j=0) const
	 { getCopy (vec, maxJaccs, j);
	 }
	void getMaxJfor(RxVecd &vec, int j=0) const
	 { getCopy (vec, maxJfors, j);
	 }
	void getOutputConv(RxVecd &vec, int j=0) const
	 { getCopy (vec, outConv, j);
	 }

	void setMaxJlim(RxVecd &vec, int j=0)
	 { setCopy (maxJlims, vec, j);
	 }
	void setMinJlim(RxVecd &vec, int j=0)
	 { setCopy (minJlims, vec, j);
	 }
	void setMaxJvel(RxVecd &vec, int j=0)
	 { setCopy (maxJvels, vec, j);
	 }
	void setMaxJacc(RxVecd &vec, int j=0)
	 { setCopy (maxJaccs, vec, j);
	 }
	void setMaxJfor(RxVecd &vec, int j=0)
	 { setCopy (maxJfors, vec, j);
	 }

	int type() const
	 { return type_;
	 }
	int dof() const
	 { return numdof_;
	 }
	int numc() const
	 { return numc_;
	 }

	virtual int oscan (class RxParseObj *obj);
	virtual int cprint(class RxCoutput *co,class RxPrintSpec *ps) const;
};
#endif
