#ifndef _RXMATDN_H_WRAP_
#define _RXMATDN_H_WRAP_

#include <RxMatdBase.h>
#include <RxVecd.h>
#include <stdarg.h>

class RxMatd : public RxMatdBase
{
	friend class RxMatdLu;
	friend class RxVecd;
	friend class RxMatdSvd;	
	friend class RxMatdChol;
	friend class RxMatd3x3;
	friend class RxSpatialMat;

	int nr;
	int nc;
	int width;
	double *store;
	int malloced;

	void doCopy (const RxMatd &m1);

    public:

	// constructors

	RxMatd (int nrow, int ncol) 
	 { nr = nrow;
	   nc = ncol;
	   width = nc;
	   store = new double [nr*nc];
	   malloced = 1;
	 }
	RxMatd (const RxMatd &m1);
	RxMatd (const RxMatdBase &m1);
	RxMatd (const double *dp, int nrow, int ncol);
	RxMatd (int nrow, int ncol, double *buf, int w)
	 { nr = nrow;
	   nc = ncol;
	   store = buf;
	   malloced = 0;
	   width = w;
	 }
	RxMatd (int nrow, int ncol, double *buf)
	 { nr = nrow;
	   nc = ncol;
	   store = buf;
	   malloced = 0;
	   width = ncol;
	 }
	RxMatd (const RxMatd &mb, int i, int j, int nrow, int ncol)
	 { nr = nrow;
	   nc = ncol;
	   store = &mb.store[i*mb.width+j];
	   malloced = 0;
	   width = mb.width;
	 }
	RxMatd (const RxVecd &v, int column = 1);
	~RxMatd ()
	 { if (malloced) delete[] store;
	 }

	// query and access

	int nrows() const
	 { return nr;
	 }
	int ncols() const
	 { return nc;
	 }
	double *buf() const
	 { return store;
	 }
	double *buf(int *wp) const
	 { *wp = width;
	   return store;
	 }
	double& operator()(int i, int j)
	 { return store[width*i+j];
	 };
	double operator()(int i, int j) const
	 { return store[width*i+j];
	 };
	RxMatd operator()(int i, int j, int m, int n)
	 { return (RxMatd (*this, i, j, m, n));
	 }
	
	// assignment

	RxMatd& set (const RxMatd& m2);
	RxMatd& set (const double *dp);
	RxMatd& set (const RxVecd &v1);
	RxMatd& set (double, ...);
	RxMatd& setDiag (const RxVecd &v1);
	RxMatd& setDiag (double *buf);

	RxMatd& operator=(const RxMatd& m2)
	 { return (set (m2));
	 }	   
	void get (double *dp);
	RxVecd &get (RxVecd *vr);

	// T <- T,T operations

	RxMatd& add (const RxMatd &m1, const RxMatd &m2);
	RxMatd& add (const RxMatd &m2);
	RxMatd& operator+= (const RxMatd &m2)
	 { return (add (m2));
	 }
	friend RxMatd operator+(const RxMatd &m1, const RxMatd &m2);

	RxMatd& sub (const RxMatd &m1, const RxMatd &m2);
	RxMatd& sub (const RxMatd &m2);
	RxMatd& operator-= (const RxMatd &m2)
	 { return (sub (m2));
	 }
	friend RxMatd operator-(const RxMatd &m1, const RxMatd &m2);

	RxMatd& mul (const RxMatd &m1, const RxMatd &m2);
	RxMatd& mul (const RxMatd &m2)
	 { return (mul (*this, m2));
	 }
	RxMatd& Tmul (const RxMatd &m1, const RxMatd &m2);
	RxMatd& Tmul (const RxMatd &m2)
	 { return (Tmul (*this, m2));
	 }
	RxMatd& mulT (const RxMatd &m1, const RxMatd &m2);
	RxMatd& mulT (const RxMatd &m2)
	 { return (mulT (*this, m2));
	 }
	RxMatd& operator*= (const RxMatd &m2)
	 { return (mul (*this, m2));
	 }
	friend RxMatd operator*(const RxMatd &m1, const RxMatd &m2);

	// T <- T,O operations

	RxMatd& scale (double a, const RxMatd &m1);
	RxMatd& scale (double a);
	RxMatd& operator *= (double a)
	 { return (scale (a));
	 }
	RxMatd& operator /= (double a)
	 { return (scale (1/a));
	 }
	friend RxMatd operator*(const RxMatd &m1, double a);	   
	friend RxMatd operator*(double a, const RxMatd &m1);
	friend RxMatd operator/(const RxMatd &m1, double a);

	// O <- T,O operations

	RxVecd &Vmul (RxVecd *vr, const RxVecd &v1) const;
	RxVecd &mulV (RxVecd *vr, const RxVecd &v1) const;
	friend RxVecd operator*(const RxMatd &m1, const RxVecd &v1);
	friend RxVecd operator*(const RxVecd &v1, const RxMatd &m1);
	   
	// T <- T operations

	RxMatd &tsp(const RxMatd &m1);
	RxMatd &tsp();
	RxMatd T()
	 { RxMatd mr(nc, nr);
	   return (mr.tsp(*this));
	 }

	RxMatd &inv(const RxMatd &m1);
	RxMatd &inv()
	 { return (inv (*this));
	 }
	RxMatd I()
	 { RxMatd mr(nr,nc);
	   return (mr.inv(*this));
	 }

	RxMatd &neg(const RxMatd &m1);
	RxMatd &neg()
	 { return (neg (*this));
	 }
	RxMatd operator-()
	 { RxMatd mr(nr,nc);
	   return (mr.neg(*this));
	 }

	// T <- O operations

	RxMatd& setZero();
	RxMatd& setIdent();

	// O <- T operations

	double det() const;
	double norm1() const;
	double normInf() const;
	double normF() const;
	double trace() const;	
	double condest() const;	
};

inline RxVecd operator*(const RxMatd &m1, const RxVecd &v1)
{
	RxVecd vr(m1.nr);
	return (m1.mulV(&vr,v1));
}

inline RxVecd operator*(const RxVecd &v1, const RxMatd &m1)
{
	RxVecd vr(m1.nc);
	return (m1.Vmul(&vr,v1));
}

#endif
