#ifndef _RXMATDSVD_H_WRAP_
#define _RXMATDSVD_H_WRAP_

#include <RxMatd.h>

#define RX_SVD_DEFAULT_EPS	(1e-13)
#define RX_SVD_OMIT		(&RxMatdSvd::none)
#define RX_SVD_DEFAULT		(NULL)

class RxMatdSvd
{
	friend class RxMatd;

	int doSvd ();
	int calcSetup(const RxMatd &A, RxVecd *r, const RxVecd *b);
	void zeroRow (double *buf, int ib, int p, int w);
	void takeStep (double *buf, int ib, int nn, int w);
	void swapUV();

	double tol;
#define RX_SVD_CALCDONE			0x01
#define RX_SVD_SIG_MALLOCED		0x02
#define RX_SVD_U_MALLOCED		0x04
#define RX_SVD_V_MALLOCED		0x08
	int flags;
	int m, n;

	double sigmax;
	double sigdet;
	double sigmin;

	RxMatd *U;
	RxVecd *sig;
	RxMatd *V;
	double *buf;
	RxVecd *rv;
	const RxVecd *bv;
public:
	static RxMatd none;

	// initialization

	int calc (const RxMatd &A);
	int calc (const RxMatd &A, RxVecd *r, const RxVecd &b);
	int setSUV (RxVecd *s, RxMatd *u, RxMatd *v);
	void setTol (double eps)
	 { tol = eps;
	 }
	void getSUV (RxVecd **s, RxMatd **u, RxMatd **v);
	RxVecd *getSig ()
	 { return sig;
	 }
	RxMatd *getV ()
	 { return V;
	 }
	RxMatd *getU ()
	 { return U;
	 }
	double getTol () const
	 { return (tol);
	 }

	// constructors 

	RxMatdSvd ()
	 { flags = 0;
	   U = NULL;
	   V = NULL;
	   sig = NULL;
	   buf = NULL;
	   tol = RX_SVD_DEFAULT_EPS;
	 }
	RxMatdSvd (const RxMatd &A,RxVecd *s=NULL, RxMatd *u=NULL, 
			 RxMatd *v=NULL, double eps=RX_SVD_DEFAULT_EPS)
	 { flags = 0;
	   U = NULL;
	   V = NULL;
	   sig = NULL;
	   buf = NULL;  
	   setSUV (s, u, v);
	   tol = eps;
	   calc (A);
	 }
	~RxMatdSvd ();

	RxVecd &solve (RxVecd *x, const RxVecd &b) const;
	RxVecd &Tsolve (RxVecd *x, const RxVecd &b) const;
	RxMatd &solve (RxMatd *X, const RxMatd &B) const;
	RxMatd &Tsolve (RxMatd *X, const RxMatd &B) const;
	double cond() const;
	double norm2() const;
	double det() const;

	RxMatd &inv (RxMatd *M) const;
};

#endif
