#ifndef _RXMATERIAL_H_WRAP_
#define _RXMATERIAL_H_WRAP_

#include <RxPrintable.h>
#include <RxMuxedNameTable.h>
#include <RxRefManager.h>

#define RGB_PACK(r,g,b)	 (((unsigned)(b)<<16) + ((unsigned)(g)<<8) + \
			  (unsigned)(r))

#define RGBA_PACK(r,g,b,a) (((unsigned)(b)<<16) + ((unsigned)(g)<<8) + \
			  (unsigned)(r) + ((unsigned)(a)<<24))

class RxMaterial : public RxPrintable
{
    private:
	RxRefManager ref;

	float ambient_[4];
	float diffuse_[4];
	float specular_[4];
	float emission_[4];
	float shine_;
	char *name_;

	static RxMuxedNameTable nameTable;
	void unpack (float *vec, unsigned long val) const;
	unsigned long pack (const float *vec) const;

	// this pointer, normally NULL, is used in case the graphics
	// system want to allocate its own material data structure.
	// In that case, the allocated structure can be stored here.
	void *gobj;
    public:

	void *operator new (size_t n)
	 { return (RxMaterial*)RxRefManager::alloc(n);
	 }
	void operator delete (void *p, size_t n)
	 { RxRefManager::dealloc (((RxMaterial*)p)->ref, p, n);
	 }
	int reference()
	 { return ref.inc();
	 }

	void set (unsigned long amb, unsigned long dif, unsigned long spe,
		  unsigned long emi, float shine);
	void set (const RxMaterial &mat);

	RxMaterial ()
	 { set (0, 0, 0, 0, 0);
	   gobj = NULL;
	   name_ = NULL;
	 }
	RxMaterial (unsigned long amb, unsigned long dif, unsigned long spe,
		    unsigned long emi, float shine)
	 { set (amb, dif, spe, emi, shine);
	   gobj = NULL;
	   name_ = NULL;
	 }
	RxMaterial (const RxMaterial &mat)
	 { set (mat);
	   gobj = mat.gobj;
	   name_ = NULL;
	 }

	void setAmbient (unsigned long pack)
	 { unpack (ambient_, pack);
	 }
	void setAmbient (float r, float g, float b, float a=1.0);

	unsigned long getAmbient () const
	 { return pack (ambient_);
	 }
	void getAmbient (float *rp, float *gp, float *bp, float *ap=NULL) const;

	void setDiffuse (unsigned long pack)
	 { unpack (diffuse_, pack);
	 }
	void setDiffuse (float r, float g, float b, float a=1.0);

	unsigned long getDiffuse () const
	 { return pack (diffuse_);
	 }
	void getDiffuse (float *rp, float *gp, float *bp, float *ap=NULL) const;

	void setSpecular (unsigned long pack)
	 { unpack (specular_, pack);
	 }
	void setSpecular (float r, float g, float b, float a=1.0);

	unsigned long getSpecular () const
	 { return pack (specular_);
	 }
	void getSpecular (float *rp, float *gp, float *bp, float *ap=NULL) const;

	void setEmission (unsigned long pack)
	 { unpack (emission_, pack);
	 }
	void setEmission (float r, float g, float b, float a=1.0);

	unsigned long getEmission () const
	 { return pack (emission_);
	 }
	void getEmission (float *rp, float *gp, float *bp, float *ap=NULL) const;

	void setShine (float val)
	 { shine_ = val;
	 } 
	float getShine () const
	 { return shine_;
	 }

	int setName (const char *name);
	const char *name() const
	 { return name_;
	 }
	static RxMaterial *findByName(const char *name)
	 { return (RxMaterial*)nameTable.getEntry (name);
	 }

	void *getGraphicObj() const
	 { return gobj;
	 }
	void setGraphicObj(void *ptr)
	 { gobj = ptr;
	 }

	virtual int oscan (class RxParseObj *obj);
	virtual int cprint(class RxCoutput *co,class RxPrintSpec *ps) const;
};

#endif
