#ifndef _RXPLANAR2RKIN_H_WRAP_
#define _RXPLANAR2RKIN_H_WRAP_

#include <RxGenKin.h>
#include <RxVecd2.h>

#define RX_PLANAR2RKIN_SING_ELBOW	 0x0100

#define RX_PLANAR2RKIN_CONF_DOWN	 0x0001

#define RX_PLANAR2RKIN_CLIP_OUTER	 0x0001
#define RX_PLANAR2RKIN_CLIP_INNER	 0x0002

class RxP2RKin : public RxGenKin
{
	void p2Rinit();
	int setP2RJnts (const RxVecd2 &jnts);

	double s12;
	double c12;

	void buildJacobian();
	void setDefaultP2R ();

	double j00, j10;
	double j01, j11;
public:
	RxP2RKin () : RxGenKin()
	 { setDefaultP2R();
	   p2Rinit();
	 }
	RxP2RKin (const char *fileName) : RxGenKin (fileName)
	 { p2Rinit();
	 }
	RxP2RKin (const DHParams *params, int n) : RxGenKin (params, n)
	 { p2Rinit();
	 }
	RxP2RKin (const RxP2RKin &kin) : RxGenKin (kin)
	 { p2Rinit();
	 }

	int setJnts (const RxVecd &jnts)
	 { return (setP2RJnts (*(const RxVecd2 *)&jnts));
	 }

	int fwdK (RxTrsf *T, int *conf);
	int fwdK (RxTrsf *T, int *conf, const RxVecd &jnts)
	 { setP2RJnts (*(const RxVecd2 *)&jnts);
	   return (fwdK (T, conf));
	 }

	RxMatd& Jset (RxMatd *J);
	RxMatd& Jset (RxMatd *J, const RxVecd jnts)
	 { setP2RJnts (*(const RxVecd2 *)&jnts);
	   return (Jset (J));
	 }
	RxTwist& Jmul (RxTwist *tw, const RxVecd &dj);
	RxTwist& Jmul (RxTwist *tw, const RxVecd &dj, 
			       const RxVecd jnts)
	 { setP2RJnts (*(const RxVecd2 *)&jnts);
	   return (Jmul (tw, dj));
	 }
	RxVecd& JTmul (RxVecd *tor, const RxWrench &wr);
	RxVecd& JTmul (RxVecd *tor, const RxWrench &wr, 
			       const RxVecd jnts)
	 { setP2RJnts (*(const RxVecd2 *)&jnts);
	   return (JTmul (tor, wr));
	 }

	// inverse kinematics

	int invK (RxVecd *jnts, const RxTrsf &T, int conf, 
			  const RxVecd *jref=NULL, int update=0);

	int Jsolve (RxVecd *dj, const RxTwist &tw);
	int Jsolve (RxVecd *dj, const RxTwist &tw,
		   const RxVecd jnts)
	 { setP2RJnts (*(const RxVecd2 *)&jnts);
	   return (Jsolve (dj, tw));
	 }
	int JTsolve (RxWrench *wr, const RxVecd &tor);
	int JTsolve (RxWrench *wr, const RxVecd &tor,
		     const RxVecd jnts)
	 { setP2RJnts (*(const RxVecd2 *)&jnts);
	   return (JTsolve (wr, tor));
	 }
};

#endif
