#ifndef _RXPARSEOBJ_H_WRAP_
#define _RXPARSEOBJ_H_WRAP_

#include <iostream.h>
#include <stddef.h>
#include <RxDlist.h>
#include <RxCio.h>
#include <RxRefManager.h>

class RxParseObj;

class RxParseNode : public RxDnodeBase
{
	friend class RxParseObj;
	friend class RxParseTest;
	friend class RxParser;

	RxParseObj *obj_;
public:
	RxParseNode()
	 { obj_ = NULL;
	 }
	RxParseNode(RxParseObj *node)
	 { obj_ = node;
	 }
	~RxParseNode();
	RxParseObj *obj()
	 { return obj_;
	 }
	RX_DNODE_DECLS (RxParseNode);
};

class RxParseObj : public RxDnodeBase
{
	friend class RxParser;
private:
	void init();
public:
	enum {
	  BOOLEAN = 0x01,
	  INTEGER = 0x02,
	  DOUBLE = 0x04,
	  NUMBER = (INTEGER | DOUBLE),
	  STRING = 0x08,
	  IDENTIFIER = 0x10,
	  LIST = 0x100,
	  BOOLEAN_LIST = (BOOLEAN | LIST),
	  INTEGER_LIST = (INTEGER | LIST),
	  DOUBLE_LIST = (DOUBLE | LIST),
	  STRING_LIST = (STRING | LIST),
	  NUMBER_LIST = (NUMBER | LIST),
	  IDENTIFIER_LIST = (IDENTIFIER | LIST),
	  LIST_IN_LIST = 0x200,
	  };
	union Value {
	  double d;
	  int i;
	  char *s;
	  class RxDlist<RxParseNode> *listp;
	  };

public:
	RxRefManager ref;

	int typeFlag;
	union Value val;	
	class RxDlist<RxParseNode> list_;
	int depth;
	char *typeName;

	char *tagName;
	int lineno;
	void *refptr;
	RxParseObj *refobj;

	RxParseNode *linkNode;

	void clearType();
	RxParseNode *curNode;		// used for iteration
	void setRef (RxParseObj &ref);

public:
	void *operator new (size_t n)
	 { allocCnt++;
	   return (RxParseObj*)RxRefManager::alloc(n);
	 }
	void operator delete (void *p, size_t n)
	 { if (RxRefManager::dealloc (((RxParseObj*)p)->ref, p, n))
	    { allocCnt--;
	    }
	 }
	int reference()
	 { return ref.inc();
	 }
	static int allocCnt;

	static RxParseObj *eof;

	int getValue (union Value *valp);
	int getType ()
	 { return typeFlag;
	 }
	
	int isBoolean()
	 { return (typeFlag == BOOLEAN);
	 }
	int isInteger()
	 { return (typeFlag == INTEGER);
	 }
	int isDouble()
	 { return (typeFlag == DOUBLE);
	 }
	int isString()
	 { return (typeFlag == STRING);
	 }
	int isIdentifier()
	 { return (typeFlag == IDENTIFIER);
	 }
	int isNumber()
	 { return ((typeFlag & ~NUMBER) == 0);
	 }
	int isList()
	 { return ((typeFlag & LIST) != 0);
	 }
	int isListList()
	 { return (typeFlag == LIST|LIST_IN_LIST);
	 }
	int isIntegerList()
	 { return (typeFlag == INTEGER_LIST);
	 }
	int isDoubleList()
	 { return (typeFlag == DOUBLE_LIST);
	 }
	int isStringList()
	 { return (typeFlag == STRING_LIST);
	 }
	int isIdentifierList()
	 { return (typeFlag == IDENTIFIER_LIST);
	 }
	int isBooleanList()
	 { return (typeFlag == BOOLEAN_LIST);
	 }
	int isNumberList()
	 { return ((typeFlag & LIST) && (typeFlag & ~NUMBER_LIST)==0);
	 }

	const char *getTagName() const
	 { return tagName;
	 }
	const char *getTypeName() const
	 { return typeName;
	 }

	int hasTagName (const char *s)
	 { return (getTagName() && strcmp(getTagName(), s)==0);
	 }
	int hasTypeName (const char *s)
	 { return (typeName && strcmp(typeName, s)==0);
	 }

	int getLine () const
	 { return lineno;
	 }
	void *getRefPtr() const
	 { return refobj ? refobj->refptr : refptr;
	 }
	void setRefPtr(void *ptr)
	 { if (refobj)
	    { refobj->refptr = ptr;
	    }
	   else
	    { refptr = ptr;
	    }
	 }

	int getRefCnt() const
	 { return ref.get();
	 }
	const char *getStringValue() const
	 { return val.s;
	 }
	int getIntegerValue() const
	 { return val.i;
	 }
	double getDoubleValue() const
	 { return val.d;
	 }
	double getNumberValue() const;

	int getNumberListValues (double *vec, int n, RxParseObj **objp=NULL);

	const RxDlist<RxParseNode> &getList() const
	 { return *val.listp;
	 }
	int getListDepth ()
	 { return depth;
	 }
	int getListLength ()
	 { return val.listp->length();
	 }
	RxParseObj *firstChild();
	RxParseObj *nextChild();

	RxParseObj()
	 { init();
	 }
	~RxParseObj();

	static RxParseObj *scan (class RxCinput *ci, int *linenop=NULL);
	static RxParseObj *scan (FILE *fp, int useXscan=0)
	 { RxCinput ci(fp, useXscan);
	   return scan (&ci);
	 }
	static RxParseObj *scan (const char *buf, int *cntp=NULL)
	 { RxCinput ci(buf, cntp);
	   return scan (&ci);
	 }
	static RxParseObj *scan (istream &is)
	 { RxCinput ci(&is);
	   return scan (&ci);
	 }

	static void freeHashedObjs (class RxCinput *ci);

	int print (class RxCoutput *co, class RxPrintSpec *fs);
	int print (FILE *fp, const char *fmt)
	 { RxCoutput co(fp);
	   RxPrintSpec fs(fmt);
	   return print (&co, &fs);
	 }
	int print (char *buf, const char *fmt, int n=0)
	 { RxCoutput co(buf, n);
	   RxPrintSpec fs(fmt);
	   return print (&co, &fs);
	 }
	int print (ostream &os)
	 { RxCoutput co(&os);
	   return print (&co, NULL);
	 }

	RX_DNODE_DECLS (RxParseObj);
};

#endif
