#ifndef _RXQTRN_H_WRAP_
#define _RXQTRN_H_WRAP_

#include <RxVecdBase.h>
#include <RxVecd3.h>
#include <mathx.h>

class RxRot;

class RxQtrn : public RxVecdBase
{
public:
	double c;
	RxVecd3 v;
	
	// construction

	RxQtrn ()
	 { 
	 };
	RxQtrn (double cc, double vx, double vy, double vz)
	 { c = cc; v.set(vx,vy,vz);
	 }
	RxQtrn (const RxQtrn &q2)
         { c = q2.c;
	   v = q2.v;
	 };
	// query and access
	
	int length() const
	 { return 4;
	 }
	double& operator()(int i);
	double operator()(int i) const;

	// assignment 

	RxQtrn& operator=(const RxQtrn& q2)
	 { c = q2.c;
	   v = q2.v;
	   return (*this);
	 };
	RxQtrn& set (double cc, double vx, double vy, double vz)
	 { c = cc; v.set(vx,vy,vz);
	   return (*this);
	 }
	void get (double *cp, double *xp, double *yp, double *zp) const
	 { *cp = c; v.get(xp, yp, zp);
	 }
	RxQtrn& set (const double *dp)
	 { c = dp[0]; v.set(&dp[1]);
	   return (*this);
	 }
	void get (double *dp) const
	 { dp[0] = c; v.get(&dp[1]);
	 }

	RxQtrn &setRot (const RxRot &R1);
	void getRot (RxRot *R1) const;

	RxQtrn &setEul (double phi, double the, double psi);
	RxQtrn &setEulD (double phi, double the, double psi)
	 { return (setEul (DTOR*phi, DTOR*the, DTOR*psi));
	 }
	void getEul (double *phip, double *thep, double *psip) const;
	void getEulD (double *phip, double *thep, double *psip) const
	 { getEul (phip, thep, psip);
	   *phip *= RTOD; *thep *= RTOD; *psip *= RTOD;
	 }
	RxQtrn &setRpy (double roll, double pitch, double yaw);
	RxQtrn &setRpyD (double roll, double pitch, double yaw)
	 { return (setRpy (DTOR*roll, DTOR*pitch, DTOR*yaw));
	 }
	void getRpy (double *rollp, double *pitchp, double *yawp) const;
	void getRpyD (double *rollp, double *pitchp, double *yawp) const
	 { getRpy (rollp, pitchp, yawp);
	   *rollp *= RTOD; *pitchp *= RTOD; *yawp *= RTOD;
	 }
	RxQtrn &setAxis (const RxVecd3 &v, double ang);
	RxQtrn &setAxisD (const RxVecd3 &v, double ang)
	 { return (setAxis (v, DTOR*ang));
	 }
	void getAxis (RxVecd3 *v, double *angp) const;
	void getAxisD (RxVecd3 *v, double *angp) const
	 { getAxis (v, angp);
	   *angp *= RTOD;
	 }
	RxQtrn &setVecs (const char *axes,const RxVecd3 &v1,const RxVecd3 &v2);
	RxQtrn &setAngs (const char *axes, ...);
	RxQtrn &setAngsD (const char *axes, ...);

	// extra constructors 

	RxQtrn (double roll, double pitch, double yaw)
	 { setRpyD (roll, pitch, yaw);
	 }
	RxQtrn (const double *dp)
         { set (dp);
	 }

	// T <- T,T operations

	RxQtrn& add (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& add (const RxQtrn &q2)
	 { return (add(*this,q2));
	 }
	RxQtrn& operator+=(const RxQtrn &q2)
	 { return (add(q2));
	 }
	friend RxQtrn operator+(const RxQtrn &q1, const RxQtrn &q2);

	RxQtrn& sub (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& sub (const RxQtrn &q2)
	 { return (sub(*this,q2));
	 }
	RxQtrn& operator-=(const RxQtrn &q2)
	 { return (sub(q2));
	 }
	friend RxQtrn operator-(const RxQtrn &q1, const RxQtrn &q2);

	RxQtrn& mul (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& mul (const RxQtrn &q2)
	 { return (mul (*this, q2));
	 }
	RxQtrn& operator*=(const RxQtrn &q2)
	 { return (mul(q2));
	 }
	friend RxQtrn operator*(const RxQtrn &q1, const RxQtrn &q2);

	RxQtrn& mulI (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& mulI (const RxQtrn &q2)
	 { return (mulI (*this, q2));
	 }
	RxQtrn& operator/=(const RxQtrn &q2)
	 { return (mulI(q2));
	 }
	friend RxQtrn operator/(const RxQtrn &q1, const RxQtrn &q2);
	
	RxQtrn& Imul (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& Imul (const RxQtrn &q2)
	 { return (Imul (*this, q2));
	 }
	RxQtrn& ImulI (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& ImulI (const RxQtrn &q2)
	 { return (ImulI (*this, q2));
	 }

	RxQtrn& slerp (const RxQtrn &q1, const RxQtrn &q2, double u);
	RxQtrn& slerp (const RxQtrn &q2, double u)
	 { return (slerp (*this, q2, u));
	 }

	RxQtrn& mulC (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& mulC (const RxQtrn &q2)
	 { return (mulC (*this, q2));
	 }
	RxQtrn& Cmul (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& Cmul (const RxQtrn &q2)
	 { return (Cmul (*this, q2));
	 }
	RxQtrn& CmulC (const RxQtrn &q1, const RxQtrn &q2);
	RxQtrn& CmulC (const RxQtrn &q2)
	 { return (CmulC (*this, q2));
	 }

	// T <- T,O operations

	RxQtrn& scale (double a, const RxQtrn &q1);
	RxQtrn& scale (double a)
	 { return (scale (a, *this));
	 }
	RxQtrn& operator*=(double a)
	 { return (scale(a));
	 }
	RxQtrn& operator/=(double a)
	 { return (scale(1/a));
	 }
	friend RxQtrn operator*(const RxQtrn &q1, double a);
	friend RxQtrn operator*(double a, const RxQtrn &q1);
	friend RxQtrn operator/(const RxQtrn &q1, double a);

	// O <- T,T operations

	double dot(const RxVecd3 &q2) const;

	// T <- T operations

	RxQtrn &unit(const RxQtrn &q1);
	RxQtrn &unit()
	 { return (unit(*this));
	 }
	RxQtrn &neg(const RxQtrn &q1)
	 { c = -q1.c; v.x() = -q1.v.x(); v.y() = -q1.v.y(); v.z() = -q1.v.z();
	   return (*this);
	 }
	RxQtrn &neg()
	 { return (neg(*this));
	 }
	friend RxQtrn operator-(const RxQtrn &q1);

	RxQtrn& inv(const RxQtrn &q1);
	RxQtrn& inv(void)
	 { return inv(*this);
	 }
	RxQtrn I(void)
	 { RxQtrn mr;
	   mr.inv(*this);
	   return (mr);
	 }
	RxQtrn& conj(const RxQtrn &q1)
	 { v.x() = -q1.v.x(); v.y() = -q1.v.y(); v.z() = -q1.v.z();
	   return (*this);
	 }
	RxQtrn& conj()
	 { return (conj(*this));
	 }
	RxQtrn C()
	 { RxQtrn mr;
	   mr.conj(*this);
	   return (mr);
	 }

	// T <- O operations

	RxQtrn &setZero()
	 { c=0; v.x() = 0; v.y() = 0; v.z() = 0;
	   return (*this);
	 }
	RxQtrn &setIdent()
	 { c=1; v.x() = 0; v.y() = 0; v.z() = 0;
	   return (*this);
	 }

	// O <- T operations

	double norm2() const;
	
	int equal (const RxQtrn &Q1, double eps);

	// constants

	static const RxQtrn ZERO;
	static const RxQtrn IDENT;
};

inline RxQtrn operator+(const RxQtrn &q1, const RxQtrn &q2)
{
	RxQtrn qr;
	return (qr.add(q1, q2));
}

inline RxQtrn operator-(const RxQtrn &q1, const RxQtrn &q2)
{
	RxQtrn qr;
	return (qr.sub(q1, q2));
}

inline RxQtrn operator*(const RxQtrn &q1, const RxQtrn &q2)
{
	RxQtrn qr;
	return (qr.mul(q1, q2));
}

inline RxQtrn operator/(const RxQtrn &q1, const RxQtrn &q2)
{
	RxQtrn qr;
	return (qr.mulI(q1, q2));
}

inline RxQtrn operator*(const RxQtrn &q1, double a)
{
	RxQtrn qr;
	return (qr.scale(a, q1));
}

inline RxQtrn operator*(double a, const RxQtrn &q1)
{
	RxQtrn qr;
	return (qr.scale(a, q1));
}

inline RxQtrn operator/(const RxQtrn &q1, double a)
{
	RxQtrn qr;
	return (qr.scale(1/a, q1));
}

inline RxQtrn operator-(const RxQtrn &q1)
{
	RxQtrn qr;
	return (qr.neg(q1));
}

#endif
