#ifndef _RXRIGIDBODY_H_WRAP_
#define _RXRIGIDBODY_H_WRAP_

#include <RxFrameNode.h>
#include <RxSpatialInertia.h>
#include <RxBoundingRegion.h>
#include <RxCvxPolyhed3.h>
#include <RxHashTab.h>
#include <mutexdefs.h>
#include <RxVolProps.h>

class RxRigidBody : public RxFrameNode
{
protected:
	RxSpatialInertia *inertia;
	RxBoundingRegion *breg;
	RxCvxPolyhed3 *poly_;
	RxRigidBody *body_;
	RxVolProps volp;
	int volpValid;
	int printT; 

	void init();
	virtual void clearAll();
	virtual int oscanField (class RxParseObj *obj);
	virtual int checkFields (class RxParseObj *obj);
	virtual int cprintFields (class RxCoutput *co,class RxPrintSpec *ps,
				  int init=1) const;

	void computeVolsRecurs (RxVolProps &volpTot, const RxTrsf &TRP);
	void computeDensityInfoRecurs (double &knownMass, double &freeVol);
	void computeDensityInfo (double &knownMass, double &freeVol);
	void computeInertiaFromVols (RxSpatialInertia &Itot,
				     const RxTrsf &TRP, double density);
public:
	static const int ID;
	static RxRigidBody *findByName(const char *s);

	virtual const char *typeName () const
	 { return "RigidBody";
	 }
	RxRigidBody ()
	 { init();
	 } 
	RxRigidBody (RxCvxPolyhed3 *poly, const RxTrsf *Tp=NULL)
	 { init();
	   setPoly (poly, Tp);
	 }
	RxRigidBody (const RxRigidBody *bdy, const RxTrsf *Tp=NULL)
	 { init();
	   setBody (bdy, Tp);
	 }
	~RxRigidBody()
	 { clearObj();
	 }

	RxCvxPolyhed3 *poly() const
	 { return poly_;
	 }
	RxRigidBody *body() const
	 { return body_;
	 }

	void setInertia (const RxSpatialInertia &I);
	void clearInertia ();
	void getInertia (RxSpatialInertia &I) const;
	const RxSpatialInertia *getInertia() const
	 { return inertia;
	 }

	void computeVols ();
	void computeInertia (double totalMass);
	double getVolume ();
	const RxVolProps *getVolProps() const
	 { return &volp;
	 }

	void clearObj();
	void setPoly (RxCvxPolyhed3 *poly, const RxTrsf *Tp=NULL);
// copies the entire sub-object and consolidates it into one rigid body
	void setBody (const RxRigidBody *bdy, const RxTrsf *Tp=NULL);

	virtual int setJnts (const RxVecd &vec, int j=0);
	virtual int setJvels (const RxVecd &vec, int j=0);
};

#endif
