#ifndef _RXROT_H_WRAP_
#define _RXROT_H_WRAP_

#include <RxMatdBase.h>
#include <RxVecd3.h>
#include <mathx.h>
#include <stdarg.h>

class RxRot : public RxMatdBase
{
    private:
	void doCopy (const RxRot &m1);
    public:
	RxVecd3 n;
	RxVecd3 o;
	RxVecd3 a;

	// constructors

	RxRot ()
	 {
	 }
	RxRot (const RxRot &m2)
         { doCopy (m2);
	 }

	// query and access

	int nrows() const
	 { return 3;
	 }
	int ncols() const
	 { return 3;
	 }
	double& operator()(int i, int j);
	double operator()(int i, int j) const;

	// assignment

	RxRot& operator=(const RxRot& m2)
	 { doCopy (m2);
	   return (*this);
	 }
	RxRot &set (const double *dp);
	void get (double *dp) const;

	RxRot &setEul (double phi, double the, double psi);
	RxRot &setEulD (double phi, double the, double psi)
	 { return (setEul (DTOR*phi, DTOR*the, DTOR*psi));
	 }
	void getEul (double *phip, double *thep, double *psip) const;
	void getEulD (double *phip, double *thep, double *psip) const
	 { getEul (phip, thep, psip);
	   *phip *= RTOD; *thep *= RTOD; *psip *= RTOD;
	 }
	RxRot &setRpy (double roll, double pitch, double yaw);
	RxRot &setRpyD (double roll, double pitch, double yaw)
	 { return (setRpy (DTOR*roll, DTOR*pitch, DTOR*yaw));
	 }
	void getRpy (double *rollp, double *pitchp, double *yawp) const;
	void getRpyD (double *rollp, double *pitchp, double *yawp) const
	 { getRpy (rollp, pitchp, yawp);
	   *rollp *= RTOD; *pitchp *= RTOD; *yawp *= RTOD;
	 }
	RxRot &setAxis (const RxVecd3 &v, double ang);
	RxRot &setAxisD (const RxVecd3 &v, double ang)
	 { return (setAxis (v, DTOR*ang));
	 }
	void getAxis (RxVecd3 *v, double *angp) const;
	void getAxisD (RxVecd3 *v, double *angp) const
	 { getAxis (v, angp);
	   *angp *= RTOD;
	 }
	RxRot &setVecs (const char *axes,const RxVecd3 &v1,const RxVecd3 &v2);
	RxRot &setAngs (const char *axes, ...);
	RxRot &setAngsD (const char *axes, ...);
	int vsetAngs (const char *fmt, int inDegrees, va_list ap);

	// extra constructors 

	RxRot (double roll, double pitch, double yaw)
	 { setRpyD (roll, pitch, yaw);
	 }
	RxRot (const double *dp)
         { set (dp);
	 }
	RxRot (const RxVecd3 &v, double ang)
	 { setAxis (v, ang);
	 }

	// T <- T,T operations
	
	RxRot& mul (const RxRot &m1, const RxRot &m2);
	RxRot& mul (const RxRot &m2)
	 { return (mul (*this, m2));
	 }
	RxRot& operator*=(const RxRot &m2)
	 { return (mul (*this, m2));
	 }
	friend RxRot operator*(const RxRot &m1, const RxRot &m2);

	RxRot& mulI (const RxRot &m1, const RxRot &m2);
	RxRot& mulI (const RxRot &m2)
	 { return (mulI (*this, m2));
	 }
	RxRot& Imul (const RxRot &m1, const RxRot &m2);
	RxRot& Imul (const RxRot &m2)
	 { return (Imul (*this, m2));
	 }
	RxRot& ImulI (const RxRot &m1, const RxRot &m2);
	RxRot& ImulI (const RxRot &m2)
	 { return (ImulI (*this, m2));
	 }
	RxRot& slerp (const RxRot &m1, const RxRot &m2, double u);
	RxRot& slerp (const RxRot &m2, double u)
	 { return (slerp (*this, m2, u));
	 }

	// O <- T,O operations

#if 1
	RxVecd3 &mul (RxVecd3 &vr, const RxVecd3 &v1) const;
	RxVecd3 &mul (RxVecd3 &vr) const
	 { return (mul (vr, vr));
	 }
	RxVecd3 &Imul (RxVecd3 &vr, const RxVecd3 &v1) const;
	RxVecd3 &Imul (RxVecd3 &vr) const
	 { return (Imul (vr, vr));
	 }
	friend RxVecd3 operator*(const RxRot &m1, const RxVecd3 v2);
#endif

	// T <- T operations

	RxRot &tsp(const RxRot &m1);
	RxRot &tsp();
	RxRot T()
	 { RxRot mr;
	   return (mr.tsp(*this));
	 }

	RxRot &inv(const RxRot &m1)
	 { return (tsp(m1));
	 }
	RxRot &inv()
	 { return (tsp());
	 }
	RxRot I()
	 { RxRot mr;
	   return (mr.tsp(*this));
	 }
	
	RxRot &normalize(const RxRot &m1);
	RxRot &normalize()
	 { return normalize(*this);
	 }

	// T <- O operations

	RxRot& setIdent();

	// constants

	static const RxRot IDENT;
#if 0
	RxVecd3 &xform (RxVecd3 &vr, const RxVecd3 &v1) const
	 { return (Imul (vr, v1));
	 }
	RxVecd3 xform (const RxVecd3 &v1) const
	 { RxVecd3 vr;
	   return (xform (vr, v1));
	 }
	RxTwist &xform (RxTwist &tr, const RxTwist &t1) const
	 { Imul (tr.w, t1.w);
	   Imul (tr.v, t1.v);
	   return (tr);
	 }
	RxTwist xform (const RxTwist &t1) const
	 { RxTwist tr;
	   return (xform (tr, t1));
	 }
	RxWrench &xform (RxWrench &wr, const RxWrench &w1) const
	 { Imul (wr.f, w1.f);
	   Imul (wr.m, w1.m);
	   return (wr);
	 }
	RxWrench xform (const RxWrench &w1) const
	 { RxWrench wr;
	   return (xform (wr, w1));
	 }

	RxVecd3 &Ixform (RxVecd3 &vr, const RxVecd3 &v1) const
	 { return (mul (vr, v1));
	 }
	RxVecd3 Ixform (const RxVecd3 &v1) const
	 { RxVecd3 vr;
	   return (Ixform (vr, v1));
	 }
	RxTwist &Ixform (RxTwist &tr, const RxTwist &t1) const
	 { mul (tr.w, t1.w);
	   mul (tr.v, t1.v);
	   return (tr);
	 }
	RxTwist Ixform (const RxTwist &t1) const
	 { RxTwist tr;
	   return (Ixform (tr, t1));
	 }
	RxWrench &Ixform (RxWrench &wr, const RxWrench &w1) const
	 { mul (wr.f, w1.f);
	   mul (wr.m, w1.m);
	   return (wr);
	 }
	RxWrench Ixform (const RxWrench &w1) const
	 { RxWrench wr;
	   return (Ixform (wr, w1));
	 }
#endif
	virtual int oscan (class RxParseObj *obj);
	virtual int cprint (class RxCoutput *co, class RxPrintSpec *fs) const;
};

inline RxRot operator*(const RxRot &m1, const RxRot &m2)
{
	RxRot mr;
	mr.mul (m1, m2);
	return (mr);
}

#if 1
inline RxVecd3 operator*(const RxRot &m1, const RxVecd3 &v2)
{
	RxVecd3 vr;
	return (m1.mul(vr, v2));
}
#endif

#endif
