#ifndef _RXSLIST_H_WRAP_
#define _RXSLIST_H_WRAP_

#include <RxFastAlloc.h>

#ifndef NULL
#define NULL 0
#endif

// Usage:
// 
// class XXXNode : public RxSnodeBase
// {
//  	... constructors, destructors, whatever ...
//
//	RX_NAMED_SNODE_DECLS (XXXNode, name);	// name is a private char *
//						// that names the node.
//	 *OR*
//
//	RX_SNODE_DECLS (XXXNode);
//
//	int compare (const void *ptr) const 	// optional ...
//	 { ... see if ptr references this node ...
//	 };
// }
//
// RxSlist<NodeType>
// 

class RxSnodeBase : public RxFastAlloc
{
	friend class RxSlistBase;

    protected:
	RxSnodeBase *next_;
    public:
	RxSnodeBase()
	 { next_ = NULL;
	 };
#if 0
	RxSnodeBase *getNext()
	 { return next_;
	 };
#endif
	RxSnodeBase *getNext(int n);
	const RxSnodeBase *getNext(int n) const
	 { return getNext(n);
	 }
	virtual int compare (const void *ptr) const
	 { return ((void*)this == ptr);
	 };
	virtual const char *getName () const
	 { return (NULL);
	 };
};

class RxSlistBase
{
    protected:
	RxSnodeBase *head_;
	RxSnodeBase *tail_;
	RxSnodeBase *findNodeByName (const char *) const;
	RxSnodeBase *findNodeByPtr (const void *ptr) const;
	int autoDelete;
	virtual void freeNode (void *)
	 { 
	 };
    public:
	RxSlistBase()
	 { head_ = NULL; 
	   tail_ = NULL;
	   autoDelete = 1;
	 };
//	~RxSlistBase();
	RxSnodeBase *addHead(RxSnodeBase* ptr);
	RxSnodeBase *addTail(RxSnodeBase* ptr);
	RxSnodeBase *addAfter(RxSnodeBase* ref, RxSnodeBase* ptr);
	RxSnodeBase *addBefore(RxSnodeBase* ref, RxSnodeBase* ptr);
	RxSnodeBase *remove(RxSnodeBase* ptr);
	void removeAndDelete(RxSnodeBase* ptr)
	 { remove (ptr);
	   freeNode (ptr);
	 }

	int removeAll();
	int removeAndDeleteAll();

	void setAutoDelete(int val)
	 { autoDelete = val;
	 }
	int getAutoDelete() const
	 { return autoDelete;
	 }

	int length() const;
};

#define RX_NAMED_SNODE_DECLS(N,name) \
    private: \
	char *name; \
    public: \
	const char *getName() const \
	 { return (name); \
	 } \
	void setName(const char *str) \
	 { free ((void*)name); \
	   name = strdup(str); \
	 } \
	N *next() \
	 { return ((N*)next_); \
	 } \
	N *next(int n) \
	 { return ((N*)getNext(n)); \
	 } \
	const N *next() const \
	 { return ((N*)next_); \
	 } \
	const N *next(int n) const \
	 { return ((N*)getNext(n)); \
	 }

#define RX_SNODE_DECLS(N) \
	N *next() \
	 { return ((N*)next_); \
	 } \
	N *next(int n) \
	 { return ((N*)getNext(n)); \
	 } \
	const N *next()	const \
	 { return ((N*)next_); \
	 } \
	const N *next(int n) const \
	 { return ((N*)getNext(n)); \
	 }

template<class T>
class RxSnode : public RxSnodeBase
{
public:
	T val;
	RX_SNODE_DECLS(RxSnode);	
	RX_FAST_ALLOC_DECLS_L(RxSnode,next_);
};

template<class N>
class RxSlist : public RxSlistBase
{
    protected:
	virtual void freeNode(void *ptr)
	 { delete (N*)ptr;
	 };
	int testNode(const RxSnodeBase *node, const void *ptr)
	 { return (((N*)node)->compare(ptr));
	 };
    public:
	RxSlist() : RxSlistBase()
	 {
	 }
#if 1
	~RxSlist()
	 { if (autoDelete) removeAndDeleteAll();
	 };
#endif
	N* head()
	 { return (N*)head_;
	 };
	N* tail()
	 { return (N*)tail_;
	 };
	const N* head() const
	 { return (N*)head_;
	 };
	const N* tail() const
	 { return (N*)tail_;
	 };
	N* findByPtr(const void *ptr)
	 { return (N*)findNodeByPtr(ptr);
	 };
	N* findByName(const char *ptr)
	 { return (N*)findNodeByName(ptr);
	 };
};
	
#endif
