#ifndef _RXSPATIALACCEL_H_
#define _RXSPATIALACCEL_H_

#include <RxSpatialVec.h>

class RxSpatialAccel: public RxSpatialVec
{
public:
	RxVecd3 &v()
	 { return b_;
	 }
	const RxVecd3 &v() const
	 { return b_;
	 }
	RxVecd3 &w()
	 { return a_;
	 }
	const RxVecd3 &w() const
	 { return a_;
	 }

	// assignment

	RxSpatialAccel& set (double vx, double vy, double vz,
		      double wx, double wy, double wz);
	RxSpatialAccel& set (const double *dp);
	void get (double *vxp, double *vyp, double *vzp,
		  double *wxp, double *wyp, double *wzp);
	void get (double *dp) const;

	// constructors

	RxSpatialAccel () : RxSpatialVec()
	 {
	 }
	RxSpatialAccel (double vx, double vy, double vz, 
		 double wx, double wy, double wz) : 
		RxSpatialVec()
	 { set (vx, vy, vz, wx, wy, wz);
	 }
	RxSpatialAccel (RxVecd3 &vv, RxVecd3 &ww) : RxSpatialVec()
	 { w() = ww;
	   v() = vv;
	 }
	RxSpatialAccel (const double *dp) : RxSpatialVec()
	 { set (dp);
	 }
#if 0
	RxSpatialAccel (const RxSpatialAccel& a2) : RxSpatialVec(a2)
	 {
	 }
#endif
	// query and access

	double& operator()(int i);
	double operator()(int i) const;

	// T <- T,T operations

	RxSpatialAccel& add (const RxSpatialAccel &a1, const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::add (a1, a2);
	 }
	RxSpatialAccel& add (const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::add (a2);
	 }
	RxSpatialAccel& operator+=(const RxSpatialAccel &a2)
	 { return (add(a2));
	 }
	friend RxSpatialAccel operator+(const RxSpatialAccel &a1, const RxSpatialAccel &a2);

	RxSpatialAccel& sub (const RxSpatialAccel &a1, const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::sub (a1, a2);
	 }
	RxSpatialAccel& sub (const RxSpatialAccel &a1)
	 { return (RxSpatialAccel&) RxSpatialVec::sub (a1);
	 }
	RxSpatialAccel& operator-=(const RxSpatialAccel &a2)
	 { return (sub(a2));
	 }

	friend RxSpatialAccel operator-(const RxSpatialAccel &a1, const RxSpatialAccel &a2);

	RxSpatialAccel& extrap (const RxSpatialAccel &a1, double a, const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::extrap (a1, a, a2);
	 }
	RxSpatialAccel& extrap (double a, const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::extrap (a, a2);
	 }
	RxSpatialAccel& interp (const RxSpatialAccel &a1, double a, const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::interp (a1, a, a2);
	 }
	RxSpatialAccel& interp (double a, const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::interp (a, a2);
	 }
#if 0
	RxSpatialAccel& cross (const RxSpatialAccel &a1, const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::cross (a1, a2);
	 }
	RxSpatialAccel& cross (const RxSpatialAccel &a2)
	 { return (RxSpatialAccel&) RxSpatialVec::cross (a2);
	 }
#endif
	// T <- T,O operations

	RxSpatialAccel& scale (double a, const RxSpatialAccel &a1)
	 { return (RxSpatialAccel&) RxSpatialVec::scale (a, a1);
	 }
	RxSpatialAccel& scale (double a)
	 { return (RxSpatialAccel&) RxSpatialVec::scale (a);
	 }
	RxSpatialAccel& operator*=(double a)
	 { return (scale(a));
	 }
	RxSpatialAccel& operator/=(double a)
	 { return (scale(1/a));
	 }

	friend RxSpatialAccel operator*(const RxSpatialAccel &a1, double a);
	friend RxSpatialAccel operator*(double a, const RxSpatialAccel &a1);
	friend RxSpatialAccel operator/(const RxSpatialAccel &a1, double a);

	RxSpatialAccel& xform (const RxSpatialAccel &a1, const RxRot& R)
	 { return (RxSpatialAccel&) RxSpatialVec::xform (a1, R);
	 }
	RxSpatialAccel& xform (const RxRot& R)
	 { return (RxSpatialAccel&) RxSpatialVec::xform (R);
	 }
	RxSpatialAccel& Ixform (const RxSpatialAccel &a1, const RxRot& R)
	 { return (RxSpatialAccel&) RxSpatialVec::Ixform (a1, R);
	 }
	RxSpatialAccel& Ixform (const RxRot& R)
	 { return (RxSpatialAccel&) RxSpatialVec::Ixform (R);
	 }
	RxSpatialAccel& xform (const RxSpatialAccel &a1, const RxTrsf& T)
	 { return (RxSpatialAccel&) RxSpatialVec::xform (a1, T);
	 }
	RxSpatialAccel& xform (const RxTrsf& T)
	 { return (RxSpatialAccel&) RxSpatialVec::xform (T);
	 }
	RxSpatialAccel& Ixform (const RxSpatialAccel &a1, const RxTrsf& T)
	 { return (RxSpatialAccel&) RxSpatialVec::Ixform (a1, T);
	 }
	RxSpatialAccel& Ixform (const RxTrsf& T)
	 { return (RxSpatialAccel&) RxSpatialVec::Ixform (T);
	 }

	// O <- T,T operations

	// T <- T operations

	RxSpatialAccel &unit(const RxSpatialAccel &a1)
	 { return (RxSpatialAccel&) RxSpatialVec::unit(a1);
	 }
	RxSpatialAccel &unit()
	 { return (RxSpatialAccel&) RxSpatialVec::unit();
	 }
	RxSpatialAccel &neg(const RxSpatialAccel &a1)
	 { return (RxSpatialAccel&) RxSpatialVec::neg(a1);
	 }
	RxSpatialAccel &neg()
	 { return (RxSpatialAccel&) RxSpatialVec::neg();
	 }
	friend RxSpatialAccel operator-(const RxSpatialAccel &a1);

	// T <- O operations

	RxSpatialAccel &setZero()
	 { return (RxSpatialAccel&) RxSpatialVec::setZero();
	 }

	// O <- T operations

	double norm1(double cv, double cw) const
	 { return RxSpatialVec::norm1 (cw, cv);
	 }
	double norm2(double cv, double cw) const
	 { return RxSpatialVec::norm2 (cw, cv);
	 }
	double normInf(double cv, double cw) const
	 { return RxSpatialVec::normInf (cw, cv);
	 }

	// constants

	static const RxSpatialAccel ZERO;
	static const RxSpatialAccel UNIT_VX;
	static const RxSpatialAccel UNIT_VY;
	static const RxSpatialAccel UNIT_VZ;
	static const RxSpatialAccel UNIT_WX;
	static const RxSpatialAccel UNIT_WY;
	static const RxSpatialAccel UNIT_WZ;
};

inline RxSpatialAccel operator+(const RxSpatialAccel &a1, const RxSpatialAccel &a2)
{
	RxSpatialAccel vr;
	((RxSpatialVec*)&vr)->add (*(RxSpatialVec*)&a1, *(RxSpatialVec*)&a2);
	return vr;
}

inline RxSpatialAccel operator-(const RxSpatialAccel &a1, const RxSpatialAccel &a2)
{
	RxSpatialAccel vr;
	((RxSpatialVec*)&vr)->sub (*(RxSpatialVec*)&a1, *(RxSpatialVec*)&a2);
	return vr;
}

inline RxSpatialAccel operator*(const RxSpatialAccel &a1, double a)
{
	RxSpatialAccel vr;
	((RxSpatialVec*)&vr)->scale (a, *(RxSpatialVec*)&a1);
	return vr;
}

inline RxSpatialAccel operator*(double a, const RxSpatialAccel &a1)
{
	RxSpatialAccel vr;
	((RxSpatialVec*)&vr)->scale (a, *(RxSpatialVec*)&a1);
	return vr;
}

inline RxSpatialAccel operator/(const RxSpatialAccel &a1, double a)
{
	RxSpatialAccel vr;
	((RxSpatialVec*)&vr)->scale (1/a, *(RxSpatialVec*)&a1);
	return vr;
}

inline RxSpatialAccel operator-(const RxSpatialAccel &a1)
{
	RxSpatialAccel vr;
	((RxSpatialVec*)&vr)->neg (*(RxSpatialVec*)&a1);
	return vr;
}

#endif


