#ifndef _RXSPATIALMAT_H_WRAP_
#define _RXSPATIALMAT_H_WRAP_

#include <RxMatdBase.h>
#include <RxMatd3x3.h>
#include <RxSpatialVec.h>

class RxSpatialMat: public RxMatdBase
{
    protected:
	RxMatd3x3 A_;
	RxMatd3x3 B_;
	RxMatd3x3 C_;
	RxMatd3x3 D_;

	static const int width;
	static const int nr;
	static const int nc;

	void doCopy (const RxSpatialMat &m1);

    public:

	RxMatd3x3 &A()
	 { return (A_);
	 }
	const RxMatd3x3 &A() const
	 { return (A_);
	 }
	RxMatd3x3 &B()
	 { return (B_);
	 }
	const RxMatd3x3 &B() const
	 { return (B_);
	 }
	RxMatd3x3 &C()
	 { return (C_);
	 }
	const RxMatd3x3 &C() const
	 { return (C_);
	 }
	RxMatd3x3 &D()
	 { return (D_);
	 }
	const RxMatd3x3 &D() const
	 { return (D_);
	 }

	// assignment

	RxSpatialMat& set(const RxSpatialMat &m2);
	RxSpatialMat& operator=(const RxSpatialMat &m2)
	 { return (set(m2));
	 }
	RxSpatialMat& set (const double *dp);
	void get (double *dp) const;
	RxSpatialMat& set (const RxMatd3x3 &A, const RxMatd3x3 &B,
			   const RxMatd3x3 &C, const RxMatd3x3 &D)
	 { A_ = A;
	   B_ = B;
	   C_ = C;
	   D_ = D;
	 }
	RxSpatialMat& set (const RxMatdBase &m1);
	RxSpatialMat& set (const RxMatd &m1);

	// constructors

	RxSpatialMat ()
	 {
	 }
	RxSpatialMat (const RxMatd3x3 &A, const RxMatd3x3 &B,
		      const RxMatd3x3 &C, const RxMatd3x3 &D) 
	 { set (A, B, C, D);
	 }
	RxSpatialMat (const double *dp)
	 { set (dp);
	 }

	// query and access

	int nrows() const
	 { return nr;
	 }
	int ncols() const
	 { return nc;
	 }
	double& operator()(int i, int j);
	double operator()(int i, int j) const;
	
	// T <- T,T operations

	RxSpatialMat& add (const RxSpatialMat &m1, const RxSpatialMat &m2);
	RxSpatialMat& add (const RxSpatialMat &m2)
	 { return add (*this, m2);
	 }
	RxSpatialMat& operator+=(const RxSpatialMat &m2)
	 { return (add(m2));
	 }
	friend RxSpatialMat operator+(const RxSpatialMat &m1, 
				      const RxSpatialMat &m2);

	RxSpatialMat& sub (const RxSpatialMat &m1, const RxSpatialMat &m2);
	RxSpatialMat& sub (const RxSpatialMat &m2)
	 { return sub(*this, m2);
	 }
	RxSpatialMat& operator-=(const RxSpatialMat &m2)
	 { return (sub(m2));
	 }
	friend RxSpatialMat operator-(const RxSpatialMat &m1, const RxSpatialMat &m2);

	RxSpatialMat& mul (const RxSpatialMat &m1, const RxSpatialMat &m2);
	RxSpatialMat& mul (const RxSpatialMat &m2)
	 { return (mul (*this, m2));
	 }

	RxSpatialMat& operator*= (const RxSpatialMat &m2)
	 { return (mul (*this, m2));
	 }
	friend RxSpatialMat operator*(const RxSpatialMat &m1, const RxSpatialMat &m2);

	// T <- T,O operations

	RxSpatialMat& scale (double a, const RxSpatialMat &m1);
	RxSpatialMat& scale (double a)
	 { return scale (a, *this);
	 }
	RxSpatialMat& operator*=(double a)
	 { return (scale(a));
	 }
	RxSpatialMat& operator/=(double a)
	 { return (scale(1/a));
	 }
	friend RxSpatialMat operator*(const RxSpatialMat &m1, double a);
	friend RxSpatialMat operator*(double a, const RxSpatialMat &m1);
	friend RxSpatialMat operator/(const RxSpatialMat &m1, double a);

	RxSpatialMat& xform (const RxSpatialMat &m1, const RxRot& R);
	RxSpatialMat& xform (const RxRot& R)
	 { return (xform (*this, R));
	 }
	RxSpatialMat& Ixform (const RxSpatialMat &m1, const RxRot& R);
	RxSpatialMat& Ixform (const RxRot& R)
	 { return (Ixform (*this, R));
	 }
	RxSpatialMat& xform (const RxSpatialMat &m1, const RxTrsf& T);
	RxSpatialMat& xform (const RxTrsf& T)
	 { return (xform (*this, T));
	 }
	RxSpatialMat& Ixform (const RxSpatialMat &m1, const RxTrsf& T);
	RxSpatialMat& Ixform (const RxTrsf& T)
	 { return (Ixform (*this, T));
	 }

	// O <- T,O operations

	RxSpatialVec &Vmul (RxSpatialVec &vr, const RxSpatialVec &v1) const;
	RxSpatialVec &mulV (RxSpatialVec &vr, const RxSpatialVec &v1) const;
	friend RxSpatialVec operator*(const RxSpatialMat &m1, 
				      const RxSpatialVec &v1);
	friend RxSpatialVec operator*(const RxSpatialVec &v1, 
				      const RxSpatialMat &m1);
	   
	// T <- T operations

	RxSpatialMat &stp(const RxSpatialMat &m1);
	RxSpatialMat &stp();
	RxSpatialMat S()
	 { RxSpatialMat mr;
	   return (mr.stp(*this));
	 }

	RxSpatialMat &inv(const RxSpatialMat &m1);
	RxSpatialMat &inv()
	 { return (inv (*this));
	 }
	RxSpatialMat I()
	 { RxSpatialMat mr;
	   return (mr.inv(*this));
	 }

	RxSpatialMat &neg(const RxSpatialMat &m1);
	RxSpatialMat &neg()
	 { return (neg (*this));
	 }
	RxSpatialMat operator-()
	 { RxSpatialMat mr;
	   return (mr.neg(*this));
	 }

	// T <- O operations

	RxSpatialMat& setZero();
	RxSpatialMat& setIdent();

	// O <- T operations

	double det() const;
	double norm1() const;
	double normInf() const;
	double normF() const;
	double trace() const;	
	double condest() const;	

	// constants

	static const RxSpatialMat IDENT;
	static const RxSpatialMat ZERO;
};

inline RxSpatialVec operator*(const RxSpatialMat &m1, const RxSpatialVec &v1)
{
	RxSpatialVec vr;
	return (m1.mulV(vr,v1));
}

inline RxSpatialVec operator*(const RxSpatialVec &v1, const RxSpatialMat &m1)
{
	RxSpatialVec vr;
	return (m1.Vmul(vr,v1));
}

#endif
