#ifndef _RXTRSF_H_WRAP_
#define _RXTRSF_H_WRAP_

#include <RxMatdBase.h>
#include <RxTrsfObj.h>
#include <RxVecd3.h>
#include <RxRot.h>
#include <mathx.h>
#include <stdarg.h>

class RxTrsf : public RxMatdBase, private RxTrsfObj
{
    public:
	static const int ID;

	RxRot R;
	RxVecd3 p;

	RxTrsf ()
	 : RxMatdBase ()
	 { type_ = ID;
	 };
	RxTrsf (const RxTrsf &m2);

	RxTrsf& operator=(const RxTrsf& m2);

	int nrows() const
	 { return 3;
	 };
	int ncols() const
	 { return 4;
	 }

	void mul (const RxTrsf &m1, const RxTrsf &m2);
	void mul (const RxTrsf &m2)
	 { mul (*this, m2);
	 }
	void mulI (const RxTrsf &m1, const RxTrsf &m2);
	void Imul (const RxTrsf &m1, const RxTrsf &m2);
	void ImulI (const RxTrsf &m1, const RxTrsf &m2);

#if 0
	RxVecd3 &mul (RxVecd3 &vr, const RxVecd3 &v1) const;
	RxVecd3 mul (const RxVecd3 &v1) const
	 { RxVecd3 vr;
	   return (mul (vr, v1));
	 }
	RxVecd3 &Imul (RxVecd3 &vr, const RxVecd3 &v1) const;
	RxVecd3 Imul (const RxVecd3 &v1) const
	 { RxVecd3 vr;
	   return (Imul (vr, v1));
	 }
#endif

	friend RxTrsf operator*(const RxTrsf &m1, const RxTrsf &m2);
	friend RxVecd3 operator*(const RxTrsf &m1, const RxVecd3 v2);

	void inv(const RxTrsf &m1);
	void setIdent();

	double& operator()(int i, int j);
	double operator()(int i, int j) const;

	RxTrsf inv()
	 { RxTrsf mr;
	   mr.inv(*this);
	   return (mr);
	 }

	RxTrsf& operator*=(const RxTrsf &m2)
	 { mul (*this, m2);
	   return (*this);
	 }
#if 0
	RxTrsf& operator-()
	 { this->doScale(-1.0); return (*this);
	 };
#endif

	RxTrsf &set (const RxRot &R_, const RxVecd3 &p_)
	 { R = R_;
	   p = p_;
	   return *this;
	 }
	RxTrsf (const RxRot &R_, const RxVecd3 &p_)
	 { set (R_, p_);
	 }
#if 0
	RxTrsf &setEul (double phi, double the, double psi);
	RxTrsf &setEulD (double phi, double the, double psi)
	 { return (setEul (DTOR*phi, DTOR*the, DTOR*psi));
	 }
	RxTrsf &setRpy (double roll, double pitch, double yaw);
	RxTrsf &setRpyD (double roll, double pitch, double yaw)
	 { return (setRpy (DTOR*roll, DTOR*pitch, DTOR*yaw));
	 }
	RxTrsf &setAxis (const RxVecd3 &v, double ang);
	RxTrsf &setAxisD (const RxVecd3 &v, double ang)
	 { return (setAxis (v, DTOR*ang));
	 }
	RxTrsf &setVecs (const char *axes,const RxVecd3 &v1,const RxVecd3 &v2);
	RxTrsf &setAngs (const char *axes, ...);
	RxTrsf &setAngsD (const char *axes, ...);

	void getEul (double *phip, double *thep, double *psip) const;
	void getEulD (double *phip, double *thep, double *psip) const
	 { getEul (phip, thep, psip);
	   *phip *= RTOD; *thep *= RTOD; *psip *= RTOD;
	 }
	   
	void getRpy (double *rollp, double *pitchp, double *yawp) const;
	void getRpyD (double *rollp, double *pitchp, double *yawp) const
	 { getRpy (rollp, pitchp, yawp);
	   *rollp *= RTOD; *pitchp *= RTOD; *yawp *= RTOD;
	 }
	void getAxis (RxVecd3 &v, double *angp) const;
	void getAxisD (RxVecd3 &v, double *angp) const
	 { getAxis (v, angp);
	   *angp *= RTOD;
	 }
#endif
#if 0
	RxVecd3 &xform (RxVecd3 &vr, const RxVecd3 &v1) const
	 { return (Imul (vr, v1));
	 }
	RxVecd3 xform (const RxVecd3 &v1) const
	 { RxVecd3 vr;
	   return (xform (vr, v1));
	 }
	RxTwist &xform (RxTwist &tr, const RxTwist &t1) const;
	RxTwist xform (const RxTwist &t1) const
	 { RxTwist tr;
	   return (xform (tr, t1));
	 }
	RxWrench &xform (RxWrench &wr, const RxWrench &w1) const;
	RxWrench xform (const RxWrench &w1) const
	 { RxWrench wr;
	   return (xform (wr, w1));
	 }

	RxVecd3 &Ixform (RxVecd3 &vr, const RxVecd3 &v1) const
	 { return (mul (vr, v1));
	 }
	RxVecd3 Ixform (const RxVecd3 &v1) const
	 { RxVecd3 vr;
	   return (Ixform (vr, v1));
	 }
	RxTwist &Ixform (RxTwist &tr, const RxTwist &t1) const;
	RxTwist Ixform (const RxTwist &t1) const
	 { RxTwist tr;
	   return (Ixform (tr, t1));
	 }
	RxWrench &Ixform (RxWrench &wr, const RxWrench &w1) const;
	RxWrench Ixform (const RxWrench &w1) const
	 { RxWrench wr;
	   return (Ixform (wr, w1));
	 }
#endif

	virtual int oscan (class RxParseObj *obj);
	virtual int cprint (class RxCoutput *co, class RxPrintSpec *fs) const;

	// constants

	static const RxTrsf IDENT;	
};

inline RxTrsf operator*(const RxTrsf &m1, const RxTrsf &m2)
{
	RxTrsf mr;
	mr.mul (m1, m2);
	return (mr);
}

#if 0
inline RxVecd3 operator*(const RxTrsf &m1, const RxVecd3 &v2)
{
	RxVecd3 vr;
	return (m1.mul(vr, v2));
}
#endif

#endif
