#ifndef _RXVECDN7_H_WRAP_
#define _RXVECDN7_H_WRAP_

#include <RxVecd.h>

class RxVecd7: public RxVecd
{
	double vbuf[7];

    public:

	int length() const
	 { return 7;
	 }
	double& operator()(int i)
	 { return vbuf[i];
	 } 
	double operator()(int i) const
	 { return vbuf[i];
	 } 

	// assignment

	RxVecd7& operator=(const RxVecd7 &v2)
	 { set (v2.v);
	   return (*this);
	 }
	RxVecd7& set (const double *dp);
	void get (double *dp);

	// constructors	

	RxVecd7 () : RxVecd (7, vbuf) 
	 { 
	 }
	RxVecd7 (const RxVecd7 &v2) : RxVecd (7, vbuf)
	 { set (v2.v);
	 }
	RxVecd7 (double *dp) : RxVecd (7, vbuf)
	 { set (dp);
	 }

	// T <- T,T operations

	RxVecd7& add (const RxVecd7 &v1, const RxVecd7 &v2);
	RxVecd7& add (const RxVecd7 &v2);

	RxVecd7& operator+=(const RxVecd7 &v2)
	 { return (add (v2));
	 }
	friend RxVecd7 operator+(const RxVecd7 &v1, const RxVecd7 &v2);

	RxVecd7& sub (const RxVecd7 &v1, const RxVecd7 &v2);
	RxVecd7& sub (const RxVecd7 &v2);
	RxVecd7& operator-=(const RxVecd7 &v2)
	 { return (sub (v2));
	 }
	friend RxVecd7 operator-(const RxVecd7 &v1, const RxVecd7 &v2);

	RxVecd7& extrap (const RxVecd7 &v1, double a, const RxVecd7 &v2);
	RxVecd7& extrap (double a, const RxVecd7 &v2)
	 { return (extrap (*this, a, v2));
	 }

	// T <- T,O operations

	RxVecd7& scale (double a, const RxVecd7 &v1);
	RxVecd7& scale (double a);
	RxVecd7& operator*=(double a)
	 { return (scale (a));
	 }
	RxVecd7& operator/=(double a)
	 { return (scale (1/a));
	 }
	friend RxVecd7 operator*(const RxVecd7 &v1, double a);
	friend RxVecd7 operator*(double a, const RxVecd7 &v1);
	friend RxVecd7 operator/(const RxVecd7 &v1, double a);

	// O <- T,T operations

	double dot(const RxVecd7 &v2) const;

	// T <- T operations

	RxVecd7 &unit(const RxVecd7 &v1);
	RxVecd7 &unit()
	 { return (unit(*this));
	 }
	RxVecd7 &neg(const RxVecd7 &v1);
	RxVecd7 &neg()
	 { return (neg (*this));
	 }
	friend RxVecd7 operator-(const RxVecd7 &v1);

	// T <- O operations

	RxVecd7 &setZero();

	// O <- T operations

	double norm1() const;
	double norm2() const;
	double normInf() const;
};

inline RxVecd7 operator+(const RxVecd7 &v1, const RxVecd7 &v2)
{
	RxVecd7 vr;
	return (vr.add (v1, v2));
}

inline RxVecd7 operator-(const RxVecd7 &v1, const RxVecd7 &v2)
{
	RxVecd7 vr;
	return (vr.sub (v1, v2));
}

inline RxVecd7 operator*(const RxVecd7 &v1, double a)
{
	RxVecd7 vr;
	return (vr.scale (a, v1));
}

inline RxVecd7 operator*(double a, const RxVecd7 &v1)
{
	RxVecd7 vr;
	return (vr.scale (a, v1));
}

inline RxVecd7 operator/(const RxVecd7 &v1, double a)
{
	RxVecd7 vr;
	return (vr.scale (1/a, v1));
}

inline RxVecd7 operator-(const RxVecd7 &v1)
{
	RxVecd7 vr;
	return (vr.neg (v1));
}

#endif
