#ifndef _RXVECDSPLINE_H_WRAP_
#define _RXVECDSPLINE_H_WRAP_

#include <RxVecd.h>
#include <RxDlist.h>

class RxVecdSknot : public RxDnodeBase
{
	friend class RxVecdSpline;

	RxVecd *v;
	double u;
	double w;
	int m;
	int idx;
	RX_DNODE_DECLS (RxVecdSknot);
};

typedef RxDlist<RxVecdSknot> RxVecdSknotList;

class RxVecdSpline
{

    private:
	int degree;
	int rational;

	RxVecdSknotList knots;
	int vlen;
	int headIdx;

	void deBoor (RxVecd *v, RxVecd *dv, RxVecd *ddv, 
		     RxVecdSknot *k0, RxVecdSknot *k1, double u, int k);
	RxVecdSknot *findLeftKnot (double u, RxVecdSknot *start=NULL);
	RxVecdSknot *findRightKnot (double u, RxVecdSknot *start=NULL);
	RxVecdSknot *lastLeftKnot;
	RxVecdSknot *curPnt;

    public:
	RxVecdSpline (int len, int deg, int rat);

	void reset ();
	void reset (int deg, int rat);

	void addPnt (RxVecd *v, double u);
	void addPnt (RxVecd *v, double w, double u);
	void removePnts (double u);
#define RX_SPLINE_RIGHT	1
#define RX_SPLINE_LEFT	2
	int calcPnt (RxVecd *v, double u);
	int calcPnt (RxVecd *v, RxVecd *dv, RxVecd *ddv, 
		double u, int side=RX_SPLINE_RIGHT);

	int getDegree()
	 { return (degree);
	 }
	int getVlength()
	 { return (vlen);
	 }
	int isRational()
	 { return (rational ? 1 : 0);
	 }
	int numKnots();
	RxVecd *firstPnt (double *w=NULL, double *u=NULL, int *mp=NULL);
	RxVecd *nextPnt (double *w=NULL, double *u=NULL, int *mp=NULL);
};

#endif
