/* Description of functions:

	ad = createArgDef (var)
	ad = argDefAdd (ad, var)

	n = argDefMatch (ad, argc, argv)

	printArgs (fp, ad)

	%i,<low>,<high>,addr
	%x,<low>,<high>,addr
	%o,<low>,<high>,addr
	%f,<low>,<high>,addr
	%s

	foo%V,<value>,addr
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#ifdef HAS_STDARG
#include <stdarg.h>
#define VA_ARGDEF(x) x
#else
#include <varargs.h>
#define VA_ARGDEF(x) (va_alist) va_dcl
#endif
#include <errno.h>

#include "argDef.h"
#include "getWord.h"

#ifndef NULL
#define NULL 0
#endif
#define MIN(x,y)	((x)<(y) ? (x) : (y))

#define SKIP_WHITE_SPACE(p)	while(isspace(*(p))) p++;

static char ebuf[256];

static char *copystr(str)
char *str;
{
	return (strcpy (malloc(strlen(str)+1), str));
}

static addDefsError(ad, str)
ArgDef *ad;
{
	sprintf (ebuf, "argDefAdd(), option %s: %s", ad->name, str);
}

static parseNumericOptions (ad, str, npp)
ArgDef *ad;
char *str;
char **npp;
{
	char *cp, *np;
	int openLimit;

	cp = str;
	SKIP_WHITE_SPACE(cp);
	if (*cp != ',')
	 { addDefsError (ad, "missing comma after format");
	   return (-1);
	 }
	cp++;	/* skip the comma */
	SKIP_WHITE_SPACE(cp);
	if (*cp != ',' && *cp != '\0')
	 { if (*cp != '(' && *cp != '[')
	    { addDefsError (ad, "range spec must begin with '[' or '('");
	      return (-1);
	    }
	   openLimit = (*cp == '(');
	   cp++;	/* skip opening delimiter */
	   SKIP_WHITE_SPACE(cp);
	   if (*cp != ',')
	    { if (ad->type == ADEF_FLOAT)
	       { ad->u.f.low = strtod (cp, &np);
	       }
	      else if (ad->type == ADEF_INT)
	       { ad->u.i.low = strtol (cp, &np, 0);
	       }
	      else
	       { ad->u.u.low = strtoul (cp, &np, 0);
	       }
	      if (np == cp)
	       { addDefsError (ad, "malformed low range spec");
		 return (-1);
	       }
	      cp = np;
	      SKIP_WHITE_SPACE(cp);
	      if (*cp != ',')
	       { addDefsError (ad, "missing comma after low range spec");
		 return (-1);
	       }
	      ad->flags |= ADEF_LOW_RANGE;
	      if (openLimit)
	       { ad->flags |= ADEF_LOW_OPEN;
	       }
	    }
	   cp++;	/* skip the comma */
	   SKIP_WHITE_SPACE(cp);
	   if (*cp != ')' && *cp != ']')
	    { if (ad->type == ADEF_FLOAT)
	       { ad->u.f.high = strtod (cp, &np);
	       }
	      else if (ad->type == ADEF_INT)
	       { ad->u.i.high = strtol (cp, &np, 0);
	       }
	      else
	       { ad->u.u.high = strtoul (cp, &np, 0);
	       }
	      if (np == cp)
	       { addDefsError (ad, "malformed high range spec");
		 return (-1);
	       }
	      cp = np;
	      ad->flags |= ADEF_HIGH_RANGE;
	      if (*cp == ')')
	       { ad->flags |= ADEF_HIGH_OPEN;
	       }
	    }
	   if (*cp != ')' && *cp != ']')
	    { addDefsError (ad, "range spec must end with ']' or ')'");
	      return (-1);
	    }
	   cp++;
	 }

	if (npp)
	 { *npp = cp;
	 }
	return (1);
}

static parseValOptions (ad, str, npp)
ArgDef *ad;
char *str;
char **npp;
{
	char *cp, *np;

	cp = str;
	SKIP_WHITE_SPACE(cp);
	if (*cp != ',')
	 { addDefsError (ad, "missing comma after format");
	   return (-1);
	 }
	cp++;
	SKIP_WHITE_SPACE(cp);
	ad->u.v.val = strtol(cp, &np, 0);
	if (cp == np)
	 { addDefsError (ad, "missing or malformed value for 'v' format");
	   return (-1);
	 }
	cp = np;
	if (npp)
	 { *npp = cp;
	 }
	return (1);
}

static parseCommentStr (ad, str, npp)
ArgDef *ad;
char *str;
char **npp;
{
	char *cp, *sp;
	char *cstr;

	cp = str;
	SKIP_WHITE_SPACE(cp);
	if (*cp != '"')
	 { return (-1);
	 }
	cp++;	/* skip the leading quote */
	cstr = (char*)malloc(strlen(cp)+1); /* longer than necessary, but OK */
	sp = cstr;
	/* Pass escaped quotes, and continue until we hit a 
	   closing (non-escaped) quote */
	while (*cp != '"' && *cp != '\0')
	 { if (cp[0] == '\\')
	    { if (cp[1] == '"')
	       { *sp++ = '"';
		 cp += 2;
	       }
	      else if (cp[1] == '\\')
	       { *sp++ = '\\';
		 cp += 2;
	       }
	      else if (cp[1] == 'n')
	       { *sp++ = '\n';
		 cp += 2;
	       }
	      else if (cp[1] == 't')
	       { *sp++ = '\t';
		 cp += 2;
	       }
	      else if (cp[1] == 'f')
	       { *sp++ = '\f';
		 cp += 2;
	       }
	      else if (cp[1] == 'r')
	       { *sp++ = '\r';
		 cp += 2;
	       }
	      else if (cp[1] == 'b')
	       { *sp++ = '\b';
		 cp += 2;
	       }
	      else if (isdigit(cp[1]))
	       { int v, i;
		 v = cp[1] - '0';
		 cp += 2;
		 for (i=0; isdigit(*cp) && i<2; i++)
		  { v = v*8 + *cp++ - '0';
		  }
		 *sp++ = MIN(v,255);
	       }
	    }
	   else
	    { *sp++ = *cp++;
	    }
	 }	   

	if (*cp != '"') /* verify the closing quote */
	 { free (cstr);
	   return (-1);
	 }
	*sp = '\0';	/* terminate the new string */
	ad->comment = cstr;
	if (npp)
	 { *npp = cp+1;	/* skip closing quote */
	 }
	return (1);
}

static parseArgDef (ad, str, list)
ArgDef *ad;
char *str;
ArgDefList *list;
{
	char *cp, *op, *np;
	ArgDef *ap;
	
	ad->flags = 0;
	cp = str;
	SKIP_WHITE_SPACE(cp);
	if (*cp == '%')
	 { sprintf (ebuf, "argDefAdd(): no option name given");
	   return (-1);
	 }
	op = cp;
	while (!isspace(*cp) && *cp != '%')
	 { cp++;
	 }
	ad->name = (char*)malloc(cp-op+1);
	strncpy (ad->name, op, cp-op);
	ad->name[cp-op] = '\0';

	for (ap=list->head; ap; ap=ap->next)
	 { if (strcmp(ap->name, ad->name) == 0)
	    { addDefsError (ad, "name already in use");
	      return (-1);
	    }
	 }

	if (isspace(*cp))
	 { SKIP_WHITE_SPACE(cp);
	   if (*cp != '%')
	    { addDefsError (ad, "no format spec");
	      return (-1);
	    }
	 }
	else
	 { ad->flags |= ADEF_ONE_WORD;
	 }

	cp++;	/* skip the '%' */
	if (isdigit (*cp))
	 { ad->num = 0;
	   do
	    { ad->num = ad->num*10 + (*cp++ -'0');
	    }
	   while (isdigit (*cp));
	   if (ad->num == 0)
	    { addDefsError (ad, "bad format spec");
	      return (-1);
	    }
	   if (ad->num > 1)
	    { ad->flags &= ~ADEF_ONE_WORD;
	    }
	 }
	else
	 { ad->num = 1;
	 }
	if (*cp == 'l')
	 { cp++;	/* skip the 'l' */
	   ad->flags |= ADEF_LONG;
	 }
	switch (*cp)
	 { default:
	    { addDefsError (ad, "bad format spec");
	      return (-1);
	      break;
	    }
	   case 'f':
	    { ad->type = ADEF_FLOAT;
	      if (parseNumericOptions (ad, &cp[1], &np) < 0)
	       { return (-1);
	       }
	      break;
	    }
	   case 'i':
	    { ad->type = ADEF_INT;
	      if (parseNumericOptions (ad, &cp[1], &np) < 0)
	       { return (-1);
	       }
	      break;
	    }
	   case 'x':
	    { ad->type = ADEF_HEX;
	      if (parseNumericOptions (ad, &cp[1], &np) < 0)
	       { return (-1);
	       }
	      break;
	    }
	   case 'o':
	    { ad->type = ADEF_OCTAL;
	      if (parseNumericOptions (ad, &cp[1], &np) < 0)
	       { return (-1);
	       }
	      break;
	    }
	   case 's':
	    { ad->type = ADEF_STRING;
	      np = &cp[1];
	      break;
	    }
	   case 'v':
	    { ad->type = ADEF_VALUE;
	      if (parseValOptions (ad, &cp[1], &np) < 0)
	       { return (-1);
	       }
	      break;
	    }
	 }
	if (ad->type != ADEF_VALUE)
	 { ad->flags |= ADEF_NEEDS_ARG;
	 }
	cp = np;
	SKIP_WHITE_SPACE(cp);
	if (*cp == ',')
	 { if (parseCommentStr (ad, &cp[1], &np) < 0)
	    { addDefsError (ad, "malformed comment string");
	      return (-1);
	    }
	   cp = np;
	   SKIP_WHITE_SPACE(cp);
	 }
	if (*cp != '\0')
	 { addDefsError (ad, "garbage at end of spec");
	   return (-1);
	 }
	return (1);
}

static cmparg (argstr, name)
char *argstr, *name;
{
	char *cps, *cpn;

	for (cps=argstr, cpn=name; *cpn!='\0'; )
	 { if (*cps++ != *cpn++)
	    { return (1);
	    }
	 }
	return (0);
}

static char *typeName(ad)
ArgDef *ad;
{
	switch (ad->type)
	 { case ADEF_FLOAT:
	    { return ("float");
	      break;
	    }
	   case ADEF_INT:
	    { return ("int");
	      break;
	    }
	   case ADEF_OCTAL:
	    { return ("octal");
	      break;
	    }
	   case ADEF_HEX:
	    { return ("hex");
	      break;
	    }
	   case ADEF_STRING:
	    { return ("string");
	      break;
	    }
	   case ADEF_VALUE:
	    { return ("value");
	      break;
	    }
	   default:
	    { return ("");
	      break;
	    }
	 }
}

static intBoundToStr (str, ad, val)
char *str;
ArgDef *ad;
int val;
{
	switch (ad->type)
	 { default:
	    { *str = '\0';
	    }
	   case ADEF_INT:
	    { sprintf (str, "%d", val);
	      break;
	    }
	   case ADEF_OCTAL:
	    { if (val == 0)
	       { strcpy (str, "0");
	       }
	      else
	       { sprintf (str, "0%o", (unsigned)val);
	       }
	      break;
	    }
	   case ADEF_HEX:
	    { sprintf (str, "0x%x", (unsigned)val);
	      break;
	    }
	 }
}

static char *highBoundToStr (str, ad)
char *str;
ArgDef *ad;
{
	if (ad->type == ADEF_FLOAT)
	 { sprintf (str, "%g", ad->u.f.high);
	 }
	else
	 { intBoundToStr (str, ad, ad->u.i.high);
	 }
	return (str);
}

static char *lowBoundToStr (str, ad)
char *str;
ArgDef *ad;
{
	if (ad->type == ADEF_FLOAT)
	 { sprintf (str, "%g", ad->u.f.low);
	 }
	else
	 { intBoundToStr (str, ad, ad->u.i.low);
	 }
	return (str);
}

static lowBoundErr (ad)
ArgDef *ad;
{
	char str[64];

	sprintf (ebuf, "Option %s: associated value must be %s %s", 
			ad->name, (ad->flags & ADEF_LOW_OPEN) ? ">" : ">=",
			lowBoundToStr(str,ad));
	return (-1);
}

static highBoundErr (ad)
ArgDef *ad;
{
	char str[64];

	sprintf (ebuf, "Option %s: associated value must be %s %s", 
			ad->name, (ad->flags & ADEF_HIGH_OPEN) ? "<" : "<=",
			highBoundToStr(str,ad));
	return (-1);
}

static floatRangeCheck (ad, fval)
ArgDef *ad;
double fval;
{
	if (ad->flags & ADEF_LOW_RANGE)
	 { if ((ad->flags & ADEF_LOW_OPEN) ? (fval <= ad->u.f.low) : 
					     (fval < ad->u.f.low))
	    { return (lowBoundErr(ad));
	    }
	 }
	if (ad->flags & ADEF_HIGH_RANGE)
	 { if ((ad->flags & ADEF_HIGH_OPEN) ? (fval >= ad->u.f.high) : 
					      (fval > ad->u.f.high))
	    { return (highBoundErr(ad));
	    }
	 }
	return (1);
}

static intRangeCheck (ad, ival)
ArgDef *ad;
int ival;
{
	if (ad->flags & ADEF_LOW_RANGE)
	 { if ((ad->flags & ADEF_LOW_OPEN) ? (ival <= ad->u.i.low) : 
					     (ival < ad->u.i.low))
	    { return (lowBoundErr(ad));
	    }
	 }
	if (ad->flags & ADEF_HIGH_RANGE)
	 { if ((ad->flags & ADEF_HIGH_OPEN) ? (ival >= ad->u.i.high) : 
					      (ival > ad->u.i.high))
	    { return (highBoundErr(ad));
	    }
	 }
	return (1);
}

static unsignedRangeCheck (ad, uval)
ArgDef *ad;
unsigned int uval;
{
	if (ad->flags & ADEF_LOW_RANGE)
	 { if ((ad->flags & ADEF_LOW_OPEN) ? (uval <= ad->u.u.low) : 
					     (uval < ad->u.u.low))
	    { return (lowBoundErr(ad));
	    }
	 }
	if (ad->flags & ADEF_HIGH_RANGE)
	 { if ((ad->flags & ADEF_HIGH_OPEN) ? (uval >= ad->u.u.high) : 
					      (uval > ad->u.u.high))
	    { return (highBoundErr(ad));
	    }
	 }
	return (1);
}

ArgDefList *argDefCreate ()
{
	ArgDefList *list;

	list = (ArgDefList*)malloc(sizeof(ArgDefList));
	list->head = list->tail = NULL;
	return (list);
}


int argDefAdd VA_ARGDEF((ArgDefList* list, ...))
{
	va_list pvar;
	char *str;
	ArgDef *ad;
#if !defined(HAS_STDARG)
	ArgDefList *list;

	va_start(pvar);
	list = va_arg(pvar, ArgDefList*);
#else
	va_start(pvar, list);
#endif
	for (str=va_arg(pvar,char*); str; str=va_arg(pvar,char*))
	 { ad = (ArgDef*)calloc(1,sizeof(ArgDef));
	   if (parseArgDef (ad, str, list) < 0)
	    { if (ad->name)
	       { free (ad->name);
	       }
	      free (ad);
	      return (-1);
	    }
	   if (list->head)
	    { list->tail->next = ad;
	      list->tail = ad;
	    }
	   else	/* this is the first item on the list */
	    { list->head = ad;
	      list->tail = ad;
	    }
	   ad->next = NULL;
	   switch (ad->type)
	    { case ADEF_INT:
	      case ADEF_VALUE:
	       { ad->addr = va_arg(pvar, int*);
		 break;
	       }
	      case ADEF_HEX:
	      case ADEF_OCTAL:
	       { ad->addr = va_arg(pvar, unsigned int*);
		 break;
	       }
	      case ADEF_FLOAT:
	       { if (ad->flags & ADEF_LONG)
		  { ad->addr = va_arg(pvar, double*);
		  }
		 else
		  { ad->addr = va_arg(pvar, float*);
		  }
		 break;
	       }
	      case ADEF_STRING:
	       { ad->addr = va_arg(pvar, char**);
		 break;
	       }
	    }
	 }
	va_end(pvar);
	return (1);
}

argDefMatch (list, argc, argv)
ArgDefList *list;
int argc;
char **argv;
{
	double fval;
	int ival;
	unsigned int uval;

	ArgDef *ad;
	char *cp, *np;
	int rcnt;
	int k;

	for (ad=list->head; ad; ad=ad->next)
	 { if (ad->flags & ADEF_ONE_WORD)
	    { if (cmparg (*argv, ad->name) == 0)
	       { break;
	       }
	    }
	   else
	    { if (strcmp (*argv, ad->name) == 0)
	       { break;
	       }
	    }
	 }
	if (!ad)
	 { return (0);
	 }
	if (ad->flags & ADEF_NEEDS_ARG)
	 { if (!(ad->flags & ADEF_ONE_WORD))
	    { if (argc < ad->num+1)
	       { if (ad->num == 1)
		  { sprintf (ebuf,
"Option %s needs an additional argument", ad->name);
		  }
		 else
		  { sprintf (ebuf,
"Option %s needs %d additional arguments", ad->name, ad->num);
		  }
		 return (-1);
	       }
	      cp = argv[1];
	      rcnt = ad->num+1;
	    }
	   else
	    { cp = &(*argv)[strlen(ad->name)];
	      rcnt = 1;
	    }
	 }
	else
	 { rcnt = 1;
	 }
	if (ad->type == ADEF_VALUE)
	 { *(int*)(ad->addr) = ad->u.v.val;
	   return (rcnt);
	 }
	for (k=0; k<ad->num; k++)
	 { 
	   if (k > 0)
	    { cp = argv[k+1];
	    }
	   switch (ad->type)
	    { case ADEF_FLOAT:
	       { fval = strtod(cp, &np);
		 break;
	       }
	      case ADEF_INT:
	       { ival = strtol(cp, &np, 0);
		 break;
	       }
	      case ADEF_HEX:
	       { uval = strtoul(cp, &np, 16);
		 break;
	       }
	      case ADEF_OCTAL:
	       { uval = strtoul(cp, &np, 8);
		 break;
	       }
	      case ADEF_STRING:
	       { *(char**)(ad->addr) = copystr(cp);
		 np = cp + strlen(cp);
		 break;
	       }
	    }
	   if (cp == np)
	    { sprintf (ebuf,
"Option %s: an associated argument %s is malformed or missing", ad->name, cp);
	      return (-1);
	    }
	   switch (ad->type)
	    { case ADEF_FLOAT:
	       { if (floatRangeCheck (ad, fval) < 0)
		  { return (-1);
		  }
		 if (ad->flags & ADEF_LONG)
		  { ((double*)(ad->addr))[k] = fval;
		  }
		 else
		  { ((float*)(ad->addr))[k] = fval;
		  }
		 break;
	       }
	      case ADEF_INT:
	       { if (intRangeCheck (ad, ival) < 0)
		  { return (-1);
		  }
		 ((int*)(ad->addr))[k] = ival;
		 break;
	       }
	      case ADEF_OCTAL:
	      case ADEF_HEX:
	       { if (unsignedRangeCheck (ad, uval) < 0)
		  { return (-1);
		  }
		 ((unsigned int*)(ad->addr))[k] = uval;
		 break;
	       }
	    }
	 }
	return (rcnt);
}

#define TAB_SPACES	8

argDefPrintUsage (fp, list, maxcols)
FILE *fp;
ArgDefList *list;
int maxcols;
{
	char buf[256];
	ArgDef *ad;
	int col, len;

	if (list->head)
	 { col = TAB_SPACES;
	   putc ('\t', fp);
	 }
	for (ad=list->head; ad; ad=ad->next)
	 { if (ad->type != ADEF_VALUE)
	    { if (ad->flags & ADEF_ONE_WORD)
	       { sprintf (buf, "[%s<%s>] ", ad->name, typeName(ad));
	       }
	      else if (ad->num == 1)
	       { sprintf (buf, "[%s <%s>] ", ad->name, typeName(ad));
	       }
	      else
	       { sprintf (buf, "[%s %d X <%s>] ", ad->name, ad->num, 
				typeName(ad));
	       }
	    }
	   else
	    { sprintf (buf, "[%s] ", ad->name);
	    }
	   len = strlen(buf);

	   /* maxcols == 0 => maxcols is specified.
	      (col+=len) > (maxcols-1) => we will spill over edge. we use
		      (maxcols-1) because if we go right to the edge
		      (maxcols), we get wrap new line inserted "for us".
	      ad != list->head means we print the first entry, no matter
		      how long it is. Subsequent entries are printed
		      full length anyway. */		     

	   if (maxcols > 0 && (col+=len) > (maxcols-1) && ad != list->head)
	    { col = TAB_SPACES+len;
	      putc ('\n', fp);
	      putc ('\t', fp);
	    }
	   fprintf (fp, "%s", buf);
	 }
	if (list->head)
	 { putc ('\n', fp);
	 }
}

argDefPrintHelp (fp, list)
FILE *fp;
ArgDefList *list;
{
	char lowStr[64];
	char highStr[64];
	ArgDef *ad;

	for (ad=list->head; ad; ad=ad->next)
	 { fprintf (fp, "%s", ad->name);
	   if (!(ad->flags & ADEF_ONE_WORD))
	    { putc (' ', fp);
	    }
	   if (ad->num > 1)
	    { fprintf (fp, "%d X ", ad->num);
	    }
	   if (ad->type != ADEF_VALUE)
	    { if (ad->flags & ADEF_LOW_RANGE || ad->flags & ADEF_HIGH_RANGE)
	       { fprintf (fp, "<%s %c%s,%s%c>", typeName(ad), 
			       (ad->flags & ADEF_LOW_OPEN) ? '(' : '[',
			       lowBoundToStr (lowStr, ad), 
			       highBoundToStr (highStr, ad),
			       (ad->flags & ADEF_HIGH_OPEN) ? ')' : ']');
	       }
	      else
	       { fprintf (fp, "<%s>", typeName(ad));
	       }
	    }
	   if (ad->comment)
	    { fprintf (fp, " -- %s", ad->comment);
	    }
	   putc ('\n', fp);
	 }
}

char *argDefErrorStr()
{
	return (ebuf);
}

void argDefArgvPrependFile (int *argcp, char ***argvp, const char *name)
{
	char ibuf[256];
	char *cppArgVec[256];
	int cppN;
	int ac, ac0;
	char **av, **av0;
	char **ap;

	if (!name)
	 { return;
	 }
	ac0 = *argcp;
	av0 = *argvp;

	cppN = 0;
	while (ac0 > 0)
	 { if (strncmp (*av0, "-D", 2) == 0)
	    { cppArgVec[cppN++] = *av0;
	    }
	   ac0--; av0++;
	 }
	cppArgVec[cppN] = 0;

	ac0 = *argcp;
	av0 = *argvp;

	if (getWordOpen(name, cppArgVec) < 0)
	 { if (errno != ENOENT)
	    { perror ("open");
	      fprintf (stderr, "argDef: can't open file %s\n", name);
	      exit (1);
	    }
	   else
	    { return;
	    }
	 }
	for (ac=ac0-cppN; getWord (ibuf)==1; ac++)
	   ;
	av = (char**)malloc(ac*sizeof(char**));
	getWordReopen();
	for (ap=av; getWord (ibuf)==1; )
	 { *ap++ = strcpy ((char*)malloc(strlen(ibuf)+1), ibuf);
	 }
	getWordClose ();
	while (ac0 > 0)
	 { if (strncmp (*av0, "-D", 2) != 0)
	    { *ap++ = *av0;
	    }
	   ac0--; av0++;
	 }
	*argcp = ac;
	*argvp = av;
}

#define NUM_COLUMNS	80

void argDefParseArgs (int argc, char **argv, ArgDefList *argDefs, 
		 const char *argFile, const char *helparg)
{
	char **argvec;
	char buf[256];
	int help = 0;

	if (helparg)
	 { sprintf (buf, "%s %%v,1, \"print detailed help\"", helparg);
	   if (argDefAdd (argDefs, buf, &help, 0) < 0)
	    { fprintf (stderr, "%s\n", argDefErrorStr());
	      exit (1);
	    }
	 }
	argvec = argv;
	argc--; argv++;
	if (argFile)
	 { argDefArgvPrependFile (&argc, &argv, argFile);
	 }
	while (argc > 0)
	 { int n;
	   if ((n = argDefMatch (argDefs, argc, argv)) <= 0)
	    { if (n < 0)
	       { fprintf (stderr, "%s\n", argDefErrorStr());
	       }
	      else
	       { fprintf (stderr,
"%s: illegal or malformed argument(s)\n", *argvec);
		 fprintf (stderr, "Valid arguments");
		 if (helparg)
		  { fprintf (stderr, " (use '%s' for more info):\n", helparg);
		  }
		 else
		  { fprintf (stderr, ":\n");
		  }
		 argDefPrintUsage (stderr, argDefs, NUM_COLUMNS);
	       }
	      exit (1);
	    }
	   argc -= n; argv += n;
	 }
	if (help)
	 { argDefPrintHelp (stdout, argDefs);
	   exit (0);
	 }
}

#if TEST

int iarg = -1000;
unsigned uarg = -1;
int earg = -1000;
double farg = -1000;
char *sarg;
int printHelp;

char **argvec;
char *aVec[256];
int aCnt;

initArgs()
{
	iarg = -1000;
	uarg = -1;
	earg = -1000;
	farg = -1000;
	sarg = NULL;
	printHelp = 0;
}


void buildArgList VA_ARGDEF((char *arg0, ...))
{
	va_list pvar;
	char *arg;

	aCnt = 0;
#if !defined(HAS_STDARG)
	va_start(pvar);
	arg = va_arg(pvar, char*);
#else
	va_start(pvar, arg0);
	arg = arg0;
#endif
	while (arg)
	 { aVec[aCnt++] = arg;
	   arg = va_arg(pvar, char*);
	 }
	va_end(pvar);
}

checkError (loc, status, str)
int loc, status;
char *str;
{
	if (status != -1 || strcmp(argDefErrorStr(), str) != 0)
	 { fprintf (stderr, "Error, location (%d):\n", loc);
	   if (status != -1)
	    { fprintf (stderr, "No error reported\n");
	    }
	   else
	    { fprintf (stderr, "%s\nvs.\n%s\n", argDefErrorStr(), str);
	    }
	   exit (-1);
	 }
}

main(argc, argv)
int argc;
char **argv;
{
	ArgDefList *dlist;
	int status;
	int n;

	char **av[10];
	int ac;

	argvec = argv;

	status = argDefAdd (argDefCreate(), "%i,[-1,1]", &iarg, 0);
	checkError ((1), status, 
		"argDefAdd(): no option name given");

	status = argDefAdd (argDefCreate(), "-i %i [-1,1]", &iarg, 0);
	checkError ((2), status, 
		"argDefAdd(), option -i: missing comma after format");

	status = argDefAdd (argDefCreate(), "-i %i, [a,1]", &iarg, 0);
	checkError ((3), status, 
		"argDefAdd(), option -i: malformed low range spec");

	status = argDefAdd (argDefCreate(), "-i %i, (-1 1)", &iarg, 0);
	checkError ((4), status, 
		"argDefAdd(), option -i: missing comma after low range spec");

	status = argDefAdd (argDefCreate(), "-i %i, (-1,a)", &iarg, 0);
	checkError ((5), status, 
		"argDefAdd(), option -i: malformed high range spec");

	status = argDefAdd (argDefCreate(), "-i %v -1", &iarg, 0);
	checkError ((6), status, 
		"argDefAdd(), option -i: missing comma after format");

	status = argDefAdd (argDefCreate(), "-i %v, bb", &iarg, 0);
	checkError ((7), status, 
"argDefAdd(), option -i: missing or malformed value for 'v' format");

	status = argDefAdd (argDefCreate(), "-i %v,1", &iarg, 
					      "-i %v,1", &iarg, 0);
	checkError ((8), status, 
		"argDefAdd(), option -i: name already in use");

	status = argDefAdd (argDefCreate(), "-i i, [-1,1]", &iarg, 0);
	checkError ((9), status, 
		"argDefAdd(), option -i: no format spec");

	status = argDefAdd (argDefCreate(), "-i %p, -1,1", &iarg, 0);
	checkError ((10), status, 
		"argDefAdd(), option -i: bad format spec");

	status = argDefAdd (argDefCreate(), "-i %", &iarg, 0);
	checkError ((11), status, 
		"argDefAdd(), option -i: bad format spec");

	status = argDefAdd (argDefCreate(), "-i %i, (-1,1],\"hi", &iarg, 0);
	checkError ((12), status, 
		"argDefAdd(), option -i: malformed comment string");

	status = argDefAdd (argDefCreate(), "-i %i, [-1,1),hi\"", &iarg, 0);
	checkError ((13), status, 
		"argDefAdd(), option -i: malformed comment string");

	status = argDefAdd (argDefCreate(), "-i %i, [-1,1] \"hi", &iarg);
	checkError ((14), status, 
		"argDefAdd(), option -i: garbage at end of spec");

	status = argDefAdd (argDefCreate(), "-i %i, -1,1] \"hi", &iarg);
	checkError ((15), status, 
	     "argDefAdd(), option -i: range spec must begin with '[' or '('");

	status = argDefAdd (argDefCreate(), "-i %i, [-1,1| \"hi", &iarg);
	checkError ((16), status, 
	     "argDefAdd(), option -i: range spec must end with ']' or ')'");

	dlist = argDefCreate();
	status = argDefAdd (dlist, "-i %i, [-1,10]", &iarg,
				    "-ff %f, [-10.5, 20.5]", &farg,
				    "-fb %f, (-10.5, 20.5]", &farg,
				    "-fe %f, [-10.5, 20.5)", &farg,
				    "-xa %x, [0xa, 0xff]", &uarg,
				    "-oa %o, [0100, 200]", &uarg,
				    "-ie=%i, [-1, 10]", &iarg,
				    "-v %v, 123", &iarg,
				    "-s %s", &sarg,
				    0);
	if (status != 1)
	 { fprintf (stderr, "Error, location (15)\n");
	   fprintf (stderr, "%s\n", argDefErrorStr());
	   exit (-1);
	 }

	buildArgList ("-i", "-2", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((20), status,
		"Option -i: associated value must be >= -1");

	buildArgList ("-i", "11", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((21), status,
		"Option -i: associated value must be <= 10");

	buildArgList ("-ff", "-10.6", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((22), status,
		"Option -ff: associated value must be >= -10.5");

	buildArgList ("-ff", "20.6", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((23), status,
		"Option -ff: associated value must be <= 20.5");

	buildArgList ("-fb", "-10.5", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((24), status,
		"Option -fb: associated value must be > -10.5");

	buildArgList ("-fe", "20.5", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((25), status,
		"Option -fe: associated value must be < 20.5");

	buildArgList ("-xa", "0x9", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((26), status,
		"Option -xa: associated value must be >= 0xa");

	buildArgList ("-xa", "0x100", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((27), status,
		"Option -xa: associated value must be <= 0xff");

	buildArgList ("-oa", "077", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((28), status,
		"Option -oa: associated value must be >= 0100");

	buildArgList ("-oa", "0311", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((29), status,
		"Option -oa: associated value must be <= 0310");

	buildArgList ("-i", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((30), status,
		"Option -i needs an additional argument");

	buildArgList ("-ie=", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((31), status,
		"Option -ie=: associated argument is malformed or missing");

	buildArgList ("-i", "aa", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((32), status,
		"Option -i: associated argument is malformed or missing");

	buildArgList ("-xa", "gg", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((33), status,
		"Option -xa: associated argument is malformed or missing");

	buildArgList ("-oa", "88", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	checkError ((34), status,
		"Option -oa: associated argument is malformed or missing");


	initArgs();
	buildArgList ("-i", "9", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	if (status != 2 || iarg != 9)
	 { fprintf (stderr, "Error, location (40)\n");
	   exit (-1);
	 }

	initArgs();
	buildArgList ("-ff", "11.1", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	if (status != 2 || farg != 11.1)
	 { fprintf (stderr, "Error, location (41)\n");
	   exit (-1);
	 }

	initArgs();
	buildArgList ("-xa", "0xba", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	if (status != 2 || uarg != 0xba)
	 { fprintf (stderr, "Error, location (41)\n");
	   exit (-1);
	 }

	initArgs();
	buildArgList ("-oa", "0102", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	if (status != 2 || uarg != 0102)
	 { fprintf (stderr, "Error, location (42)\n");
	   exit (-1);
	 }

	initArgs();
	buildArgList ("-ie=7", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	if (status != 1 || iarg != 7)
	 { fprintf (stderr, "Error, location (43)\n");
	   exit (-1);
	 }

	initArgs();
	buildArgList ("-v", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	if (status != 1 || iarg != 123)
	 { fprintf (stderr, "Error, location (44)\n");
	   exit (-1);
	 }

	initArgs();
	buildArgList ("-s", "foobar", 0);
	status = argDefMatch (dlist, aCnt, aVec);
	if (status != 2 || strcmp(sarg, "foobar") != 0)
	 { fprintf (stderr, "Error, location (45)\n");
	   exit (-1);
	 }

	initArgs();

	dlist = argDefCreate();
	if (argDefAdd (dlist, 
		    "-iarg %i,[-1,10] ", &iarg,
		    "-earg=%i,	    ", &earg,
		    "-str %s", &sarg,
		    "-help %v,1,\"print out \\\"help\\\" information\"", 
		    &printHelp,
		    0) < 0)
	 { fprintf (stderr, "%s\n", argDefErrorStr());
	   exit (1);
	 }

	argc--; argv++;
	while (argc)
	 { n = argDefMatch (dlist, argc, argv);
	   if (n < 0)
	    { fprintf (stderr, "%s\n", argDefErrorStr());
	      exit (1);
	    }
	   else if (n > 0)
	    { argc -= n;
	      argv += n;
	    }
	   else
	    { fprintf (stderr, "Usage: %s\n", argvec[0]);
	      argDefPrintUsage (stderr, dlist, 80);
	      exit (1);
	    } 
	 }
	if (printHelp)
	 { argDefPrintHelp(stdout, dlist);
	 }
	printf ("iarg=%d\n", iarg);
	printf ("earg=%d\n", earg);
	printf ("sarg=%s\n", sarg);
}

#endif
