/*---------McGill-Computer-Vision-and-Robotics-Laboratory--C-Module---------
 *		
 * NAME:	cputime, int_cputime
 *
 * USAGE:	include "/u1/robot/util/useful.h"
 *
 *		time = cputime (option);
 *		  double time;
 *		  enum cputime_option;
 *
 * DESCRIPTION: Returns the time consumed by a process. The options are as
 *		follows:
 *
 *		  System  : Time spent executing system calls for the process.
 *		  User    : Time spent executing the processes own code.
 *		  All	  : All time spent executing for the process.
 * 
 * FILES:	/u1/robot/util/cputime.c		: source
 *		/u1/robot/util/useful.a			: archive
 * 
 * NOTES:
 *
 * AUTHOR/DATE: John Lloyd and Mike Parker, September 1984
 *
 * UPDATES:
 *
 *-------------McGill-Computer-Vision-and-Robotics-Laboratory--------------*/

#ifdef HAS_TIMES
#include <sys/times.h>
#include <limits.h>

double cputime()
{
 struct tms info;

 (void) times(&info);
 return(info.tms_utime / CLK_TCK);
}

int int_cputime()
{
 return (1000000*cputime());
}

#else

#include <sys/time.h>
#include <sys/resource.h>

#include <useful.h>

double cputime(option)
enum cputime_option option;
{
 static struct rusage ru;
 double result;

 getrusage(RUSAGE_SELF,&ru);
 
 result = 0;

 if (option == System || option == All)
    result = ru.ru_stime.tv_sec + (ru.ru_stime.tv_usec/1000000.);

 if (option == User || option == All)
    result = ru.ru_utime.tv_sec + (ru.ru_utime.tv_usec/1000000.);
    
 return (result);
}

#endif
