/*=~=~=~=~ McGill-Computer-Vision-and-Robotics-Laboratory--C-Module ~=~=~=~=
 *		
 * NAME:	fatal, warn, usage -- library routines to print diagnistics
 *		   on stderr.
 *
 * SYNOPSIS:	fatal (code, s, arg1, arg2, ...)
 *		   int code;
 *		   char *s;
 *		   args 1 to n (combined size not to exceed 32 bytes)
 *		   
 *		usage (code, s, arg1, arg2, ...)
 *		   int code;
 *		   char *s;
 *		   args 1 to n (combined size not to exceed 32 bytes)
 * 
 *		warn (s, arg1, arg2, ...)
 *		   char *s;
 *		   args 1 to n (combined size not to exceed 32 bytes)
 *
 * DESCRIPTION:	These routines print diagnostic messages on stderr. The usage
 *		for each routine is analogous to printf: a format string,
 *		followed by a list of arguments to be printed. The number of 
 *		arguments is currently limited such that combined size of 
 *		all the	arguments may not exceed 32 bytes. This was done so 
 *		that the software can be written machine independently.
 *
 *		'fatal' prints the diagnostic preappended with the string
 *		"ERROR -- ", and then exits with the value 'code'.
 *
 *		'usage' prints the diagnostic preappended with the string
 *		"USAGE: ", and then exits with the value 'code'. 
 *		It is used when interpreting argument strings.
 *
 *		'warn' prints the diagnostic preappended with the string
 *		"WARNING -- ", but does not exit.
 * 
 * ARGUMENTS:	code ---	exit code
 *
 *		s ---		format string
 *
 *		arg1 ... argn	something the print
 *
 * NOTES:	
 *
 * AUTHOR/DATE:	John Lloyd, November 1986
 *
 * UPDATES:	
 *
 *~=~=~=~=~=~= McGill-Computer-Vision-and-Robotics-Laboratory ~=~=~=~=~=~=*/

#include <stdio.h>

struct _space {
  unsigned long l[8]; } ;

fatal (code, s, space)
int code;
char *s;
struct _space space;
{
	fprintf (stderr, "ERROR -- ");
	fprintf (stderr, s, space);
	exit (code);
}

usage (code, s, space)
int code;
char *s;
struct _space space;
{
	fprintf (stderr, "USAGE: ");
	fprintf (stderr, s, space);
	exit (code);
}

warn (s, space)
char *s;
struct _space space;
{
	fprintf (stderr, "USAGE: ");
	fprintf (stderr, s, space);
}
