#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cdefs.h>
#include <unistd.h>
#include <getWord.h>

#ifdef HAS_UNION_WAIT
#define WAIT_T union wait
#else
#define WAIT_T int
#endif

static char getWordIbuf[1024];
static int getWordIdx = 0;
static int getWordLine;
static char getWordFileName[256];
static FILE *getWordFp;

#ifndef CPP
#ifdef sun
#define CPP	"/usr/ccs/lib/cpp"
#else
#define CPP	"/usr/lib/cpp"
#endif
#endif

int getWordClose ()
{
	if (getWordFp)
	 { fclose (getWordFp);
	   unlink (getWordFileName);
	   getWordFp = 0;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

int getWordOpen (char *name, char **args)
{
	WAIT_T status;
	int rfd, wfd;	
	int pid;
	char **argv;
	int k, argc;

	getWordClose();

	/* count the number of arguments */
	argc = 2;
	for (k=0; args[k]; k++)
	   ;
	argc += k;

	/* now build the arg list */
	argv = (char**)malloc(sizeof(char*)*(argc+1));
	k = 0;
	argv[k++] = CPP;
	argv[k++] = "-P";
	while (argv[k++] = *args++)
	   ;
#if 0
for (k=0; k<=argc; k++)
 { printf ("argv[%d] = %s\n", k, argv[k] ? argv[k] : "null");
 }
#endif
	if ((rfd = open(name, O_RDONLY)) == -1)
	 { fprintf (stderr, "getWordOpen(): can't open file %s\n",name);
	   perror ("open");
	   exit (1);
	 }
	sprintf (getWordFileName, "/tmp/getWord%d", getpid());
	if ((wfd = open(getWordFileName, (O_WRONLY|O_CREAT), 0666)) == -1)
	 { fprintf (stderr, "getWordOpen(): can't open file %s\n",
		    getWordFileName);
	   perror ("open");
	   exit (1);
	 }
	if ((pid = fork()) == 0)
	 { /* in the child process */
	   dup2 (rfd, fileno(stdin));
	   dup2 (wfd, fileno(stdout));
	   if (execvp (CPP, argv) == -1)
	    { fprintf (stderr,"getWordOpen(): Can't execlp %s\n", CPP);

	      perror ("execlp");
	      exit (1);
	    }
	 }
	else
	 { /* in the parent process */
	   wait (&status);
	   if ((WIFEXITED(status) && WEXITSTATUS(status) != 0) ||
	       (WIFSIGNALED(status)))
	    { fprintf (stderr, "getWordOpen(): Error executing %s\n", CPP);
	      exit (1);
	    }
	 }
	if ((getWordFp = fopen(getWordFileName, "r")) == NULL)
	 { fprintf (stderr,"getWordOpen(): can't fopen file %s\n",
			   getWordFileName);
	   perror ("fopen");
	   exit (1);
	 }
	close (rfd);
	close (wfd);
	getWordIdx = 0;
	getWordLine = 0;
	free (argv);
	return (1);
}

int getWordReopen ()
{
	if (getWordFp == NULL)
	 { fprintf (stderr, "getWordReopen(): no file open\n");
	   exit (1);
	 }
	if ((freopen (getWordFileName, "r", getWordFp)) == NULL)
	 { fprintf (stderr,"getWordReopen(): can't freopen file %s\n",
			   getWordFileName);
	   perror ("freopen");
	   exit (1);
	 }
	getWordIdx = 0;
	getWordLine = 0;
}

int getWord (buf)
char *buf;
{
	char *cp, *bp;

	if (getWordFp == NULL)
	 { fprintf (stderr, "getWord(): no file open\n");
	   exit (1);
	 }
	do
	 { for (cp=&getWordIbuf[getWordIdx]; *cp!='#' && *cp!='\0'; cp++)
	    { if (!isspace(*cp))
	       { bp = buf;
		 if (*cp == '"')
		  { cp++;
		    while (*cp != '"' && *cp != '\n')
		     { if (cp[0] == '\\')
			{ if (cp[1] == '"')
			   { *bp++ = '"';
			     cp += 2;
			   }
			  else if (cp[1] == '\\')
			   { *bp++ = '\\';
			     cp += 2;
			   }
			  else if (cp[1] == 'n')
			   { *bp++ = '\n';
			     cp += 2;
			   }
			  else if (cp[1] == 't')
			   { *bp++ = '\t';
			     cp += 2;
			   }
			  else if (cp[1] == 'f')
			   { *bp++ = '\f';
			     cp += 2;
			   }
			  else if (cp[1] == 'r')
			   { *bp++ = '\r';
			     cp += 2;
			   }
			  else if (cp[1] == 'b')
			   { *bp++ = '\b';
			     cp += 2;
			   }
			  else if (isdigit(cp[1]))
			   { int v, i;
			     v = cp[1] - '0';
			     cp += 2;
			     for (i=0; isdigit(*cp) && i<2; i++)
			      { v = v*8 + *cp++ - '0';
			      }
			     *bp++ = MIN(v,255);
			   }
			}
		       else
			{ *bp++ = *cp++;
			}
		     }
		    /* verify the closing quote */
		    if (*cp != '"')
		     { printf ("%c\n", *cp);
		       fprintf (stderr, "getWord(): unterminated string, line %d\n",
				    getWordLine);
		       exit (1);
		     }
		    cp++;
		  }
		 else
		  { while (!isspace(*cp) && *cp!='#' && *cp!='\0')
		     { *bp++ = *cp++;
		     }
		  }
		 *bp = '\0';
		 getWordIdx = cp - getWordIbuf;
		 return (1);
	       }
	    }
	   getWordIdx = 0;
	   getWordIbuf[0] = '\0';
	 }
	while (fgets (getWordIbuf, 1024, getWordFp) && ++getWordLine);
	return (0);
}

int getWordLineNo()
{
	return (getWordLine);
}
