/* A version of sprintf that sets a limit on the size of the final
   string, UP TO MAXSTRING; beyond that, at the moment, we get a
   segmenttation fault; this is because stdio doesn't provide
   the proper support. */

#if !HAS_SNPRINTF
#include <snprintf.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#if !HAS_STDARG
#include <varargs.h>	/* hope we have varargs.h if no stdarg */
#endif

int vsnprintf(char *str, int nmax, const char *fmt, va_list ap)
{
	/*
	// nmax is the maximum number of characters allowed in the
	// string, including the terminating NULL. If the output
	// is truncated (i.e., the number of characters preceeding
	// the NULL is >= nmax), -1 is returned
	*/
	char buf[MAXSTRING];
	char *bufp = buf;
	int len;

	if (nmax >= MAXSTRING)
	 { bufp = (char*)malloc(nmax+1);
	 }
	vsprintf (bufp, fmt, ap);
	bufp[nmax] = '\0';
	len = strlen(bufp);
	if (len >= nmax)
	 { strncpy (str, bufp, nmax-1);
	   str[nmax-1] = '\0';
	   len = -1;
	 }
	else
	 { strcpy (str, bufp);
	 }
	if (bufp != buf)
	 { free (bufp);
	 }
	return len;
}

#if !HAS_STDARG
int snprintf (va_alist) va_dcl
#else
int snprintf (char* str, int nmax, const char* fmt, ...)
#endif
{
	va_list ap;
	int n;
#if !HAS_STDARG
	char *str;
	int nmax;
	char *fmt;

	va_start(ap);

	str = va_arg(ap, char*);
	nmax = va_arg(ap, int);
	fmt = va_arg(ap, char*);
#else
	va_start(ap,fmt);
#endif
	n = vsnprintf (str, nmax, fmt, ap);
	va_end(ap);
	return (n);	
}

#endif
