#include <string.h>
#include <stringx.h>
#include <ctype.h>
#include <cdefs.h>

char *strndup (const char *s, int n)
{
	char *cp;
	int l;

	if ((l = strlen(s)) < n)
	 { n = l;
	 }
	cp = (char *)malloc (sizeof(char)*(n+1));
	if (cp != NULL)
	 { strncpy (cp, s, n);
	   cp[n] = '\0';
	 }
	return ((char*)cp);
}

int fmtindex (char **beginp, char **endp, const char *s)
{
	/* searches a string for a valid printf() format conversion
	   sequences, and then returns the beginning and ending pointers
	   associated with that sequence.

	   return values:

		1: a format was found
		0: no format was found
	       -1: an illegal format was found
	*/

	static const char *flags = "-+ #0";
	static const char *intconvs = "diouxX";
	static const char *floatconvs = "feEgG";
	static const char *charconvs = "cs";
	static const char *allconvs = "diouxXfeEgGcsnp";

	int rcode = 0;
	const char *begin = NULL;
	const char *end = NULL;
	const char *convchars;

	const char *cp;
	for (cp=s; *cp != '\0'; cp++)
	 { if (*cp == '%')
	    { begin = cp++;
	      if (*cp == '%')
	       { continue;
	       }
	      while (strchr (flags, *cp))
	       { cp++;
	       }
	      while (isdigit(*cp))
	       { cp++;
	       }
	      if (*cp == '.')
	       { cp++;
		 while (isdigit(*cp))
		  { cp++;
		  }
	       }
	      if (*cp == 'h' || *cp == 'l')
	       { cp++;
		 if (*cp == 'l')
		  { cp++;
		  }
		 convchars = intconvs;
	       }
	      else if (*cp == 'L')
	       { cp++;
		 convchars = floatconvs;
	       }
	      else if (*cp == 'w')
	       { cp++;
		 convchars = charconvs;
	       }
	      else
	       { convchars = allconvs;
	       }
	      if (strchr (convchars, *cp))
	       { rcode = 1;
	       }
	      else
	       { rcode = -1;
	       }
	      end = cp;
	      break;
	    }
	 }
	if (beginp) *beginp = (char*)begin;
	if (endp) *endp = (char*)end;
	return (rcode);
}
