#ifdef Sun4sol_TARGET

#include <unistd.h>
#include <thread.h>
#include <sys/types.h>
#include <sys/priocntl.h>
#include <sys/rtpriocntl.h>
#include <sys/tspriocntl.h>
#include <sys/iapriocntl.h>
#include <sys/lwp.h>
#include <sys/mman.h>
#include <errno.h>
#include <sys/wait.h>

static id_t rtID;
static id_t tsID;
static id_t iaID;

id_t schedinfo (name, maxpri)
char *name;
int *maxpri;
{
	pcinfo_t info;
	int nc;

	(void) strcpy (info.pc_clname, name);
	nc = priocntl (0L, 0L, PC_GETCID, (caddr_t)&info);

	if (nc == -1L)
	 { return (-1);
	 }
	if (strcmp ("RT", name) == 0)
	 { rtinfo_t *rtinfop;
	   rtinfop = (rtinfo_t*)&info.pc_clinfo;
	   if (maxpri)
	    { *maxpri = rtinfop->rt_maxpri;
	    }
	 }
	else if (strcmp ("TS", name) == 0)
	 { tsinfo_t *tsinfop;
	   tsinfop = (tsinfo_t*)&info.pc_clinfo;
	   if (maxpri)
	    { *maxpri = tsinfop->ts_maxupri;
	    }
	 }
	else if (strcmp ("IA", name) == 0)
	 { iainfo_t *iainfop;
	   iainfop = (iainfo_t*)&info.pc_clinfo;
	   if (maxpri)
	    { *maxpri = iainfop->ia_maxupri;
	    }
	 }
	else
	 { errno = EINVAL;
	   return (-1);
	 }
	return (info.pc_cid);
}

showPri(char *str, idtype_t type, id_t id)
{
	pcparms_t pcparms;
	rtparms_t *rtparmsp;
	tsparms_t *tsparmsp;
	iaparms_t *iaparmsp;

	pcparms.pc_cid = PC_CLNULL;
	if (priocntl (type, id, PC_GETPARMS, (caddr_t)&pcparms) == -1L)
	 { perror ("priocntl PC_GETPARMS");
	   exit (1);
	 }
	if (pcparms.pc_cid == rtID)
	 { rtparmsp = (rtparms_t *)&pcparms.pc_clparms;
	   printf ("%s: RT at %d\n", str, rtparmsp->rt_pri);
	 }
	else if (pcparms.pc_cid == tsID)
	 { tsparmsp = (tsparms_t *)&pcparms.pc_clparms;
	   printf ("%s: TS at %d\n", str, tsparmsp->ts_upri);
	 }
	else if (pcparms.pc_cid == iaID)
	 { iaparmsp = (iaparms_t *)&pcparms.pc_clparms;
	   printf ("%s: IA at %d\n", str, iaparmsp->ia_upri);
	 }
	else
	 { printf ("Unknown scheduling type (%d)\n", pcparms.pc_cid);
	   exit (1);
	 }
}

#define PATH	"/tmp/setpri"

boostPri ()
{
	char idbuf[8];
	id_t child;
	
	sprintf (idbuf, "%d", getpid());
	if (child = fork())
	 { wait ((int*)0);
	 }
	else
	 { if (execl (PATH, PATH, idbuf, (char*)NULL) < 0)
	    { perror ("Unable to execute %s\n", PATH);
	      exit (1);
	    }
	 }
}

id_t lwpId = -1;

threadFxn()
{
	printf ("here\n");
	lwpId = _lwp_self();
	sleep(1000);
}

main()
{
	pcparms_t pcparms;
	rtparms_t *rtparmsp;
	id_t schedinfo();
	int maxrtpri;
	int maxtspri;
	int maxiapri;
	thread_t tid;

	printf ("pid=%d\n", getpid());

	if ((rtID = schedinfo ("RT", &maxrtpri)) == -1)
	 { perror ("schedinfo for RT");
	   exit (1);
	 }
	if ((tsID = schedinfo ("TS", &maxtspri)) == -1)
	 { perror ("schedinfo for TS");
	   exit (1);
	 }
	if ((iaID = schedinfo ("IA", &maxiapri)) == -1)
	 { perror ("schedinfo for IA");
	   exit (1);
	 }
	pcparms.pc_cid = PC_CLNULL;
	if (priocntl (P_LWPID, _lwp_self(), PC_GETPARMS, 
	    (caddr_t)&pcparms) == -1L)
	 { perror ("priocntl P_LWPID PC_GETPARMS");
	   exit (1);
	 }
	showPri("proc before", P_PID, getpid());
	boostPri();
	showPri("proc after ", P_PID, getpid());	

	if (thr_create ((void*)0, 0, threadFxn, (void*)0, 
			 THR_BOUND, &tid) != 0)
	 { perror ("thr_create");
	   exit (1);
	 }
	while (lwpId == -1)
	 { nap (0.1);
	 }
	showPri("thr 0 before", P_LWPID, _lwp_self());
	showPri("thr 1 before", P_LWPID, lwpId);
	if (priocntl (P_LWPID, _lwp_self(), PC_SETPARMS, 
	    (caddr_t)&pcparms) == -1L)
	 { perror ("priocntl P_LWPID PC_SETPARMS");
	   exit (1);
	 }
	showPri("thr 0 after ", P_LWPID, _lwp_self());
	showPri("thr 1 after ", P_LWPID, lwpId);
}
#endif

