/*---------McGill-Computer-Vision-and-Robotics-Laboratory--C-Module---------
 *		
 * NAME:	testchar(), testinput() -- see if input is ready
 *
 * USAGE:	status = testchar(time)
 *		  int status;
 *		  float time;
 *
 *		status = testinput(fd, time)
 *		  int status, fd;
 *		  float time;
 *
 * DESCRIPTION: These functions return TRUE (1) whenever there is input on the
 * 		indicated stream that has yet to be read. This can be used to
 *		effect polled asynchronous IO. 'testinput()' examines the
 *		stream indicated by the file descriptor 'fd'; 'testchar()' 
 *		uses stdin by default.
 *
 *		The variable 'time' is indicated in seconds and tells the
 *		routine how long to wait for a character to appear before
 *		returning. The routine thus returns whenever a character 
 *		appears, or the indicated amount of time expires.
 *
 * NOTES:	Timevalues are truncated as necessary to make sure the fit
 *		into the range allowed by a struct timeval; this range is
 *		2.14748e9 to 0 seconds. Negative values are assumed zero.
 *	
 * DIAGNOSTICS: The functions returns 1 or 0 normally. If they return -1, that
 *		means the call to select() is failing unexpectedly, and a
 *		guru should be consulted.
 *
 * GLOBAL VARS: None.
 *
 * FILES:	/u1/robot/util/testchar.c			: source
 *		/u1/robot/util/useful.a				: archive
 * 
 * AUTHOR/DATE:	John Lloyd, November, 1984
 *
 * UPDATES:
 *
 *-------------McGill-Computer-Vision-and-Robotics-Laboratory--------------*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>

#include <useful.h>

#define NFDS 32

#if defined(_IO_stdin)	/* Linux */
#define IO_COUNT(_fp)	((_fp)->_IO_read_end - (_fp)->_IO_read_ptr)
#else
#define IO_COUNT(_fp)   ((_fp)->_cnt)
#endif

testchar (time)
double time;
{
	return (testc (stdin, time));
}

testc (fp, time)
FILE *fp;
double time;
{
	/* First check to see if there is any data in the buffer */

	if (IO_COUNT(fp) > 0)
	 { return (1);
	 }
	else
	 { return (testinput(fileno(fp), time));
	 }
}

testinput (fd, time)
int fd;
double time;
{
	struct timeval timeout;		/* timeout variable for select()   */
	fd_set readmask;		/* IO detection mask for select()  */
	int status;			/* status returned by select()	   */
 
	/* turn 'time' into the appropriate entry in struct timeval */
 
	if (time > 2.14748e9)
	 { time = 2.14748e9;
	 }
	if (time < 0)
	 { time = 0;
	 }
	timeout.tv_sec = time;
	timeout.tv_usec = (time - timeout.tv_sec) * 1000000;

	/* Now perform the test on the stdin stream */
 
	FD_ZERO (&readmask);
	FD_SET (fd, &readmask);
	status = select (NFDS, &readmask, NULL, NULL, &timeout);
 
	if (status > 1 || status < 0)
	 { return (-1);
	 }
	else
	 { return (status);
	 }
}
