#include <RxDlist.h>
#include <string.h>

int RxDlistBase::removeAndDelete(RxDnodeBase* ptr1, RxDnodeBase* ptr2)
{
	RxDnodeBase *node;
	RxDnodeBase *nextNode = NULL;
	RxDnodeBase *lastNode;
	int cnt = 0;

	if (ptr1->prev_)
	 { ptr1->prev_->next_ = ptr2->next_;
	 }
	else
	 { head_ = ptr2->next_;
	 }
	if (ptr2->next_)
	 { ptr2->next_->prev_ = ptr1->prev_;
	 }
	else
	 { tail_ = ptr1->prev_;
	 }
	lastNode = ptr2->next_;
	for (node=ptr1; node!=lastNode; node=nextNode)
	 { nextNode = node->next_;
	   freeNode (node);
	   cnt++;
	 }
	return cnt;
}

int RxDlistBase::remove(RxDnodeBase* ptr1, RxDnodeBase* ptr2)
{
	RxDnodeBase *node;
	RxDnodeBase *nextNode = NULL;
	RxDnodeBase *lastNode;
	int cnt = 0;

	if (ptr1->prev_)
	 { ptr1->prev_->next_ = ptr2->next_;
	 }
	else
	 { head_ = ptr2->next_;
	 }
	if (ptr2->next_)
	 { ptr2->next_->prev_ = ptr1->prev_;
	 }
	else
	 { tail_ = ptr1->prev_;
	 }
	lastNode = ptr2->next_;
	for (node=ptr1; node!=lastNode; node=nextNode)
	 { nextNode = node->next_;
	   cnt++;
	 }
	return cnt;
}

int RxDlistBase::removeAll()
{
	int notEmpty;

	notEmpty = (head_ != NULL);
	head_ = NULL;
	tail_ = NULL;
	return notEmpty;
}

int RxDlistBase::removeAndDeleteAll()
{
	RxDnodeBase *node, *nextNode = NULL;
	int cnt = 0;

	for (node=head_; node; node=nextNode)
	 { nextNode = node->next_;
	   freeNode (node);
	   cnt++;
	 }
	head_ = NULL;
	tail_ = NULL;
	return cnt;
}

RxDnodeBase *RxDlistBase::remove(RxDnodeBase* ptr)
{
	if (ptr->next_)
	 { ptr->next_->prev_ = ptr->prev_;
	 }
	else
	 { tail_ = ptr->prev_;
	 }
	if (ptr->prev_)
	 { ptr->prev_->next_ = ptr->next_;
	 }
	else
	 { head_ = ptr->next_;
	 }
	return (ptr);
}

#if 0
RxDlistBase::~RxDlistBase()
{
	if (autoDelete) removeAndDeleteAll ();
}
#endif

RxDnodeBase *RxDlistBase::addTail(RxDnodeBase* ptr)
{
	ptr->next_ = NULL;
	if (tail_ == NULL)
	 { head_ = ptr;
	 }
	else
	 { tail_->next_ = ptr;
	 }
	ptr->prev_ = tail_;
	tail_ = ptr;
	return (ptr);
}

RxDnodeBase *RxDlistBase::addHead(RxDnodeBase* ptr)
{
	ptr->prev_ = NULL;
	if (head_ == NULL)
	 { tail_ = ptr;
	 }
	else
	 { head_->prev_ = ptr;
	 }
	ptr->next_ = head_;
	head_ = ptr;
	return (ptr);
}

RxDnodeBase *RxDlistBase::addAfter(RxDnodeBase* ref, RxDnodeBase* ptr)
{
	if (ref == NULL)
	 { return addHead (ptr);
	 }
	else
	 { ptr->prev_ = ref;
	   if (ref->next_)
	    { ref->next_->prev_ = ptr;
	    }
	   else
	    { tail_ = ptr;
	    }
	   ptr->next_ = ref->next_;
	   ref->next_ = ptr;
	   return (ptr);
	 }
}

RxDnodeBase *RxDlistBase::addBefore(RxDnodeBase* ref, RxDnodeBase* ptr)
{
	if (ref == NULL)
	 { return addTail (ptr);
	 }
	else
	 { ptr->next_ = ref;
	   if (ref->prev_)
	    { ref->prev_->next_ = ptr;
	    }
	   else
	    { head_ = ptr;
	    }
	   ptr->prev_ = ref->prev_;
	   ref->prev_ = ptr;
	   return (ptr);
	 }
}

RxDnodeBase *RxDlistBase::findNodeByPtr (const void *ptr) const
{
	RxDnodeBase *node;

	for (node=head_; node; node=node->next_)
	 { if (node->compare (ptr))
	    { return (node);
	    }
	 }
	return (NULL);
}

RxDnodeBase *RxDlistBase::findNodeByName (const char *name) const
{
	RxDnodeBase *node;

	for (node=head_; node; node=node->next_)
	 { if (strcmp (node->getName(), name) == 0)
	    { return (node);
	    }
	 }
	return (NULL);
}

int RxDlistBase::length () const
{
	RxDnodeBase *node;
	int cnt = 0;

	for (node=head_; node; node=node->next_)
	 { cnt++;
	 }
	return (cnt);
}

RxDnodeBase *RxDnodeBase::getNext(int n)
{
	RxDnodeBase *node, *nextNode;
	int k;

	if (n < 0)
	 { return (getPrev(-n));
	 }
	else
	 { node = this;
	   for (k=n; k; k--)
	    { if ((nextNode = node->next_))
	       { node = nextNode;
	       }
	      else
	       { break;
	       }
	    }
	   return (node);
	 }
}

RxDnodeBase *RxDnodeBase::getPrev(int n)
{
	RxDnodeBase *node, *prevNode;
	int k;

	if (n < 0)
	 { return (getNext(-n));
	 }
	else
	 { node = this;
	   for (k=n; k; k--)
	    { if ((prevNode = node->prev_))
	       { node = prevNode;
	       }
	      else
	       { break;
	       }
	    }
	   return (node);
	 }
}
