#include <RxDlist.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <stdarg.h>

static int itemCnt = 0;

class MyNode : public RxDnodeBase 
{
    public:
	MyNode (const char *s)
	 { str = strdup(s);
	   itemCnt++;
	 };
	~MyNode() { printf ("freeing %s\n", str); free (str); itemCnt--; };
	RX_NAMED_DNODE_DECLS (MyNode, str)
};

typedef RxDlist<MyNode> MyList;
typedef RxDnode<int> IntNode;
typedef RxDlist<IntNode> IntList;

char *makeStr(int i)
{
	static char buf[64];
	sprintf (buf, "%d%d", i, i);
	return (buf);
}

testList (MyList *list, ...)
{
	MyNode *node;
	char *names[256];
	char *cp;
	int len, k;

	va_list ap;
	va_start (ap, list);
	len = 0;
	do
	 { cp = va_arg (ap, char *);
	   if (cp )
	    { names[len++] = cp;
	    }
	   else
	    { break;
	    }
	 }
	while (1);
	va_end (ap);

	if (list->length() != len)
	 { fprintf (stderr, "length = %d vs. %d\n", list->length(), len);
	   return (0);
	 }
	if (len != itemCnt)
	 { fprintf (stderr, "length = %d, itemCnt = %d\n", len, itemCnt);
	   return (0);
	 }
	k = 0;
	for (node=list->head(); node; node=node->next())
	 { if (strcmp (names[k], node->getName()) != 0)
	    { fprintf (stderr, "A: names[%d]=%s vs. %s\n", 
				k, names[k], node->getName());
	      return (0);
	    }
	   if (node->prev())
	    { if (node->prev()->next() != node)
	       { fprintf (stderr, "Bad prev loop\n");
		 return (0);
	       }
	    }
	   else
	    { if (k != 0)
	       { fprintf (stderr, "Bad prev\n");
		 return (0);
	       }
	    }
	   k++;
	 }
	k = len-1;
	for (node=list->tail(); node; node=node->prev())
	 { if (strcmp (names[k], node->getName()) != 0)
	    { fprintf (stderr, "B: names[%d]=%s vs. %s\n", 
				k, names[k], node->getName());
	      return (0);
	    }
	   if (node->next())
	    { if (node->next()->prev() != node)
	       { fprintf (stderr, "Bad next loop\n");
		 return (0);
	       }
	    }
	   else
	    { if (k != len-1)
	       { fprintf (stderr, "Bad next\n");
		 return (0);
	       }
	    }
	   k--;
	 }
	if (list->head())
	 { if (list->head()->prev())
	    { fprintf (stderr, "bad head\n");
	      return (0);
	    }
	   if (list->tail() == NULL || list->tail()->next())
	    { fprintf (stderr, "bad tail\n");
	      return (0);
	    }
	 }
	else
	 { if (list->tail())
	    { fprintf (stderr, "tail should be NULL\n");
	      return (0);
	    }
	 }
	return (1);
}

int testStr (int i, const char *s)
{
	return ((strcmp(makeStr(i), s)) == 0);
}

RX_FAST_ALLOC_DEFS(RxDnode<int>);

main()
{
	MyList *list = new MyList;
	MyNode *node;
	MyNode *nextNode = NULL;
	int i;

	/* make sure list is properly initailized */
	assert (list->head() == NULL && list->tail() == NULL);

	for (i=0; i<10; i++)
	 { list->addTail (new MyNode(makeStr(i)));
	 }
	assert (list->length() == 10);
	assert ((node = list->findByName("00")) != NULL);
	assert (strcmp(node->getName(), "00") == 0);

	i = 0;
	for (node=list->head(); node; node=nextNode)
	 { nextNode = node->next();
	   printf ("%s\n", node->getName());
	   assert (testStr (i, node->getName()));
	   if (i%2 != 0)
	    { list->remove(node);
	      delete node;
	    }
	   i++;
	 }
	assert ((node = list->findByName("11")) == NULL);
	assert (itemCnt == 5);
	assert (list->length() == 5);
	assert (testStr (0, list->head()->getName()));
	assert (testStr (8, list->tail()->getName()));

	/* list consistency checks */

	list->removeAndDeleteAll();
	assert (testList (list, 0));

	for (i=1; i<=3; i++)
	 { list->addTail (new MyNode(makeStr(i)));
	 }
	assert (testList (list, "11", "22", "33", 0));
	list->addTail (new MyNode("44"));
	assert (testList (list, "11", "22", "33", "44", 0));
	list->addHead (new MyNode("00"));
	assert (testList (list, "00", "11", "22", "33", "44", 0));
	list->addAfter (list->findByName("00"), new MyNode("AA"));
	assert (testList (list, "00", "AA", "11", "22", "33", "44", 0));
	list->addAfter (list->findByName("44"), new MyNode("BB"));
	assert (testList (list, "00", "AA", "11", "22", "33", "44", "BB", 0));
	list->removeAndDelete  (list->findByName ("00"));
	assert (testList (list, "AA", "11", "22", "33", "44", "BB", 0));
	list->removeAndDelete  (list->findByName ("11"));
	assert (testList (list, "AA", "22", "33", "44", "BB", 0));
	list->removeAndDelete  (list->findByName ("BB"));
	assert (testList (list, "AA", "22", "33", "44", 0));
	list->addAfter (0, new MyNode("00"));
	assert (testList (list, "00", "AA", "22", "33", "44", 0));
	list->removeAndDelete  (list->findByName ("00"));
	list->removeAndDelete  (list->findByName ("AA"));
	list->removeAndDelete  (list->findByName ("22"));
	assert (testList (list, "33", "44", 0));
	list->addBefore (0, new MyNode ("ZZ"));
	assert (testList (list, "33", "44", "ZZ", 0));
	list->addBefore (list->findByName("33"), new MyNode ("XX"));
	assert (testList (list, "XX", "33", "44", "ZZ", 0));
	list->addBefore (list->findByName("44"), new MyNode ("WW"));
	assert (testList (list, "XX", "33", "WW", "44", "ZZ", 0));

	delete list;
	assert (itemCnt == 0);

	IntList *ilist = new IntList;
	IntNode *inode;
	IntNode *nextInode = NULL;

	/* make sure ilist is properly initailized */
	assert (ilist->head() == NULL && ilist->tail() == NULL);

	for (i=0; i<10; i++)
	 { inode = new IntNode;
	   inode->val = i;
	   ilist->addTail (inode);
	 }
	assert (ilist->length() == 10);

	i = 0;
	for (inode=ilist->head(); inode; inode=nextInode)
	 { nextInode = inode->next();
	   assert (inode->val == i);
	   if (i%2 != 0)
	    { ilist->remove(inode);
	      delete inode;
	    }
	   i++;
	 }
	i = 0;
	for (inode=ilist->head(); inode; inode=inode->next())
	 { assert (inode->val == 2*i);
	   i++;
	 }
	assert (ilist->length() == 5);
	assert (ilist->head()->val == 0);
	assert (ilist->tail()->val == 8);

	/* ilist consistency checks */

	ilist->removeAndDeleteAll();
	assert (ilist->length() == 0);

	/* test fast allocator */
	 { RxFastAlloc **availp;

	  availp = IntNode::getAvailPtr();
	  inode = IntNode::alloc();
	  ilist->addTail (inode);
	  ilist->remove (inode);
	  IntNode::free (inode);

	  inode = IntNode::alloc(availp);
	  ilist->addTail (inode);
	  ilist->remove (inode);
	  IntNode::free (availp, inode);
	 }	

	printf ("\nROUTINES OK\n\n");
	return (0);
}
