#include <RxHashTab.h>
#include <stdlib.h>
#include <prime.h>

unsigned int RxHashTabBase::hashStr(const char *s) const
{
	const char *cp;
	unsigned int val = 0;

	for (cp = s; *cp != '\0'; cp++)
	 { val = (val<<2) + *cp;
	 }
	return (val % size_);	
}

RxHashTabBase::RxHashTabBase(int size)
{
	size_ = primeGetClose(size);
	table = (RxHashNodeBase**)calloc(size_,sizeof(RxHashNodeBase*));
	cur_idx = 0;
	cur_pair = NULL;
	numEnts = 0;
}

RxHashNodeBase *RxHashTabBase::getFirst()
{
	int i;

	for (i=0; i<size_; i++)
	 { if (table[i] != NULL)
	    { cur_idx = i;
	      cur_pair = table[i];
	      return (cur_pair);
	    }
	 }
	cur_pair = NULL;
	return (NULL);
}

RxHashNodeBase *RxHashTabBase::getNext()
{
	int i;

	if (cur_pair == NULL)
	 { return (NULL);
	 }
	else if (cur_pair->next)
	 { cur_pair = cur_pair->next;
	   return (cur_pair);
	 }
	else
	 { for (i=cur_idx+1; i<size_; i++)
	    { if (table[i])
	       { cur_idx = i;
		 cur_pair = table[i];
		 return (cur_pair);
	       }
	    }
	   cur_pair = NULL;
	   return (NULL);
	 }
}

int RxHashTabBase::remove_ (int idx, void *p)
{
	RxHashNodeBase *last, *pair;

	last = NULL;
	for (pair=table[idx]; pair; pair=pair->next)
	 { if (pair->match(p))
	    { if (last)
	       { last->next = pair->next;
	       }
	      else
	       { table[idx] = pair->next;
	       }
	      delete pair;
	      numEnts--;
	      return 1;
	    }
	   last = pair;
	 }
	return 0;
}

RxHashNodeBase *RxHashTabBase::insert_ (int idx, RxHashNodeBase *pair)
{
	pair->next = table[idx];
	table[idx] = pair;
	numEnts++;
	return (pair);
}

void RxHashTabBase::clear()
{
	RxHashNodeBase *pair, *next;
	int i;

	for (i=0; i<size_; i++)
	 { if ((pair = table[i]))
	    { do
	       { next = pair->next;
		 delete pair;
		 pair = next;
	       }
	      while (pair);
	      table[i] = NULL;
	    }
	 }
	numEnts = 0;
}

RxHashNodeBase *RxHashTabBase::lookup(int idx, void *p) const
{
	RxHashNodeBase *pair;

	for (pair=table[idx]; pair; pair=pair->next)
	 { if (pair->match(p))
	    { return (pair);
	    }
	 }
	return (NULL);
}
