#ifndef _RXHASHTAB_H_WRAP_
#define _RXHASHTAB_H_WRAP_

#include <string.h>

class RxHashNodeBase {
	friend class RxHashTabBase;
public:
	RxHashNodeBase *next;
	virtual ~RxHashNodeBase()
	 {
	 }
	virtual int match(void *) const
	 { return 0;
	 }
};

class RxHashTabBase
{
protected:
	int cur_idx;
	RxHashNodeBase *cur_pair;
	int size_;
	RxHashNodeBase **table;
	int numEnts;

	unsigned int hashStr(const char *s) const;
	unsigned int hashPtr(const void *ptr) const
	 { return ((unsigned int)ptr) % size_;
	 }
	unsigned int hashInt(int i) const
	 { return (i % size_);
	 }
	RxHashNodeBase *getFirst();
	RxHashNodeBase *getNext();
	RxHashNodeBase *lookup(int idx, void *p) const;
	int remove_ (int idx, void *p);
	RxHashNodeBase *insert_ (int idx, RxHashNodeBase *pair);
public:
	RxHashTabBase(int size);
	void clear();
	~RxHashTabBase()
	 { clear();
	   delete table;
	 }
	int numEntries()
	 { return numEnts;
	 }
	int size()
	 { return (size_);
	 }
};

#define RX_HASH_TAB_DEFS(HP,KEYTYPE,DATATYPE)		\
public:							\
	int remove (KEYTYPE s)				\
	 { return remove_ (hash(s), (void*)s);		\
	 }						\
	HP *find (KEYTYPE s) const			\
	 { return (HP*)lookup (hash(s), (void*)s);	\
	 }						\
	HP *insert (KEYTYPE s, DATATYPE p)		\
	 { int idx = hash(s);				\
	   HP *pair = new HP (s, p);			\
	   return ((HP*)insert_ (idx, pair));		\
	 }						\
	HP *get (KEYTYPE s, DATATYPE p)			\
	 { int idx = hash(s);				\
	   HP *pair = (HP*)lookup (idx, (void*)s);	\
	   if (pair == NULL)				\
	    { pair = new HP (s, p);			\
	      return ((HP*)insert_ (idx, pair));	\
	    }						\
	   else						\
	    { return pair;				\
	    }						\
	 }						\
	HP *first()					\
	 { return (HP*)getFirst();			\
	 }						\
	HP *next()					\
	 { return (HP*)getNext();			\
	 }

class RxStrHashNode : public RxHashNodeBase
{
	char *key_;
	void *ptr_;
public:
	RxStrHashNode (const char *key, void *ptr)
	 { set (key, ptr);
	 }
	void set (const char *key, void *ptr)
	 { key_ = strdup(key);
	   ptr_ = ptr;
	 }
	void setKey(char *key)
	 { key_ = key;
	 }
	void setPtr(void *ptr)
	 { ptr_ = ptr;
	 }
	const char *key() const
	 { return key_;
	 }
	void *ptr() const
	 { return ptr_;
	 }
	int match(void *s) const
	 { return (strcmp(key_,(char*)s) == 0);
	 }
};

class RxStrHashTab : public RxHashTabBase
{
public:
	RxStrHashTab(int size=100) : RxHashTabBase(size) { }
	int hash(const char *s) const
	 { return (hashStr(s));
	 }
	RX_HASH_TAB_DEFS(RxStrHashNode, const char *, void *);
};

class RxPtrHashNode : public RxHashNodeBase
{
	const void *key_;
	void *ptr_;
public:
	RxPtrHashNode (const void *key, void *ptr)
	 { set (key, ptr);
	 }
	void set (const void *key, void *ptr)
	 { key_ = key;
	   ptr_ = ptr;
	 }
	void setKey(void *key)
	 { key_ = key;
	 }
	void setPtr(void *ptr)
	 { ptr_ = ptr;
	 }
	const void *key() const
	 { return key_;
	 }
	void *ptr() const
	 { return ptr_;
	 }
	int match(void *key) const
	 { return (key_ == key);
	 }
};

class RxPtrHashTab : public RxHashTabBase
{
public:
	RxPtrHashTab(int size=100) : RxHashTabBase(size) { }
	int hash(const void *key) const
	 { return (hashPtr(key));
	 }
	RX_HASH_TAB_DEFS(RxPtrHashNode, const void *, void *);
};

#endif
