#ifndef _RXPRINTSPEC_H_WRAP_
#define _RXPRINTSPEC_H_WRAP_

#include <RxSlist.h>
#include <stdlib.h>
#include <string.h>
#include <stringx.h>

class RxFmtNode : public RxSnodeBase
{
	friend class RxPrintSpec;

    private:
	char fmtIsAlloced;
	char type_;
	char *fmt_;
	int width_;

    public:
	enum {
	  CONV,
	  STR
	  };

	RxFmtNode (const char *s, int n, int type);
	RxFmtNode (const char *s1, const char *s2, int type);
	RxFmtNode (const RxFmtNode &node);
	~RxFmtNode()
	 { if (fmt_)
	    { free (fmt_);
	    }
	 };
	char type() const
	 { return type_;
	 }
	
	int width() const
	 { return width_;
	 }
	char *fmt() const
	 { return fmt_;
	 }

	RX_SNODE_DECLS(RxFmtNode);
	const RxFmtNode *nextConv() const;
	const RxFmtNode *nextStr() const;
};

typedef RxSlist<RxFmtNode> RxFmtList;

class RxPrintSpec
{
	void init(int leadspc);
	void clear();

	RxFmtNode *prefixNode_;
	RxFmtNode *suffixNode_;
	RxFmtList fmtList;
	RxFmtList specialList;
	int leadSpacing;
	int leadSpacingBase;
	int nconv;
	int nstr;
	int nspecial;

	RxFmtNode *addConvNode (const char *s, int n);
	RxFmtNode *addStrNode (const char *s, int n);
	RxFmtNode *addSpecialNode (const char *s, int n);
		
public:
	int set (const char *s);
	int set (const RxFmtNode *n1, const RxFmtNode *n2);
	int set (const RxPrintSpec &ref);

	void clearPrefixAndSuffix();

	RxPrintSpec ()
	 { init(0);
	 };
	RxPrintSpec (const char *s, int leadspc=0)
	 { init(leadspc);
	   set (s);
	 }
	RxPrintSpec (const RxFmtNode *n1, const RxFmtNode *n2, int leadspc=0)
	 { init(leadspc);
	   set (n1, n2);
	 }	   
	~RxPrintSpec ();

	const RxFmtNode *fmtNode(int n) const;
	const RxFmtNode *convNode(int n) const;
	const RxFmtNode *strNode(int n) const;
	const RxFmtNode *specialNode(int n) const;

	const RxFmtNode *findSpecial (const char *s) const;

	int getLeadSpacing() const
	 { return leadSpacing;
	 }
	int setLeadSpacing(int n)
	 { int old = leadSpacing;
	   leadSpacing = n;
	   return old;
	 }
	int addLeadSpacing(int n)
	 { int old = leadSpacing;
	   leadSpacing += n;
	   return old;
	 }

	int numFmts()
	 { return nconv+nstr;
	 }
	int numSpecials()
	 { return nspecial;
	 }
	int numConvs()
	 { return nconv;
	 }

	const RxFmtNode *prefixNode() const
	 { return prefixNode_;
	 }
	const RxFmtNode *suffixNode() const
	 { return suffixNode_;
	 }

	const char *prefix() const
	 { return (prefixNode_ ? prefixNode_->fmt() : NULL);
	 }
	const char *suffix() const
	 { return (suffixNode_ ? suffixNode_->fmt() : NULL);
	 }
};

#endif
