#include <RxPrintable.h>
#include <RxCio.h>
#include <RxParseObj.h>

int RxPrintable::scan (FILE *fp)
{
	RxCinput ci(fp, /*useXscan=*/0);
	return scan (&ci);
}

int RxPrintable::xscan (FILE *fp)
{
	RxCinput ci(fp, /*useXscan=*/1);
	return scan (&ci);
}

int RxPrintable::scan (const char *buf, int *cntp)
{
	RxCinput ci(buf, cntp);
	return scan (&ci);
}

istream &RxPrintable::scan (istream &is)
{
	RxCinput ci(&is);
	int status;

	status = scan (&ci);
	if (status == EOF)
	 { is.clear (ios::eofbit);
	 }
	else if (status == 0)
	 { is.clear (ios::failbit);
	 }
	return is;
}

int RxPrintable::scan (class RxCinput *ci)
{
	RxParseObj *obj;

	obj = RxParseObj::scan (ci);
	if (obj == RxParseObj::eof)
	 { return EOF;
	 }
	else if (obj == NULL)
	 { return 0;
	 }
	else
	 { return oscan (obj);
	 }
}

int RxPrintable::print (FILE *fp, const char *s) const
{
	RxCoutput co(fp);
	RxPrintSpec fs(s);
	return cprint (&co, &fs);
}

int RxPrintable::print (char *buf, const char *s, int n) const
{
	RxCoutput co(buf, n);
	RxPrintSpec fs(s);
	return cprint (&co, &fs);
}

ostream &RxPrintable::print (ostream &os) const
{
	RxCoutput co(&os);
	RxPrintSpec fs("%g");	// fake print spec -- format will be ignored
	cprint (&co, &fs);
	return os;
}
