#include <RxTimeval.h>
#include <cdefs.h>
#include <stdio.h>

bail (int loc, char *s)
{
	printf ("%d::%s\n", loc, s);
	exit (1);
}

check (int loc, const RxTimeval &tv1, const RxTimeval &tv2)
{
	if (tv1 != tv2)
	 { bail (loc, "");
	 }
}

main()
{
	struct timeval ts;
	long s, u;
	double t;

	ts.tv_sec = 111;
	ts.tv_usec = 222;

	RxTimeval tv1;

	tv1 = 111.000222;
	tv1.get(s,u);
	printf ("s=%d u=%d\n", s, u);
	tv1.setZero();

	check (100, tv1, RxTimeval(0,0));
	check (101, RxTimeval(ts), RxTimeval(111,222));
	check (102, RxTimeval(111.000222), RxTimeval(111,222));

	RxTimeval tvx;

	tvx.set(111,222);
	tvx.get(s,u);
	if (s != 111 || u != 222)
	 { bail (103, "get");
	 }
	tvx.setZero();
	check (104, tvx, RxTimeval(0,0));

	tvx.set(111,222);

	RxTimeval tvy;
	tvy.set(111,222);

	if (tvx != tvy)
	 { bail (105, "cmp");
	 }
	if (tvx < tvy)
	 { bail (106, "cmp");
	 }
	if (tvx > tvy)
	 { bail (107, "cmp");
	 }
	if (!(tvx == tvy))
	 { bail (108, "cmp");
	 }
	if (!(tvx >= tvy))
	 { bail (109, "cmp");
	 }
	if (!(tvx <= tvy))
	 { bail (110, "cmp");
	 }

	tvy.set (111,111);

	if (!(tvx != tvy))
	 { bail (111, "cmp");
	 }
	if (tvx < tvy)
	 { bail (112, "cmp");
	 }
	if (!(tvx > tvy))
	 { bail (113, "cmp");
	 }
	if (tvx == tvy)
	 { bail (114, "cmp");
	 }
	if (!(tvx >= tvy))
	 { bail (115, "cmp");
	 }
	if (tvx <= tvy)
	 { bail (116, "cmp");
	 }

	tvy.set (111,333);

	if (!(tvx != tvy))
	 { bail (117, "cmp");
	 }
	if (!(tvx < tvy))
	 { bail (118, "cmp");
	 }
	if (tvx > tvy)
	 { bail (119, "cmp");
	 }
	if (tvx == tvy)
	 { bail (120, "cmp");
	 }
	if (tvx >= tvy)
	 { bail (121, "cmp");
	 }
	if (!(tvx <= tvy))
	 { bail (122, "cmp");
	 }

	check (130, tvx.set(-100, -999999), tvy.set (-100, -999999));
	check (131, tvx.set(-100, -1000000), tvy.set (-101, 0));
	check (132, tvx.set(-100, -1000013), tvy.set (-101, -13));
	check (133, tvx.set(-100, -20000013), tvy.set (-120, -13));

	check (134, tvx.set(100, -999999), tvy.set (99, 1));
	check (135, tvx.set(100, -1000000), tvy.set (99, 0));
	check (136, tvx.set(100, -1000010), tvy.set (98, 999990));
	check (137, tvx.set(100, -20000010), tvy.set (79, 999990));

	check (138, tvx.set(-100, 1), tvy.set (-99, -999999));
	check (139, tvx.set(-100, 1000000), tvy.set (-99, 0));
	check (140, tvx.set(-100, 1000010), tvy.set (-98, -999990));

	tvx = 111.000222;
	check (141, tvx, tvy.set(111,222));

	tvx = ts;
	check (142, tvx, tvy.set(111,222));
	
	ts = tvx;
	if (ts.tv_sec != 111 || ts.tv_usec != 222)
	 { bail (143, "assign");
	 }

	t = tvx;
	if (ABS(t-111.000222) > 1e-12)
	 { bail (144, "assign");
	 }

	check (145, tvx+tvx+tvx, RxTimeval(333,666));
	check (146, tvx-tvx+tvx, RxTimeval(111,222));
	check (147, 2*tvx-tvx+tvx*3, RxTimeval(444,888));
	check (148, (tvx+tvx)/2, RxTimeval(111,222));

	tvx *= 2;
	check (149, tvx, RxTimeval(222,444));
	tvx += tvx;
	check (150, tvx, RxTimeval(444,888));
	tvx /= 2;
	check (151, tvx, RxTimeval(222,444));
	tvx -= tvx;
	check (152, tvx, RxTimeval(0,0));
	
	tvx = -RxTimeval(100,200);
	check (153, tvx, RxTimeval(-100,-200));

	printf ("\nRoutines OK\n\n");
	return 0;
}
