#include <fastmath.h>
#include <cdefs.h>

extern RxAtan2TabEnt rxAtan2Tab[];

#define ATAN2_TAB_ORIGIN	(ATAN2_TAB_SIZE/2+1)

double ATAN2 (double y, double x)
{
	RxAtan2TabEnt *tab;
	double r;
	int idx;

	if (x == 0.0 && y == 0.0)
	 { return (atan2(0.0, 0.0));
	 }

	/* octants are labelled 1 to 8, counter clockwise from the +X axis */

	if (y >= 0.)			/* top half */
	 { if (x >= 0.)
	    { if (x >= y)		/* octant 1, take y over x */
	       { 
		 idx = (r = ATAN2_TAB_SCALE * (y/x));
		 tab = &rxAtan2Tab [ idx+ATAN2_TAB_ORIGIN ];
		 r -= idx;
		 return ( ((tab->a3*r + tab->a2)*r + tab->a1)*r + tab->a0 );	
	       }
	      else			/* octant 2, take x over y */
	       {
		 idx = (r = ATAN2_TAB_SCALE * (x/y));
		 tab = &rxAtan2Tab [ idx+ATAN2_TAB_ORIGIN ];
		 r -= idx;
		 return ( M_PI/2 - (((tab->a3*r + tab->a2)*r + tab->a1)*r + tab->a0) );
	       }
	    }
	   else
	    { if (y >= -x)		/* octant 3, take -x over y */
	       {
		 idx = (r = ATAN2_TAB_SCALE * (x/y));
		 tab = &rxAtan2Tab [ idx+ATAN2_TAB_ORIGIN ];
		 r -= idx;
		 return ( M_PI/2 - (((tab->a3*r + tab->a2)*r + tab->a1)*r + tab->a0) );
	       }
	      else			/* octant 4, take y over -x */
	       {
		 idx = (r = ATAN2_TAB_SCALE * (y/x));
		 tab = &rxAtan2Tab [ idx+ATAN2_TAB_ORIGIN ];
		 r -= idx;
		 return ( M_PI + (((tab->a3*r + tab->a2)*r + tab->a1)*r + tab->a0) );
	       }
	    }
	 }
	else
	 { if (x < 0.)
	    { if (-x >= -y)		/* octant 5, take -y over -x */
	       {
		 idx = (r = ATAN2_TAB_SCALE * (y/x));
		 tab = &rxAtan2Tab [ idx+ATAN2_TAB_ORIGIN ];
		 r -= idx;
		 return ( -M_PI + (((tab->a3*r + tab->a2)*r + tab->a1)*r + tab->a0) );
	       }
	      else			/* octant 6, take -x over -y */
	       {
		 idx = (r = ATAN2_TAB_SCALE * (x/y));
		 tab = &rxAtan2Tab [ idx+ATAN2_TAB_ORIGIN ];
		 r -= idx;
		 return ( -M_PI/2 - (((tab->a3*r + tab->a2)*r + tab->a1)*r + tab->a0) );
	       }
	    }
	   else
	    { if (-y >= x)		/* octant 7, take x over -y */
	       {
		 idx = (r = ATAN2_TAB_SCALE * (x/y));
		 tab = &rxAtan2Tab [ idx+ATAN2_TAB_ORIGIN ];
		 r -= idx;
		 return ( -M_PI/2 - (((tab->a3*r + tab->a2)*r + tab->a1)*r + tab->a0) );
	       }
	      else			/* octant 8, take -y over x */
	       {
		 idx = (r = ATAN2_TAB_SCALE * (y/x));
		 tab = &rxAtan2Tab [ idx+ATAN2_TAB_ORIGIN ];
		 r -= idx;
		 return (((tab->a3*r + tab->a2)*r + tab->a1)*r + tab->a0);
	       }
	    }
	 }
}
