#include <fastmath.h>
#include <cdefs.h>

extern RxSincosTabEnt rxSincosTab[];

double SIN (double a)
{
	RxSincosTabEnt *tab;
	double x;
	int idx;

	idx = (x = SINCOS_TAB_SCALE*a);

	if (x >= 0)
	 { tab = &rxSincosTab [(idx & (SINCOS_TAB_SIZE-1))+2];
	   x -= idx;
	 }
	else
	 { tab = &rxSincosTab [(idx & (SINCOS_TAB_SIZE-1))+1];
	   x -= (idx-1);
	 }
	return (((tab->a3*x + tab->a2)*x + tab->a1)*x + tab->a0);
}

#define COS_OFFSET	SINCOS_TAB_SIZE/4

double COS (double a)
{
	RxSincosTabEnt *tab;
	double x;
	int idx;

	idx = (x = SINCOS_TAB_SCALE*a);

	if (x >= 0)
	 { tab = &rxSincosTab [((idx+COS_OFFSET) & (SINCOS_TAB_SIZE-1))+2];
	   x -= idx;
	 }
	else
	 { tab = &rxSincosTab [((idx+COS_OFFSET) & (SINCOS_TAB_SIZE-1))+1];
	   x -= (idx-1);
	 }
	return (((tab->a3*x + tab->a2)*x + tab->a1)*x + tab->a0);
}

void SINCOS (double *sinp, double *cosp, double a)
{
	RxSincosTabEnt *tab;
	double x;
	int idx;

	idx = (x = SINCOS_TAB_SCALE*a);

	if (x >= 0)
	 { tab = &rxSincosTab [(idx & (SINCOS_TAB_SIZE-1))+2];
	   x -= idx;
	 }
	else
	 { tab = &rxSincosTab [(idx & (SINCOS_TAB_SIZE-1))+1];
	   x -= (idx-1);
	 }
	*sinp = ((tab->a3*x + tab->a2)*x + tab->a1)*x + tab->a0;
	if (tab >= &rxSincosTab[3*SINCOS_TAB_SIZE/4])
	 { tab -= 3*SINCOS_TAB_SIZE/4;
	 }
	else
	 { tab += SINCOS_TAB_SIZE/4;
	 }
	*cosp = ((tab->a3*x + tab->a2)*x + tab->a1)*x + tab->a0;
}
