#include <RxTimeval.h>
#include <fastmath.h>

#include <stdio.h>
#include <cdefs.h>
#include <stdlib.h>

/* 1.5e-14 */

char **argvec;

int interactive = 0;

#define NRANDOM	10000
#define TEST_RNG	20*M_PI

void checkError (double *maxerrp, double a1, double a2)
{
	if (ABS(a1-a2) > *maxerrp)
	 { *maxerrp = ABS(a1-a2);
	 }
}

main(int argc, char **argv)
{
	double cnt= 100000;
	double err, maxerr;
	double a;
	double s;
	double c;
	int i;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (ARGUMENT("-i"))
	    { interactive = 1;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s [-i]\n", argvec[0]);
	   exit (1);
	 }

	if (interactive)
	 { while (1)
	    { printf ("> ");
	      scanf ("%lf", &a);
	      printf ("SIN(x)=%19.16f\n", SIN(a));
	      printf ("sin(x)=%19.16f\n", sin(a));
	      printf ("COS(x)=%19.16f\n", COS(a));
	      printf ("cos(x)=%19.16f\n", cos(a));
	    }
	 }
	else
	 {
	   maxerr = 0;
	   for (i=-SINCOS_TAB_SIZE; i<SINCOS_TAB_SIZE+2; i++)
	    { a = i/SINCOS_TAB_SCALE;
	      checkError (&maxerr, sin(a), SIN(a));
	      a = (i+0.5)/SINCOS_TAB_SCALE;
	      checkError (&maxerr, sin(a), SIN(a));
	    }
	   for (i=0; i<NRANDOM; i++)
	    { a = 2*TEST_RNG*(drand48()-0.5);
	      checkError (&maxerr, sin(a), SIN(a));
	    }
	   printf ("max SIN error = %g\n", maxerr);
	   
	   maxerr = 0;
	   for (i=-SINCOS_TAB_SIZE; i<SINCOS_TAB_SIZE+2; i++)
	    { a = i/SINCOS_TAB_SCALE;
	      checkError (&maxerr, cos(a), COS(a));
	      a = (i+0.5)/SINCOS_TAB_SCALE;
	      checkError (&maxerr, cos(a), COS(a));
	    }
	   for (i=0; i<NRANDOM; i++)
	    { a = 2*TEST_RNG*(drand48()-0.5);
	      checkError (&maxerr, cos(a), COS(a));
	    }
	   printf ("max COS error = %g\n", maxerr);
	   
	   maxerr = 0;
	   for (i=-SINCOS_TAB_SIZE; i<SINCOS_TAB_SIZE+2; i++)
	    { a = i/SINCOS_TAB_SCALE;
	      SINCOS (&s, &c, a);
	      checkError (&maxerr, cos(a), c);
	      checkError (&maxerr, sin(a), s);
	      a = (i+0.5)/SINCOS_TAB_SCALE;
	      SINCOS (&s, &c, a);
	      checkError (&maxerr, cos(a), c);
	      checkError (&maxerr, sin(a), s);
	    }
	   for (i=0; i<NRANDOM; i++)
	    { a = 2*TEST_RNG*(drand48()-0.5);
	      SINCOS (&s, &c, a);
	      checkError (&maxerr, cos(a), c);
	      checkError (&maxerr, sin(a), s);
	    }
	   printf ("max SINCOS error = %g\n", maxerr);
	   
	   TIMEIT (s = SIN(M_PI/3), cnt);
	   TIMEIT (s = sin(M_PI/3), cnt);
	   TIMEIT (c = COS(M_PI/3), cnt);
	   TIMEIT (c = cos(M_PI/3), cnt);
	   TIMEIT (SINCOS(&s, &c, M_PI/3), cnt);
	   TIMEIT ((c = cos(M_PI/3), s =sin(M_PI/3)), cnt);
	   TIMEIT (a = atan2(1.1,2.2), cnt);
	 }
}
