#include <RxCvxPolygon2.h>
#include <stdio.h>
#include <stdiox.h>
#include <stdlib.h>
#include <cdefs.h>
#include <useful.h>
#ifdef HAS_GRF4
#include <grf4.h>
#endif
#include <math.h>
#include <RxTimeval.h>

#define PREC (1e-12)
#define EPS  (1e-7)
char **argvec;

typedef struct {
	double lam[2];
	int n;
} LineIsec;

typedef struct {
	int com;
	RxLine2 hp;
	RxCvxPolygon2 testPoly;
	LineIsec testIsec;	
	RxCvxPolygon2 *testPolyPtr;
	RxCvxPolygon2 poly1;
	RxCvxPolygon2 poly2;
} CommandRec;

CommandRec comRec;
RxCvxPolygon2 workPoly;
RxCvxPolygon2 polyRes;
LineIsec workIsec;
char *testFile = "RxCvxPolygon2Test.txt";

#define DO_QUIT		0
#define DO_ADD		1
#define DO_ISECT	2
#define DO_RESET	3
#define DO_INTERSECT	4

#ifdef HAS_GRF4
int useGfx = 0;
int timeTest = 0;
int winw = 400;
int winh = 400;
double xrange = 200.0;
double yrange = 200.0;
double tickSpacing = 10;
double tickHeight = 5;

volatile int cmdFlag = 0;
volatile int oldCmdFlag;

void drawAxes ()
{
	double x, y;

	grf4Move2 (-xrange/2, 0);
	grf4Draw2 ( xrange/2, 0);
	grf4Move2 (0, -yrange/2);
	grf4Draw2 (0,  yrange/2);
	for (x=tickSpacing; x<xrange; x+=tickSpacing)
	 { grf4Move2 ( x, -tickHeight);
	   grf4Draw2 ( x,  tickHeight);
	   grf4Move2 (-x, -tickHeight);
	   grf4Draw2 (-x,  tickHeight);
	 }
	for (y=tickSpacing; y<yrange; y+=tickSpacing)
	 { grf4Move2 (-tickHeight,  y);
	   grf4Draw2 ( tickHeight,  y);
	   grf4Move2 (-tickHeight, -y);
	   grf4Draw2 ( tickHeight, -y);
	 }
}

RxVecd2 computePnt (const RxVecd2 &p, const RxVecd2 &u, int dir)
{
#define MAG	(10*MAX(xrange,yrange))
	return (p + MAG*dir*u);
#undef MAG
}

int equalIsec (LineIsec *isec1, LineIsec *isec2)
{
	double diff;
	int i;

	if (isec1->n != isec2->n)
	 { return 0;
	 }
	for (i=0; i<isec1->n; i++)
	 { diff = isec1->lam[i] - isec2->lam[i];
	   if (ABS(diff) > EPS)
	    { return 0;
	    }
	 }
}
	
int xscanIsec (FILE *fp, LineIsec *isec)
{
	char *checkp;
	char buf[32];
	int i;

	if (xscanf (fp, "%d:", &isec->n) != 1)
	 { return -1;
	 }
	if (isec->n < 0 || isec->n > 2)
	 { return -1;
	 }
	for (i=0; i<isec->n; i++)
	 { if (xscanf (fp, "%s", buf) != 1)
	    { return -1;
	    }
	   if (strcmp (buf, "-I") == 0)
	    { isec->lam[i] = -HUGE_VAL;
	    }
	   else if (strcmp (buf, "+I") == 0)
	    { isec->lam[i] = HUGE_VAL;
	    }
	   else
	    { isec->lam[i] = strtod (buf, &checkp);
	      if (checkp == buf)
	       { return -1;
	       }
	    }
	 }
	return 1;
}

int fprintIsec (FILE *fp, LineIsec *isec)
{
	int i;

	fprintf (fp, "%d: ", isec->n);
	for (i=0; i<isec->n; i++)
	 { if (isec->lam[i] == -HUGE_VAL)
	    { fprintf (fp, "-I");
	    }
	   else if (isec->lam[i] == HUGE_VAL)
	    { fprintf (fp, "+I");
	    }
	   else
	    { fprintf (fp, "%13.8f", isec->lam[i]);
	    }
	   if (i<isec->n-1)
	    { fprintf (fp, " ");
	    }
	 }
	return 1;
}

void drawIsec (RxLine2 &hp, LineIsec *isec)
{
	RxVecd2 pnt;
	int infty;

	if (isec->n == 0)
	 { return;
	 }
	infty = (isec->lam[0] == HUGE_VAL ? 1 : 
		 isec->lam[0] == -HUGE_VAL ? -1 : 0);
	if (infty)
	 { pnt = computePnt (hp.q(), hp.u(), infty);
	 }
	else
	 { pnt = hp.q() + isec->lam[0]*hp.u();
	 }
	if (isec->n == 1)
	 { grf4Point2 (pnt.x(), pnt.y());
	   return;
	 }
	grf4Move2 (pnt.x(), pnt.y());

	infty = (isec->lam[1] == HUGE_VAL ? 1 : 
		 isec->lam[1] == -HUGE_VAL ? -1 : 0);
	if (infty)
	 { pnt = computePnt (hp.q(), hp.u(), infty);
	 }
	else
	 { pnt = hp.q() + isec->lam[1]*hp.u();
	 }
	grf4Draw2 (pnt.x(), pnt.y());
}

void drawPolytope (RxCvxPolygon2 *poly)
{
	RxEdge2 *ep;
	RxVrtx2 *vp;

	if ((ep = poly->headEdge()))
	 { vp = ep->nextVrtx();
	   grf4Move2 (vp->q().x(), vp->q().y());
	   do
	    { ep = ep->next();
	      vp = ep->nextVrtx();
	      grf4Draw2 (vp->q().x(), vp->q().y());
	    }
	   while (ep != poly->headEdge());
	 }
}

void drawCommand (CommandRec *rec)
{
	RxVecd2 pnt;

	switch (rec->com)
	 { case DO_ADD:
	    { pnt = computePnt (rec->hp.q(), rec->hp.u(), 1);
	      grf4Move2 (pnt.x(), pnt.y());
	      pnt = computePnt (rec->hp.q(), rec->hp.u(), -1);
	      grf4Draw2 (pnt.x(), pnt.y());
	      drawPolytope (&workPoly);
	      break;
	    }
	   case DO_ISECT:
	    { pnt = computePnt (rec->hp.q(), rec->hp.u(), 1);
	      grf4Move2 (pnt.x(), pnt.y());
	      pnt = computePnt (rec->hp.q(), rec->hp.u(), -1);
	      grf4Draw2 (pnt.x(), pnt.y());
	      drawPolytope (&workPoly);
	      break;
	    }
	   case DO_RESET:
	    {
	      break;
	    }
	   case DO_INTERSECT:
	    { drawPolytope (&rec->poly1);
	      drawPolytope (&rec->poly2);
	      break;
	    }
	 }
}

void doDraw(void *)
{
	grf4Clear();
	drawAxes ();
	if ((oldCmdFlag % 2) == 0)
	 { drawCommand (&comRec);
	 }
	else
	 { switch (comRec.com)
	    { case DO_ADD:
	       { drawPolytope (&workPoly);
		 break;
	       }
	      case DO_ISECT:
	       { drawIsec (comRec.hp, &workIsec);
		 break;
	       }
	      case DO_INTERSECT:
	       { drawPolytope (&polyRes);
		 break;
	       }
	    }
	 }
	grf4Flush();
}

void onDo (int n, int code, long)
{
	int flags;

	if (code == BF_PRESSED)
	 { cmdFlag++;
	   flags = grf4GetButtonFlags (n);
	   if ((cmdFlag%2)==0)
	    { grf4SetButtonFlags (n, flags & ~(BF_ILLUMINATED));
	    }
	   else
	    { grf4SetButtonFlags (n, flags | BF_ILLUMINATED);
	    }
	 }
}

#endif

RxCvxPolygon2 *getNamedPoly (const char *name, CommandRec *rec)
{
	if (strcmp (name, "poly1") == 0)
	 { return (&rec->poly1);
	 }
	else if (strcmp (name, "poly2") == 0)
	 { return (&rec->poly2);
	 }
	else if (strcmp (name, "res") == 0)
	 { return (&polyRes);
	 }
	else
	 { return (NULL);
	 }
}

const char *getPolyName (const RxCvxPolygon2 *poly, CommandRec *rec)
{
	if (poly == &rec->poly1)
	 { return ("poly1");
	 }
	else if (poly == &rec->poly2)
	 { return ("poly2");
	 }
	else if (poly == &polyRes)
	 { return ("res");
	 }
	else
	 { return (NULL);
	 }
}

void readTestPoly (FILE *fp, CommandRec *rec)
{
	char buf[32];

	if (xscanf (fp, "%s", buf) != 1)
	 { fprintf (stderr,
"Error reading test polytope description, line %d\n", xscanfLine (fp));
	   exit (1);
	 }
	if ((rec->testPolyPtr = getNamedPoly (buf, rec)) == NULL)
	 { fprintf (stderr,
"Bogus test polytope description %s, line %d\n", buf, xscanfLine (fp));
	   exit (1);
	 }
	if (rec->testPolyPtr == &polyRes)
	 { if (rec->testPoly.xscan (fp) != 1)
	    { fprintf (stderr,
"Error reading test polytope, line %d\n", xscanfLine (fp));
	      exit (1);
	    }
	 }
	else
	 { rec->testPoly.set (*rec->testPolyPtr);
	 }
}

void readCommand (FILE *fp, CommandRec *rec)
{
	char buf[64];
	int n;

	if ((n = xscanf (fp, "%s", buf)) < 0)
	 { if (n == EOF)
	    { rec->com = DO_QUIT;
	      return;
	    }
	   else
	    { fprintf (stderr, 
"Error reading command, line %d\n", xscanfLine (fp));
	      exit (1);
	    }
	 }
	if (strcmp (buf, "ADD") == 0)
	 { rec->com = DO_ADD;
	   if (rec->hp.xscan (fp) != 1)
	    { fprintf (stderr,
"Error reading half plane, line %d\n", xscanfLine (fp));
	      exit (1);
	    }
	   if (rec->testPoly.xscan (fp) != 1)
	    { fprintf (stderr,
"Error reading test polytope, line %d\n", xscanfLine (fp));
	      exit (1);
	    }
	 }
	else if (strcmp (buf, "INTERSECT") == 0)
	 { rec->com = DO_INTERSECT;
	   if (rec->poly2.xscan (fp) != 1)
	    { fprintf (stderr,
"Error reading poly2, line %d\n", xscanfLine (fp));
	      exit (1);
	    }
	   readTestPoly (fp, rec);
	   rec->com = DO_INTERSECT;
	 }
	else if (strcmp (buf, "ISECT") == 0)
	 { char buf1[32], buf2[32];
	   
	   rec->com = DO_ISECT;
	   if (rec->hp.xscan (fp) != 1)
	    { fprintf (stderr,
"Error reading half plane, line %d\n", xscanfLine (fp));
	      exit (1);
	    }
	   if (xscanIsec (fp, &rec->testIsec) < 0)
	    { fprintf (stderr,
"Error reading test isec, line %d\n", xscanfLine (fp));
	      exit (1);
	    }
	 }
	else if (strcmp (buf, "RESET") == 0)
	 { if (workPoly.xscan (fp) != 1)
	    { fprintf (stderr,
"Error reading woirk polytope, line %d\n", xscanfLine (fp));
	      exit (1);
	    }	
	   rec->poly1.set (workPoly);
	   rec->com = DO_RESET;
	 }
	else
	 { fprintf (stderr,
"Bad command %s, line %d\n", buf, xscanfLine (fp));
	   exit (1);
	 }
}

printCommand (CommandRec *rec)
{
	switch (rec->com)
	 { case DO_ADD:
	    { rec->hp.print ("ADD %12.7f %12.9f\n");
	      break;
	    }
	   case DO_ISECT:
	    { rec->hp.print ("ISECT %12.7f %12.9f\n");
	      break;
	    }
	   case DO_RESET:
	    { printf ("RESET\n");
	      workPoly.print ("%12.7f %12.9f\n");
	      break;
	    }
	   case DO_INTERSECT:
	    { printf ("INTERSECT\n");
	      rec->poly2.print ("%12.7f\n");
	      break;
	    }
	 }
}

int doCommand (CommandRec *rec, int check, int line)
{
	switch (rec->com)
	 { case DO_ADD:
	    { 
	      if (workPoly.consistencyCheck (/*print=*/1) == 0)
	       { printf ("Error: ADD, near line %d: bogus input\n",line);
		 workPoly.print ("   Got: %12.7f");
		 return 0;
	       }
	      workPoly.intersect (rec->hp);
	      if (check)
	       {
		 if (workPoly.consistencyCheck (/*print=*/1) == 0)
		  { printf ("Error: ADD, near line %d: bogus result\n",line);
		    rec->testPoly.print ("Wanted: %12.7f %12.9f");
		    workPoly.print ("   Got: %12.7f");
		    return (0);
		  }
		 if (workPoly.equal (rec->testPoly, PREC) == 0)
		  { printf ("Error ADD, near line %d:\n", line);
		    rec->testPoly.print ("Wanted: %12.7f %12.9f");
		    workPoly.print ("   Got: %12.7f");
		    return (0);
		  }
	       }
	      else
	       { workPoly.print ("%12.7f");
	       }
	      break;
	    }
	   case DO_ISECT:
	    { workIsec.n = workPoly.intersectLine (workIsec.lam, rec->hp);
	      if (check)
	       { if (equalIsec (&workIsec, &rec->testIsec) == 0)
		  { printf ("Error in ISECT near line %d:\n", line);
		    printf ("Wanted: ");
		    fprintIsec (stdout, &comRec.testIsec);
		    printf ("\n   Got: ");
		    fprintIsec (stdout, &workIsec);
		    printf ("\n");
//		    return (0);
		  }
	       }
	      else
	       { fprintIsec (stdout, &workIsec);
		 printf ("\n");
	       }
	      break;
	      return (equalIsec (&workIsec, &rec->testIsec));
	    }
	   case DO_INTERSECT:
	    { if (check)
	       { int i, j;
		 RxCvxPolygon2 poly1Save;
		 RxCvxPolygon2 poly2Save;
		 RxCvxPolygon2 xformTestPoly;
		 RxCvxPolygon2 *resPtr;
		 double ang = M_PI/3;

		 poly1Save.set (rec->poly1);
		 poly2Save.set (rec->poly2);

		 if (rec->poly1.consistencyCheck (/*print=*/1) == 0)
		  { printf ("Error: ADD, near line %d: bogus input 1\n",line);
		    rec->poly1.print ("   Got: %12.7f");
		    return 0;
		  }
		 if (rec->poly2.consistencyCheck (/*print=*/1) == 0)
		  { printf ("Error: ADD, near line %d: bogus input 2\n",line);
		    rec->poly2.print ("   Got: %12.7f");
		    return 0;
		  }
#if 1
		 for (i=0; i<rec->poly1.numVtxs(); i++)
		  { if (i > 0) rec->poly1.shiftVtxs (1);
		    for (j=0; j<rec->poly2.numVtxs(); j++)
		     { if (j > 0) rec->poly2.shiftVtxs (1);
		       resPtr = polyRes.intersect (rec->poly1, rec->poly2);
		       if  (resPtr->equal (rec->testPoly, PREC) == 0)
			{ printf ("Error in INTERSECT (%d,%d) near line %d:\n",
				   i, j, line);
			  printf ("Wanted: ");
			  printf ("%s\n", getPolyName(rec->testPolyPtr,rec));
			  comRec.testPoly.print ("%12.7\n");
			  printf ("\n   Got: ");
			  printf ("%s\n", getPolyName(resPtr,rec));
			  polyRes.print ("%12.7\n");
			  printf ("\n");
			  rec->poly1.print ("poly1\n");
			  rec->poly2.print ("poly2\n");
			  return (0);
			}
		       if (polyRes.consistencyCheck (/*print=*/1) == 0)
			{ printf (
"Error: INTERSECT, near line %d: bogus result\n", line);
			  return (0);
			}
		     }
		  }
#endif
#if 1
		 rec->poly1.intersect (rec->poly2);
		 if (!rec->poly1.equal (rec->testPoly, PREC))
		  { printf ("Error in self INTERSECT near line %d:\n", line);
		    printf ("Wanted: ");
		    printf ("%s\n", getPolyName(rec->testPolyPtr,rec));
		    comRec.testPoly.print ("%12.7\n");
		    printf ("\n   Got: ");
		    printf ("%s\n", getPolyName(&rec->poly1,rec));
		    rec->poly1.print ("%12.7\n");
		    printf ("\n");
		    return (0);
		  }
		 xformTestPoly.xform (rec->testPoly, 0, 0, ang);
		 rec->poly1.xform (poly1Save, 0, 0, ang);
		 rec->poly2.xform (poly2Save, 0, 0, ang);

		 resPtr = polyRes.intersect (rec->poly1, rec->poly2);
#endif

#if 1
		 if (!polyRes.equal (xformTestPoly, PREC))
		  { printf ("Error in xform INTERSECT near line %d:\n", line);
		    printf ("Wanted: ");
		    printf ("%s\n", getPolyName(rec->testPolyPtr,rec));
		    xformTestPoly.print ("%12.7\n");
		    printf ("\n   Got: ");
		    printf ("%s\n", getPolyName(resPtr,rec));
		    polyRes.print ("%12.7\n");
		    printf ("\n");
		    rec->poly1.print ("poly1\n");
		    rec->poly2.print ("poly2\n");
		    return (0);
		  }
#endif
		 rec->poly1.set (poly1Save);
		 rec->poly2.set (poly2Save);
	       }
	      else
	       { polyRes.intersect (rec->poly1, rec->poly2);
		 polyRes.print ("%12.7\n");
	       }
	      break;
	    }
	 }	
	return 1;
}

main(int argc, char **argv)
{
	FILE *fp = NULL;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (ARGUMENT("-gfx"))
	    { 
#ifdef NO_GRAPHICS
	      fprintf (stderr, "Error: graphics support not compiled in\n");
	      exit (1);
#else
	      useGfx = 1;
#endif
	    }
	   else if (ARGUMENT ("-time"))
	    { timeTest = 1;
	    }
	   else if (ARGUMENT ("-dbg"))
	    { RxCvxPolygon2::setDebug (1);
	    }
	   else if (ARGUMENT ("-file"))
	    { argc--; argv++;
	      if (argc < 1)
	       { goto usage;
	       }
	      testFile = *argv;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s [-gfx] [-time] [-file <fileName>] [-dbg]\n", argvec[0]);
	   exit (1);
	 }

	if (testFile && strcmp(testFile,"-") != 0)
	 { if ((fp = fopen (testFile, "r")) == NULL)
	    { fprintf (stderr, "%s: can't open file %s\n", *argvec, testFile);
	      perror ("fopen");
	      exit (1);
	    }
	 }
	else
	 { fp = stdin;
	 }

	workPoly.setCharDist (xrange);

#ifdef HAS_GRF4
	if (useGfx)
	 { 
	   grf4AddButton ("do cmd", 0, 1, onDo, 0);
	   grf4PlaceWindow (200, 400, winw, winh);
	   grf4Init ("DAO", GRF_WHITE, GRF_BLACK, GRF_2DONLY);

	   grf4Ortho (-xrange/2, xrange/2, -yrange/2, yrange/2, 
		      -1.0, 1.0, /*aspectRatio=*/1);
	   grf4SetDrawFxn (doDraw, (void*)0);

	   doDraw(0);

	   oldCmdFlag = cmdFlag;
	   while (1)
	    { while (oldCmdFlag == cmdFlag)
	       { nap (0.1);
		 grf4Cycle();
	       }
	      while (oldCmdFlag < cmdFlag)
	       { if ((oldCmdFlag % 2) != 0)
		  { doCommand (&comRec, /*check=*/0, /*line=*/0);
		    switch (comRec.com)
		     { case DO_ADD:
			{ workPoly.print ("%12.7f %12.9f");
			  break;
			}
		       case DO_ISECT:
			{ fprintIsec (stdout, &workIsec);
			  printf ("\n");
			  break;
			}
		     }
		  }
		 else
		  { do
		     { readCommand (fp, &comRec);
		       if (comRec.com == DO_QUIT)
			{ exit (0);
			}
		       printCommand (&comRec);
		     }
		    while (comRec.com == DO_RESET);
		  }
		 doDraw (0);
		 oldCmdFlag++;
	       }
	    }
	 }
	else
#endif
	if (timeTest)
	 { RxCvxPolygon2 poly1;
	   RxCvxPolygon2 poly2;
	   int count = 3000;
	   RxEdge2 *ep;

	   /* polygon vertices for doing canonical intersection speed test */
	   static double vtx1_isect[] = { 
	    -25.0, 20.0,   25.0, 20.0,  25.0, 30.0,  0.0, 45.0, -25.0, 30.0
	   };
	   static double vtx2_isect[] = {
	      0.0,  0.0,   20.0, 20.0,  10.0, 40.0, -15.0, 40.0, -15.0,-15.0
	   };

	   /* polygon vertices for doing the inside speed test */
	   static double vtx1_inside[] = { 
	    -13.0, 18.0,   17.0, 18.0,  17.0, 23.0,    2.0, 38.0,  -13.0, 23.0
	   };
	   static double vtx2_inside[] = {
	      0.0,  0.0,   20.0, 20.0,  10.0, 40.0,  -15.0, 40.0,  -15.0,-15.0
	   };


	   RxCvxPolygon2::setDebug (0);

	   poly1.setVertices (vtx1_isect, ARRAY_SIZE(vtx1_isect)/2);
	   poly2.setVertices (vtx2_isect, ARRAY_SIZE(vtx2_isect)/2);
	   printf ("7 vertex intersection:\n");
	   TIMEIT(polyRes.intersect (poly1, poly2), count);

 TIMEIT(poly1.setVertices (vtx1_isect, ARRAY_SIZE(vtx1_isect)/2),count);

	   TIMEIT((ep=RxEdge2::alloc(), RxEdge2::free(ep)), 100000);

	   poly1.setVertices (vtx1_inside, ARRAY_SIZE(vtx1_inside)/2);
	   poly2.setVertices (vtx2_inside, ARRAY_SIZE(vtx2_inside)/2);
	   printf ("%d vertex polygon inside a %d vertex polygon:\n",
		 ARRAY_SIZE(vtx1_inside)/2, ARRAY_SIZE(vtx2_inside)/2);
	   TIMEIT(polyRes.intersect (poly1, poly2), count);
	 }
	else
	 { 
	   xscanfBegin (fp);
	   while (1)
	    { do
	       { readCommand (fp, &comRec);
		 if (comRec.com == DO_QUIT)
		  { printf ("\nROUTINES OK\n\n");
		    exit (0);
		  }
	       }
	      while (comRec.com == DO_RESET);
	      if (doCommand (&comRec, /*check=*/1, xscanfLine(fp)) == 0)
	       { printf ("here\n");
		 exit (1);
	       }
	    }
	 }
}
