#ifndef _RXLINE2_H_WRAP_
#define _RXLINE2_H_WRAP_

#include <RxVecd2.h>
#include <iostream.h>
#include <stdio.h>
#include <cdefs.h>

#define NORMALIZE_U	0

class RxLine2
{
	// 2 dimensional line

	double ulen;
	RxVecd2 u_;	// (normalized) vector in the direction of the line
	RxVecd2 q_;	// some point on the line

public:
	static const double EPS;

	// access 

#if 1
	RxVecd2 &u()
	 { return u_;
	 }
	RxVecd2 &q()
	 { return q_;
	 }
#endif
	const RxVecd2 &u() const
	 { return u_;
	 }
	const RxVecd2 &q() const
	 { return q_;
	 }

	// assignment

	RxLine2 &set (const RxLine2 &l1)
	 { u_ = l1.u_;
	   q_ = l1.q_;
	   ulen = l1.ulen;
	   return (*this);
	 }
	RxLine2& operator=(const RxLine2 &l1) 
	 { return (set(l1));
	 }
	RxLine2 &set (const RxVecd2 &q1, const RxVecd2 &u1);
	void get (RxVecd2 *q1, RxVecd2 *u1) const
	 { *u1 = u_;
	   *q1 = q_;
	 }
	// Plucker coordinates, where x1 = q() X u().
	RxLine2 &setPlucker (double x1, const RxVecd2 &u1);
	void getPlucker (double *x1, RxVecd2 *u1);

	RxLine2& set(double qx, double qy, double ux, double uy)
	 { u_.x() = ux;
	   u_.y() = uy;
	   q_.x() = qx;
	   q_.y() = qy;
	   return (*this);
	 }

	// Half plane representation, by which the line
	// is defined as all points p for which a.p = b
	RxLine2& setHP (const RxVecd2 &a, double b);
	void getHP (RxVecd2 *a, double *b);

	// constructors

	RxLine2 ()
	 {
	 }
	RxLine2 (const RxVecd2 &q1, const RxVecd2 &u1)
	 { set (q1, u1);
	 }
	RxLine2 (const RxVecd2 &a, double b)
	 { setHP (a, b);
	 }

	// get the normal

	void normal(RxVecd2 *n)
	 { n->x() =  u_.x();
	   n->y() = -u_.y();
	 }

	// distance computations

	double dist (const RxVecd2 &p1)
	 { double d = sdist(p1);
	   return d < 0 ? -d : d;
	 }
	double dist (RxVecd2 *p, const RxVecd2 &p1)
	 { double d = sdist (p, p1);
	   return d < 0 ? -d : d;
	 }
	// signed distance computation
	double sdist (const RxVecd2 &p1);
	double sdist (RxVecd2 *p, const RxVecd2 &p1);
//	double lineto (RxLine2 *lr, const RxVecd2 &p) const;

	double project (RxVecd2 *p, const RxVecd2 &p1) const;
	int side (const RxVecd2 &p, double tol=EPS) const;
	int sideQ (const RxVecd2 &p, double tol=EPS) const
	 { double x = u_.x()*(p.y()-q_.y()) - u_.y()*(p.x()-q_.x());
	   return (x > tol ? 1 : x < -tol ? -1 : 0);
	 }
	double intersect (RxVecd2 *p, const RxLine2 &l2, double tol=EPS) const;

	// I/O

	int scan()
	 { return this->scan(stdin); };
	int scan (FILE *fp);
	int xscan (FILE *fp);
	int scan (const char *buf, int *cntp=NULL);
	int print(const char *fmt) const
	 { return this->print(stdout, fmt); };
	int print(FILE *fp, const char *fmt) const;
	int print(char *buf, const char *fmt) const;

	ostream &print(ostream &os) const;
	istream &read(istream &is);

	// comparison

	int equal (const RxLine2 &l1, double tol=EPS) const;
};

inline ostream &operator<<(ostream &os, const RxLine2 &l1)
{
	return l1.print(os);
}

inline istream &operator>>(istream &is, RxLine2 &l1)
{
	return l1.read(is);
}

#endif
