#ifndef _RXLINE3_H_WRAP_
#define _RXLINE3_H_WRAP_

#include <RxVecd3.h>
#include <iostream.h>
#include <stdio.h>
#include <cdefs.h>

class RxLine3
{
	// 3 dimensional line

private:
	RxVecd3 u_;	// (normalized) vector in the direction of the line
	RxVecd3 q_;	// any point on the line

public:
	const double EPS = 1e-13;

	// access

	RxVecd3 &u()
	 { return u_;
	 }
	RxVecd3 &q()
	 { return q_;
	 }
	const RxVecd3 &u() const
	 { return u_;
	 }
	const RxVecd3 &q() const
	 { return q_;
	 }

	// assignment

	RxLine3 &set (const RxLine3 &l1)
	 { u_ = l1.u_;
	   q_ = l1.q_;
	   return (*this);
	 }
	RxLine3& operator=(const RxLine3 &l1) 
	 { return (set(l1));
	 }
	RxLine3 &set (const RxVecd3 &u1, const RxVecd3 &q1);
	void get (RxVecd3 *u1, RxVecd3 *q1) const
	 { *u1 = u_;
	   *q1 = q_;
	 }
	// Plucker coordinates, where x1 = q() X u().
	RxLine3 &setPlucker (const RxVecd3 &u1, const RxVecd3 &x1);
	void getPlucker (RxVecd3 *u1, RxVecd3 *x1) const;

	// constructors	

	RxLine3 ()
	 { 
	 }
	RxLine3 (const RxVecd3 &u1, const RxVecd3 &q1)
	 { set (u1, q1);
	 }

	// distance computation

	double dist (const RxVecd3 &p1) const;
	double dist (RxVecd3 *n, RxVecd3 *p, const RxVecd3 &p1, 
		     double tol=EPS) const;

	double dist (const RxLine3 &l1) const;
	double dist (RxVecd3 *n, RxVecd3 *p, const RxLine3 &l1, 
		     double tol=EPS) const;

	// I/O

	int scan()
	 { return this->scan(stdin); };
	int scan (FILE *fp);
	int xscan (FILE *fp);
	int scan (const char *buf, int *cntp=NULL);
	int print(const char *fmt) const
	 { return this->print(stdout, fmt); };
	int print(FILE *fp, const char *fmt) const;
	int print(char *buf, const char *fmt) const;

	ostream &print(ostream &os) const;
	istream &read(istream &is);

	// comparison

	int equal (const RxLine3 l1, double tol) const;
};

inline ostream &operator<<(ostream &os, const RxLine3 &l1)
{
	return l1.print(os);
}

inline istream &operator>>(istream &is, RxLine3 &l1)
{
	return l1.read(is);
}

#endif
