#include <RxSpatialAccel.h>
#include <RxRot.h>
#include <RxTrsf.h>
#include <RxMathError.h>
#include <math.h>
#include <cdefs.h>

const RxSpatialAccel RxSpatialAccel::ZERO    (0,0,0, 0,0,0);
const RxSpatialAccel RxSpatialAccel::UNIT_VX (1,0,0, 0,0,0);
const RxSpatialAccel RxSpatialAccel::UNIT_VY (0,1,0, 0,0,0);
const RxSpatialAccel RxSpatialAccel::UNIT_VZ (0,0,1, 0,0,0);
const RxSpatialAccel RxSpatialAccel::UNIT_WX (0,0,0, 1,0,0);
const RxSpatialAccel RxSpatialAccel::UNIT_WY (0,0,0, 0,1,0);
const RxSpatialAccel RxSpatialAccel::UNIT_WZ (0,0,0, 0,0,1);

RxSpatialAccel& RxSpatialAccel::set (double vx, double vy, double vz,
		       double wx, double wy, double wz)
{
	w().x() = wx;
	w().y() = wy;
	w().z() = wz;
	v().x() = vx;
	v().y() = vy;
	v().z() = vz;
	return (*this);
}

RxSpatialAccel& RxSpatialAccel::set (const double *dp)
{
	w().x() = dp[3];
	w().y() = dp[4];
	w().z() = dp[5];
	v().x() = dp[0];
	v().y() = dp[1];
	v().z() = dp[2];
	return (*this);
}

void RxSpatialAccel::get (double *vxp, double *vyp, double *vzp,
		   double *wxp, double *wyp, double *wzp)
{
	*wxp = v().x();
	*wyp = v().y();
	*wzp = v().z();
	*vxp = w().x();
	*vyp = w().y();
	*vzp = w().z();
}

void RxSpatialAccel::get (double *dp) const
{
	dp[3] = w().x();
	dp[4] = w().y();
	dp[5] = w().z();
	dp[0] = v().x();
	dp[1] = v().y();
	dp[2] = v().z();
}

double& RxSpatialAccel::operator()(int i)
{
	static double dum;

	switch (i)
	 { case 0:
	    { return (v().x());
	    }
	   case 1:
	    { return (v().y());
	    }
	   case 2:
	    { return (v().z());
	    }
	   case 3:
	    { return (w().x());
	    }
	   case 4:
	    { return (w().y());
	    }
	   case 5:
	    { return (w().z());
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxSpatialAccel::(%d): %K\n", i).raise();	      
	      return (dum);
	    }
	 }
}

double RxSpatialAccel::operator()(int i) const
{
	static double dum;

	switch (i)
	 { case 0:
	    { return (v().x());
	    }
	   case 1:
	    { return (v().y());
	    }
	   case 2:
	    { return (v().z());
	    }
	   case 3:
	    { return (w().x());
	    }
	   case 4:
	    { return (w().y());
	    }
	   case 5:
	    { return (w().z());
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxSpatialAccel::(%d): %K\n", i).raise();	      
	      return (dum);
	    }
	 }
}
