#include <RxSpatialInertia.h>
#include <assert.h>

#define EPS	(1e-10)

int testAdd (const RxSpatialInertia &mt, const RxSpatialInertia &m1, 
	     const RxSpatialInertia &m2, double eps)
{
	RxSpatialInertia mc;
	RxVecd3 com;	
	double m;

	m = m1.getMass() + m2.getMass();
	if (m == 0)
	 { mc.setZero();
	 }
	else
	 { 
	   mc.I() = m1.I() + m2.I();
	   com = (m1.getMass()*m1.getCom() + m2.getMass()*m2.getCom())/m;
	   mc.setMass (m);
	   mc.setCom (com);
	 }
	return (mt.equal(mc, eps));
}

int testSub (const RxSpatialInertia &mt, const RxSpatialInertia &m1, 
	     const RxSpatialInertia &m2, double eps)
{
	RxSpatialInertia mc;
	RxVecd3 com;
	double m;

	m = m1.getMass() - m2.getMass();
	if (m == 0)
	 { mc.setZero();
	 }
	else
	 { 
	   mc.I() = m1.I() - m2.I();
	   com = (m1.getMass()*m1.getCom() - m2.getMass()*m2.getCom())/m;
	   mc.setMass (m);
	   mc.setCom (com);
	 }
	return (mt.equal(mc, eps));
}

int testScale (const RxSpatialInertia &mt, double a, 
	       const RxSpatialInertia &m2, double eps)
{
	RxSpatialInertia mc;
	double m;

	m = m2.getMass()*a;
	if (m == 0)
	 { mc.setZero();
	 }
	else
	 { mc.I() = a*m2.I();
	   mc.setCom (m2.getCom());
	   mc.setMass (m);
	 }
	return (mt.equal(mc, eps));
}

typedef struct {
	  double x, y, z;
	  double roll, pitch, yaw;
} Pose;

Pose poseList[] = {
	0,   0, 0,     0, 0, 0,
	10, 20, 30,	0, 0, 0,
	11, 44,  0,    90, 20, 45,
       -3,  -9, -10,  -180, 30, 20
};

char *spec1 = "\
[  2.0 [ 10.0 11.0 12.0 ]\n\
  [ 32.0 33.0 34.0 35.0 36.0 37.0 ] ]";

char *spec2 = "\
[  0.0  24.0 -22.0   2.0   0.0   0.0\n\
 -24.0   0.0  20.0   0.0   2.0   0.0\n\
  22.0 -20.0   0.0   0.0   0.0   2.0\n\
  32.0  37.0  36.0   0.0 -24.0  22.0\n\
  37.0  33.0  35.0  24.0   0.0 -20.0\n\
  36.0  35.0  34.0 -22.0  20.0   0.0]";

main()
{
	RxSpatialInertia M1(RxMatd3x3(10,6,5,6,20,4,5,4,30), 10.0, RxVecd3 (1,2,3));
	RxSpatialInertia M2(RxMatd3x3(11,9,8,9,22,7,8,7,33), 44.0, RxVecd3 (3,2,1));
	RxSpatialInertia M3(RxMatd3x3(10,6,5,6,20,4,5,5,30), 1.0);
	RxSpatialInertia M4;
	RxSpatialInertia MU(RxMatd3x3(1,0,0,0,1,0,0,0,1), 1.0);
	RxSpatialInertia *testVals[] = { &M1, &M2, &M3, &MU };
	RxVecd3 com;
	double m;
	int p, q, i;

	// test equal, copy, set

	assert (MU.equal(RxSpatialInertia::UNIT, EPS));
	M4 = M1;
	assert (M4.equal(M1, EPS));
	RxSpatialInertia MT(M1);
	assert (MT.equal(M1, EPS));

	M4.set (M3.I(), M3.getMass(), M3.getCom());
	assert (!M4.equal(M1, EPS));

	// test basic arithemtic 

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { for (q=0; q<ARRAY_SIZE(testVals); q++)
	    { RxSpatialInertia ma(*testVals[p]);
	      RxSpatialInertia mb(*testVals[q]);
	      RxSpatialInertia mc;

	      assert (testAdd (ma+mb, ma, mb, EPS));
	      assert (testAdd (mc.add (ma,mb), ma, mb, EPS));

	      mc = ma;
	      assert (testAdd (mc.add (mb), ma, mb, EPS));
	      mc = ma;
	      assert (testAdd (mc += mb, ma, mb, EPS));

	      assert (testSub (ma-mb, ma, mb, EPS));
	      assert (testSub (mc.sub (ma,mb), ma, mb, EPS));
	      mc = ma;
	      assert (testSub (mc.sub (mb), ma, mb, EPS));
	      mc = ma;
	      assert (testSub (mc -= mb, ma, mb, EPS));

	      mc.scale (-1, mb);
	      assert (testSub (ma+mc, ma, mb, EPS));
	      mc.neg (mb);
	      assert (testSub (ma+mc, ma, mb, EPS));
	      mc = mb;
	      mc.neg();
	      assert (testSub (ma+mc, ma, mb, EPS));
	      assert (testSub (ma+(-mb), ma, mb, EPS));
	    }
	 }

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { RxSpatialInertia ma(*testVals[p]);
	   RxSpatialInertia mb, mc;

	   assert (testScale (ma*M_PI, M_PI, ma, EPS));
	   assert (testScale (M_PI*ma, M_PI, ma, EPS));
	   assert (testScale (ma/M_PI, 1/M_PI, ma, EPS));
	   assert (testScale (mc.scale(M_PI, ma), M_PI, ma, EPS));
	   mc = ma;
	   assert (testScale (mc.scale(M_PI), M_PI, ma, EPS));
	   mc = ma;			
	   assert (testScale (mc *= M_PI, M_PI, ma, EPS));
	   mc = ma;			
	   assert (testScale (mc /= M_PI, 1/M_PI, ma, EPS));

#if 0
	   assert (testInv (mc.inv(ma), ma, EPS));
	   assert (testInv (ma.I(), ma, EPS));
	   mb = ma;
	   assert (testInv (mb.inv(), ma, EPS));
#endif
	   for (i=0; i<ARRAY_SIZE(poseList); i++)
	    { RxTrsf T, TI;
	      
	      T.p.set (poseList[i].x, poseList[i].y, poseList[i].z);
	      T.R.setRpy (poseList[i].roll,poseList[i].pitch,poseList[i].yaw);
	      TI.inv(T);

	      mb = ma;
	      mc.xform (mb, T.R);
	      mb.xform (T.R);
	      assert (mb.equal (mc, EPS));

	      mb.Ixform (mc, T.R);
	      mc.Ixform (T.R);
	      assert (mb.equal (mc, EPS));
	      assert (ma.equal (mc, EPS));

	      mb.xform (ma, T.R);
	      mc.xform (mb, TI.R);
	      assert (ma.equal (mc, EPS));
	
	      mb = ma;
	      mc.xform (mb, T);
	      mb.xform (T);
	      assert (mb.equal (mc, EPS));

	      mb.Ixform (mc, T);
	      mc.Ixform (T);
	      assert (mb.equal (mc, EPS));
	      assert (ma.equal (mc, EPS));
	      assert (ma.equal (mc, EPS));
	    }		
	 }

	M4.setUnit();
	assert (M4.equal (RxSpatialInertia::UNIT, EPS));
	M4.setZero();
	assert (M4.equal (RxSpatialInertia::ZERO, EPS));

	// IO testing

	 { char buf[1024];
	   int n;

	   M4.scan (spec1);
	   n = M4.print (buf, "%4.1f");
	   assert (n == strlen (spec1));
	   assert (strcmp (spec1, buf) == 0);
	
	   n = M4.print (buf, "%M%5.1f");
	   assert (n == strlen (spec2));
	   assert (strcmp (spec2, buf) == 0);
	   
	   M4.scan (spec2);
	   n = M4.print (buf, "%M%5.1f");
	   assert (n == strlen (spec2));
	   assert (strcmp (spec2, buf) == 0);
	 }

	printf ("\nROUTINES OK\n\n");

}
