#include <RxSpatialVec.h>
#include <RxRot.h>
#include <RxTrsf.h>
#include <RxMathError.h>
#include <math.h>
#include <cdefs.h>

const RxSpatialVec RxSpatialVec::ZERO    (0,0,0, 0,0,0);

RxSpatialVec& RxSpatialVec::set (const RxSpatialVec& v2)
{
	a_.x() = v2.a_.x();
	a_.y() = v2.a_.y();
	a_.z() = v2.a_.z();
	b_.x() = v2.b_.x();
	b_.y() = v2.b_.y();
	b_.z() = v2.b_.z();
	return *this;
}

RxSpatialVec& RxSpatialVec::set (double ax, double ay, double az,
				 double bx, double by, double bz)
{
	a_.x() = ax;
	a_.y() = ay;
	a_.z() = az;
	b_.x() = bx;
	b_.y() = by;
	b_.z() = bz;
	return (*this);
}

RxSpatialVec& RxSpatialVec::set (const double *dp)
{
	a_.x() = dp[0];
	a_.y() = dp[1];
	a_.z() = dp[2];
	b_.x() = dp[3];
	b_.y() = dp[4];
	b_.z() = dp[5];
	return (*this);
}

void RxSpatialVec::get (double *axp, double *ayp, double *azp,
		   double *bxp, double *byp, double *bzp)
{
	*bxp = b_.x();
	*byp = b_.y();
	*bzp = b_.z();
	*axp = a_.x();
	*ayp = a_.y();
	*azp = a_.z();
}

void RxSpatialVec::get (double *dp) const
{
	dp[0] = a_.x();
	dp[1] = a_.y();
	dp[2] = a_.z();
	dp[3] = b_.x();
	dp[4] = b_.y();
	dp[5] = b_.z();
}

double& RxSpatialVec::operator()(int i)
{
	static double dum;

	switch (i)
	 { case 0:
	    { return (a_.x());
	    }
	   case 1:
	    { return (a_.y());
	    }
	   case 2:
	    { return (a_.z());
	    }
	   case 3:
	    { return (b_.x());
	    }
	   case 4:
	    { return (b_.y());
	    }
	   case 5:
	    { return (b_.z());
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxSpatialVec::(%d): %K\n", i).raise();	      
	      return (dum);
	    }
	 }
}

double RxSpatialVec::operator()(int i) const
{
	static double dum;

	switch (i)
	 { case 0:
	    { return (a_.x());
	    }
	   case 1:
	    { return (a_.y());
	    }
	   case 2:
	    { return (a_.z());
	    }
	   case 3:
	    { return (b_.x());
	    }
	   case 4:
	    { return (b_.y());
	    }
	   case 5:
	    { return (b_.z());
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxSpatialVec::(%d): %K\n", i).raise();	      
	      return (dum);
	    }
	 }
}

RxSpatialVec& RxSpatialVec::add(const RxSpatialVec &v1, const RxSpatialVec &v2)
{
	a_.x() = v1.a_.x() + v2.a_.x();
	a_.y() = v1.a_.y() + v2.a_.y();
	a_.z() = v1.a_.z() + v2.a_.z();

	b_.x() = v1.b_.x() + v2.b_.x();
	b_.y() = v1.b_.y() + v2.b_.y();
	b_.z() = v1.b_.z() + v2.b_.z();
	return (*this);
}

RxSpatialVec& RxSpatialVec::add(const RxSpatialVec &v2)
{
	a_.x() += v2.a_.x();
	a_.y() += v2.a_.y();
	a_.z() += v2.a_.z();

	b_.x() += v2.b_.x();
	b_.y() += v2.b_.y();
	b_.z() += v2.b_.z();
	return (*this);
}

RxSpatialVec& RxSpatialVec::sub(const RxSpatialVec &v1, const RxSpatialVec &v2)
{
	a_.x() = v1.a_.x() - v2.a_.x();
	a_.y() = v1.a_.y() - v2.a_.y();
	a_.z() = v1.a_.z() - v2.a_.z();

	b_.x() = v1.b_.x() - v2.b_.x();
	b_.y() = v1.b_.y() - v2.b_.y();
	b_.z() = v1.b_.z() - v2.b_.z();
	return (*this);
}

RxSpatialVec& RxSpatialVec::sub(const RxSpatialVec &v2)
{
	a_.x() -= v2.a_.x();
	a_.y() -= v2.a_.y();
	a_.z() -= v2.a_.z();

	b_.x() -= v2.b_.x();
	b_.y() -= v2.b_.y();
	b_.z() -= v2.b_.z();
	return (*this);
}

RxSpatialVec& RxSpatialVec::extrap(const RxSpatialVec &v1, double a, const RxSpatialVec &v2)
{
	a_.x() = v1.a_.x() + a*v2.a_.x();
	a_.y() = v1.a_.y() + a*v2.a_.y();
	a_.z() = v1.a_.z() + a*v2.a_.z();

	b_.x() = v1.b_.x() + a*v2.b_.x();
	b_.y() = v1.b_.y() + a*v2.b_.y();
	b_.z() = v1.b_.z() + a*v2.b_.z();
	return (*this);
}

RxSpatialVec& RxSpatialVec::interp(const RxSpatialVec &v1, double a, const RxSpatialVec &v2)
{
	a_.x() = (1-a)*v1.a_.x() + a*v2.a_.x();
	a_.y() = (1-a)*v1.a_.y() + a*v2.a_.y();
	a_.z() = (1-a)*v1.a_.z() + a*v2.a_.z();

	b_.x() = (1-a)*v1.b_.x() + a*v2.b_.x();
	b_.y() = (1-a)*v1.b_.y() + a*v2.b_.y();
	b_.z() = (1-a)*v1.b_.z() + a*v2.b_.z();
	return (*this);
}

RxSpatialVec& RxSpatialVec::cross(const RxSpatialVec &v1, const RxSpatialVec &v2)
{
	RxVecd3 ar, br, tmp;

	ar.cross(v1.a_, v2.a_);
	br.cross(v1.a_, v2.b_);
	tmp.cross (v1.b_, v2.a_);
	br.add (tmp);
	a_ = ar;
	b_ = br;
	return (*this);
}

RxSpatialVec& RxSpatialVec::scale(double a, const RxSpatialVec &v1)
{
	a_.x() = a*v1.a_.x();
	a_.y() = a*v1.a_.y();
	a_.z() = a*v1.a_.z();

	b_.x() = a*v1.b_.x();
	b_.y() = a*v1.b_.y();
	b_.z() = a*v1.b_.z();
	return (*this);
}

RxSpatialVec& RxSpatialVec::scale(double a)
{
	a_.x() *= a;
	a_.y() *= a;
	a_.z() *= a;

	b_.x() *= a;
	b_.y() *= a;
	b_.z() *= a;
	return (*this);
}

RxSpatialVec &RxSpatialVec::xform (const RxSpatialVec &v1, const RxRot &R)
{
	/*         
	   Let R be a rotation from frame A to frame B. Then spatial vector 
	   v1 = (a,b) is converted from A into B by multiplying each
	   component by the transpose of R.
	 */

	a_.xform (v1.a_, R);
	b_.xform (v1.b_, R);
	return (*this);
}

RxSpatialVec &RxSpatialVec::xform (const RxSpatialVec &v1, const RxTrsf &T)
{
	/*         [ R  p ]
	   Let T = [ 0  1 ] be the transform from frame A to frame B.

	   Then spatial vector v = (a,b) is converted from A into B via 
	   multiplication by the following adjoint transform:

			      T
			 [   R     0   ]       [  0  -pz  py ]
			 [             ]       [ 	     ]
		adj(T) = [   T      T  ],  P = [  pz  0  -px ]
			 [ -R  P   R   ]       [ 	     ]
					       [ -py  px  0  ]
	*/
	RxVecd3 k;

	/*		              T
	    Compute k = [ P  I ] (a b)
	 */

	k.x() = v1.a_.y() * T.p.z() - v1.a_.z() * T.p.y() + v1.b_.x();
	k.y() = v1.a_.z() * T.p.x() - v1.a_.x() * T.p.z() + v1.b_.y();
	k.z() = v1.a_.x() * T.p.y() - v1.a_.y() * T.p.x() + v1.b_.z();

	b_.xform (k, T.R);
	a_.xform (v1.a_, T.R);
	return (*this);
}

RxSpatialVec &RxSpatialVec::Ixform (const RxSpatialVec &v1, const RxRot &R)
{
	/*         
	   Let R be a rotation from frame A to frame B. Then spatial vector 
	   v1 = (a,b) is converted from B into A by multiplying each
	   component by R.
	 */

	a_.Ixform (v1.a_, R);
	b_.Ixform (v1.b_, R);
	return (*this);
}

RxSpatialVec &RxSpatialVec::Ixform (const RxSpatialVec &v1, const RxTrsf &T)
{
	/*         [ R  p ]
	   Let T = [ 0  1 ] be the transform from frame A to frame B.

	   Then a spatial vector v = (a,b) is converted from B into A via 
	   multiplication by the inverse adjoint transform:

			 [   R     0  ]       [  0  -pz  py ]
		 -1      [            ]       [ 	    ]
	       adj (T) = [            ],  P = [  pz  0  -px ]
			 [  P R    R  ]       [ 	    ]
					      [ -py  px  0  ]
	*/

	RxVecd3 k;

	a_.Ixform (v1.a_, T.R);
	k.Ixform (v1.b_, T.R);

	b_.x() = T.p.y() * a_.z() - T.p.z() * a_.y() + k.x();
	b_.y() = T.p.z() * a_.x() - T.p.x() * a_.z() + k.y();
	b_.z() = T.p.x() * a_.y() - T.p.y() * a_.x() + k.z();
	return (*this);
}

double RxSpatialVec::dot(const RxSpatialVec &v2) const
{
	return (a_.x()*v2.b_.x() + a_.y()*v2.b_.y() + a_.z()*v2.b_.z() +
		b_.x()*v2.a_.x() + b_.y()*v2.a_.y() + b_.z()*v2.a_.z());
}

RxSpatialVec &RxSpatialVec::unit(const RxSpatialVec &v1)
{
	a_.unit(v1.a_);
	b_.unit(v1.b_);
	return *this;
}

RxSpatialVec &RxSpatialVec::neg(const RxSpatialVec &v1)
{
	a_.x() = -v1.a_.x();
	a_.y() = -v1.a_.y();
	a_.z() = -v1.a_.z();

	b_.x() = -v1.b_.x();
	b_.y() = -v1.b_.y();
	b_.z() = -v1.b_.z();
	return (*this);
}

RxSpatialVec &RxSpatialVec::setZero()
{
	a_.x() = 0;
	a_.y() = 0;
	a_.z() = 0;

	b_.x() = 0;
	b_.y() = 0;
	b_.z() = 0;
	return (*this);
}

double RxSpatialVec::norm1(double ca, double cb) const
{
	return (ca*(ABS(a_.x()) + ABS(a_.y()) + ABS(a_.z())) + 
		cb*(ABS(b_.x()) + ABS(b_.y()) + ABS(b_.z())));
}

double RxSpatialVec::norm2(double ca, double cb) const
{
	return (ca*a_.norm2() + cb*b_.norm2());
}

double RxSpatialVec::normInf(double ca, double cb) const
{
	double amax, bmax;

	amax = ca*a_.normInf();
	bmax = cb*b_.normInf();
	return ((amax > bmax) ? amax : bmax);
}
