#include <RxTrsf.h>
#include <RxMathError.h>
#include <RxParseObj.h>
#include <RxSysError.h>

const int RxTrsf::ID = 0x1110;

// need to allocate a local ident RxRot since RxRot::IDENT
// may not yet be initialized
static double identBuf[9] = {
	1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0 };
static RxRot RI(identBuf);
const RxTrsf RxTrsf::IDENT (RI, RxVecd3::ZERO);

void RxTrsf::mul (const RxTrsf &m1, const RxTrsf &m2)
{
	register double x, y, z;

	x = m2.p.x(); y = m2.p.y(); z = m2.p.z();
	p.x() = x * m1.R.n.x() + y * m1.R.o.x() + z * m1.R.a.x() + m1.p.x();
	p.y() = x * m1.R.n.y() + y * m1.R.o.y() + z * m1.R.a.y() + m1.p.y();
	p.z() = x * m1.R.n.z() + y * m1.R.o.z() + z * m1.R.a.z() + m1.p.z();
 
	R.mul (m1.R, m2.R);
}

void RxTrsf::mulI (const RxTrsf &m1, const RxTrsf &m2)
{
	register double x, y, z;

	R.mulI (m1.R, m2.R);

	x = -m2.p.x(); y = -m2.p.y(); z = -m2.p.z();
	p.x() = m1.p.x() + x * R.n.x() + y * R.o.x() + z * R.a.x();
	p.y() = m1.p.y() + x * R.n.y() + y * R.o.y() + z * R.a.y();
	p.z() = m1.p.z() + x * R.n.z() + y * R.o.z() + z * R.a.z();
}

void RxTrsf::Imul (const RxTrsf &m1, const RxTrsf &m2)
{
	register double x, y, z;

	x = m2.p.x() - m1.p.x(); y = m2.p.y() - m1.p.y(); z = m2.p.z() - m1.p.z();
	p.x() = x * m1.R.n.x() + y * m1.R.n.y() + z * m1.R.n.z();
	p.y() = x * m1.R.o.x() + y * m1.R.o.y() + z * m1.R.o.z();
	p.z() = x * m1.R.a.x() + y * m1.R.a.y() + z * m1.R.a.z();

	R.Imul (m1.R, m2.R);
}

void RxTrsf::ImulI (const RxTrsf &m1, const RxTrsf &m2)
{
	register double x, y, z;
	double p_x, p_y, p_z;

	x = m1.p.x(); y = m1.p.y(); z = m1.p.z();
	p_x = x * m1.R.n.x() + y * m1.R.n.y() + z * m1.R.n.z();
	p_y = x * m1.R.o.x() + y * m1.R.o.y() + z * m1.R.o.z();
	p_z = x * m1.R.a.x() + y * m1.R.a.y() + z * m1.R.a.z();

	R.ImulI (m1.R, m2.R);

	x = -m2.p.x(); y = -m2.p.y(); z = -m2.p.z();
	p.x() = x * R.n.x() + y * R.o.x() + z * R.a.x() + p_x;
	p.y() = x * R.n.y() + y * R.o.y() + z * R.a.y() + p_y;
	p.z() = x * R.n.z() + y * R.o.z() + z * R.a.z() + p_z;
}

#if 0
RxVecd3 &RxTrsf::mul (RxVecd3 &vr, const RxVecd3 &v1) const
{
	register double x, y, z;

	x = v1.x(); y = v1.y(); z = v1.z();
	vr.x() = x * R.n.x() + y * R.o.x() + z * R.a.x() + p.x();
	vr.y() = x * R.n.y() + y * R.o.y() + z * R.a.y() + p.y();
	vr.z() = x * R.n.z() + y * R.o.z() + z * R.a.z() + p.z();
}

RxVecd3 &RxTrsf::Imul (RxVecd3 &vr, const RxVecd3 &v1) const
{
	register double x, y, z;

	x = v1.x() - p.x(); y = v1.y() - p.y(); z = v1.z() - p.z();
	vr.x() = x * R.n.x() + y * R.n.y() + z * R.n.z();
	vr.y() = x * R.o.x() + y * R.o.y() + z * R.o.z();
	vr.z() = x * R.a.x() + y * R.a.y() + z * R.a.z();
}
#endif

void RxTrsf::inv(const RxTrsf &m1)
{
	register double x, y, z;

	x = -m1.p.x(); y = -m1.p.y(); z = -m1.p.z();
	p.x() = x * m1.R.n.x() + y * m1.R.n.y() + z * m1.R.n.z();
	p.y() = x * m1.R.o.x() + y * m1.R.o.y() + z * m1.R.o.z();
	p.z() = x * m1.R.a.x() + y * m1.R.a.y() + z * m1.R.a.z();

	R.inv (m1.R);
}

void RxTrsf::setIdent()
{
	R.setIdent();
	p.x() = 0;
	p.y() = 0;
	p.z() = 0;
}

double& RxTrsf::operator()(int i, int j)
{
	static double dum;

	switch (j)
	 { case 0:
	    { return (R.n(i));
	    }
	   case 1:
	    { return (R.o(i));
	    }
	   case 2:
	    { return (R.a(i));
	    }
	   case 3:
	    { return (p(i));
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
			   "RxTrsf(%d,%d): %K\n", i, j).raise();
	      return dum;
	    }
	 }
}

double RxTrsf::operator()(int i, int j) const
{
	static double dum;

	switch (j)
	 { case 0:
	    { return (R.n(i));
	    }
	   case 1:
	    { return (R.o(i));
	    }
	   case 2:
	    { return (R.a(i));
	    }
	   case 3:
	    { return (p(i));
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
			   "RxTrsf(%d,%d): %K\n", i, j).raise();
	      return dum;
	    }
	 }
}

RxTrsf& RxTrsf::operator=(const RxTrsf& m2)
{
	R.n.x() = m2.R.n.x();
	R.n.y() = m2.R.n.y();
	R.n.z() = m2.R.n.z();
	R.o.x() = m2.R.o.x();
	R.o.y() = m2.R.o.y();
	R.o.z() = m2.R.o.z();
	R.a.x() = m2.R.a.x();
	R.a.y() = m2.R.a.y();
	R.a.z() = m2.R.a.z();
	p.x() = m2.p.x();
	p.y() = m2.p.y();
	p.z() = m2.p.z();
	return (*this);
}

RxTrsf::RxTrsf (const RxTrsf &m2)
{
	type_ = ID;
	R.n.x() = m2.R.n.x();
	R.n.y() = m2.R.n.y();
	R.n.z() = m2.R.n.z();
	R.o.x() = m2.R.o.x();
	R.o.y() = m2.R.o.y();
	R.o.z() = m2.R.o.z();
	R.a.x() = m2.R.a.x();
	R.a.y() = m2.R.a.y();
	R.a.z() = m2.R.a.z();
	p.x() = m2.p.x();
	p.y() = m2.p.y();
	p.z() = m2.p.z();
}

int RxTrsf::oscan (RxParseObj *obj)
{
	RxParseObj *lobj;
	int len;

	len = obj->getListLength();
	if (!obj->isNumberList() || 
	    (len != 3 && len != 4 && len != 7 && len != 12))
	 { RxSysError (RxSysError::ParsingError,
"RxTrsf::oscan: object not a list of 3, 4, 7 or 12 numbers\n").post();
	   return 0;
	 }
	if (len == 12)
	 { return RxMatdBase::oscan (obj);
	 }
	else
	 { lobj = obj->firstChild();
	 }
	if (len == 3)
	 { R.setIdent();
	 }
	else if (len == 4)
	 { p.set (0,0,0);
	 }
	if (len == 7 || len == 3)
	 { RxVecd3 axis;

	   p.x() = lobj->getNumberValue();
	   lobj = obj->nextChild();
	   p.y() = lobj->getNumberValue();
	   lobj = obj->nextChild();
	   p.z() = lobj->getNumberValue();
	   lobj = obj->nextChild();
	 }
	if (len == 7 || len == 4)
	 { RxVecd3 axis;
	   double ang;

	   axis.x() = lobj->getNumberValue();
	   lobj = obj->nextChild();
	   axis.y() = lobj->getNumberValue();
	   lobj = obj->nextChild();
	   axis.z() = lobj->getNumberValue();
	   lobj = obj->nextChild();
	   ang = lobj->getNumberValue();
	   
	   R.setAxisD (axis, ang);
	 }
	return 1;
}

int RxTrsf::cprint (RxCoutput *co, RxPrintSpec *ps) const
{
	const RxFmtNode *cf0, *cf1;
	const char *fmtT = "%8.3f";	// format for translational component
	const char *fmtR;		// format for rotational component
	const char *fmtS = " ";		// format for matrix spacing

	// figure out format strings
	cf0 = ps->convNode(0);
	if (cf0)
	 { fmtT = cf0->fmt();
	 }
	if (cf0 && (cf1 = ps->convNode(1)))
	 { fmtR = cf1->fmt();
	   if (cf0->next() != cf1)
	    { // then there is a matrix spacing format as well
	      fmtS = cf0->next()->fmt();
	    }
	 }
	else
	 { fmtR = fmtT;
	 }
	
	co->output_str (ps->prefix());
	if (ps->findSpecial ("%M"))
	 { double vec[3];
	   int i;
	   co->output_char ('[');
	   for (i=0; i<3; i++)
	    { vec[0] = (*this)(i,0);
	      vec[1] = (*this)(i,1);
	      vec[2] = (*this)(i,2);
	      co->output_doubleArray (fmtR, fmtS, NULL, vec, 3);
	      co->output_str (fmtS);
	      co->output_double (fmtT, (*this)(i,3));
	      if (i<2)
	       { co->output_newline (ps->getLeadSpacing()+1);
	       }
	    }
	   co->output_char (']');
	 }
	else
	 { RxVecd3 axis;
	   double ang;

	   co->output_char ('[');
	   co->output_doubleArray (fmtT, NULL, NULL, p.buf(), 3);
	   co->output_str ("  ");
	   R.getAxisD (&axis, &ang);
	   co->output_doubleArray (fmtR, NULL, NULL, axis.buf(), 3);
	   co->output_char (' ');
	   co->output_double (fmtT, ang);
	   co->output_char (']');
	 }
	if (ps->numConvs() == 0)
	 { co->output_char ('\n');
	 }
	else
	 { co->output_str (ps->suffix());
	 }
	return co->nchars();
}
