#include <RxTwist.h>
#include <RxVecdBaseTest.h>
#include <RxTrsf.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>

#define EPS	(1e-14)
#define EQ(a,b)	(ABS(a-b) < EPS)

#define LEN	6

double testVals[][LEN] = {
 { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 },
 { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0 },
 {-3.0, -2.0, -1.0, -0.5, 1.1, 2.2 }};

double rotVals[][3] = {
 { 90.0, -90.0, 45.0 },
 { 50.0,  60.0, 70.0 },
 { 90.0,  90.0,  0.0 },
 {  0.0,  90.0, 90.0 },
 { 90.0,   0.0, 90.0 },
 { 90.0,   0.0,  0.0 },
 {  0.0,  90.0,  0.0 },
 {  0.0,   0.0, 90.0 }};

int testSvUnit (const RxTwist &tu, const RxTwist &t1,
		double eps)
{
	RxTwist chk;

	chk.v().unit(t1.v());
	chk.w().unit(t1.w());
	return (chk.equal(tu, eps));	
}

int testSvNorm1 (double nc, const RxTwist &t1,
		 double cv, double cw, double eps)
{
	double chk;

	chk = cv*t1.v().norm1() + cw*t1.w().norm1();
	return (ABS(chk-nc) <= eps);
}

int testSvNorm2 (double nc, const RxTwist &t1,
		 double cv, double cw, double eps)
{
	double chk;

	chk = cv*t1.v().norm2() + cw*t1.w().norm2();
	return (ABS(chk-nc) <= eps);
}

int testSvNormInf (double nc, const RxTwist &t1,
		 double cv, double cw, double eps)
{
	double chk, amax, bmax;

	amax = cv*t1.v().normInf();
	bmax = cw*t1.w().normInf();
	return (ABS(MAX(amax,bmax)-nc) <= eps);
}

main()
{
	double vec[LEN];
	RxTwist t1(10,20,30,40,50,60);
	RxTwist t2(100,200,300,400,500,600);
	int i, j, k;

	assert (EQ(t1.v().x(),10) && EQ(t1.v().y(),20) && EQ(t1.v().z(),30));
	assert (EQ(t1.w().x(),40) && EQ(t1.w().y(),50) && EQ(t1.w().z(),60));
	
	RxTwist t3(t2);

	assert (EQ(t3.v().x(),100) && EQ(t3.v().y(),200) && EQ(t3.v().z(),300));
	assert (EQ(t3.w().x(),400) && EQ(t3.w().y(),500) && EQ(t3.w().z(),600));

	assert (t3.length() == LEN);

	assert (t3.equal(t2,EPS));
	assert (!t3.equal(t1,EPS));
	for (i=0;i<LEN;i++)
	 { t1(i) = i*10+1;
	 }
	assert (t1.equal(RxTwist(1,11,21,31,41,51),EPS));
	
	t1 = t3;

	assert (t3.equal(t1,EPS));
	
	t1.set(11,22,33,44,55,66);

	assert (t1.equal(RxTwist(11,22,33,44,55,66),EPS));
	
	t1.get(vec);
	for (i=0; i<LEN; i++)
	 { assert (EQ(vec[i],(i+1)*11));
	 }
	for (i=0; i<LEN; i++)
	 { vec[i] = (i+1)*111;
	 }
	assert ((t3.set(vec)).equal(RxTwist(111,222,333,444,555,666),EPS));

	t3 = t1;
	assert (testAdd (t3.add(t3.add(t1)), 2*t1, 2*t1, EPS));
	
	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { for (j=i; j<ARRAY_SIZE(testVals); j++)
	    { t1.set(testVals[i]);
	      t2.set(testVals[j]);

	      assert (testAdd (t1+t2, t1, t2, EPS));
	      assert (testAdd (t3.add(t1,t2), t1, t2, EPS));
	      t3 = t1;
	      assert (testAdd (t3.add(t2), t1, t2, EPS));
	      t3 = t1;					    
	      assert (testAdd (t3 += t2, t1, t2, EPS));

	      assert (testSub (t1-t2, t1, t2, EPS));
	      assert (testSub (t3.sub(t1,t2), t1, t2, EPS));
	      t3 = t1;
	      assert (testSub (t3.sub(t2), t1, t2, EPS));
	      t3 = t1;					    
	      assert (testSub (t3 -= t2, t1, t2, EPS));

	      assert (testExtrap (t3.extrap(t1,11.1,t2), t1, 11.1, t2, EPS));
	      t3 = t1;
	      assert (testExtrap (t3.extrap(11.1,t2), t1, 11.1, t2, EPS));

#if 0
	      t3 = t1;
	      assert (testDot (t3.dot(t2), t1, t2, EPS));
#endif
	    }
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { t1.set(testVals[i]);

	   assert (testScale (t3.scale(M_PI,t1), M_PI, t1, EPS));
	   t3 = t1;
	   assert (testScale (t3.scale(M_PI), M_PI, t1, EPS));
	   t3 = t1;
	   assert (testScale (t3 *= M_PI, M_PI, t1, EPS));
	   t3 = t1;
	   assert (testScale (t3 /= M_PI, 1/M_PI, t1, EPS));
	   assert (testScale (M_PI*t1, M_PI, t1, EPS));
	   assert (testScale (t1*M_PI, M_PI, t1, EPS));
	   assert (testScale (t1/M_PI, 1/M_PI, t1, EPS));

	   assert (testSvUnit (t3.unit(t1), t1, EPS));
	   t3 = t1;
	   assert (testSvUnit (t3.unit(), t1, EPS));
	   assert (testScale (t3.neg(t1), -1.0, t1, EPS));
	   t3 = t1;
	   assert (testScale (t3.neg(), -1.0, t1, EPS));
	   assert (testScale (-t1, -1.0, t1, EPS));

	   assert (testSvNorm1 (t1.norm1(1, M_PI), t1, 1, M_PI, EPS));
	   assert (testSvNorm2 (t1.norm2(1, M_PI), t1, 1, M_PI, EPS));
	   assert (testSvNormInf (t1.normInf(1, M_PI), t1, 1, M_PI, EPS));
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { t1.set(testVals[i]);
 
	   for (j=0; j<ARRAY_SIZE(rotVals); j++)
	    { RxTrsf T;

	      T.R.setRpyD (rotVals[j][0], rotVals[j][1], rotVals[j][2]);

	      t3.xform (t1, T.R);
	      t3.Ixform (t3, T.R);
	      assert (t3.equal(t1,EPS));
	      t3 = t1;
	      t3.xform (T.R);
	      t3.Ixform (T.R);
	      assert (t3.equal(t1,EPS));

	      for (k=0; k<ARRAY_SIZE(testVals); k++)
	       { T.p.set (testVals[k][0], testVals[k][1], testVals[k][2]);

		 t3.xform (t1, T);
		 t3.Ixform (t3, T);
		 assert (t3.equal(t1,EPS));
		 t3 = t1;
		 t3.xform (T);
		 t3.Ixform (T);
		 assert (t3.equal(t1,EPS));
	       }
	    }
	 }

	t1.set(0,0,0,0,0,0);
	assert ((t2.setZero()).equal(t1, EPS));

	printf ("\nROUTINES OK\n\n");
	return 0;
}
