#include <RxWrench.h>
#include <RxRot.h>
#include <RxTrsf.h>
#include <RxMathError.h>
#include <math.h>
#include <cdefs.h>

const RxWrench RxWrench::ZERO    (0,0,0, 0,0,0);
const RxWrench RxWrench::UNIT_FX (1,0,0, 0,0,0);
const RxWrench RxWrench::UNIT_FY (0,1,0, 0,0,0);
const RxWrench RxWrench::UNIT_FZ (0,0,1, 0,0,0);
const RxWrench RxWrench::UNIT_MX (0,0,0, 1,0,0);
const RxWrench RxWrench::UNIT_MY (0,0,0, 0,1,0);
const RxWrench RxWrench::UNIT_MZ (0,0,0, 0,0,1);

RxWrench& RxWrench::set (double fx, double fy, double fz,
			 double mx, double my, double mz)
{
	m().x() = mx;
	m().y() = my;
	m().z() = mz;
	f().x() = fx;
	f().y() = fy;
	f().z() = fz;
	return (*this);
}

RxWrench& RxWrench::set (const double *dp)
{
	f().x() = dp[0];
	f().y() = dp[1];
	f().z() = dp[2];
	m().x() = dp[3];
	m().y() = dp[4];
	m().z() = dp[5];
	return (*this);
}

void RxWrench::get (double *fxp, double *fyp, double *fzp,
		   double *mxp, double *myp, double *mzp)
{
	*mxp = f().x();
	*myp = f().y();
	*mzp = f().z();
	*fxp = m().x();
	*fyp = m().y();
	*fzp = m().z();
}

void RxWrench::get (double *dp) const
{
	dp[0] = f().x();
	dp[1] = f().y();
	dp[2] = f().z();
	dp[3] = m().x();
	dp[4] = m().y();
	dp[5] = m().z();
}

double& RxWrench::operator()(int i)
{
	static double dum;

	switch (i)
	 { case 0:
	    { return (f().x());
	    }
	   case 1:
	    { return (f().y());
	    }
	   case 2:
	    { return (f().z());
	    }
	   case 3:
	    { return (m().x());
	    }
	   case 4:
	    { return (m().y());
	    }
	   case 5:
	    { return (m().z());
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxWrench::(%d): %K\n", i).raise();	      
	      return (dum);
	    }
	 }
}

double RxWrench::operator()(int i) const
{
	static double dum;

	switch (i)
	 { case 0:
	    { return (f().x());
	    }
	   case 1:
	    { return (f().y());
	    }
	   case 2:
	    { return (f().z());
	    }
	   case 3:
	    { return (m().x());
	    }
	   case 4:
	    { return (m().y());
	    }
	   case 5:
	    { return (m().z());
	    }
	   default:
	    { RxMathError (RxMathError::BadIndex,
		 "RxWrench::(%d): %K\n", i).raise();	      
	      return (dum);
	    }
	 }
}
