#include <RxWrench.h>
#include <RxVecdBaseTest.h>
#include <RxTrsf.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>

#define EPS	(1e-14)
#define EQ(a,b)	(ABS(a-b) < EPS)

#define LEN	6

double testVals[][LEN] = {
 { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0 },
 { 1.0, 2.0, 3.0, 4.0, 5.0, 6.0 },
 {-3.0, -2.0, -1.0, -0.5, 1.1, 2.2 }};

double rotVals[][3] = {
 { 90.0, -90.0, 45.0 },
 { 50.0,  60.0, 70.0 },
 { 90.0,  90.0,  0.0 },
 {  0.0,  90.0, 90.0 },
 { 90.0,   0.0, 90.0 },
 { 90.0,   0.0,  0.0 },
 {  0.0,  90.0,  0.0 },
 {  0.0,   0.0, 90.0 }};

int testSvUnit (const RxWrench &fu, const RxWrench &f1,
		double eps)
{
	RxWrench chk;

	chk.f().unit(f1.f());
	chk.m().unit(f1.m());
	return (chk.equal(fu, eps));	
}

int testSvNorm1 (double nc, const RxWrench &f1,
		 double cf, double cm, double eps)
{
	double chk;

	chk = cf*f1.f().norm1() + cm*f1.m().norm1();
	return (ABS(chk-nc) <= eps);
}

int testSvNorm2 (double nc, const RxWrench &f1,
		 double cf, double cm, double eps)
{
	double chk;

	chk = cf*f1.f().norm2() + cm*f1.m().norm2();
	return (ABS(chk-nc) <= eps);
}

int testSvNormInf (double nc, const RxWrench &f1,
		 double cf, double cm, double eps)
{
	double chk, amax, bmax;

	amax = cf*f1.f().normInf();
	bmax = cm*f1.m().normInf();
	return (ABS(MAX(amax,bmax)-nc) <= eps);
}

main()
{
	double vec[LEN];
	RxWrench f1(10,20,30,40,50,60);
	RxWrench f2(100,200,300,400,500,600);
	int i, j, k;

	assert (EQ(f1.f().x(),10) && EQ(f1.f().y(),20) && EQ(f1.f().z(),30));
	assert (EQ(f1.m().x(),40) && EQ(f1.m().y(),50) && EQ(f1.m().z(),60));
	
	RxWrench f3(f2);

	assert (EQ(f3.f().x(),100) && EQ(f3.f().y(),200) && EQ(f3.f().z(),300));
	assert (EQ(f3.m().x(),400) && EQ(f3.m().y(),500) && EQ(f3.m().z(),600));

	assert (f3.length() == LEN);

	assert (f3.equal(f2,EPS));
	assert (!f3.equal(f1,EPS));
	for (i=0;i<LEN;i++)
	 { f1(i) = i*10+1;
	 }
	assert (f1.equal(RxWrench(1,11,21,31,41,51),EPS));
	
	f1 = f3;

	assert (f3.equal(f1,EPS));
	
	f1.set(11,22,33,44,55,66);

	assert (f1.equal(RxWrench(11,22,33,44,55,66),EPS));
	
	f1.get(vec);
	for (i=0; i<LEN; i++)
	 { assert (EQ(vec[i],(i+1)*11));
	 }
	for (i=0; i<LEN; i++)
	 { vec[i] = (i+1)*111;
	 }
	assert ((f3.set(vec)).equal(RxWrench(111,222,333,444,555,666),EPS));

	f3 = f1;
	assert (testAdd (f3.add(f3.add(f1)), 2*f1, 2*f1, EPS));
	
	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { for (j=i; j<ARRAY_SIZE(testVals); j++)
	    { f1.set(testVals[i]);
	      f2.set(testVals[j]);

	      assert (testAdd (f1+f2, f1, f2, EPS));
	      assert (testAdd (f3.add(f1,f2), f1, f2, EPS));
	      f3 = f1;
	      assert (testAdd (f3.add(f2), f1, f2, EPS));
	      f3 = f1;					    
	      assert (testAdd (f3 += f2, f1, f2, EPS));

	      assert (testSub (f1-f2, f1, f2, EPS));
	      assert (testSub (f3.sub(f1,f2), f1, f2, EPS));
	      f3 = f1;
	      assert (testSub (f3.sub(f2), f1, f2, EPS));
	      f3 = f1;					    
	      assert (testSub (f3 -= f2, f1, f2, EPS));

	      assert (testExtrap (f3.extrap(f1,11.1,f2), f1, 11.1, f2, EPS));
	      f3 = f1;
	      assert (testExtrap (f3.extrap(11.1,f2), f1, 11.1, f2, EPS));

#if 0
	      f3 = f1;
	      assert (testDot (f3.dot(f2), f1, f2, EPS));
#endif
	    }
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { f1.set(testVals[i]);

	   assert (testScale (f3.scale(M_PI,f1), M_PI, f1, EPS));
	   f3 = f1;
	   assert (testScale (f3.scale(M_PI), M_PI, f1, EPS));
	   f3 = f1;
	   assert (testScale (f3 *= M_PI, M_PI, f1, EPS));
	   f3 = f1;
	   assert (testScale (f3 /= M_PI, 1/M_PI, f1, EPS));
	   assert (testScale (M_PI*f1, M_PI, f1, EPS));
	   assert (testScale (f1*M_PI, M_PI, f1, EPS));
	   assert (testScale (f1/M_PI, 1/M_PI, f1, EPS));

	   assert (testSvUnit (f3.unit(f1), f1, EPS));
	   f3 = f1;
	   assert (testSvUnit (f3.unit(), f1, EPS));
	   assert (testScale (f3.neg(f1), -1.0, f1, EPS));
	   f3 = f1;
	   assert (testScale (f3.neg(), -1.0, f1, EPS));
	   assert (testScale (-f1, -1.0, f1, EPS));

	   assert (testSvNorm1 (f1.norm1(1, M_PI), f1, 1, M_PI, EPS));
	   assert (testSvNorm2 (f1.norm2(1, M_PI), f1, 1, M_PI, EPS));
	   assert (testSvNormInf (f1.normInf(1, M_PI), f1, 1, M_PI, EPS));
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { f1.set(testVals[i]);
 
	   for (j=0; j<ARRAY_SIZE(rotVals); j++)
	    { RxTrsf T;

	      T.R.setRpyD (rotVals[j][0], rotVals[j][1], rotVals[j][2]);

	      f3.xform (f1, T.R);
	      f3.Ixform (f3, T.R);
	      assert (f3.equal(f1,EPS));
	      f3 = f1;
	      f3.xform (T.R);
	      f3.Ixform (T.R);
	      assert (f3.equal(f1,EPS));

	      for (k=0; k<ARRAY_SIZE(testVals); k++)
	       { T.p.set (testVals[k][0], testVals[k][1], testVals[k][2]);

		 f3.xform (f1, T);
		 f3.Ixform (f3, T);
		 assert (f3.equal(f1,EPS));
		 f3 = f1;
		 f3.xform (T);
		 f3.Ixform (T);
		 assert (f3.equal(f1,EPS));
	       }
	    }
	 }

	f1.set(0,0,0,0,0,0);
	assert ((f2.setZero()).equal(f1, EPS));

	printf ("\nROUTINES OK\n\n");
	return 0;
}
