#include <stdio.h>
#include <math.h>
#include <string.h>	
#include <stdlib.h>
#ifdef HAS_STDARG
#include <stdarg.h>
#define VA_ARGDEF(x) x
#else
#include <varargs.h>
#define VA_ARGDEF(x) (va_alist) va_dcl
#endif
#include <sys/types.h>
#include <sys/time.h>

#include <cdefs.h>
#include <grf4.h>
#include "grf4Internal.h"
#include "grf4Support.h"
#include "grf4GenDraw.h"

#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif

#define X11_H		1
#define SUNVIEW_H	2
#define GL_H		3
#define GL_WF_H		4
#define OPEN_GL_H	5

#if HLS
int grf4HiddenLineMode = 0;
#endif

int grf4GeneralFlags;

int grf4HostType;
static int displayFlags;
int grf4InitDone;

static int winw;
static int winh;
static int directionOfY;

static void (*redrawFxn)(void *);
static void *redrawArg;

#define fxntab grf4FunctionTable
struct grf4fxnTable fxntab;

static float desiredLeft;
static float desiredRight;
static float desiredBottom;
static float desiredTop;
static float desiredNear;
static float desiredFar;
static float desiredAspectRatio;

static float virtL;
static float virtB;
static float virtW;
static float virtH;

static int perspective;

static WinPlaceDesc placeDesc;

int grf4PanelWidth;
int grf4PanelHeight;

BUTTON_DESC *buttonList;
int numButtons;
int rowNumber = 1;

int buttonSpacing;
int buttonWidth;
int buttonHeight;

#define NO_VIEW_CTRL		0
#define YAW_PITCH_CTRL		1
#define X_Y_CTRL		2
#define ZOOM_CTRL		3
#define PANY_PANX_CTRL		4
#define TILT_CTRL		5

static int viewCtrlMode = NO_VIEW_CTRL;
static int yCtrlButton;
static int xCtrlButton;
static int yawCtrlButton;
static int pitchCtrlButton;
static int panXCtrlButton;
static int panYCtrlButton;
static int tiltCtrlButton;
static int zoomCtrlButton;
static int xMotionEnable = 1;
static int yMotionEnable = 1;

#define ROLL_BUTTON	0
#define PITCH_BUTTON	1
#define YAW_BUTTON	2

static struct rpyButton {
	float val;
	int dispButton;
	int upButton;
	int downButton;
  } rpyButtons[3];

static void updateRpyButtons();

typedef struct _TRSF_BLK {
	struct _TRSF_BLK *next;
	TRSF T;
  } TRSF_BLK;

static TRSF_BLK *trsfStack;

/* -----------------------------------------------------------------------

   GRF4 coordinate frame conventions.

						^
						|
	Y					|
	^					+---->
	|				      object
	|		  | 		      frame		    |
 Z <----+ - - - - - - - - | - - - - - - - - - - - - - - - - - - - - |
			  |					    |
       eye	         near					   far
      frame	       clipping					 clipping
			plane					  plane

    T is the transform from the eye frame to the object frame.
	
    E is the portion of T from the eye frame to the viewing center,
    so that 

	T = E D

    where D is a transform from the viewing center to object coordinates.

    When E is set, T must be reset as well, according to 

	T = Enew Eold^-1 D


    A point vo in object coordinates is transformed into eye coordinates by

	ve = T vo

    After a point is transformed into eye coordinates, it is projected
    into the window plane (which corresponds to the near clipping plane)
    using either the perspective transform 

	wx = ex near / ez
	wy = ey near / ez

    or the orthogonal projection

	wx = ex
	wy = ey

    window points are then mapped to pixels according to the virtual
    size characteristics of the window.
*/

static TRSF E;
TRSF T;
static TRSF L;	/* not used right now */

static drawObjCnt = 0;

static int determineHostType();
static void popMaterial();
static void pushMaterial();
static void adjustScreenTrans(int, int);
static BUTTON_DESC *getButton(int );
static void buttonCompile(int w, int h);

static void buttonResize(int dw, int dh);
static void drawAllButtons();
static BUTTON_DESC *overButton(int x, int y);

static void onButton (int n, int code, long mode);
static void bumpRpyValue (int idx, int inc);
static void onRpyButton(int n, int code, long idx);
static void onDrag(int button, int x1, int y1, int x2, int y2);
static void updateRpyButtons(double r, double p, double y);

static double range180(double a)
{
	/* puts an angle in the range -180 < a <= 180 */

	double norm;
	
	norm = a/(2*180);
	norm -= (int)norm;
	if (norm > 0.5)
	 { return (180*(2*(norm-1)));
	 }
	else if (norm <= -0.5)
	 { return (180*(2*(norm+1)));
	 }
	else
	 { return (180*(2*(norm)));
	 }
}

int grf4GetDrawObjCnt()
{
	return (drawObjCnt);
}

int grf4IncDrawObjCnt()
{
	return (drawObjCnt++);
}

void grf4PlaceWindow (int x, int y, int w, int h)
{
	placeDesc.x = x;
	placeDesc.y = y;
	placeDesc.w = w;
	placeDesc.h = h;
	placeDesc.placed = 1;
}

void grf4SetMinSize (int w, int h)
{
	placeDesc.minW = w;
	placeDesc.minH = h;
}

static int determineHostType(int flags)
{
	if (flags & GRF_X11)
	 {
#ifdef HAS_X_WINDOWS
	   if (getenv ("DISPLAY"))
	    { return (X11_H);
	    }
#endif
	   return (0);
	 }
#ifdef HAS_OPEN_GL
	if (!(flags & GRF_OLD_GL))
	 { if (flags & GRF_WIRE_FRAME)
	    { return (X11_H);
	    }
	   else
	    { return (OPEN_GL_H);
	    }
	 }
#endif
#ifdef HAS_GL_WINDOWS
	if (flags & GRF_WIRE_FRAME)
	 { return (GL_WF_H);
	 }
	else
	 { return (GL_H);
	 }
#else
	if (flags & GRF_OLD_GL)
	 { return 0;
	 }
#endif
	if (getenv("DISPLAY"))
	 { return (X11_H);
	 }
#ifdef sun
	if (getenv("WINDOW_PARENT"))
	 { return (SUNVIEW_H);
	 }
#endif
	return 0;	/* default! */
}

typedef struct matStackElem {
	struct matStackElem *next;
	void *m;
} MatStackElem;

static MatStackElem *matStackFree;
static MatStackElem *matStack;

static void* currentMaterial;

static void popMaterial()
{
	MatStackElem *se;

	if (se = matStack)
	 { matStack = se->next;
	   currentMaterial = se->m;
	   (*fxntab.setMaterial)(se->m);
	   se->next = matStackFree;
	   matStackFree = se;
	 }
}

static void pushMaterial()
{
	MatStackElem *se;

	if ((se = matStackFree) == 0)
	 { se = ALLOC(MatStackElem);
	 }
	else
	 { matStackFree = se->next;
	 }
	se->next = matStack;
	matStack = se;
	se->m = currentMaterial;
}

int grf4FxnInitDone;

int grf4DoFxnInit (int flags)
{
	grf4FxnInitDone = 1;

	if ((grf4HostType = determineHostType(flags)) == 0)
	 { return (-1);
	 }

	/* instantiate functions depending on host type */

	switch (grf4HostType)
	 {
	   case GL_H:
	    { if (grf4glSetupFxns (&fxntab) < 0)
	       { return (-1);
	       }
	      displayFlags = 0;
	      break;
	    }
	   case GL_WF_H:
	    { if (grf4glwfSetupFxns (&fxntab) < 0)
	       { return (-1);
	       }
	      gdrSetupFxns (&fxntab);
	      displayFlags = GRF_WIRE_FRAME;
	      break;
	    }
	   case OPEN_GL_H:
	    { if (grf4GLSetupFxns (&fxntab) < 0)
	       { return (-1);
	       }
	      displayFlags = 0;
	      break;
	    }
	   case SUNVIEW_H:
	    { if (grf4SvSetupFxns (&fxntab) < 0)
	       { return (-1);
	       }
	      gdrSetupFxns (&fxntab);
	      displayFlags = GRF_WIRE_FRAME;
	      break;
	    }
	   case X11_H:
	    { if (grf4X11SetupFxns (&fxntab) < 0)
	       { return (-1);
	       }
	      gdrSetupFxns (&fxntab);
	      displayFlags = GRF_WIRE_FRAME | GRF_X11;
	      break;
	    }
	 }
	return (1);
}

int grf4SetDisplay (int flags)
{
	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(flags) < 0)
	    { return (-1);
	    }
	 }
	if (grf4HostType == 0)
	 { return (-1);
	 }
	return (displayFlags | (flags & ~(GRF_WIRE_FRAME | GRF_X11)));
}

int grf4Init (char *name, int fgcolor, int bgcolor, int flags)
{
	int x,y,w,h;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(flags) < 0)
	    { return (-1);
	    }
	 }
	if (grf4HostType == 0)
	 { return (-1);
	 }

	if (flags & GRF_VIEW_CONTROL)
	 { 
	   grf4AddButton("View Control", BF_NOTBOXED|BF_INACTIVE, 1, 0, 0);
	   yawCtrlButton = grf4AddButton("yaw", BF_ILLUMINATED, 1,
					  onButton, YAW_PITCH_CTRL);
	   pitchCtrlButton = grf4AddButton("pitch", BF_SAMEROW|BF_ILLUMINATED,
					  1, onButton, YAW_PITCH_CTRL);
	   xCtrlButton = grf4AddButton("x", 0, 1, 
					  onButton, X_Y_CTRL);
	   yCtrlButton = grf4AddButton("y", BF_SAMEROW, 1, 
					  onButton, X_Y_CTRL);
	   zoomCtrlButton = grf4AddButton("zoom", 0, 1, onButton, ZOOM_CTRL);
	   panXCtrlButton = grf4AddButton("pan x", 0, 1, 
					  onButton, PANY_PANX_CTRL);
	   panYCtrlButton = grf4AddButton("pan y", BF_SAMEROW, 1, 
					  onButton, PANY_PANX_CTRL);
	   tiltCtrlButton = grf4AddButton("tilt", 0, 1, onButton, TILT_CTRL);

	   rpyButtons[ROLL_BUTTON].upButton = 
	      grf4AddButton("+", 0, 1, onRpyButton, ROLL_BUTTON);
	   rpyButtons[ROLL_BUTTON].dispButton = 
	      grf4AddButton("roll:     0.0", BF_INACTIVE|BF_SAMEROW, 3, 0,0);
	   rpyButtons[ROLL_BUTTON].downButton = 
	      grf4AddButton("-", BF_SAMEROW, 1, onRpyButton, ROLL_BUTTON);

	   rpyButtons[PITCH_BUTTON].upButton = 
	      grf4AddButton("+", 0, 1, onRpyButton, PITCH_BUTTON);
	   rpyButtons[PITCH_BUTTON].dispButton = 
	      grf4AddButton("pitch:    0.0", BF_INACTIVE|BF_SAMEROW, 3, 0,0);
	   rpyButtons[PITCH_BUTTON].downButton = 
	      grf4AddButton("-", BF_SAMEROW, 1, onRpyButton, PITCH_BUTTON);

	   rpyButtons[YAW_BUTTON].upButton = 
	      grf4AddButton("+", 0, 1, onRpyButton, YAW_BUTTON);
	   rpyButtons[YAW_BUTTON].dispButton = 
	      grf4AddButton("yaw:      0.0", BF_INACTIVE|BF_SAMEROW, 3, 0,0);
	   rpyButtons[YAW_BUTTON].downButton = 
	      grf4AddButton("-", BF_SAMEROW, 1, onRpyButton, YAW_BUTTON);

	   viewCtrlMode = YAW_PITCH_CTRL;
	 }

	switch (grf4HostType)
	 {
	   case GL_H:
	    { if (grf4glInit (name, fgcolor, bgcolor, &placeDesc, flags) < 0)
	       { fprintf (stderr, "Could not initialize GL graphics\n");
		 return (-1);
	       }
	      break;
	    }
	   case GL_WF_H:
	    { if (grf4glwfInit (name, fgcolor, bgcolor, &placeDesc, flags) < 0)
	       { fprintf (stderr, "Could not initialize GL wireframe graphics\n");
		 return (-1);
	       }
	      break;
	    }
	   case OPEN_GL_H:
	    { if (grf4GLInit (name, fgcolor, bgcolor, &placeDesc, flags) < 0)
	       { fprintf (stderr, "Could not initialize GL graphics\n");
		 return (-1);
	       }
	      break;
	    }
	   case SUNVIEW_H:
	    { if (grf4SvInit (name, fgcolor, bgcolor, &placeDesc, flags) < 0)
	       { fprintf (stderr, "Could not initialize SunView graphics\n");
		 return (-1);
	       }
	      break;
	    }
	   case X11_H:
	    { if (grf4X11Init (name, fgcolor, bgcolor, &placeDesc, flags) < 0)
	       { fprintf (stderr, "Could not initialize X11 graphics\n");
		 return (-1);
	       }
	      break;
	    }
	 }
	(*fxntab.getWindowSize) (&winw, &winh, &directionOfY);
	buttonCompile (grf4PanelWidth, grf4PanelHeight);

	if (grf4PanelWidth)
	 { flags |= GRF_CONTROL_PANEL;
	 }

	if (grf4HostType != GL_H)
	 { gdrInit (winw, winh, directionOfY);
	 }
	(*fxntab.ortho)(-0.5, winw+0.5, -0.5, winh+0.5, -1.0, 1.0);
	identTrsf (&E);
	identTrsf (&L);	/* L not used right now */
	identTrsf (&T);
	(*fxntab.loadTrans)(&T);

	if (viewCtrlMode != NO_VIEW_CTRL)
	 { updateRpyButtons(0.0, 0.0, 0.0);
	 }

	drawAllButtons ();

	currentMaterial = (*fxntab.defMaterial)(0xffffff, 0, 0, 0, 0.0, 0.0);
	(*fxntab.setMaterial)(currentMaterial);

	grf4InitEvents(flags, winw, winh, grf4PanelWidth, grf4PanelHeight);

	grf4InitDone = 1;
	grf4GeneralFlags = flags;

	return (displayFlags | (flags & ~(GRF_WIRE_FRAME | GRF_X11)));
}

long grf4Capacity()
{
	return (*fxntab.capacity)();
}

void grf4HideLines(int on)
{
	grf4HiddenLineMode = on;
}

void grf4DumpPostScript(FILE *fp, char *name, int flags, double w, double h)
{
	PostScriptInit (fp, name, flags, w, h);
}

void grf4PostScriptInit (FILE *fp, char *name, int flags, double w, double h)
{
	PostScriptInit (fp, name, flags, w, h);
}

void grf4PostScriptFinish (FILE *fp)
{
	PostScriptFinish (fp);
}

void grf4PostScriptBoxSet (double llx, double lly, double urx, double ury)
{
	PostScriptBoxSet (llx, lly, urx, ury);
}

void grf4PostScriptXform (float *xmp, float *xbp, float *ymp, float *ybp)
{
	PostScriptXform (xmp, xbp, ymp, ybp);
}

void grf4GetWindowSize (int *wp, int *hp)
{
	if (wp) *wp = winw;
	if (hp) *hp = winh;
}

void grf4GetVirtWindowSize (double *vwp, double *vhp)
{
	if (vwp) *vwp = virtW;
	if (vhp) *vhp = virtH;
}

void grf4GetVirtWindowPos (double *lp, double *bp)
{
	if (lp) *lp = virtL;
	if (bp) *bp = virtB;
}

void grf4SetDrawFxn (void (*fxn)(void *), void *arg)
{
	redrawFxn = fxn;
	redrawArg = arg;
}

void grf4Window (double left, double right, double bottom, double top, 
		 double near, double far, double aspectSet)
{
	float winAr;

	desiredLeft = left;
	desiredRight = right;
	desiredBottom = bottom;
	desiredTop = top;
	desiredNear = near;
	desiredFar = far;
	desiredAspectRatio = aspectSet;

	perspective = 1;

	adjustScreenTrans (winw, winh);

	identTrsf (&E);
	identTrsf (&T);
	(*fxntab.loadTrans)(&T);

	if (viewCtrlMode != NO_VIEW_CTRL)
	 { updateRpyButtons(0.0, 0.0, 0.0);
	 }
}

void grf4Ortho (double left, double right, double bottom, double top, 
		double near, double far, double aspectSet)
{
	desiredLeft = left;
	desiredRight = right;
	desiredBottom = bottom;
	desiredTop = top;
	desiredNear = near;
	desiredFar = far;
	desiredAspectRatio = aspectSet;

	perspective = 0;

	adjustScreenTrans (winw, winh);

	identTrsf (&E);
	identTrsf (&T);
	(*fxntab.loadTrans)(&T);

	if (viewCtrlMode != NO_VIEW_CTRL)
	 { updateRpyButtons(0.0, 0.0, 0.0);
	 }
}

static void adjustScreenTrans(int w, int h)
{
	/* Adjust either the width or the height so that the aspect
	   ratio defined by

	((desiredTop-desiredBottom)*w)/((desiredRight-desiredLeft)*h)

	   remains constant.
	 */

	float right   = desiredRight;
	float left    = desiredLeft;
	float bottom  = desiredBottom;
	float top     = desiredTop;
	float delta;
	float ar;

	ar = ((top-bottom)*w)/((right-left)*h);

	if (desiredAspectRatio > 0)
	 { 
	   if (ar > desiredAspectRatio)
	    { /* adjust by increasing the width */

	      delta = (ar/desiredAspectRatio - 1.0)*(right-left)/2.0;
	      right += delta;
	      left -= delta;
	    }
	   else if (ar < desiredAspectRatio)
	    { /* adjust by increasing the height */

	      delta = (desiredAspectRatio/ar - 1.0)*(top-bottom)/2.0;
	      top += delta;
	      bottom -= delta;
	    }
	    }

	if (perspective)
	 { (*fxntab.window) (left, right, bottom, top, 
			     desiredNear, desiredFar);
	 }
	else
	 { (*fxntab.ortho) (left, right, bottom, top, 
			     desiredNear, desiredFar);
	 }
	virtL = left;
	virtB = bottom;
	virtH = top - bottom;
	virtW = right - left;
}

void grf4DesiredWindow (float *lp, float *rp, float *bp, float *tp, 
			float *np, float *fp, float *ap)
{
	/* hack for internal use */

	if (lp) *lp = desiredLeft;
	if (rp) *rp = desiredRight;
	if (bp) *bp = desiredBottom;
	if (tp) *tp = desiredTop;
	if (np) *np = desiredNear;
	if (fp) *fp = desiredFar;
	if (ap) *ap = desiredAspectRatio;
}

void grf4Color (int color)
{
	(*fxntab.color)(color);
}

void grf4Clear ()
{
	(*fxntab.clear)();
}

void grf4Flush ()
{
	drawObjCnt = 0;

	(*fxntab.flush)();
}

static int doSetView (TRSF *Enew, int updateButtons)
{
	(*fxntab.loadTrans)(Enew);
	(*fxntab.resetLights)();
	multLiTrsf (&T, &E, &T);
	multTrsf (&T, Enew, &T);
	(*fxntab.loadTrans)(&T);
	E = *Enew;

	if (viewCtrlMode != NO_VIEW_CTRL && updateButtons==YES)
	 { double r, p, y;

	   trsfToRpy (&r, &p, &y, &E);
	   r = range180 (r);
	   p = range180 (p);
	   y = range180 (y);
	   updateRpyButtons(r, p, y);
	 }
	return (1);
}

void grf4SetView (double x, double y, double z, 
		  double roll, double pitch, double yaw)
{
	TRSF Enew;

	rpyToTrsf (&Enew, roll,pitch,yaw);
	Enew.p.x = x;
	Enew.p.y = y;
	Enew.p.z = z;

	updateRpyButtons (roll, pitch, yaw);
	doSetView (&Enew, /*updateButtons=*/NO);
}

void grf4SetViewTrsf (TRSF *Enew)
{
	doSetView (Enew, /*updateButtons=*/YES);
}

void grf4PushTrans ()
{
	TRSF_BLK *tb;

	tb = (TRSF_BLK*)malloc(sizeof(TRSF_BLK));
	tb->next = trsfStack;
	trsfStack = tb;
	tb->T = T;
}

void grf4PopTrans ()
{
	TRSF_BLK *tb;

	if (tb = trsfStack)
	 { T = tb->T;
	   trsfStack = tb->next;
	   free ((char*)tb);
	   (*fxntab.loadTrans)(&T);
	 }
}

void grf4ResetTrans ()
{
	identTrsf (&T);
	(*fxntab.loadTrans)(&T);
}

void grf4SetTrans (const TRSF *Tnew)
{
	/* Sets the 4X4 transform describing the spatial position of the
	   indicated object. 
	 */
	T = *Tnew;
	(*fxntab.loadTrans)(&T);
}

TRSF *grf4GetTrans (TRSF *Tr)
{
	/* Gets the 4X4 transform describing the spatial position of the
	   indicated object. 
	 */
	*Tr = T;
	return (Tr);
}

TRSF *grf4GetTransPtr()
{
	return &T;
}

void grf4MultTrans (TRSF *D)
{
	multTrsf (&T, &T, D);
	(*fxntab.loadTrans)(&T);
}

void grf4Translate (double x, double y, double z)
{
	TRSF D;

	identTrsf (&D);
	D.p.x = x;
	D.p.y = y;
	D.p.z = z;

	multTrsf (&T, &T, &D);
	(*fxntab.loadTrans)(&T);
}

void grf4Rotate (double roll, double pitch, double yaw)
{
	TRSF D;

	rpyToTrsf (&D, roll,pitch,yaw);
	D.p.x = 0.0;
	D.p.y = 0.0;
	D.p.z = 0.0;

	multTrsf (&T, &T, &D);
	(*fxntab.loadTrans)(&T);
}

TRSF *grf4GetView (TRSF *Er)
{
	*Er = E;
	return (Er);
}

TRSF *grf4GetViewTrsfPtr()
{
	return &E;
}
	
void grf4Move (double x, double y, double z)
{
	(*fxntab.move)(x, y, z);
}

void grf4Draw (double x, double y, double z)
{
	(*fxntab.draw)(x, y, z);
}

void grf4Point (double x, double y, double z)
{
	(*fxntab.point)(x, y, z);
}

void grf4Move2 (double x, double y)
{
	(*fxntab.move2)(x, y);
}

void grf4Draw2 (double x, double y)
{
	(*fxntab.draw2)(x, y);
}

void grf4Point2 (double x, double y)
{
	(*fxntab.point2)(x, y);
}

void grf4MoveRaw (double x, double y)
{
	(*fxntab.moveRaw)(x, y);
}

void grf4DrawRaw (double x, double y)
{
	(*fxntab.drawRaw)(x, y);
}

void grf4PointRaw (double x, double y)
{
	(*fxntab.pointRaw)(x, y);
}

void grf4PutText (char *str, int x, int y)
{
	(*fxntab.putText)(str, x, y);
}

void grf4GetTextExtent (int *xp, int *yp, int *wp, int *hp, char *str)
{
	(*fxntab.getTextExtent)(xp, yp, wp, hp, str);
}

Grf4Object *grf4GenDisc (int npts, double radius, double depth, void *material)
{
	/* creates a disc defined by a regular polygon of 
	   'npts' in the x-y plane, projected upwards along
	   z by a distance 'depth'. The top face lies in the
	   plane z=depth, while the bottom face lies in the
	   plane z=0. The regular polygon is centered on
	   the x-y plane origin. 

	   'material' points to a grf4 material definition.
	 */

	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));
	obj->ptr.prism = (*fxntab.genDisc)(npts, radius, depth);
	obj->type = GRF_PRISM;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

Grf4Object *grf4GenCone (int npts, double radius1, double radius2, 
			 double depth, void *material)
{
	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));
	obj->ptr.prism = (*fxntab.genCone)(npts, radius1, radius2, depth);
	obj->type = GRF_PRISM;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

Grf4Object *grf4GenPrism(int npts, float *bot, float *top, 
			 char *hidden, double depth, void *material)
{
	/* Creates a prism formed by extruding a polygon in the
	   x-y plane along the z coordinate by an amount given by
	   'depth'. The bottom face corresponds to z=0, while the
	   top face corresponds to z=depth.

	   'npts' is the number of points in the defining polygon.

	   'bot' is an 2*npts array of floats defining the x-y vertices
	   of the bottom polygon (x1, y1, x2, y2, ... ). The vertices
	   are assumed to be ordered counter-clockwise when viewed from
	   *inside* the prism.

	   'top' is an optional argument giving the vertices of the
	   top polygon (which may be different from bottom polygon but
	   which still must have npts vertices). If top=NULL, then 
	   the top polygon is assumed to be identical to the bottom
	   polygon.

	   'hidden' is an array of 'npts' chars which determines
	   how the side edges of the prism should be rendered. If
	   hidden[i] == '1', then the edge corresponding to the i-th
	   polygon vertex is assumed to be smooth, and should be
	   rendered using shading. Otherwise, if hidden[i] == '0',
	   the edge is assumed to be crisp. For example, all the
	   vertices of a smooth disk would have hidden[i] == '1'.

	   'material' points to a grf4 material definition.
	 */

	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));
	obj->ptr.prism = (*fxntab.genPrism)(npts, bot, top, hidden, depth);
	obj->type = GRF_PRISM;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

Grf4Object *grf4GenBox(double lenx, double leny, double lenz, void *material)
{
	
	float bot[8];
	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));

	/* generate the bottom face */

	bot[0] = 0.0;   bot[1] = 0.0;
	bot[2] = lenx;  bot[3] = 0.0;
	bot[4] = lenx;  bot[5] = leny;
	bot[6] = 0.0;   bot[7] = leny;

	obj->ptr.prism = (*fxntab.genPrism)(4, bot, (float*)0, (char*)0, lenz);
	obj->type = GRF_PRISM;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

Grf4Object *grf4GenSphere (int npts, double radius, void *material)
{
	/* Creates a sphere of given 'radius', centered on the
	   origin, using 'npts'*'npts' polygons. 'material' is
	   a GRF4 material definition. */

	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));
	obj->ptr.sphere = (*fxntab.genSphere)(npts, radius);
	obj->type = GRF_SPHERE;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

Grf4Object *grf4GenPolygon (int npts, float *vtxs, void *material)
{
	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));
	obj->ptr.polygon = (*fxntab.genPolygon)(npts, vtxs);
	obj->type = GRF_POLYGON;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

Grf4Object *grf4GenRectangle (double width, double height, void *material)
{
	float vtxs[8];

	vtxs[0] = -width/2;
	vtxs[1] = -height/2;
	vtxs[2] =  width/2;
	vtxs[3] = -height/2;
	vtxs[4] =  width/2;
	vtxs[5] =  height/2;
	vtxs[6] = -width/2;
	vtxs[7] =  height/2;
	return (grf4GenPolygon (4, vtxs, material));
}

Grf4Object *grf4GenRegPolygon (int npts, double size, void *material)
{
	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));
	obj->ptr.polygon = (*fxntab.genRegPolygon)(npts, size);
	obj->type = GRF_POLYGON;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

Grf4Object *grf4GenCircle (int npts, double radius, void *material)
{
	/* This is the same as grf4GenRegPolygon(), except that
	   the size is controlled by a radius instead of a side length */

	double ang;	/* half the angle subtended by one side */
	double side;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	ang = M_PI/npts;
	/* largest radius is given by (side/2)/sin(ang), and the
	   smallest radius is given by (side/2)/tan(ang), where 'side'
	   is the length of one side of the approximating polygon.
	   We calculate side so that radius reflects the average
	   of these two amounts. */
	side = 4*radius/(1/sin(ang) + 1/tan(ang));

	return (grf4GenRegPolygon(npts, side, material));
}

Grf4Object *grf4GenLineSet (int size, float *lines, void *material)
{
	Grf4Object *obj;

	if (!grf4FxnInitDone)
	 { if (grf4DoFxnInit(0) < 0)
	    { return (0);
	    }
	 }
	if (grf4HostType == 0)
	 { return (0);
	 }
	obj = (Grf4Object*)calloc(1,sizeof(Grf4Object));
	obj->ptr.lineSet = (*fxntab.genLineSet)(size, lines);
	obj->type = GRF_LINESET;
	if (material)
	 { obj->material = material;
	 }
	return (obj);
}

double grf4SetPrismDepth (Grf4Object *obj, double depth)
{
	if (obj->type != GRF_PRISM)
	 { return (-1.0);
	 }
	else
	 { return (*fxntab.setDepth)(obj, depth);
	 }	
}

void grf4SetObjMaterial (Grf4Object *obj, void *material)
{
	obj->material = material;
}

void grf4DrawObject (Grf4Object *obj)
{
	/* Renders the indicated object */

	int matSaved = 0;

	drawObjCnt++;

	if (obj->material && 
	    (obj->material != currentMaterial ||
	     obj->type == GRF_RIGIDBODY ||
	     obj->type == GRF_FRAMENODE))
	 { matSaved = 1;
	   pushMaterial();
	   currentMaterial = obj->material;
	   (*fxntab.setMaterial)(obj->material);
	 }
	switch (obj->type)
	 { case GRF_PRISM:
	    { (*fxntab.drawPrism)(obj);
	      break;
	    }
	   case GRF_DISC:
	    { (*fxntab.drawPrism)(obj);
	      break;
	    }
	   case GRF_CONE:
	    { (*fxntab.drawPrism)(obj);
	      break;
	    }
	   case GRF_SPHERE:
	    { (*fxntab.drawSphere)(obj);
	      break;
	    }
	   case GRF_POLYGON:
	    { (*fxntab.drawPolygon)(obj);
	      break;
	    }
	   case GRF_LINESET:
	    { (*fxntab.drawLineSet)(obj);
	      break;
	    }
#ifdef LIMA_SUPPORT
	   case GRF_POLYHED3:
	    { (*fxntab.drawPolyhed3)(obj->ptr.gen);
	      break;
	    }
	   case GRF_RIGIDBODY:
	    { grf4DrawCompoundBody(obj);
	      break;
	    }
	   case GRF_FRAMENODE:
	    { grf4DrawFrameNode(obj);
	      break;
	    }
#endif
	 }
	if (matSaved)
	 { popMaterial();
	 }
}

void grf4DrawPolyhed (void *ptr)
{
	(*fxntab.drawPolyhed3)(ptr);
}

void *grf4DefMaterialD(MATERIAL_DEF *matd)
{
	return (*fxntab.defMaterialD)(matd);
}

void *grf4DefMaterial(unsigned long diffuse, unsigned long ambient, 
		      unsigned long emission, unsigned long specular, 
		      double shine, double alpha)
{
	/* used to define GRF4 materials. The format and arguments
	   are virtually identical to those used in an old GL
	   material definition, except that RGB values are expressed
	   using bitmasks packed into the lower 24 bits of a long word.
	 */
	return (*fxntab.defMaterial)(diffuse, ambient, emission,
				     specular, shine, alpha);
}

void grf4SetMaterial(void *m)
{
	currentMaterial = m;
	(*fxntab.setMaterial)(m);
}

void *grf4GetMaterial()
{
	return currentMaterial;
}

void grf4PushMaterial()
{
	pushMaterial();
}

void grf4PopMaterial()
{
	popMaterial();
}

void *grf4DefLight(unsigned long color, unsigned long ambient, 
		   double x, double y, double z, double w)
{
	return (*fxntab.defLight)(color, ambient, x,y,z,w);
}

void grf4SetLight (int n, void *l)
{
	(*fxntab.loadTrans)(&E);
	(*fxntab.setLight)(n,l);
	(*fxntab.loadTrans)(&T);
}


int alistCmp(const void *p1, const void *p2)
{
	AlistElem *a1 = (AlistElem*)p1;
	AlistElem *a2 = (AlistElem*)p2;

	if (a1->left < a2->left)
	 { return (-1);
	 }
	else if (a1->left > a2->left)
	 { return (1);
	 }
	else
	 { return (0);
	 }
}

void alistBuild (AlistElem **fwdp, AlistElem **revp, AlistElem *base, int n)
{
	AlistElem *fwd, *rev;
	int i;

	fwd = (AlistElem*)malloc(n*sizeof(AlistElem));
	for (i=0; i<n; i++)
	 { fwd[i] = base[i];
	 }
	qsort (fwd, n, sizeof(AlistElem), alistCmp);

	rev = (AlistElem*)malloc(n*sizeof(AlistElem));
	for (i=0; i<n; i++)
	 { rev[i].right = base[i].left;
	   rev[i].left = base[i].right;
	 }
	qsort (rev, n, sizeof(AlistElem), alistCmp);

	*fwdp = fwd;
	*revp = rev;
}

int alistSearch (int *rp, int key, AlistElem *base, int n)
{
	AlistElem keyElem;
	AlistElem *ptr;

	keyElem.left = key;
	ptr = (AlistElem*)bsearch(&keyElem,base,n,sizeof(AlistElem),alistCmp);
	if (ptr)
	 { *rp = ptr->right;
	   return (1);
	 }
	else
	 { return (0);
	 }	
}

void grf4Fatal VA_ARGDEF((char* fmt, ...))
{
	va_list ap;
#if !HAS_STDARG
	char *fmt;

	va_start(ap);
	fmt = va_arg(ap, char *);
#else
	va_start(ap,fmt);
#endif
	fprintf (stderr, "ERROR -- ");
	vfprintf (stderr, fmt, ap);
	va_end (ap);
	exit (1);
}

void grf4CalcVirtualCursor (double *vxp, double *vyp, int x, int y)
{
	*vxp = virtL + x*(virtW)/(double)winw;
	*vyp = virtB + y*(virtH)/(double)winh;
}

/* -------------------- button module ------------------ */

/* The button panel is laid out, horizontally, as follows:
	
	|<------ winw ------>||<-----panelWidth------>|
	XXXXXXXXXXXXXXXXXXXXXXO.....BBBBBBBBBBBBBB.....
			       |<->||<-- b->w -->|
		    buttonSpacing|

   The vertical layout looks like this:
	._
	.|
	.| buttonSpacing
	.- 
	B- 
	B| b->h
	B| 
	B- b->y
*/

static BUTTON_DESC *getButton(int num)
{
	BUTTON_DESC *b;

	for (b=buttonList; b; b=b->next)
	 { if (b->num == num)
	    { return (b);
	    }
	 }
	return(0);
}

int grf4AddButton(char *name, int flags, int sizeK, 
		  void (*fxn)(int, int, long), long arg)
{
	BUTTON_DESC *b;

	if (grf4InitDone)
	 { fprintf (stderr, 
		    "grf4AddButton(): must be called before grf4Init()\n");
	   exit(-1);
	 }
	if (sizeK < 1)
	 { fprintf (stderr, "grf4AddButton(): bad value for sizeK\n");
	   exit (-1);
	 }

	b = ALLOC(BUTTON_DESC);
	b->next = buttonList;
	buttonList = b;

	b->name = (char*)malloc(strlen(name)+1);
	strcpy (b->name, name);
	b->flags = flags;
	b->fxn = fxn;
	b->arg = arg;
	b->num = ++numButtons;
	b->sizeK = sizeK;
	return (b->num);
}

void grf4SetButtonFlags (int num, int flags)
{
	BUTTON_DESC *b;

	if (b = getButton (num))
	 { b->flags = flags;
	   if (grf4InitDone)	/* then grf4Init() has been called */
	    { (*fxntab.drawButton)(b);
	    }
	 }
}

int grf4GetButtonFlags (int num)
{
	BUTTON_DESC *b;

	if (b = getButton (num))
	 { return (b->flags);
	 }
	return (-1);
}

void grf4SetButtonName (int num, char *name)
{
	BUTTON_DESC *b;

	if (b = getButton (num))
	 { if (b->name)
	    { free (b->name);
	    }
	   b->name = (char*)malloc(strlen(name)+1);
	   strcpy (b->name, name);
	   if (grf4InitDone)	/* then grf4Init() has been called */
	    { (*fxntab.getButtonExtent)(b);
	      (*fxntab.drawButton)(b);
	    }
	 }
}

void grf4ButtonInit(int *w, int *h)
{
	/* Called by the local init function whenever it is safe
	   to call getButtonExtent(). It returns the desired height
	   and width of the button panel. */

	BUTTON_DESC *b, *b_next, *bRowBegin;
	int rowSizeK, n, kw, kwMax, row, maxH, maxW;

	/* If there are no buttons, return a panelWidth of 0 */

	if (buttonList == 0)
	 { *w = grf4PanelWidth = 0;
	   *h = grf4PanelHeight = 0;
	   return;
	 }

	/* Reverse the direction of the button list */

	b = buttonList;
	buttonList = 0;
	for ( ; b; b=b_next)
	 { b_next = b->next;
	   b->next = buttonList;
	   buttonList = b;
	 }

	buttonSpacing = 4; /* MAX(4,0.2*maxH) */

	/* Compute the  minimum height and width for each button,
	   and row number, for each button. Compute a few width
	   parameters relevant for each row as well. */

	row = -1;
	maxW = maxH = 0;
	for (b=buttonList; b; b=b->next)
	 { if (!(b->flags & BF_SAMEROW) || row == -1)
	    { row++;
	      bRowBegin = b;
	      rowSizeK = kwMax = n = 0;
	    }
	   (*fxntab.getButtonExtent)(b);
	   if (b->minh > maxH)
	    { maxH = b->minh;
	    }
	   b->row = row;
	   rowSizeK += b->sizeK;
	   n++;
	   if ((kw = (int)ceil(b->minw/(double)b->sizeK)) > kwMax)
	    { kwMax = kw;
	    }
	   if (!b->next || !(b->next->flags & BF_SAMEROW))
	    { bRowBegin->rowSizeK = rowSizeK;
	      bRowBegin->rowMinW = rowSizeK*kwMax + (n-1)*buttonSpacing;
	      bRowBegin->rowN = n;
	      if (bRowBegin->rowMinW > maxW)
	       { maxW = bRowBegin->rowMinW;
	       }
	    }
	 }

	buttonWidth = maxW;
	buttonHeight = maxH;
	/* the extra 1 on panelWidth is for the dividing line */
	grf4PanelWidth = maxW + 2*buttonSpacing + 1;
	grf4PanelHeight = (row+2)*buttonSpacing + (row+1)*buttonHeight;

	*w = grf4PanelWidth;
	*h = grf4PanelHeight;
}

static void buttonCompile(int w, int h)
{
	/* Called from the init routine once we know the overall width 
	   of the window. Sets the sizes and offsets so that drawButton() 
	   will work.

	   'w' is the entire width of the window, including the 
	    button panel width. */

	BUTTON_DESC *b;
	int baseX, baseY;
	int row;
	int kernW, dkw, dsw, ns, deltaW, n;
	
	baseY = h;
	row = -1;
	for (b=buttonList; b; b=b->next)
	 { if (b->row != row)
	    { row = b->row;
	      baseY -= (buttonHeight + buttonSpacing);
	      baseX = w-(buttonSpacing+buttonWidth);
	      kernW = (b->rowMinW - (b->rowN-1)*buttonSpacing)/(b->rowSizeK);
	      deltaW = buttonWidth - b->rowMinW;
	      dkw = deltaW/b->rowSizeK;
	      n = 0;
	      if (b->rowN > 1)
	       { deltaW = deltaW%b->rowSizeK;
		 dsw = deltaW/(b->rowN-1);
		 ns = deltaW%(b->rowN-1);
	       }
	      else
	       { dsw = ns = 0;
	       }
	    }
	   b->w = b->sizeK*(kernW + dkw);
	   b->h = buttonHeight;
	   b->x = baseX;
	   b->y = baseY;
	   baseX += (buttonSpacing + dsw + b->w);
	   if (n++ < ns)
	    { baseX += 1;
	    }
	 }
}

static void buttonResize(int dw, int dh)
{
	/* Called whenever we resize the window, to update the 
	   button offsets. */

	BUTTON_DESC *b;

	for (b=buttonList; b; b=b->next)
	 { b->x += dw;
	   b->y += dh;
	 }
}

static void drawAllButtons()
{
	BUTTON_DESC *b;

	for (b=buttonList; b; b=b->next)
	 { (*fxntab.drawButton)(b);
	 }
}

static BUTTON_DESC *overButton(int x, int y)
{
	register BUTTON_DESC *b;

	if (x <= buttonSpacing ||
	    x >= grf4PanelWidth-buttonSpacing)
	 { return (0);
	 }
	for (b=buttonList; b; b=b->next)
	 { if (!(b->flags & BF_INACTIVE) &&
		(x >= b->x && x < b->x+b->w) && (y >= b->y && y < b->y+b->h))
	    { return (b);
	    }
	 }
	return (0);
}

static int dragX0;
static int dragY0;
static int dragging;
static BUTTON_DESC *buttonPressed;
static BUTTON_DESC *buttonInFocus;
static int mouseButtonDown;

#define BUTTON1	1
#define BUTTON2	2
#define BUTTON3	3

void grf4OnButtonEvent(int win, int num, int pressed, int x, int y)
{
	BUTTON_DESC *bd;

	if (pressed)
	 { if (mouseButtonDown)
	    { return;
	    }
	   if (win == MAINWIN)
	    { if (num == BUTTON1 || num == BUTTON2)
	       { dragX0 = x;
		 dragY0 = y;
		 dragging = 1;
	       }
	    }
	   else if (win == CTRLWIN)
	    { if (num == BUTTON1)
	       { bd = overButton(x, y);
		 if (bd && bd->fxn)
		  { (*bd->fxn)(bd->num, BF_PRESSED, bd->arg);
		  }
		 buttonPressed = bd;
	       }
	    }
	   mouseButtonDown = num;
	 }
	else
	 { if (mouseButtonDown != num)
	    { return;
	    }
	   if (dragging > 0)
	    { dragging = 0;
	    }
	   else if (dragging == 1)
	    { /* onSelect (dragX0, dragY0); */
	      dragging = 0;
	    }
	   else if (bd = buttonPressed)
	    { if (bd->fxn)
	       { (*bd->fxn)(bd->num, BF_RELEASED, bd->arg);
	       }
	      buttonPressed = 0;
	    }
	   mouseButtonDown = 0;
	 }
}

void grf4OnMotionEvent (int win, int x, int y)
{
	BUTTON_DESC *bd;

	if (dragging && viewCtrlMode)
	 { onDrag (mouseButtonDown, dragX0, dragY0, x, y);
	   dragX0 = x;
	   dragY0 = y;
	   dragging++;
	 }
	bd = NULL;
	if (win == CTRLWIN)
	 { bd = overButton (x, y);
	 }
	if (bd != buttonInFocus)
	 { if (buttonInFocus && buttonInFocus->fxn)
	    { (*buttonInFocus->fxn)(buttonInFocus->num, BF_OUTFOCUS,
				    buttonInFocus->arg);
	    }
	   if (bd && bd->fxn)
	    { (*bd->fxn)(bd->num, BF_INFOCUS, bd->arg);
	    }
	   buttonInFocus = bd;	     
	 }
}

void grf4OnFocusChangeEvent (int win)
{
	if (win != MAINWIN)
	 { if (dragging)
	    { dragging = 0;
	    }
	 }
	if (win != CTRLWIN)
	 { if (buttonInFocus && buttonInFocus->fxn)
	    { (*buttonInFocus->fxn)(buttonInFocus->num, BF_OUTFOCUS,
				    buttonInFocus->arg);
	      buttonInFocus = NULL;
	    }
	 }
	if (mouseButtonDown)
	 { mouseButtonDown = 0;
	 }
	if (win == CTRLWIN)
	 { BUTTON_DESC *bd;
	   Grf4Event evx;

	   (*fxntab.pollEvent)(win, GRF_CURSOR, &evx);
	   bd = overButton (evx.val1, evx.val2);
	   if (bd && bd->fxn)
	    { (*bd->fxn)(bd->num, BF_INFOCUS, bd->arg);
	      buttonInFocus = bd;	     
	    }
	 }
}

static void onButton (int n, int code, long mode)
{
	if (code == BF_PRESSED)
	 { if (viewCtrlMode != mode)
	    { switch (viewCtrlMode)
	       { case YAW_PITCH_CTRL:
		  { grf4ClearButton (yawCtrlButton);
		    grf4ClearButton (pitchCtrlButton);
		    break;
		  }
		 case X_Y_CTRL:
		  { grf4ClearButton (xCtrlButton);
		    grf4ClearButton (yCtrlButton);
		    break;
		  }
		 case PANY_PANX_CTRL:
		  { grf4ClearButton (panXCtrlButton);
		    grf4ClearButton (panYCtrlButton);
		    break;
		  }
		 case ZOOM_CTRL:
		  { grf4ClearButton (zoomCtrlButton);
		    break;
		  }
		 case TILT_CTRL:
		  { grf4ClearButton (tiltCtrlButton);
		    break;
		  }
	       }
	      switch (mode)
	       { case YAW_PITCH_CTRL:
		  { grf4LightButton (yawCtrlButton);
		    grf4LightButton (pitchCtrlButton);
		    break;
		  }
		 case X_Y_CTRL:
		  { grf4LightButton (xCtrlButton);
		    grf4LightButton (yCtrlButton);
		    break;
		  }
		 case PANY_PANX_CTRL:
		  { grf4LightButton (panXCtrlButton);
		    grf4LightButton (panYCtrlButton);
		    break;
		  }
		 case ZOOM_CTRL:
		  { grf4LightButton (zoomCtrlButton);
		    break;
		  }
		 case TILT_CTRL:
		  { grf4LightButton (tiltCtrlButton);
		    break;
		  }
	       }
	      viewCtrlMode = mode;
	      xMotionEnable = yMotionEnable = 1;
	    }
	   else
	    { int butx, buty;

	      switch (viewCtrlMode)
	       { case YAW_PITCH_CTRL:
		  { butx = yawCtrlButton;
		    buty = pitchCtrlButton;
		    break;
		  }
		 case X_Y_CTRL:
		  { butx = xCtrlButton;
		    buty = yCtrlButton;
		    break;
		  }
		 case PANY_PANX_CTRL:
		  { butx = panXCtrlButton;
		    buty = panYCtrlButton;
		    break;
		  }
		 default:
		  { return;
		  }
	       }
	      if (n == butx)
	       { if (xMotionEnable)
		  { grf4ClearButton(butx);
		    xMotionEnable = 0;
		  }
		 else
		  { grf4LightButton(butx);
		    xMotionEnable = 1;
		  }
	       }
	      else if (n == buty)
	       { if (yMotionEnable)
		  { grf4ClearButton(buty);
		    yMotionEnable = 0;
		  }
		 else
		  { grf4LightButton(buty);
		    yMotionEnable = 1;
		  }
	       }
	      else
	       { grf4Fatal ("onButton(): bogus button no. %d\n", n);
	       }
	    }
	 }
}


void grf4OnResizeEvent(int win, int w, int h)
{
	winw = w;
	winh = h;

	if (grf4HostType != GL_H)
	 { gdrResize (winw, winh);
	 }
	adjustScreenTrans (winw, winh);
}

void grf4OnRedrawEvent(int win)
{
	if (win == MAINWIN)
	 { if (redrawFxn)
	    { (*redrawFxn)(redrawArg);
	    }
	 }
	else if (win == CTRLWIN)
	 { drawAllButtons();
	 }
}

static void bumpRpyValue (int idx, int inc)
{
	struct rpyButton *bd;
	char buf[32];
	TRSF Enew;
	int n;

	bd = &rpyButtons[idx];

	Enew = E;
	n = (NINT(bd->val)+inc)/inc;

	bd->val = range180 ((double)n*inc);
	rpyToTrsf (&Enew, rpyButtons[ROLL_BUTTON].val, 
			  rpyButtons[PITCH_BUTTON].val, 
			  rpyButtons[YAW_BUTTON].val);
	switch (idx)
	 { case ROLL_BUTTON:
	    { sprintf (buf, "roll:  %6.1f", bd->val);
	      break;
	    }
	   case PITCH_BUTTON:
	    { sprintf (buf, "pitch: %6.1f", bd->val);
	      break;
	    }
	   case YAW_BUTTON:
	    { sprintf (buf, "yaw:   %6.1f", bd->val);
	      break;
	    }
	 }
	grf4SetButtonName (bd->dispButton, buf);
	doSetView (&Enew, /*updateButtons=*/NO);
	if (redrawFxn)
	 { (*redrawFxn)(redrawArg);
         }
}

static void onRpyButton(int n, int code, long idx)
{
	struct rpyButton *bd;

	bd = &rpyButtons[idx];

	if (code == BF_PRESSED)
	 { if (n == bd->upButton)
	    { bumpRpyValue (idx, 90);
	    }
	   else if (n == bd->downButton)
	    { bumpRpyValue (idx, -90);
	    }
	   else
	    { grf4Fatal ("onRpyButton(): bogus button: n=%d idx=%d\n",n,idx);
	    }			  
	 }
}

void grf4ClearButton(int num)
{
	int flags;
	
	flags = grf4GetButtonFlags (num);
	grf4SetButtonFlags (num, flags & ~(BF_ILLUMINATED));
}

void grf4LightButton(int num)
{
	int flags;
	
	flags = grf4GetButtonFlags (num);
	grf4SetButtonFlags (num, flags | BF_ILLUMINATED);
}

static void onDrag(int button, int x1, int y1, int x2, int y2)
{
	TRSF D, Enew;
	float roll, pitch, yaw;
	float px, py, pz;
	int deltaX, deltaY;
	float scale;

	deltaX = x2-x1;
	deltaY = y2-y1;
	
	if (viewCtrlMode == NO_VIEW_CTRL)
	 { return;
	 }

	if (button == BUTTON1)
	 { scale = 1.0;
	 }
	else
	 { scale = 0.1;
	 }
#define DEADBAND	5

	switch (viewCtrlMode)
	 { case YAW_PITCH_CTRL:
	    { 
	      yaw = pitch = 0.0;
	      if (yMotionEnable)
	       { yaw = -scale*90.0*deltaY/winh;
	       }
	      if (xMotionEnable)
	       { pitch = scale*90.0*deltaX/winw;
	       }
	      zeroVect (&D.p);
	      rpyToTrsf (&D, 0.0, pitch, yaw);
	      Enew = E;
	      zeroVect (&Enew.p);
	      multTrsf (&Enew, &D, &Enew);
	      Enew.p = E.p;
	      break;
	    }
	   case PANY_PANX_CTRL:
	    { 
	      yaw = pitch = 0.0;
	      if (yMotionEnable)
	       { yaw = scale*90.0*deltaY/winh;
	       }
	      if (xMotionEnable)
	       { pitch = -scale*90.0*deltaX/winw;
	       }
	      zeroVect (&D.p);
	      rpyToTrsf (&D, 0.0, pitch, yaw);
	      Enew = E;
	      multTrsf (&Enew, &D, &Enew);
	      break;
	    }
	   case TILT_CTRL:
	    { 
	      roll = scale*90.0*deltaX/winw;
	      zeroVect (&D.p);
	      rpyToTrsf (&D, roll, 0.0, 0.0);
	      Enew = E;
	      multTrsf (&Enew, &D, &Enew);
	      break;
	    }
	   case X_Y_CTRL:
	    { if (perspective)
	       { scale *= (desiredNear+desiredFar)/(2*desiredNear);
	       }
	      else
	       { scale = 1.0;
	       }
	      Enew = E;
	      if (xMotionEnable)
	       { Enew.p.x += scale*virtW*deltaX/winw;
	       }
	      if (yMotionEnable)
	       { Enew.p.y += scale*virtH*deltaY/winh;
	       }
	      break;
	    }
	   case ZOOM_CTRL:
	    { float r1, r2;

	      scale *= (desiredFar-desiredNear)/(2*virtW);
	      Enew = E;
	      x1 -= winw/2;
	      x2 -= winw/2;
	      y1 -= winh/2;
	      y2 -= winh/2;
	      r1 = sqrt((double)(x1*x1+y1*y1));
	      r2 = sqrt((double)(x2*x2+y2*y2));
	      Enew.p.z += scale*(r2-r1)*virtW/winw;
	    }
	 }
	doSetView (&Enew, /*updateButtons=*/YES);
	if (redrawFxn)
	 { (*redrawFxn)(redrawArg);
         }
}

static void updateRpyButtons(double r, double p, double y)
{
	char buf[64];
	struct rpyButton *bd;

	bd = &rpyButtons[ROLL_BUTTON];
	if (r != bd->val)
	 { sprintf (buf, "roll:  %6.1f", r);
	   grf4SetButtonName (bd->dispButton, buf);
	   bd->val = r;
	 }

	bd = &rpyButtons[PITCH_BUTTON];
	if (p != bd->val)
	 { sprintf (buf, "pitch: %6.1f", p);
	   grf4SetButtonName (bd->dispButton, buf);
	   bd->val = p;
	 }

	bd = &rpyButtons[YAW_BUTTON];
	if (y != bd->val)
	 { sprintf (buf, "yaw:   %6.1f", y);
	   grf4SetButtonName (bd->dispButton, buf);
	   bd->val = y;
	 }
}

void grf4PointToScr (int *sxp, int *syp, double x, double y, double z)
{
	double ex, ey, ez;
	double wx, wy;

	/* first convert to eye coordinates */
	ex = T.n.x*x + T.o.x*y + T.a.x*z + T.p.x;
	ey = T.n.y*x + T.o.y*y + T.a.y*z + T.p.y;
	ez = T.n.z*x + T.o.z*y + T.a.z*z + T.p.z;

	/* then convert to window coordinates */
	if (perspective)
	 { wx = ex*(-desiredNear/ez);
	   wy = ey*(-desiredNear/ez);
	 }
	else
	 { wx = ex;
	   wy = ey;
	 }

	/* finally, compute screen coordinates */
	*sxp = (wx-virtL)*(winw/virtW);
	*syp = (wy-virtB)*(winh/virtH);
}

double grf4ScrToPlane (VECT *p, int sx, int sy, VECT *nrm, double off)
{
	/* Plane is defined to be those points p for which

	   p . nrm + off = 0
	*/
	VECT nrmE, pntE;
	double offE;
	double lam;

	pntE.x = sx*(virtW/(double)winw) + virtL;
	pntE.y = sy*(virtH/(double)winh) + virtB;

	/* map plane into eye coordinates */
	invRotateVect (&nrmE, nrm, &T);
	offE = off - dotVect(&T.p, &nrmE);

	/* now intersect pick line with the plane */
	
	if (perspective)
	 { 
	   pntE.z = -desiredNear;
	   lam = -offE/dotVect(&pntE, &nrmE);
	   scaleVect (&pntE, lam, &pntE);
	 }
	else
	 { lam = -(-offE - pntE.x*nrmE.x - pntE.y*nrmE.y)/nrmE.z;
	   pntE.z = -lam;
	 }
	transVect (p, &pntE, &T);
	return lam;
}

#define EPS (1e-10)

static int lineClosestPoints (double *lam, VECT *u1, VECT *b1, VECT *u2, VECT *b2)
{
	VECT db;
	double u1u1, u1u2, u2u2;
	double dbu1, dbu2;
	double denom;

	subtractVect (&db, b2, b1);

	u2u2 = dotVect (u2, u2);
	u1u1 = dotVect (u1, u1);
	u1u2 = dotVect (u1, u2);
	dbu1 = dotVect (&db, u1);
	dbu2 = dotVect (&db, u2);

	denom = u1u2*u1u2 - u1u1*u2u2;

	if (ABS(denom) < EPS)
	 { lam[0] = HUGE_VAL;
	   lam[1] = HUGE_VAL;
	   return 0;
	 }
	else
	 { lam[0] = (-u2u2*dbu1 + u1u2*dbu2)/denom;
	   lam[1] = (-u1u2*dbu1 + u1u1*dbu2)/denom;
	   return 1;
	 }	
}

double grf4ScrToLine (VECT *p, int sx, int sy, VECT *u, VECT *b)
{
	/* Line is defined to be those points p for which

	   p  = b + x u
	*/
	VECT uE, bE;
	VECT u0, b0;
	double wx, wy;
	double lam[2];

	/* map line into eye coordinates */
	invTransVect (&bE, b, &T);
	invRotateVect (&uE, u, &T);
/*	unitVect (&uE, &uE); */

	wx = sx*(virtW/(double)winw) + virtL;
	wy = sy*(virtH/(double)winh) + virtB;

	if (perspective)
	 { u0.x = wx;
	   u0.y = wy;
	   u0.z = -desiredNear;
	   unitVect (&u0, &u0);
	   zeroVect (&b0);
	 }
	else
	 { u0.x = 0;
	   u0.y = 0;
	   u0.z = -1;
	   b0.x = wx;
	   b0.y = wy;
	   b0.z = 0;
	 }

	lineClosestPoints (lam, &u0, &b0, &uE, &bE);
	if (lam[0] != HUGE_VAL)
	 { 
	   combineVect (p, 1.0, &b0, lam[0], &u0);
	   transVect (p, p, &T);
	 }
	return lam[1];
}
