#ifdef HAS_OPEN_GL

/* Check:

	font WRT contexts
*/

#define USE_RAW_X_TEXT_CALLS	0

/* Check ~mouse/X/Xlib.doc */

#include <stdio.h>
#include <errno.h>
#include <ctype.h>	
#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>

#include <GL/gl.h>
//#include <GL/glu.h> 
#include <GL/glx.h>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>
#include <X11/keysym.h>

#include <cdefs.h>
#include "grf4Tval.h"
#include <grf4.h>
#include "grf4Internal.h"
#include "grf4GLObject.h"

/* -------- X11 support functions --------- */

static XrmDatabase db;
static char *defaults = "\
*Foreground: white\n\
*Background: black\n\
*BorderColor: white\n\
*BorderWidth: 1\n\
*BorderMargin: 0\n\
*Font: fixed\n\
*Name: robotsim\n\
*IconName: robotsim\n\
";

static char *displayname;
static char *geometryspec;
static char *foreground;
static char *background;
static char *bordercstr;
static char *borderwstr;
static char *bordermstr;
static char *margincstr;
static char *name;
static char *iconname;
static char *fontname;

static int argc;
static char **argv;

static Display *disp;
static Screen *scr;
static int width;
static int height;
static int depth;
static Window rootwin;
static Colormap defcmap;
static Visual *visual;

static XColor fgcolor;
static XColor bgcolor;
static XColor bdcolor;
static XColor mgcolor;

static int reverse = 0;

static Window topwin;
static Colormap wincmap;
static int topx;
static int topy;
static int topw;
static int toph;
static int margin;
static int borderwidth;
static XFontStruct *font;
static long eventMask;

static Pixmap pixmap = None;

static saveargv();
static handleargs();
static maybeset();
static setup_db();
static char *get_default_value();
static char *defaultValue();
static setup_color();
static setup_font();
static setup_colors();
static setup_numbers();
static setup_windows();
static resize();

static int iGLGetEventFd();
static int iGLEnableEvent();
static int iGLDisableEvent();
static int iGLPollEvent();
static int iGLReadEvent();
static void iGLInitDeviceMaps();

#if USE_RAW_X_TEXT_CALLS
static GC topWinGc;
static GC panWinGc;
static GC panRevGc;
#endif

typedef struct {
	Window id;	/* id returned from XCreateWindow() */
	long x, y;	/* position of the window (WRT topwin) */
	int w, h;	/* height and width */
	int active;	/* true if the window is active */
	unsigned long eventMask; /* selected events */
} WinDesc;

static WinDesc wdesc[NUMWINS];

#define MAIN	wdesc[MAINWIN]
#define CTRL	wdesc[CTRLWIN]

static saveargv(int ac, char **av)
{
	int i;
	int nc;
	char *abuf;

	argc = ac;
	argv = (char **)malloc((ac+1)*sizeof(char *));
	nc = 1;
	for (i=0;i<ac;i++)
	 { nc += strlen(av[i]) + 1;
	 }
	abuf = (char *) malloc(nc*sizeof(char));
	for (i=0;i<ac;i++)
	 { argv[i] = abuf;
	   strcpy(abuf,av[i]);
	   while (*abuf) abuf ++;
	   abuf ++;
	 }
	*abuf = '\0';
	argv[ac] = 0;
}

static handleargs(int ac, char **av)
{
	int skip;
	int errs;

	skip = 0;
	errs = 0;
	for (ac--,av++;ac;ac--,av++)
	 { if (skip > 0)
	    { skip --;
	      continue;
	    }
	   if (**av != '-')
	    { fprintf(stderr,"unrecognized argument `%s'\n",*av);
	      errs ++;
	      continue;
	    }
	   if (0)
	    {
needarg:;
	      fprintf(stderr,"%s needs a following argument\n",*av);
	      errs ++;
	      continue;
	    }
#define WANTARG() do { if (++skip >= ac) goto needarg; } while (0)
	   if (!strcmp(*av,"-display"))
	    { WANTARG();
	      displayname = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-geometry"))
	    { WANTARG();
	      geometryspec = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-foreground") || !strcmp(*av,"-fg"))
	    { WANTARG();
	      foreground = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-background") || !strcmp(*av,"-bg"))
	    { WANTARG();
	      background = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-bordercolor") || !strcmp(*av,"-bd"))
	    { WANTARG();
	      bordercstr = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-borderwidth") || !strcmp(*av,"-bw"))
	    { WANTARG();
	      borderwstr = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-bordermargin") || !strcmp(*av,"-bm"))
	    { WANTARG();
	      bordermstr = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-margincolor") || !strcmp(*av,"-mc"))
	    { WANTARG();
	      margincstr = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-name"))
	    { WANTARG();
	      name = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-iconname"))
	    { WANTARG();
	      iconname = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-font") || !strcmp(*av,"-fn"))
	    { WANTARG();
	      fontname = av[skip];
	      continue;
	    }
	   if (!strcmp(*av,"-reverse"))
	    { reverse = 1;
	      continue;
	    }
#undef WANTARG
	   fprintf(stderr,"unrecognized option `%s'\n",*av);
	   errs ++;
	 }
	if (errs)
	 { exit(1);
	 }
}

static maybeset (char **strp, char *str)
{
	if (str && !*strp) *strp = str;
}

static setup_db()
{
	char *str;
	char *home;
	XrmDatabase db2;

	db = XrmGetStringDatabase(defaults);
	str = XResourceManagerString(disp);
	if (str)
	 { db2 = XrmGetStringDatabase(str);
	   XrmMergeDatabases(db2,&db);
	 }
	else
	 { home = getenv("HOME");
	   if (home)
	    { str = (char*)malloc(strlen(home)+1+10+1);
	      sprintf(str,"%s/.Xdefaults",home);
	      db2 = XrmGetFileDatabase(str);
	      if (db2)
	       { XrmMergeDatabases(db2,&db);
	       }
	      free(str);
	    }
	 }
}

static char *get_default_value(char *name, char *class)
{
	char *type;
	XrmValue value;

	if (XrmGetResource(db,name,class,&type,&value) == False) return(0);
	return(value.addr);
}

static char *defaultValue(char *name, char *prop, char *Name, char *class)
{
	char nbuf[256];
	char cbuf[256];

	sprintf (nbuf, "%s.%s", name, prop);
	sprintf (cbuf, "%s.%s", Name, class);
	return (get_default_value(nbuf,cbuf));
}

static setup_color(char *str, XColor *col)
{
	if (XParseColor(disp,wincmap,str,col) == 0)
	 { fprintf(stderr,"bad color `%s'\n",str);
	   exit(1);
	 }
	while (1)
	 { if (XAllocColor(disp,wincmap,col) == 0)
	    { if (wincmap != defcmap)
	       { fprintf(stderr,"can't allocate colormap cell for color `%s'\n",str);
		 exit(1);
	       }
	      wincmap = XCopyColormapAndFree(disp,wincmap);
	      continue;
	    }
	   break;
	 }
}

static setup_font()
{
	font = fontname ? XLoadQueryFont(disp,fontname) : 0;
	if (fontname && !font)
	 { fprintf(stderr,"can't load font %s, using server default\n",fontname);
	 }
}

static setup_colors()
{
	wincmap = defcmap;
	setup_color(foreground,&fgcolor);
	setup_color(background,&bgcolor);
	setup_color(bordercstr,&bdcolor);
	setup_color(margincstr,&mgcolor);
	if (reverse)
	 { XColor ct;
	   ct = fgcolor;
	   fgcolor = bgcolor;
	   bgcolor = ct;
	 }
}

static setup_numbers()
{
	if (bordermstr) margin = atoi(bordermstr);
	if (borderwstr) borderwidth = atoi(borderwstr);
}

static XVisualInfo *mainVis;
static XVisualInfo *ctrlVis;
static GLXContext mainGlc;
static GLXContext ctrlGlc;
static Colormap mainCmap;
static Colormap ctrlCmap;
static int mainInit = 0;
static int ctrlInit = 0;
static int iGLflags = 0;

static Bool WaitForNotify(Display *d, XEvent *e, char *arg)
{
	return (e->type == MapNotify) && (e->xmap.window == (Window)arg);
}

/* --------- GL support functions -------- */

TRSF* iGLTransPtr;
int iGLPerspectiveMode;
float iGLNearPlane;

extern int grf4GeneralFlags; 

static TRSF Trans;

static int bgColor;
static int fgColor;

static float setLeft;
static float setRight;
static float setBottom;
static float setTop;
static float setNear;
static float setFar;

static int screenCoordsSet = 0;
static int useZbuffer = 0;
static int savedZbufferState;

static int fontBase = -1;

static GLfloat identMatrix[16] = { 1.0, 0.0, 0.0, 0.0,
			      0.0, 1.0, 0.0, 0.0,
			      0.0, 0.0, 1.0, 0.0,
			      0.0, 0.0, 0.0, 1.0 };

static unsigned long getcpack();
static void iGLOrtho();
static void iGLWindow();
static void iGLLoadTrans();
static void setScreenCoords();
static void clearScreenCoords();
static do_lmdef();

static void doDrawButton();
static void iGLPxlDrawStub();
static void iGLPxlPutTextStub();
static void iGLPxlPointStub();
static void iGLDraw();
static void iGLMove();
static void iGLDraw2();
static void iGLMove2();
static void clearColor(int val);
static void iGLClear();
static void iGLSwapBuffers();

static void decompose();
static void *iGLDefMaterial();
static void *iGLDefMaterialD();
static void *iGLDefLight();
static void iGLSetLight();
static void iGLResetLights();
static long iGLCapacity();

static void iGLGetWindowSize();
static void iGLSetWindowExtent();
static void iGLDrawPanel();
static void iGLPutText();
static void iGLGetTextExtent();
static void iGLDrawButton();
static void iGLGetButtonExtent();
static int iGLGetEventFd();
static int iGLEnableEvent();
static int iGLDisableEvent();
static int iGLPollEvent();
static int iGLReadEvent();
static void iGLInitDeviceMaps();
static void iGLcharstr (char *str);
#if USE_RAW_X_TEXT_CALLS == 0
static void iGLFontInit ();
#endif

typedef struct
{
	GLfloat diffuse[4];
	GLfloat ambient[4];
	GLfloat emission[4];
	GLfloat specular[4];
	GLfloat shininess;
} MaterialDef;

typedef struct
{
	GLfloat diffuse[4];
	GLfloat ambient[4];
	GLfloat specular[4];
	GLfloat position[4];
} LightDef;

static unsigned long getcpack()
{
	short r,g,b;
	/* OGLXXX get GL_CURRENT_COLOR: scale color values */
	{float tmp[4]; glGetFloatv(GL_CURRENT_COLOR, tmp);
	 r=255*tmp[0];
	 g=255*tmp[1];
	 b=255*tmp[2];
	}
	return (RGBPACK(r,g,b));
}

static void iGLOrtho(double l, double r, double b, double t, double n, double f)
{
	setLeft = l;
	setRight = r;
	setBottom = b;
	setTop = t;
	setNear = n;
	setFar = f;

	iGLPerspectiveMode = 0;
	iGLNearPlane = n;
	 
	{ GLint mm; glGetIntegerv(GL_MATRIX_MODE, &mm);glMatrixMode(GL_PROJECTION);glLoadIdentity();glOrtho(l, r, b, t, n, f);glMatrixMode(mm);};
}
	
static void iGLWindow(double l, double r, double b, double t, double n,double f)
{
	setLeft = l;
	setRight = r;
	setBottom = b;
	setTop = t;
	setNear = n;
	setFar = f;

	iGLPerspectiveMode = 1;
	iGLNearPlane = n;

	 { GLint mm;
	   glGetIntegerv(GL_MATRIX_MODE, &mm);
	   glMatrixMode(GL_PROJECTION);
	   glLoadIdentity();
	   glFrustum(l, r, b, t, n, f);
	   glMatrixMode(mm);
	 }
}

#define LIGHT_GREY	RGBPACK(160,160,160)
#define DARK_GREY	RGBPACK(100,100,100)
#define DEEP_GREY	RGBPACK(60,60,60)
#undef WHITE
#define WHITE		RGBPACK(255,255,255)
#undef BLACK
#define BLACK		RGBPACK(0,0,0)


static cunpack (float *vf, int val)
{
	vf[0] = ((val & 0x0000ff)>>0)/255.;
	vf[1] = ((val & 0x00ff00)>>8)/255.;
	vf[2] = ((val & 0xff0000)>>16)/255.;
}

static void setcpack (unsigned long val)
{
	unsigned char ubv[4];
	ubv[0] = (val>>0)&0xff;
	ubv[1] = (val>>8)&0xff;
	ubv[2] = (val>>16)&0xff;
	ubv[3] = (val>>24)&0xff;
	glDisable (GL_LIGHTING);
	glColor4ubv(ubv);
}

static void iGLDrawPanel()
{
	if (CTRL.active)
	 { glXMakeCurrent (disp, CTRL.id, ctrlGlc);
	   clearColor (LIGHT_GREY);
	   glXMakeCurrent (disp, MAIN.id, mainGlc);
	 }
}

	/* OGLXXX
	 * Ambient:
	 * 	If this is a light model lmdef, then use glLightModelf 
	 *	and GL_LIGHT_MODEL_AMBIENT.
	 * Include ALPHA parameter with ambient
	 */
static float lmodel_1[] = { GL_AMBIENT, 0, 0, 0,
		        };

#define MIN_WINW	21
#define MIN_WINH	21

static void iGLGetWindowSize (int *wp, int *hp, int *directionOfYp)
{
	if (wp) *wp = MAIN.w;
	if (hp) *hp = MAIN.h;
	if (directionOfYp) *directionOfYp = 1;
}

static void iGLPutText (char *str, int x, int y)
{
#if USE_RAW_X_TEXT_CALLS
	XDrawImageString(disp, MAIN.id, topWinGc,
			 x, MAIN.h-1-y, str, strlen(str));
#else
	glRasterPos2i(x,  y);
	iGLcharstr (str);
#endif
}

static void putButtonText (char *str, int x, int y, int color)
{
#if USE_RAW_X_TEXT_CALLS
	/* this is the way to do things using straight X11 character
	   calls. Problem is that we need to get use graphics contexts
	   which match the foreground and background colors properly.
	 */
	GC gc = (color==BLACK ? panRevGc : panWinGc);
	XDrawImageString (disp, CTRL.id, gc, x, CTRL.h-1-y, str, strlen(str));
#else
	setcpack (color);
	glRasterPos2i (x, y);
	iGLcharstr (str);
#endif
}

static void iGLDrawButton(BUTTON_DESC *b)
{
	int x1,x2;
	int y1,y2;

	int baseColor;
	int charColor;
	int topEdgeColor;
	int bottomEdgeColor;

	int cbaseX;
	int cbaseY;

	glXMakeCurrent(disp, CTRL.id, ctrlGlc);
	x1 = b->x;
	y1 = b->y;
	x2 = x1 + b->w - 1;
	y2 = y1 + b->h - 1;

	setcpack(WHITE);	/* white */

	cbaseX = b->x + b->coffX + (b->w-b->minw)/2;
	cbaseY = b->y + b->coffY + (b->h-b->minh)/2;

	if (b->flags & BF_NOTBOXED)
	 { setcpack(LIGHT_GREY);
	   glRecti(x1, y1, x2, y2);
	   if (b->flags & BF_ILLUMINATED)
	    { charColor = WHITE;
	    }
	   else
	    { charColor = BLACK;
	    }
	 }
	else
	 { setcpack(BLACK);
	/* OGLXXX rect: remove extra PolygonMode changes */
	   glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); 
	   glRecti(x1, y1, x2, y2); 
	   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);		/* outer black box */
	   if (b->flags & BF_ILLUMINATED)
	    { baseColor = LIGHT_GREY;
	      charColor = WHITE;
	      topEdgeColor = WHITE;
	      bottomEdgeColor = DARK_GREY;
	    }
	   else
	    { baseColor = DARK_GREY;
	      charColor = BLACK;
	      topEdgeColor = DEEP_GREY;
	      bottomEdgeColor = LIGHT_GREY;
	    }
	   setcpack(topEdgeColor);
	/* OGLXXX rect: remove extra PolygonMode changes */
	   glPolygonMode(GL_FRONT_AND_BACK, GL_LINE); 
	   glRecti(x1+1, y1+1, x2-1, y2-1); 
	   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	   setcpack(bottomEdgeColor);
	   glBegin(GL_LINES); 
	   glVertex2i(x1+1, y1+1);
	   glVertex2i(x2-1, y1+1);
	   glEnd();
	   setcpack(baseColor);
	   glRecti(x1+2, y1+2, x2-2, y2-2);
	 }
	putButtonText (b->name, cbaseX, cbaseY, charColor);
	glXMakeCurrent(disp, MAIN.id, mainGlc);
}

static double mpos[3];

static void iGLPxlDrawStub(int x1, int y1, int x2, int y2)
{
	grf4Fatal ("Internal software error: call to pixel level function\n");
}

static void iGLPxlPutTextStub(char *s, int x1, int y1)
{
	grf4Fatal ("Internal software error: call to pixel level function\n");
}

static void iGLPxlPointStub(int x1, int y1)
{
	grf4Fatal ("Internal software error: call to pixel level function\n");
}

static void iGLDraw(double x, double y, double z)
{
	/* OGLXXX Add glEnd() after these vertices,  before next glBegin() */
	glBegin(GL_LINES);
        glVertex3f(mpos[0], mpos[1], mpos[2]);
	glVertex3f(x, y, z);
	glEnd();
	mpos[0] = x;
	mpos[1] = y;
	mpos[2] = z;
}

static void iGLMove(double x, double y, double z)
{
	mpos[0] = x;
	mpos[1] = y;
	mpos[2] = z;
}

static void iGLPoint(double x, double y, double z)
{
	/* OGLXXX points: put as many vertices as possible between Begin and End */
	glBegin(GL_POINTS);
	glVertex3f(x, y, z);
	glEnd();
}

static void iGLDraw2(double x, double y)
{
	/* OGLXXX Add glEnd() after these vertices,  before next glBegin() */
	glBegin(GL_LINES);
        glVertex2f(mpos[0], mpos[1]);
	glVertex2f(x, y);
	glEnd();
	mpos[0] = x;
	mpos[1] = y;
}

static void iGLMove2(double x, double y)
{
	mpos[0] = x;
	mpos[1] = y;
}

static void iGLPoint2(double x, double y)
{
	/* OGLXXX points: put as many vertices as possible between Begin and End */
	glBegin(GL_POINTS);
	glVertex2f(x, y);
	glEnd();
}

static void clearColor(int val)
{
	float colv[4];
	cunpack (colv, val);
	glClearIndex(0);
	glClearColor(colv[0], colv[1], colv[2], 0);
	glClear(GL_COLOR_BUFFER_BIT);
}

static void iGLClear()
{
	clearColor (bgColor);
	if (useZbuffer)
	 { glClearDepth(1.);
	   glClear(GL_DEPTH_BUFFER_BIT);
	 }
}

static void iGLLoadTrans(TRSF *T)
{
	GLfloat m[16];

	m[0*4+0] = T->n.x;
	m[0*4+1] = T->n.y;
	m[0*4+2] = T->n.z;
	m[0*4+3] = 0.0;

	m[1*4+0] = T->o.x;
	m[1*4+1] = T->o.y;
	m[1*4+2] = T->o.z;
	m[1*4+3] = 0.0;

	m[2*4+0] = T->a.x;
	m[2*4+1] = T->a.y;
	m[2*4+2] = T->a.z;
	m[2*4+3] = 0.0;

	m[3*4+0] = T->p.x;
	m[3*4+1] = T->p.y;
	m[3*4+2] = T->p.z;
	m[3*4+3] = 1.0;

	Trans = *T;
	iGLTransPtr = &Trans;
	glLoadMatrixf(m);
}


/* Lighting and material functions */

#define BLUE_PART(m)	(((m)&0xff0000)>>16)
#define GREEN_PART(m)	(((m)&0xff00)>>8)
#define RED_PART(m)	(((m)&0xff))

static void decompose(float *v, unsigned long m)
{
	v[0] = RED_PART(m)/255.0;
	v[1] = GREEN_PART(m)/255.0;
	v[2] = BLUE_PART(m)/255.0;
}

static MaterialDef *doDefMaterial (unsigned long d, unsigned long a, 
				   unsigned long e, unsigned long s, 
				   double shine, double alpha)
{
	MaterialDef *mdef = ALLOC(MaterialDef);

	cunpack (mdef->diffuse, d);
	mdef->diffuse[3] = alpha;
	cunpack (mdef->ambient, a);
	mdef->ambient[3] = alpha;
	cunpack (mdef->emission, e);
	mdef->emission[3] = alpha;
	cunpack (mdef->specular, s);
	mdef->specular[3] = alpha;
	mdef->shininess = shine;
	return mdef;
}

static void *iGLDefMaterial(unsigned long diffuse, unsigned long ambient, 
			    unsigned long emission, unsigned long specular, 
			    double shine, double alpha)
{
	return doDefMaterial (diffuse, ambient, emission,
			      specular, shine, alpha);
}

static void *iGLDefMaterialD(MATERIAL_DEF *md)
{
	return doDefMaterial (md->diffuse, md->ambient, md->emission,
			      md->specular, md->shine, md->alpha);
}

static void iGLSetMaterial(void *ptr)
{
	MaterialDef *mdef = (MaterialDef*)ptr;

	if (!(grf4GeneralFlags & GRF_2DONLY))
	 { glMaterialfv(GL_FRONT, GL_DIFFUSE, mdef->diffuse);
	   glMaterialfv(GL_FRONT, GL_AMBIENT, mdef->ambient);
	   glMaterialfv(GL_FRONT, GL_EMISSION, mdef->emission);
	   glMaterialfv(GL_FRONT, GL_SPECULAR, mdef->specular);
	   glMaterialf (GL_FRONT, GL_SHININESS, mdef->shininess);
	   glEnable (GL_LIGHTING);
	 }
}

void iGLSetMaterialHook(void *ptr)
{
	iGLSetMaterial (ptr);
}

struct lampDesc {
	int code;
	int idx;
  };

static struct lampDesc lampPool[] = {
	GL_LIGHT0, 0, GL_LIGHT1, 0, GL_LIGHT2, 0, GL_LIGHT3, 0, 
	GL_LIGHT4, 0, GL_LIGHT5, 0, GL_LIGHT6, 0, GL_LIGHT7, 0,
};

static int lightIdx = 0;

static void* iGLDefLight(unsigned long color, unsigned long ambient, 
			 double x, double y, double z, double w)
{
	LightDef *ldef;

	ldef = ALLOC(LightDef);
	cunpack (ldef->diffuse, color);
	ldef->diffuse[3] = 1.0;
	cunpack (ldef->ambient, color);
	ldef->ambient[3] = 1.0;
	ldef->specular[0] = 1.0;
	ldef->specular[1] = 1.0;
	ldef->specular[2] = 1.0;
	ldef->specular[3] = 1.0;
	ldef->position[0] = x;
	ldef->position[1] = y;
	ldef->position[2] = z;
	ldef->position[3] = 0;
	return ldef;
}

static void iGLSetLight(int n, void *ptr)
{
	LightDef *ldef = (LightDef*)ptr;

	if (n >= GL_MAX_LIGHTS)
	 { fprintf (stderr, "grf4SetLight(): don't have that many lamps\n");
	   return;
	 }
	if (n < 0)
	 { fprintf (stderr, "grf4SetLight(): negative lamp index\n");
	   return;
	 }
	if (ldef)
	 { glLightfv (GL_LIGHT0+n, GL_DIFFUSE, ldef->diffuse);
	   glLightfv (GL_LIGHT0+n, GL_AMBIENT, ldef->ambient);
	   glLightfv (GL_LIGHT0+n, GL_SPECULAR, ldef->specular);
	   glLightfv (GL_LIGHT0+n, GL_POSITION, ldef->position);
	   glEnable (GL_LIGHTING);
	   glEnable (GL_LIGHT0+n);
	 }
	else
	 { glDisable (GL_LIGHT0+n);
	 }
}

static void iGLResetLights()
{
	int i;
	/* XXX not sure what to do here */
}

static long iGLCapacity()
{
	return (GRF_HARDWARE_A | GRF_COLOR | GRF_LIGHTING);
}

static void iGLcharstr (char *str)
{
	/* OGLXXX charstr: check list numbering */
	glPushAttrib (GL_LIST_BIT);
	glListBase(fontBase);
	glCallLists(strlen(str), GL_UNSIGNED_BYTE, str);
	glPopAttrib ();
}

#if USE_RAW_X_TEXT_CALLS == 0
static void iGLFontInit()
{
	/* extra stuff so that GL can draw text using this font: */
	int first, last, k;

	first = font->min_char_or_byte2;
	last = font->max_char_or_byte2;
	k = glGenLists(1);
	fontBase = glGenLists(last+1);
	if (fontBase == 0)
	 { fprintf (stderr, 
"iGLFontInit(): couldn't create display lists\n");
	   fprintf (stderr, "glGenLists(): error %d\n", glGetError());
	   exit (1);
	 }
	glXUseXFont(font->fid, first, last-first+1, fontBase+first);
}
#endif

static void iGLSwapBuffers ()
{
	glXSwapBuffers (disp, MAIN.id);
	glFlush();
}

#if 0
printAttr(Window win)
{
	/* This was used for debugging */
	XWindowAttributes a;

	XGetWindowAttributes (disp, win, &a);

	printf ("x=%d\n", a.x);
	printf ("y=%d\n", a.y);
	printf ("w=%d\n", a.width);
	printf ("h=%d\n", a.height);
	printf ("border_width=%d\n", a.border_width);
	printf ("depth=%d\n", a.depth);
	printf ("class=%d\n", a.class);
	printf ("bit_gravity=%d\n", a.bit_gravity);
	printf ("win_gravity=%d\n", a.win_gravity);
	printf ("backing_store=%d\n", a.backing_store);
	printf ("backing_planes=%d\n", a.backing_planes);
	printf ("backing_pixel=%d\n", a.backing_pixel);
	printf ("save_under=%d\n", a.save_under);
	printf ("map_installed=%d\n", a.map_installed);
	printf ("map_state=%d\n", a.map_state);
	printf ("all_event_masks=%d\n", a.all_event_masks);
	printf ("your_event_mask=%d\n", a.your_event_mask);
	printf ("do_not_propagate_mask=%d\n", a.do_not_propagate_mask);
	printf ("override_redirect=%d\n", a.override_redirect);
}
#endif

static setup_windows(char *Name, char *name, int x, int y,
		     unsigned int w, unsigned int h, int minW, int minH)
{
	int vattr[32];
	int vattrIdx = 0;
#define BEGIN_VIS_ATTR()	vattrIdx = 0
#define PUT_VIS_ATTR(x)		vattr[vattrIdx++] = (x)
#define END_VIS_ATTR()		vattr[vattrIdx] = None
	int bits;
	unsigned long int attrmask;
	unsigned long int gcvalmask;
	XSetWindowAttributes attr;
	XWindowAttributes aget;
	XGCValues gcval;
#ifdef X11R3
	XSizeHints sizehints;
#else
	XTextProperty wn_prop;
	XTextProperty in_prop;
	XSizeHints normal_hints;
	XWMHints wm_hints;
	XClassHint class_hints;
#endif
	XEvent event;

	minW = MAX(minW,1);
	minH = MAX(minH,1);

	bits = XParseGeometry(geometryspec,&x,&y,&w,&h);
	if (bits & XNegative) x = width + x - w;
	if (bits & YNegative) y = height + y - h;

	/* get an appropriate visual */
	BEGIN_VIS_ATTR();
	PUT_VIS_ATTR(GLX_RGBA);
	PUT_VIS_ATTR(GLX_DEPTH_SIZE);
	PUT_VIS_ATTR(1);
	if (!(iGLflags & GRF_DOUBLEBUF_OFF))
	 { PUT_VIS_ATTR(GLX_DOUBLEBUFFER);
	 }
	END_VIS_ATTR();
	mainVis = glXChooseVisual(disp, DefaultScreen(disp), vattr);

	/* create a GLX context */
	mainGlc = glXCreateContext(disp, mainVis, 0, GL_TRUE);
	/* create a color map */
	mainCmap = XCreateColormap(disp, rootwin, mainVis->visual, AllocNone);

	attrmask = 0;
	attr.background_pixel = mgcolor.pixel;
	attrmask |= CWBackPixel;
	attr.border_pixel = bdcolor.pixel;
	attrmask |= CWBorderPixel;
	attr.backing_store = NotUseful;
	attrmask |= CWBackingStore;
	attr.event_mask = StructureNotifyMask;
	attrmask |= CWEventMask;
	w -= 2 * borderwidth;
	h -= 2 * borderwidth;

	topwin = XCreateWindow(disp,rootwin,
			       x,y,w,h,
			       borderwidth,
			       depth,
			       InputOutput,
			       CopyFromParent,
			       attrmask,&attr);
#ifdef X11R3
	sizehints.flags = 0;
	if (CTRL.active)
	 { sizehints.min_width = (CTRL.w+1) + minW;
	   sizehints.min_height = MAX(minH,CTRL.h);
	 }
	else
	 { sizehints.min_width = minW;
	   sizehints.min_height = minH;
	 }
	sizehints.flags |= PMinSize;
	sizehints.width = w;
	sizehints.height = h;
	sizehints.flags |= (bits & (WidthValue|HeightValue)) ? USSize : PSize;
	if (bits & (XValue|YValue))
	 { sizehints.x = x;
	   sizehints.y = y;
	   sizehints.flags |= USPosition;
	 }
	XSetStandardProperties(disp,topwin,Name,name,None,argv,argc,&sizehints);
#else	
	wn_prop.value = (unsigned char *)Name;
	wn_prop.encoding = XA_STRING;
	wn_prop.format = 8;
	wn_prop.nitems = strlen((char *)wn_prop.value);
	in_prop.value = (unsigned char *)name;
	in_prop.encoding = XA_STRING;
	in_prop.format = 8;
	in_prop.nitems = strlen((char *)in_prop.value);
	normal_hints.flags = PMinSize | PResizeInc;
	normal_hints.x = x;
	normal_hints.y = y;
	normal_hints.flags |= (bits & (XValue|YValue)) ? USPosition : PPosition;
	normal_hints.width = w;
	normal_hints.height = h;
	normal_hints.flags |= (bits & (WidthValue|HeightValue)) ? USSize : PSize;
	if (CTRL.active)
	 { normal_hints.min_width = (CTRL.w+1) + minW;
	   normal_hints.min_height = MAX(minH,CTRL.h);
	 }
	else
	 { normal_hints.min_width = minW;
	   normal_hints.min_height = minH;
	 }
	normal_hints.width_inc = 1;
	normal_hints.height_inc = 1;
	wm_hints.flags = InputHint;
	wm_hints.input = True;
	class_hints.res_name = name;
	class_hints.res_class = Name;
	XSetWMProperties(disp,topwin,&wn_prop,&in_prop,argv,argc,&normal_hints,&wm_hints,&class_hints);
#endif
	attrmask = 0;
	attr.colormap = mainCmap;
	attrmask |= CWColormap;
	attr.border_pixel = bdcolor.pixel;
	attrmask |= CWBorderPixel;
	attr.background_pixel = bgcolor.pixel;
	attrmask |= CWBackPixel;
	attr.backing_store = NotUseful;
	attrmask |= CWBackingStore;
	eventMask = ExposureMask;
	attr.event_mask = eventMask;
	attrmask |= CWEventMask;
	MAIN.id = XCreateWindow(disp,topwin,0,0,1,1,0,mainVis->depth,
		     InputOutput, mainVis->visual, attrmask,&attr);
	MAIN.eventMask = attr.event_mask;
	XMapRaised(disp,MAIN.id);

#if USE_RAW_X_TEXT_CALLS
	gcvalmask = 0;
	if (font)
	 { gcval.font = font->fid;
	   gcvalmask |= GCFont;
	 }
	gcval.foreground = fgcolor.pixel;
	gcvalmask |= GCForeground;
	gcval.background = bgcolor.pixel;
	gcvalmask |= GCBackground;
	topWinGc = XCreateGC(disp,topwin,gcvalmask,&gcval);
#endif
	/* set up GL stuff for the MAIN window */
	
 	if (CTRL.active)
	 { long mask;

	   BEGIN_VIS_ATTR();
	   PUT_VIS_ATTR(GLX_RGBA);
	   END_VIS_ATTR();

	   ctrlVis = glXChooseVisual(disp, DefaultScreen(disp), vattr);
	   /* create a GLX context */
	   ctrlGlc = glXCreateContext(disp, ctrlVis, 0, GL_TRUE);
	   /* create a color map */
	   ctrlCmap = XCreateColormap(disp, rootwin, ctrlVis->visual, AllocNone);
	   attr.colormap = ctrlCmap;
	   CTRL.id = XCreateWindow(disp,topwin,0,0,1,1,0,ctrlVis->depth,
				   InputOutput,ctrlVis->visual,attrmask,&attr);
	   CTRL.eventMask = attr.event_mask;
	   XMapRaised(disp,CTRL.id);
#if USE_RAW_X_TEXT_CALLS
	   panWinGc = XCreateGC(disp,CTRL.id,gcvalmask,&gcval);
	   panRevGc = XCreateGC(disp,CTRL.id,gcvalmask,&gcval);
	   XSetForeground(disp,panRevGc,bgcolor.pixel);
	   XSetBackground(disp,panRevGc,fgcolor.pixel);
#endif
	   mask = attr.event_mask;
	   mask |= EnterWindowMask|LeaveWindowMask;
	   mask |= ButtonPressMask|ButtonReleaseMask;
	   mask |= PointerMotionMask;
	   XSelectInput (disp, CTRL.id, mask);

	 }
	XMapRaised(disp,topwin);
	MAIN.x = margin;	/* this won't change */
	MAIN.y = margin;

	XGetWindowAttributes (disp, topwin, &aget);
	topx = aget.x;
	topy = aget.y;
	topw = w;
	toph = h;

	/* wait for main window */
	XIfEvent(disp, &event, WaitForNotify, (char*)topwin);

	resize(w,h);
}

#define BUTTON_BORDER_WIDTH	2

static void iGLGetTextExtent (int *xp, int *yp, int *wp, int *hp, char *str)
{
	int direction;
	int fontAscent;
	int fontDescent;
	XCharStruct cs;

	XTextExtents (font, str, strlen(str), 
		      &direction,
		      &fontAscent,
		      &fontDescent,
		      &cs);
	if (xp) *xp = -cs.lbearing;
	if (yp) *yp = MAX(cs.descent,fontDescent);
	if (wp) *wp = cs.rbearing-cs.lbearing;
	if (hp) *hp = MAX(cs.ascent,fontAscent) + MAX(cs.descent,fontDescent);
}

static void iGLGetButtonExtent(BUTTON_DESC *b)
{
	int x, y, w, h;
	int margin;

	iGLGetTextExtent (&x, &y, &w, &h, b->name);
	margin = 1 + BUTTON_BORDER_WIDTH;
	b->coffX = margin + x;
	b->coffY = margin + y;
	b->minw = 2*margin + w;
	b->minh = 2*margin + h;
}

static resize(int tw, int th)
{
	int x, y, w, h;

	x = MAIN.x;
	y = MAIN.y;

	w = tw - (2*margin);
	h = th - (2*margin);
	if (CTRL.active)
	 { w -= (CTRL.w+1);	/* extra 1 to give space for dividing line */
	 }
	XMoveResizeWindow(disp,MAIN.id,x,y,w,h);
	glXMakeCurrent (disp, MAIN.id, mainGlc);

	MAIN.w = w;
	MAIN.h = h;

	if (mainInit == 0)
	 { /* set up GL stuff for main window */
	   iGLFontInit();
	   glMatrixMode(GL_MODELVIEW);
	   if (!(iGLflags & GRF_2DONLY))
	    { glEnable(GL_DEPTH_TEST);
	      useZbuffer = 1;
	    }
/*	   glDepthFunc(GL_LEQUAL); */
	   glLoadMatrixf(identMatrix);
	   iGLOrtho (-0.5, MAIN.w-0.5, -0.5, MAIN.h-0.5, -1.0, 1.0);
	   mainInit = 1;
	 }
	if (MAIN.w-1 >= 0)
	 { glViewport(0,  0, ( MAIN.w-1)-(0)+1, ( MAIN.h-1)-( 0)+1);
	   glScissor(0,  0, ( MAIN.w-1)-(0)+1, ( MAIN.h-1)-( 0)+1);
	 }
	if (CTRL.active)
	 { XMoveResizeWindow(disp,CTRL.id,x+w+1,y,CTRL.w,CTRL.h);

	   CTRL.x = x+w+1;
	   CTRL.y = y;
	   glXMakeCurrent (disp, CTRL.id, ctrlGlc);
	   /* set up GL stuff for CTRL panel window */
	   if (ctrlInit == 0)
	    { iGLFontInit();
	      ctrlInit = 1;
	    }
	   glViewport(0,  0, ( CTRL.w-1)-(0)+1, ( CTRL.h-1)-( 0)+1);
	   glScissor(0,  0, ( CTRL.w-1)-(0)+1, ( CTRL.h-1)-( 0)+1);
	    { GLint mm; glGetIntegerv(GL_MATRIX_MODE, &mm);
	      glMatrixMode(GL_PROJECTION);
	      glLoadIdentity();
	      glOrtho(-0.5, CTRL.w-0.5, -0.5, CTRL.h-0.5,-1,1);
	      glMatrixMode(mm);
	    }
	   iGLDrawPanel();
	   glXMakeCurrent (disp, MAIN.id, mainGlc);
	 }
/*	setup_pixmap();*/
}

grf4GLInit (char *label, int fgcolor, int bgcolor, WinPlaceDesc *placing, 
	    int flags)
{
	int ac = 1;
	char *dummyProgName = "dummyProgName";
	char **av = &dummyProgName;

	char Nbuf[256];
	char nbuf[256];

	int x, y, w, h;

	iGLflags = flags;

	strcpy(nbuf, label);
	strcpy(Nbuf, label);
	nbuf[0] = tolower(nbuf[0]);
	Nbuf[0] = toupper(Nbuf[0]);

	saveargv(ac,av);
	handleargs(ac,av);
	if ((disp = XOpenDisplay(displayname)) == 0)
	 { return (-1);
	 }

	scr = XDefaultScreenOfDisplay(disp);
	width = XWidthOfScreen(scr);
	height = XHeightOfScreen(scr);
	depth = XDefaultDepthOfScreen(scr);
	rootwin = XRootWindowOfScreen(scr);
	defcmap = XDefaultColormapOfScreen(scr);
	visual = XDefaultVisualOfScreen(scr);
	setup_db();

	maybeset(&geometryspec, defaultValue(
		 nbuf, "geometry", Nbuf, "Geometry"));
	maybeset(&fontname, defaultValue(
		 nbuf, "font", Nbuf, "Font"));
	maybeset(&foreground, defaultValue(
		 nbuf, "foreground", Nbuf, "Foreground"));
	maybeset(&background, defaultValue(
		 nbuf, "background", Nbuf, "Background"));
	maybeset(&bordercstr, defaultValue(
		 nbuf, "borderColor", Nbuf, "BorderColor"));
	maybeset(&borderwstr, defaultValue(
		 nbuf, "borderWidth", Nbuf, "BorderWidth"));
	maybeset(&bordermstr, defaultValue(
		 nbuf, "borderMargin", Nbuf, "BorderMargin"));
	maybeset(&margincstr, defaultValue(
		 nbuf, "marginColor", Nbuf, "Background"));
	maybeset(&name, defaultValue(
		 nbuf, "name", Nbuf, "Name"));
	maybeset(&iconname, defaultValue(
		 nbuf, "iconName", Nbuf, "IconName"));

	setup_font();

	if (bgcolor)
	 { reverse = 1;
	 }
	setup_colors();
	setup_numbers();

	/* set up the null window with info appropriate to the screen */

	wdesc[NULLWIN].x = 0;
	wdesc[NULLWIN].y = 0;
	wdesc[NULLWIN].w = width;
	wdesc[NULLWIN].h = height;

	grf4ButtonInit (&CTRL.w, &CTRL.h);
	CTRL.active = (CTRL.w > 0);

	if (!placing->placed)
	 { x = y = 200;
	   w = h = 300;
	 }
	else
	 { x = placing->x;
	   y = placing->y;
	   w = placing->w;
	   h = placing->h;
	 }

	w = MAX(w,placing->minW);
	h = MAX(h,placing->minH);

	setup_windows(Nbuf,nbuf,x,y,
		      2*margin+w+CTRL.w,2*margin+MAX(h,CTRL.h),
		      placing->minW, placing->minH);
	iGLInitDeviceMaps();
/*	XAutoRepeatOff (disp); 	This seems to be sticky, at least on SGI's */
	return (1);
}

grf4GLSetupFxns (struct grf4fxnTable *ftab)
{
	ftab->capacity = iGLCapacity;
	ftab->getWindowSize = iGLGetWindowSize;
	ftab->ortho = iGLOrtho;
	ftab->window = iGLWindow;
	ftab->loadTrans = iGLLoadTrans;
	ftab->color = setcpack;
	ftab->clear = iGLClear;
	ftab->flush = iGLSwapBuffers;

	ftab->move = iGLMove;
	ftab->draw = iGLDraw;
	ftab->point = iGLPoint;
	ftab->move2 = iGLMove2;
	ftab->draw2 = iGLDraw2;
	ftab->point2 = iGLPoint2;
	ftab->drawPrism = iGLDrawPrism;
	ftab->drawSphere = iGLDrawSphere;
	ftab->drawPolygon = iGLDrawPolygon;
	ftab->drawLineSet = iGLDrawLineSet;
#ifdef LIMA_SUPPORT
	ftab->drawPolyhed3 = iGLDrawPolyhed3;
#endif
	ftab->genDisc = iGLGenDisc;
	ftab->genCone = iGLGenCone;
	ftab->genPrism = iGLGenPrism;
	ftab->genSphere = iGLGenSphere;
	ftab->genPolygon = iGLGenPolygon;
	ftab->genRegPolygon = iGLGenRegPolygon;
	ftab->genLineSet = iGLGenLineSet;
	ftab->setDepth = iGLSetDepth;

	ftab->putText = iGLPutText;
	ftab->getTextExtent = iGLGetTextExtent;
	ftab->drawButton = iGLDrawButton;
	ftab->getButtonExtent = iGLGetButtonExtent;

	ftab->pxlPutText = iGLPxlPutTextStub;
	ftab->pxlDraw = iGLPxlDrawStub;
	ftab->pxlPoint = iGLPxlPointStub;
	ftab->pxlFlush = iGLPxlDrawStub;
	ftab->pxlClear = iGLPxlDrawStub;

	ftab->defLight = iGLDefLight;
	ftab->defMaterial = iGLDefMaterial;
	ftab->defMaterialD = iGLDefMaterialD;
	ftab->setLight = iGLSetLight;
	ftab->resetLights = iGLResetLights;
	ftab->setMaterial = iGLSetMaterial;

	ftab->getEventFd = iGLGetEventFd;
	ftab->enableEvent = iGLEnableEvent;
	ftab->disableEvent = iGLDisableEvent;
	ftab->pollEvent = iGLPollEvent;
	ftab->readEvent = iGLReadEvent;

	return (1);
}

static AlistElem alistBase[] = { 
	GRF_KB_A,		XK_a,
	GRF_KB_B,		XK_b,
	GRF_KB_C,		XK_c,
	GRF_KB_D,		XK_d,
	GRF_KB_E,		XK_e,
	GRF_KB_F,		XK_f,
	GRF_KB_G,		XK_g,
	GRF_KB_H,		XK_h,
	GRF_KB_I,		XK_i,
	GRF_KB_J,		XK_j,
	GRF_KB_K,		XK_k,
	GRF_KB_L,		XK_l,
	GRF_KB_M,		XK_m,
	GRF_KB_N,		XK_n,
	GRF_KB_O,		XK_o,
	GRF_KB_P,		XK_p,
	GRF_KB_Q,		XK_q,
	GRF_KB_R,		XK_r,
	GRF_KB_S,		XK_s,
	GRF_KB_T,		XK_t,
	GRF_KB_U,		XK_u,
	GRF_KB_V,		XK_v,
	GRF_KB_W,		XK_w,
	GRF_KB_X,		XK_x,
	GRF_KB_Y,		XK_y,
	GRF_KB_Z,		XK_z,
	GRF_KB_0,		XK_0,
	GRF_KB_1,		XK_1,
	GRF_KB_2,		XK_2,
	GRF_KB_3,		XK_3,
	GRF_KB_4,		XK_4,
	GRF_KB_5,		XK_5,
	GRF_KB_6,		XK_6,
	GRF_KB_7,		XK_7,
	GRF_KB_8,		XK_8,
	GRF_KB_9,		XK_9,
	GRF_KB_Break,		XK_Break,
	GRF_KB_Ctrl,		XK_Control_L,
	GRF_KB_CapsLock,	XK_Caps_Lock,
	GRF_KB_RightShift,	XK_Shift_R,
	GRF_KB_LeftShift,	XK_Shift_L,
	GRF_KB_Esc,		XK_Escape,
	GRF_KB_Tab,		XK_Tab,
	GRF_KB_Ret,		XK_Return,
	GRF_KB_Space,		XK_space,
	GRF_KB_Linefeed,	XK_Linefeed,
	GRF_KB_Backspace,	XK_BackSpace,
	GRF_KB_Del,		XK_Delete,
	GRF_KB_Semicolon,	XK_semicolon,
	GRF_KB_Period,		XK_period,
	GRF_KB_Comma,		XK_comma,
	GRF_KB_Quote,		XK_apostrophe,
	GRF_KB_Backquote,	XK_grave,
	GRF_KB_Minus,		XK_minus,
	GRF_KB_Slash,		XK_slash,
	GRF_KB_Backslash,	XK_backslash,
	GRF_KB_Equal,		XK_equal,
	GRF_KB_LeftBracket,	XK_bracketleft,
	GRF_KB_RightBracket,	XK_bracketright,
	GRF_KB_LeftArrow,	XK_Left,
	GRF_KB_DownArrow,	XK_Right,
	GRF_KB_RightArrow,	XK_Up,
	GRF_KB_UpArrow,		XK_Down,
	GRF_KB_Pad0,		XK_KP_0,
	GRF_KB_Pad1,		XK_KP_1,
	GRF_KB_Pad2,		XK_KP_2,
	GRF_KB_Pad3,		XK_KP_3,
	GRF_KB_Pad4,		XK_KP_4,
	GRF_KB_Pad5,		XK_KP_5,
	GRF_KB_Pad6,		XK_KP_6,
	GRF_KB_Pad7,		XK_KP_7,
	GRF_KB_Pad8,		XK_KP_8,
	GRF_KB_Pad9,		XK_KP_9,
	GRF_KB_PadF1,		XK_KP_F1,
	GRF_KB_PadF2,		XK_KP_F2,
	GRF_KB_PadF3,		XK_KP_F3,
	GRF_KB_PadF4,		XK_KP_F4,
	GRF_KB_PadPeriod,	XK_KP_Decimal,
	GRF_KB_PadMinus,	XK_KP_Subtract,
	GRF_KB_PadComma,	XK_KP_Separator,
	GRF_KB_PadEnter,	XK_KP_Enter,
	GRF_KB_LeftAlt,		XK_Alt_L,
	GRF_KB_RightAlt,	XK_Alt_R,
	GRF_KB_RightCtrl,	XK_Control_R,
	GRF_KB_F1,		XK_F1,
	GRF_KB_F2,		XK_F2,
	GRF_KB_F3,		XK_F3,
	GRF_KB_F4,		XK_F4,
	GRF_KB_F5,		XK_F5,
	GRF_KB_F6,		XK_F6,
	GRF_KB_F7,		XK_F7,
	GRF_KB_F8,		XK_F8,
	GRF_KB_F9,		XK_F9,
	GRF_KB_F10,		XK_F10,
	GRF_KB_F11,		XK_F11,
	GRF_KB_F12,		XK_F12,
	GRF_KB_PrintScreen,	XK_Print,			  
	GRF_KB_ScrollLock,	XK_Scroll_Lock,
	GRF_KB_Pause,		XK_Pause,
	GRF_KB_Insert,		XK_Insert,
	GRF_KB_Home,		XK_Home,
	GRF_KB_PageUp,		XK_Prior,
	GRF_KB_End,		XK_End,
	GRF_KB_PageDown,	XK_Next,
	GRF_KB_NumLock,		XK_Num_Lock,
	GRF_KB_PadDivide,	XK_KP_Divide,
	GRF_KB_PadMultiply,	XK_KP_Multiply,
	GRF_KB_PadPlus,		XK_KP_Add,
};

static unsigned char keyToCodeMap[256];
static unsigned char codeToKeyMap[256];

static void iGLInitDeviceMaps()
{
	int i, key, code;

	for (i=0; i<ARRAY_SIZE(alistBase); i++)
	 { key = alistBase[i].left;
	   code = XKeysymToKeycode(disp, alistBase[i].right);
	   if (code < 0 || code > 255 || key <= 0 || key > 255)
	    { printf 
("Warning, X11 init, range troubke with key %d, code=%d\n", key, code);
	    }
	   else
	    { keyToCodeMap[key] = code;
	      if (code > 0)
	       { codeToKeyMap[code] = key;
	       }
	    }
	 }
}

static int iGLGetEventFd ()
{
	return (XConnectionNumber(disp));
}

static unsigned long getX11EventMask(int type)
{
	switch (type)
	 { default:
	    { grf4Fatal ("X11 getEventMask(): bougs type (%d)\n", type);
	    }
	   case GRF_KEY_EVENT:
	    { return (KeyPressMask|KeyReleaseMask);
	    }
	   case GRF_BUTTON_EVENT:
	    { return (ButtonPressMask|ButtonReleaseMask);
	    }
	   case GRF_MOTION_EVENT:
	    { return (PointerMotionMask);
	    }
	   case GRF_FOCUS_EVENT:
	    { return (EnterWindowMask|LeaveWindowMask);
	    }
	   case GRF_RESIZE_EVENT:
	   case GRF_REDRAW_EVENT:
	    { return (0);	/* these are always enabled */
	    }
	 }
}

static int iGLEnableEvent (int win, int type)
{
	unsigned long mask;

	mask = getX11EventMask(type);
	if (mask & wdesc[win].eventMask)
	 { return (1);
	 }
	else
	 { wdesc[win].eventMask |= mask;
	   XSelectInput (disp, wdesc[win].id, wdesc[win].eventMask);
	   return (0);
	 }
}

static int iGLDisableEvent (int win, int type)
{
	unsigned long mask;

	mask = getX11EventMask(type);
	if (!(mask & wdesc[win].eventMask))
	 { return (0);
	 }
	else
	 { wdesc[win].eventMask &= ~mask;
	   XSelectInput (disp, wdesc[win].id, wdesc[win].eventMask);
	   return (1);
	 }
}

static int winIdToNum (Window wid)
{
	if (wid == MAIN.id)
	 { return (MAINWIN);
	 }
	else if (wid == CTRL.id)
	 { return (CTRLWIN);
	 }
	else
	 { return (NULLWIN);
	 }
}

static int iGLPollEvent (int win, int code, Grf4Event *ev)
{
	int val1 = -1;
	int val2 = -1;

	/* Return values from XQueryPointer */
	Window rootWin, childWin;
	int rx, ry, x, y;
	unsigned int mask;
	
	if (GRF_IS_KEY(code))
	 { unsigned char code;
	   char map[32];

	   code = keyToCodeMap[code];
	   if (code == 0)
	    { return (-1);
	    }
	   XQueryKeymap(disp, map);
	   val1 = map[code/8] & (1<<(code%8));		 
	 }
	else
	 { Window wid;

	   wid = (win==NULLWIN) ? MAIN.id : wdesc[win].id;

	   XQueryPointer (disp, wid, &rootWin, &childWin, 
				&rx, &ry, &x, &y, &mask);
	   
	   switch (code)
	    { default:
	       { return (-1);
		 break;
	       }
	      case GRF_LEFT_MOUSE:
	       { val1 = mask & Button1Mask;
		 break;
	       }
	      case GRF_MIDDLE_MOUSE:
	       { val1 = mask & Button2Mask;
		 break;
	       }
	      case GRF_RIGHT_MOUSE:
	       { val1 = mask & Button3Mask;
		 break;
	       }
	      case GRF_CURSOR:
	       { if (win == NULLWIN)
		  { val1 = rx;
		    val2 = (height-1)-ry;
		  }
		 else
		  { val1 = x;
		    val2 = (wdesc[win].h-1)-y;
       		  }
		 break;
	       }
	      case GRF_REDRAW:
	       { if (win == NULLWIN)
		  { val1 = 0;
		    val2 = 0;
		  }
		 else
		  { XWindowAttributes aget;
		 
		    XGetWindowAttributes (disp, topwin, &aget);
		    val1 = aget.x+wdesc[win].x;
		    val2 = (height-wdesc[win].h) - (aget.y+wdesc[win].y);
		  }
		 break;
	       }
	      case GRF_RESIZE:
	       { val1 = wdesc[win].w;
		 val2 = wdesc[win].h;
		 break;
	       }
	      case GRF_FOCUS_CHANGE:
	       { val1 = winIdToNum(childWin);
		 break;
	       }
	    }
	 }
	ev->val1 = val1;
	ev->val2 = val2;
	ev->code = code;
	return (1);
}

/* The event stack is used to insert new events where necessary */

#define EVSTACKSIZE	10

static Grf4Event evStack[EVSTACKSIZE];
static int evStackp = 0;

static int pushEvent (int code, int val1, int val2, int win)
{
	if (evStackp < EVSTACKSIZE)
	 { evStack[evStackp].code = code;
	   evStack[evStackp].val1 = val1;
	   evStack[evStackp].val2 = val2;
	   evStack[evStackp].win  = win;
	   evStackp++;
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static int popEvent (Grf4Event *ev)
{
	if (evStackp)
	 { *ev = evStack[--evStackp];
	   return (1);
	 }
	else
	 { return (0);
	 }
}

static int iGLReadEvent(Grf4Event *ev)
{
	int code = GRF_NULLEVENT;
	int val1 = -1;
	int val2 = -1;
	int win;
	XEvent e, eNext;

	if (popEvent (ev))
	 { return (1);
	 }
	XSync(disp,False);
	if (XQLength(disp) <= 0)
	 { return (0);
	 }
	XNextEvent (disp, &e);

	win = winIdToNum(e.xany.window);

	switch (e.type)
	 { default:
	    { break;
	    }
	   case ConfigureNotify:
	    { int tw, th;	/* width and height of top window */
	      int w, h, x, y;	/* GRF4-relative coords for main window */
	      int doResize = 0;

	      /* absorb following expose events */

	      while (XQLength(disp) > 0)
	       { XPeekEvent (disp, &eNext);
		 if (eNext.xany.window != e.xany.window ||
		     (eNext.type != Expose && 
		      eNext.type != GraphicsExpose) )		     
		  { break;
		  }
		 XNextEvent (disp, &eNext);
	       }
	      tw = e.xconfigure.width;
	      th = e.xconfigure.height;
	       x = e.xconfigure.x+margin;
	       y = (height-MAIN.h) - (e.xconfigure.y+margin);
	       w = tw - 2*margin - CTRL.w;
	       h = th - 2*margin;

	      if (tw != topw || th != toph)
	       { /* resize event */

		 resize (tw, th);
		 doResize = 1;
	       }

	      topw = tw;
	      toph = th;
	      topx = e.xconfigure.x;
	      topy = e.xconfigure.y;

	      pushEvent (GRF_REDRAW, x+CTRL.w, y, CTRLWIN);
	      pushEvent (GRF_REDRAW, x, y, MAINWIN);

	      if (doResize)
	       { pushEvent (GRF_RESIZE, w, h, MAINWIN);
	       }
	      if (CTRL.active)
	       { iGLDrawPanel();
	       }
	      break;
	    }
	   case GraphicsExpose:
	      /* XGraphicsExposeEvent - xgraphicsexpose */
	   case Expose:
	    { /* XExposeEvent - xexpose */
	      Window win;
	      int x, y;

	      if (e.type==GraphicsExpose)
	       { /* XXX - shouldn't depend casting Drawabe into a Window */
		 win = (Drawable)e.xgraphicsexpose.drawable;
	       }
	      else
	       { win = e.xexpose.window;
		 if (e.xexpose.count)
		  { /* at least one more expose event coming, so ignore */
		    break;
		  }
	       }
	      x = topx+margin;
	      y = (height-MAIN.h) - (topy+margin);
	      if (win==MAIN.id)
	       { pushEvent (GRF_REDRAW, x, y, MAINWIN);
	       }
	      else if (win==CTRL.id)
	       { pushEvent (GRF_REDRAW, x+CTRL.w, y, CTRLWIN);
		 iGLDrawPanel();
	       }
	      break;
	    }
	   case MotionNotify:
	    { while (XQLength(disp) > 0)
	       { XPeekEvent (disp, &eNext);
		 if (eNext.type != MotionNotify || 
		     eNext.xany.window != e.xany.window)
		  { break;
		  }
		 XNextEvent (disp, &e);
	       }
	      code = GRF_CURSOR;
	      if (win == CTRLWIN)
	       { val1 = e.xmotion.x;
		 val2 = (CTRL.h-1)-e.xmotion.y;
	       }
	      else
	       { val1 = e.xmotion.x;
		 val2 = (MAIN.h-1)-e.xmotion.y;
	       }
	      break;
	    }
	   case ButtonPress:
	    { switch (e.xbutton.button)
	       { case Button1: code = GRF_LEFT_MOUSE;
		      break;
		 case Button2: code = GRF_MIDDLE_MOUSE;
		      break;
		 case Button3: code = GRF_RIGHT_MOUSE;
		      break;
	       }
	      val1 = 1;
	      break;
	    }
	   case ButtonRelease:
	    { switch (e.xbutton.button)
	       { case Button1: code = GRF_LEFT_MOUSE;
		      break;
		 case Button2: code = GRF_MIDDLE_MOUSE;
		      break;
		 case Button3: code = GRF_RIGHT_MOUSE;
		      break;
	       }
	      val1 = 0;
	      break;
	    }
	   case KeyPress:
	    { code = codeToKeyMap[e.xkey.keycode];
	      val1 = 1;
	      break;
	    }
	   case KeyRelease:
	    { code = codeToKeyMap[e.xkey.keycode];
	      val1 = 0;
	      break;
	    }
	   case EnterNotify:
	    { code = GRF_FOCUS_CHANGE;
	      val1 = win;
	      break;
	    }
	   case LeaveNotify:
	    { code = GRF_FOCUS_CHANGE;
	      val1 = NULLWIN;
	      break;
	    }
	 }
	ev->val1 = val1;
	ev->val2 = val2;
	ev->code = code;
	ev->win = win;
	return (1);
}

#else

#include <stdio.h>

grf4GLInit ()
{
	return(-1);	
}

grf4GLSetupFxns ()
{
	fprintf (stderr, "grf4Init(): Open GL support not available\n");
	return (-1);
}

#endif
