#include <grf4.h>
#include "grf4Internal.h"
#include "grf4Support.h"
#include "grf4GenObject.h"
#include "grf4GenDraw.h"
#include "grf4GlCommon.h"
#include <RxCvxPolyhed3.h>

extern TRSF* grf4TransPtr;
extern TRSF grf4IdentTrsf;
extern int grf4PerspectiveMode;
extern int grf4HiddenLineMode;

static void Vecd3ToV3f (float *fp, const RxVecd3 &vec)
{
	fp[0] = vec.x();
	fp[1] = vec.y();
	fp[2] = vec.z();
}

static int faceVisible (const RxFace3 &f, TRSF *T, float *nrml)
{
	float n[3];
	float v[3];

	Vecd3ToV3f (n, f.plane.n());
	_grf4MultRotV3f (n, T, n);
	if (nrml)
	 { nrml[0] = n[0];
	   nrml[1] = n[1];
	   nrml[2] = n[2];
	 }
	if (!grf4PerspectiveMode)
	 { return (n[2] > 0.0);
	 }
	else
	 { Vecd3ToV3f (v, f.selist.head()->vrtx()->q());
	   _grf4MultTrsfV3f (v, T, v);
	   return (_grf4DotV3f (v, n) < 0.0);
	 }
}

void gdrDrawPolyhed3 (void *ptr)
{
	RxCvxPolyhed3 *poly = (RxCvxPolyhed3 *)ptr;
	const RxFace3ListNode *fnode;
	const RxEdge3ListNode *enode;
	const RxSedg3 *se;
	TRSF *T;

	T = (grf4TransPtr ? grf4TransPtr : &grf4IdentTrsf);

#if HLS
	if (grf4HiddenLineMode)
	 { 
	   for (fnode=poly->faces().head(); fnode; fnode=fnode->next())
	    { float n[3];
	      float v[3];
	      if (faceVisible (fnode->f, T, n))
	       { gdrBeginPolygon((void*)poly);
		 gdrNrml (n[0], n[1], n[2]);
		 se=fnode->f.selist.head();
		 do
		  { int visible;
		    int leftFaceVis;
		    int rightFaceVis;
		    const RxEdge3 *e = se->next()->edge();
		    const RxVecd3 &q = se->vrtx()->q();

		    leftFaceVis = faceVisible (*e->lse.face(), T, (float*)0);
		    rightFaceVis = faceVisible (*e->rse.face(), T, (float*)0);
		    visible = 
			((e->isSoft()==1 && (leftFaceVis != rightFaceVis)) ||
			 (e->isSoft()==0 && (leftFaceVis || rightFaceVis)));

		    v[0] = q.x();
		    v[1] = q.y();
		    v[2] = q.z();
		    _grf4MultTrsfV3f (v, T, v);
		    gdrVrtx (v[0], v[1], v[2], visible);
		    se=se->next();
		  }
		 while (se != fnode->f.selist.head());
		 gdrEndPolygon();
	       }
	    }
	 }
	else
#endif
	 { 
	   for (enode=poly->edges().head(); enode; enode=enode->next())
	    { const RxVecd3 &qh = enode->e.lse.vrtx()->q();
	      const RxVecd3 &qt = enode->e.rse.vrtx()->q();
	      float vh[3];
	      float vt[3];
	      int leftFaceVis;
	      int rightFaceVis;

	      leftFaceVis = faceVisible (*enode->e.lse.face(),T,(float*)0);
	      rightFaceVis = faceVisible (*enode->e.rse.face(), T, (float*)0);

	      if ((enode->e.isSoft()==1 && (leftFaceVis != rightFaceVis)) ||
		  (enode->e.isSoft()==0 && (leftFaceVis || rightFaceVis)))
	       { 
		 vh[0] = qh.x();
		 vh[1] = qh.y();
		 vh[2] = qh.z();
		 _grf4MultTrsfV3f (vh, T, vh);

		 vt[0] = qt.x();
		 vt[1] = qt.y();
		 vt[2] = qt.z();
		 _grf4MultTrsfV3f (vt, T, vt);

		 grf4DrawVector (vh[0], vh[1], vh[2], vt[0], vt[1], vt[2]);
	       }
	    }
	 }
}
