#include <grf4.h>
#include "grf4GlCommon.h"
#include <math.h>

float *_grf4CrossV3f (float *vr, float *v1, float *v2)
{
	vr[0] = v1[1]*v2[2] - v1[2]*v2[1];
	vr[1] = v1[2]*v2[0] - v1[0]*v2[2];
	vr[2] = v1[0]*v2[1] - v1[1]*v2[0];
	return(vr);
}

float _grf4DotV3f (float *v1, float *v2)
{
	return (v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2]);
}

float *_grf4NormalizeV3f (float *vr, float *v1)
{
	double mag;

	mag = sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]);
	vr[0] = v1[0]/mag;
	vr[1] = v1[1]/mag;
	vr[2] = v1[2]/mag;
	return(vr);
}

void _grf4PrintV3f (char *fmt, float *v1)
{
	char buf[128];

	sprintf (buf, "%s %s %s\n", fmt, fmt, fmt);
	printf (buf, v1[0], v1[1], v1[2]);
}

float *_grf4SubtractV3f (float *vr, float *v1, float *v2)
{
	vr[0] = v1[0] - v2[0];
	vr[1] = v1[1] - v2[1];
	vr[2] = v1[2] - v2[2];
	return(vr);
}

float _grf4MagV3f (float *v1)
{
	return (sqrt(v1[0]*v1[0] + v1[1]*v1[1] + v1[2]*v1[2]));
}

float *_grf4MultTrsfV3f(float *vr, TRSF *t1, float *v1)
{
	double x, y, z;

	x = v1[0];
	y = v1[1];
	z = v1[2];
	
	vr[0] = t1->n.x*x + t1->o.x*y + t1->a.x*z + t1->p.x;
	vr[1] = t1->n.y*x + t1->o.y*y + t1->a.y*z + t1->p.y;
	vr[2] = t1->n.z*x + t1->o.z*y + t1->a.z*z + t1->p.z;
	return vr;
}
	
float *_grf4MultRotV3f(float *vr, TRSF *t1, float *v1)
{
	double x, y, z;

	x = v1[0];
	y = v1[1];
	z = v1[2];
	
	vr[0] = t1->n.x*x + t1->o.x*y + t1->a.x*z;
	vr[1] = t1->n.y*x + t1->o.y*y + t1->a.y*z;
	vr[2] = t1->n.z*x + t1->o.z*y + t1->a.z*z;
	return vr;
}

static SPHERE_TEMPLATE *sphereList = 0;

SPHERE_TEMPLATE *_grf4MakeSphereTemplate(int n)
{
	SPHERE_TEMPLATE *s, *l;
	int tindex;
	int pindex;
	int i;
	double theta;
	double phi;
	double ct;
	double st;
	double cp;
	double sp;

	s = (SPHERE_TEMPLATE*)malloc(sizeof(SPHERE_TEMPLATE));
	s->nTheta = n;
	s->nPhi = n/2;
	s->pnts = (float(*)[3])malloc(sizeof(float)*3*(n+1)*(n/2+1));
	s->pntNormals = (float(*)[3])malloc(sizeof(float)*3*(n+1)*(n/2+1));
	s->ctrNormals = (float(*)[3])malloc(sizeof(float)*3*(n+1)*(n/2+1));

	for (tindex=0;tindex<=s->nTheta;tindex++)
	 { theta = (tindex * 2 * M_PI) / s->nTheta;
	   ct = cos(theta);
	   st = sin(theta);
	   for (pindex=0;pindex<=s->nPhi;pindex++)
	    { phi = (pindex * M_PI) / s->nPhi;
	      cp = cos(phi);
	      sp = sin(phi);
	      i = SPHERE_IDX(s,tindex,pindex);
	      s->pnts[i][0] = sp * ct;
	      s->pnts[i][1] = sp * st;
	      s->pnts[i][2] = cp;
	      s->pntNormals[i][0] = s->pnts[i][0];
	      s->pntNormals[i][1] = s->pnts[i][1];
	      s->pntNormals[i][2] = s->pnts[i][2];
	    }
	 }
	for (theta=0;theta<s->nTheta;theta++)
	 { for (phi=0;phi<s->nPhi;phi++)
	    { 
	      float v1[3], v2[3], nrml[3];
	      int i1, i2, i3, i4;

	      i1 =      SPHERE_IDX(s,theta,phi);
	      i2 = i1 + SPHERE_IDX(s,0,1);
	      i3 = i2 + SPHERE_IDX(s,1,0);
	      i4 = i3 - SPHERE_IDX(s,0,1);

	      if (phi==0)
	       { _grf4SubtractV3f(v1, s->pnts[i2], s->pnts[i1]);
		 _grf4SubtractV3f(v2, s->pnts[i3], s->pnts[i4]);
	       }
	      else if (phi==(s->nPhi-1))
	       { _grf4SubtractV3f(v1, s->pnts[i4], s->pnts[i3]);
		 _grf4SubtractV3f(v2, s->pnts[i1], s->pnts[i2]);
	       }
	      else
	       { _grf4SubtractV3f(v1, s->pnts[i3], s->pnts[i2]);
		 _grf4SubtractV3f(v2, s->pnts[i1], s->pnts[i2]);
	       }
	      _grf4CrossV3f (nrml, v1, v2);
	      _grf4NormalizeV3f (s->ctrNormals[i1], nrml);
	    }
	 }
	return(s);
}

SPHERE_TEMPLATE *_grf4GetSphereTemplate(int n)
{
	SPHERE_TEMPLATE *s;

	for (s=sphereList; s; s = s->next)
	 { if (s->nTheta == n && s->nPhi == n/2)
	    { return(s);
	    }
	 }
	s = _grf4MakeSphereTemplate(n);
	s->next = sphereList;
	sphereList = s;	
	return(s);
}
