#include <cdefs.h>
#include "grf4Tval.h"

Tval grf4SetTval(sec, usec)
long sec;
long usec;
{
	Tval tv;

	if (ABS(usec) > 1000000)
	 { sec += (usec/1000000);
	   usec = usec % 1000000;
	 }
	if (sec > 0 && usec < 0)
	 { tv.tv_sec = sec - 1;
	   tv.tv_usec = 1000000 + usec;
	 }
	else if (sec < 0 && usec > 0)
	 { tv.tv_sec = sec + 1;
	   tv.tv_usec = -1000000 + usec;
	 }
	else
	 { tv.tv_sec = sec;
	   tv.tv_usec = usec;
	 }
	return (tv);
}

Tval grf4ZeroTval()
{
	return (grf4SetTval (0, 0));
}

Tval grf4AddTval(tv1, tv2)
Tval tv1, tv2;
{
	return (grf4SetTval (tv1.tv_sec + tv2.tv_sec, tv1.tv_usec + tv2.tv_usec));
}

Tval grf4NegTval(tv1)
Tval tv1;
{
	return (grf4SetTval (-tv1.tv_sec, -tv1.tv_usec));
}

Tval grf4SubTval(tv1, tv2)
Tval tv1, tv2;
{
	return (grf4SetTval (tv1.tv_sec - tv2.tv_sec, tv1.tv_usec - tv2.tv_usec));
}

int grf4CmpTval(tv1, tv2)
Tval tv1, tv2;
{
	Tval tv;

	tv = grf4SubTval (tv1, tv2);
	if (tv.tv_sec > 0 || tv.tv_usec > 0)
	 { return (1);
	 }
	else if (tv.tv_sec < 0 || tv.tv_usec < 0)
	 { return (-1);
	 }
	else
	 { return (0);
	 }
}

double grf4TvalToDouble (tv1)
Tval tv1;
{
	return (tv1.tv_sec + tv1.tv_usec * 1e-6);
}

Tval grf4DoubleToTval (t)
double t;
{
	Tval tv;

	tv.tv_sec = (long)t;
	tv.tv_usec = (t - tv.tv_sec)*1e6;
	return (tv);
}

Tval grf4ScaleTval (s, tv1)
double s;
Tval tv1;
{
	return (grf4DoubleToTval (s * grf4TvalToDouble (tv1)));	/* XXX ??? */
}


#if TEST

bail (loc, s)
int loc;
char *s;
{
	printf ("%d::%s\n", loc, s);
	exit (1);
}

check (loc, tv1, tv2)
int loc;
Tval tv1, tv2;
{
	if (grf4CmpTval (tv1, tv2) != 0)
	 { bail (loc, "");
	 }
}

Tval simpSet (s, u)
long s, u;
{
	Tval tv;

	tv.tv_sec = s;
	tv.tv_usec = u;
	return (tv);
}

main()
{
	check (100, grf4SetTval(100, 100), simpSet (100, 100));
	check (101, grf4SetTval(100, 999999), simpSet (100, 999999));
	check (102, grf4SetTval(100, 1000000), simpSet (101, 0));
	check (103, grf4SetTval(100, 1000010), simpSet (101, 10));
	check (104, grf4SetTval(100, 20000010), simpSet (120, 10));

	check (105, grf4SetTval(-100, -999999), simpSet (-100, -999999));
	check (106, grf4SetTval(-100, -1000000), simpSet (-101, 0));
	check (107, grf4SetTval(-100, -1000013), simpSet (-101, -13));
	check (108, grf4SetTval(-100, -20000013), simpSet (-120, -13));

	check (109, grf4SetTval(100, -999999), simpSet (99, 1));
	check (110, grf4SetTval(100, -1000000), simpSet (99, 0));
	check (111, grf4SetTval(100, -1000010), simpSet (98, 999990));
	check (112, grf4SetTval(100, -20000010), simpSet (79, 999990));

	check (113, grf4SetTval(-100, 1), simpSet (-99, -999999));
	check (114, grf4SetTval(-100, 1000000), simpSet (-99, 0));
	check (115, grf4SetTval(-100, 1000010), simpSet (-98, -999990));\

	printf ("\nRoutines OK\n\n");
}

#endif
