#include <RxMatd3x3.h>
#include <RxMatdBaseTest.h>

#include <assert.h>
#include <cdefs.h>
#include <math.h>

#define EPS	(1e-10)
#define EQ(a,b)	(ABS(a-b) < EPS)

#define LEN	12

double testVals[][LEN] = {
	{ 1.0, 0.0, 0.0,   0.0, 1.0, 0.0,   0.0, 0.0, 1.0,  0.0, 0.0, 0.0},
	{ 1.0, 2.0, 3.0,   4.0, 5.0, 6.0,   9.0, 8.0, 1.0,  3.0, 2.0, 1.0},
	{ 0.0, 0.0, 1.0,   1.0, 2.0, 3.0,   5.0, 6.6, 7.0,  1.0, 7.0,-9.0 },
	{ 1.0, 1.0, 1.0,  -1.0,-2.0,-3.0,  -5.0, 6.6,-7.0, -8.0,-4.0, 8.0 },
	{ 3.0,-2.0, 1.0,   0.0, 1.0, 0.0,  10.0,-9.0,-5.0,  2.0, 2.0, 4.0 }
};

double rotVals[][3] = {
 { 90.0, -90.0, 45.0 },
 { 50.0,  60.0, 70.0 },
 { 90.0,  90.0,  0.0 },
 {  0.0,  90.0, 90.0 },
 { 90.0,   0.0, 90.0 },
 { 90.0,   0.0,  0.0 },
 {  0.0,  90.0,  0.0 },
 {  0.0,   0.0, 90.0 }};

main()
{
	double vec[100];
	RxMatd3x3 m1;
	RxMatd3x3 m2(testVals[0]);
	RxMatd3x3 m3(testVals[1]);
	int i, j;
	int p, q;

	assert (m1.nrows() == 3);
	assert (m1.ncols() == 3);
	assert (m2.nrows() == 3);
	assert (m2.ncols() == 3);

	for (i=0; i<m3.nrows(); i++)
	 { for (j=0; j<m3.ncols(); j++)
	    { assert (EQ(testVals[1][i*m3.ncols()+j],m3(i,j)));
	      m3(i,j) = (i+1)*100+(j+1);
	      assert (EQ(m3(i,j),(double)((i+1)*100+(j+1))));
	    }
	 }

	RxMatd3x3 matA(m3);

	for (i=0; i<m3.nrows(); i++)
	 { for (j=0; j<m3.ncols(); j++)
	    { assert (EQ(m3(i,j),matA(i,j)));
	    }
	 }

	m1.setZero();
	for (i=0; i<m1.nrows(); i++)
	 { for (j=0; j<m1.ncols(); j++)
	    { assert (EQ(m1(i,j),0.0));
	    }
	 }
	assert (!m1.equal(m3,EPS));
	m1 = m3;
	assert (m1.equal(m3,EPS));

	m1.set(testVals[2]);
	for (i=0; i<m1.nrows(); i++)
	 { for (j=0; j<m1.ncols(); j++)
	    { assert (EQ(m1(i,j),testVals[2][i*m1.ncols()+j]));
	    }
	 }
	m1.get (vec);
	for (i=0; i<m1.nrows()*m1.ncols(); i++)
	 { assert (EQ(vec[i],testVals[2][i]));
	 }

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { m1.set(testVals[p]);

	   assert (testScale (m3.scale(M_PI,m1), M_PI, m1, EPS));
	   m3 = m1;
	   assert (testScale (m3.scale(M_PI), M_PI, m1, EPS));
	   m3 = m1;
	   assert (testScale (m3 *= M_PI, M_PI, m1, EPS));
	   m3 = m1;
	   assert (testScale (m3 /= M_PI, 1/M_PI, m1, EPS));
	   assert (testScale (M_PI*m1, M_PI, m1, EPS));
	   assert (testScale (m1*M_PI, M_PI, m1, EPS));
	   assert (testScale (m1/M_PI, 1/M_PI, m1, EPS));

	   assert (testTsp (m3.tsp(m1), m1, EPS));
	   m3 = m1;
	   assert (testTsp (m3.tsp(), m1, EPS));
	   assert (testTsp (m1.T(), m1, EPS));

	   RxMatd3x3 mc;
	   RxMatd3x3 m1x(testVals[p]);

	   assert (testTsp (mc.tsp(m1x), m1x, EPS));
	   mc = m1x;
	   assert (testTsp (mc.tsp(), m1x, EPS));
	   assert (testTsp (m1x.T(), m1x, EPS));

	   m3 = m1;
	   assert (testScale (m3.neg(m1), -1.0, m1, EPS));
	   m3 = m1;
	   assert (testScale (m3.neg(), -1.0, m1, EPS));
	   assert (testScale (-m1, -1.0, m1, EPS));
	 }

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { for (q=0; q<ARRAY_SIZE(testVals); q++)
	    { m1.set(testVals[p]);
	      m2.set(testVals[q]);

	      assert (testAdd (m1+m2, m1, m2, EPS));
	      assert (testAdd (m3.add(m1,m2), m1, m2, EPS));
	      m3 = m1;
	      assert (testAdd (m3.add(m2), m1, m2, EPS));
	      m3 = m1;					    
	      assert (testAdd (m3 += m2, m1, m2, EPS));

	      assert (testSub (m1-m2, m1, m2, EPS));
	      assert (testSub (m3.sub(m1,m2), m1, m2, EPS));
	      m3 = m1;
	      assert (testSub (m3.sub(m2), m1, m2, EPS));
	      m3 = m1;					    
	      assert (testSub (m3 -= m2, m1, m2, EPS));

	      RxMatd3x3 mc;
	      RxMatd3x3 m2x(testVals[q]);

	      assert (testMul (m1.T()*m2, m1.T(), m2, EPS));
	      assert (testMul (mc.mul(m1.T(),m2), m1.T(), m2, EPS));
	      m3 = m1;
	      assert (testMul (m3.mul(m2x), m1, m2x, EPS));
	      m3 = m1;
	      assert (testMul (m3 *= m2x, m1, m2x, EPS));
	    }
	 }

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { RxMatd3x3 ma(testVals[p]);
	   RxMatd3x3 mc;

	   mc.inv(ma);
	   assert (testInv (mc, ma, EPS));
	   mc = ma;
	   assert (testInv (mc.inv(), ma, EPS));
	   assert (testInv (ma.I(), ma, EPS));
	 }

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { m1.set (testVals[p]);

	   for (j=0; j<ARRAY_SIZE(rotVals); j++)
	    { RxRot R;

	      R.setRpyD (rotVals[j][0], rotVals[j][1], rotVals[j][2]);

	      m3.xform (m1, R);
	      m3.Ixform (m3, R);
	      assert (m3.equal(m1,EPS));
	      m3 = m1;
	      m3.xform (R);
	      m3.Ixform (R);
	      assert (m3.equal(m1,EPS));
	    }
	 }

	RxMatd3x3 mc;

	mc.setIdent();
	for (i=0; i<mc.nrows(); i++)
	 { for (j=0; j<mc.ncols(); j++)
	    { assert (EQ(mc(i,j), (i==j) ? 1.0 : 0.0));
	    }
	 }

#if 0
#define LENV	3

	// test new vector based routines

	RxVecd3 vtest();
	for (i=0; i<LENV; i++)
	 { vtest(i) = i;
	 }

	 { RxMatd3x3 MV;
	   assert (MV.nrows() == LENV && MV.ncols() == 1);
	   for (i=0; i<LENV; i++)
	    { assert (EQ(MV(i,0),vtest(i)));
	    }
	 }
	 { RxMatd3x3 MV;
	   assert (MV.nrows() == 1 && MV.ncols() == LENV);
	   for (i=0; i<LENV; i++)
	    { assert (EQ(MV(0,i),vtest(i)));
	    }
	 }
	 { RxMatd3x3 MV;

	   MV.setDiag (vtest);
	   for (i=0; i<LENV; i++)
	    { for (j=0; j<LENV+2; j++)
	       { assert (EQ(MV(i,j), i==j ? vtest(i) : 0));
	       }
	    }
	   MV.setDiag (vtest.buf());
	   for (i=0; i<LENV; i++)
	    { for (j=0; j<LENV+2; j++)
	       { assert (EQ(MV(i,j), i==j ? vtest(i) : 0));
	       }
	    }
	 }
	 { RxMatd3x3 MV;

	   MV.setDiag (vtest);
	   for (i=0; i<LENV+2; i++)
	    { for (j=0; j<LENV; j++)
	       { assert (EQ(MV(i,j), i==j ? vtest(i) : 0));
	       }
	    }
	   MV.setDiag (vtest.buf());
	   for (i=0; i<LENV+2; i++)
	    { for (j=0; j<LENV; j++)
	       { assert (EQ(MV(i,j), i==j ? vtest(i) : 0));
	       }
	    }
	 }

	// test vector multiply

	for (p=0; p<ARRAY_SIZE(testVals); p++)
	 { for (q=0; q<ARRAY_SIZE(testVals); q++)
	    { m1.set(testVals[p]);
	      m2.set(testVals[q]);

	      for (i=0; i<m2.nrows(); i++)
	       { RxMatd3x3 rowM (m2, i, 0, 1, m2.ncols());	// ith row of m2
		 RxVecd3 rowV ();
		 RxVecd3 resV ();
		 RxMatd3x3 resM;

		 rowM.get (&rowV);
		 for (j=0; j<m2.ncols(); j++)
		  { assert (EQ(rowV(j),m2(i,j)));
		  }
		 m1.mulV (&resV, rowV);
		 assert (testMul (resM.set(resV), m1, rowM.T(), EPS));
		 resV = m1*rowV;
		 assert (testMul (resM.set(resV), m1, rowM.T(), EPS));
	       }
	      for (j=0; j<m2.ncols(); j++)
	       { RxMatd3x3 colM (m2, 0, j, m2.nrows(), 1);	// ith col of m2
		 RxVecd3 colV (m2.nrows());
		 RxVecd3 resV (m1.ncols());
		 RxMatd3x3 resM;

		 colM.get (&colV);
		 for (i=0; i<m2.nrows(); i++)
		  { assert (EQ(colV(i),m2(i,j)));
		  }
		 m1.Vmul (&resV, colV);
		 assert (testMul (resM.set(resV), colM.T(), m1, EPS));
		 resV = colV*m1;
		 assert (testMul (resM.set(resV), colM.T(), m1, EPS));
	       }
	    }
	 }
#endif

	printf ("\nROUTINES OK\n\n");
	return 0;
}
