#include <RxMatdLu.h>
#include <RxMathError.h>
#include <stdlib.h>
#include <cdefs.h>
#include <math.h>
#include <mathx.h>

void RxMatdLu::doLu (const RxMatd &A)
{
	double tmp, max, d;
	int i, j, k, max_i;
	int w = LU->width;
	double *buf = LU->store;

	*LU = A;

#define M(ii,jj)	buf[(ii)*w+(jj)]
#define SWAP(a,b)	tmp = a; a = b; b = tmp;

	for (j=0; j<n; j++)
	 { 
	   for (k=0; k<j; k++)
	    { SWAP (M(k,j), M(perm[k],j));
	    }
	   for (k=0; k<j; k++)
	    { tmp = M(k,j);
	      for (i=k+1; i<j; i++)
	       { M(i,j) -= M(i,k)*tmp;
	       }
	    }
	   for (k=0; k<j; k++)
	    { tmp = M(k,j);
	      for (i=j; i<n; i++)
	       { M(i,j) -= M(i,k)*tmp;
	       }
	    }
	   // find pivot index 
	   max_i = j;
	   max = ABS(M(j,j));
	   for (i=j+1; i<n; i++)
	    { d = M(i,j);
	      if (ABS(d) > max)
	       { max_i = i; max = ABS(d);
	       }
	    }
	   perm[j] = max_i;
	   if (j != max_i)
	    { for (k=0; k<=j; k++)
	       { SWAP(M(j,k), M(max_i,k));
	       }
	    }
	   d = M(j,j);
	   if (ABS(d) <= tol)
	    { flags |= RX_LU_SINGULAR;
	    }
	   else
	    { for (i=j+1; i<n; i++)
	       { M(i,j) /= d;
	       }
	    }
	 }
#undef SWAP
#undef M
}

int RxMatdLu::calc (const RxMatd &A)
{
	int newn;

	newn = A.nr;
	if (A.nc != newn)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdLu::calc: %K in A\n").raise();
	   flags &= ~RX_LU_CALCDONE;
	   return -1;
	 }
	if (perm && n != newn)
	 { free (perm);
	 }
	n = newn;

	if (LU)
	 { if (LU->nr != n || LU->nc != n)
	    { if (flags & RX_LU_LU_MALLOCED)
	       { delete LU;
		 LU = new RxMatd (n, n);
	       }
	      else
	       { RxMathError (RxMathError::BadSize,
			      "RxMatdLu::calc: %K in LU\n").raise();
		 flags &= ~RX_LU_CALCDONE;
		 return -1;
	       }
	    }
	 }
	else
	 { LU = new RxMatd (n, n);
	   flags |= RX_LU_LU_MALLOCED;
	 }
	if (perm == NULL)
	 { perm = (int*)malloc(sizeof(int)*n);
	   flags |= RX_LU_PERM_MALLOCED;
	 }

	doLu (A);
	flags |= RX_LU_CALCDONE;
	return (1);
}

int RxMatdLu::setLUP (RxMatd *Lu, int *pv)
{
	if (Lu)
	 { if (flags & RX_LU_LU_MALLOCED)
	    { delete Lu;
	      flags &= ~RX_LU_LU_MALLOCED;
	    }
	   LU = Lu;
	 }
	if (pv)
	 { if (flags & RX_LU_PERM_MALLOCED)
	    { free (perm);
	      flags &= ~RX_LU_PERM_MALLOCED;
	    }
	   perm = pv;
	 }
	return (1);
}

void RxMatdLu::copyLUP (RxMatd *L, RxMatd *U, RxMatd *P) const
{
	int i, j, k;
	int w = LU->width;
	int *rowidxs;

	if (!(flags & RX_LU_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdLu::copyLU: %K\n").raise();
	   return;
	 }
	if ((L && (L->nr != n || L->nc != n)) ||
	    (U && (U->nr != n || U->nc != n)))
	 { RxMathError (RxMathError::BadSize, 
			"RxMatdLu::copyLU: %K\n").raise();
	   return;
	 }
	if (P)
	 { rowidxs = (int*)malloc(sizeof(int)*n);
	   for (i=0; i<n; i++)
	    { rowidxs[i] = i;
	    }
	   for (i=0; i<n; i++)
	    { if ((k = perm[i]) != i)
	       { j = rowidxs[k];
		 rowidxs[k] = rowidxs[i];
		 rowidxs[i] = j;
	       }
	    }
	 }
	for (i=0; i<n; i++)
	 { for (j=0; j<n; j++)
	    { if (L)
	       { L->store[L->width*i+j] = 
		    (i>j) ? LU->store[i*w+j] : (i==j) ? 1.0 : 0.0;
	       }
	      if (U)
	       { U->store[U->width*i+j] = 
		    (i<=j) ? LU->store[i*w+j] : 0.0;
	       }
	      if (P)
	       { P->store[P->width*i+j] = 
		    (j==rowidxs[i]) ? 1.0 : 0.0;
	       }
	    }
	 }
	if (P)
	 { free (rowidxs);
	 }
}

RxMatdLu::~RxMatdLu ()
{
	if (flags & RX_LU_LU_MALLOCED)
	 { delete LU;
	 }
	if (flags & RX_LU_PERM_MALLOCED)
	 { free (perm);
	 }
}

RxVecd &RxMatdLu::solve (RxVecd *x, const RxVecd &b) const
{
	double sum;
	int w = LU->width;
	int i, j;
	int i0 = -1;

	if (x->len != n || b.len != n)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdLu::solve: %K\n").raise();
	   return *x;
	 }
	if (!(flags & RX_LU_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdLu::solve: %K\n").raise();
	   return *x;
	 }

	for (i=0; i<n; i++)
	 { x->v[i] = b.v[i];
	 }
	for (i=0; i<n; i++)
	 { sum = x->v[perm[i]];
	   x->v[perm[i]] = x->v[i];
	   if (i0 >= 0)
	    { for (j=i0; j<i; j++)
	       { sum -= x->v[j]*LU->store[i*w+j];
	       }
	    }
	   else if (sum)
	    { i0 = i;
	    }
	   x->v[i] = sum;
	 }
	for (i=n-1; i>=0; i--)
	 { sum = x->v[i];
	   for (j=i+1; j<n; j++)
	    { sum -= x->v[j]*LU->store[i*w+j];
	    }
	   x->v[i] = sum/LU->store[i*w+i];
	 }
	return *x;
}

RxVecd &RxMatdLu::Tsolve (RxVecd *x, const RxVecd &b) const
{
	double sum;
	int w = LU->width;
	int i, j;

	if (x->len != n || b.len != n)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdLu::solve: %K\n").raise();
	   return *x;
	 }
	if (!(flags & RX_LU_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdLu::solve: %K\n").raise();
	   return *x;
	 }

	for (i=0; i<n; i++)
	 { x->v[i] = b.v[i];
	 }

	// solve U^T y = b

	for (i=0; i<n; i++)
	 { sum = x->v[i];
	   for (j=0; j<i; j++)
	    { sum -= x->v[j]*LU->store[j*w+i];
	    }
	   x->v[i] = sum/LU->store[i*w+i];
	 }
	// solve L^T x = y

	for (i=n-1; i>=0; i--)
	 { sum = x->v[i];
	   for (j=i+1; j<n; j++)
	    { sum -= x->v[j]*LU->store[j*w+i];
	    }
	   x->v[i] = sum;
	 }

	// and now undo the permutations

	for (i=n-1; i>=0; i--)
	 { if ((j = perm[i]) != i)
	    { sum = x->v[j];
	      x->v[j] = x->v[i];
	      x->v[i] = sum;
	    }
	 }
	return *x;
}

#define MAXELEM	16

RxMatd &RxMatdLu::solve (RxMatd *X, const RxMatd &B) const
{
	double buf[MAXELEM];
	double *tmp;
	double sum;
	int w = LU->width;
	int i, j, k;
	int i0;

	if (X->nr != n || B.nr != n || X->nc != B.nc)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdLu::solve: %K\n").raise();
	   return *X;
	 }
	if (!(flags & RX_LU_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdLu::solve: %K\n").raise();
	   return *X;
	 }

	tmp = (n > MAXELEM) ? (double*)malloc(sizeof(double)*n) : buf;

	for (k=0; k<X->nc; k++)
	 { i0 = -1;
	   for (i=0; i<n; i++)
	    { tmp[i] = B.store[i*B.width+k];
	    }
	   for (i=0; i<n; i++)
	    { sum = tmp[perm[i]];
	      tmp[perm[i]] = tmp[i];
	      if (i0 >= 0)
	       { for (j=i0; j<i; j++)
		  { sum -= tmp[j]*LU->store[i*w+j];
		  }
	       }
	      else if (sum)
	       { i0 = i;
	       }
	      tmp[i] = sum;
	    }
	   for (i=n-1; i>=0; i--)
	    { sum = tmp[i];
	      for (j=i+1; j<n; j++)
	       { sum -= tmp[j]*LU->store[i*w+j];
	       }
	      tmp[i] = sum/LU->store[i*w+i];
	      X->store[i*X->width+k] = tmp[i];
	    }
	 }
	if (n > MAXELEM) free (tmp);
	return (*X);
}

double RxMatdLu::condest(const RxMatd &A) const
{
	double pvbuf[MAXELEM];
	double pnbuf[MAXELEM];
	double ppbuf[MAXELEM];
	double yvbuf[MAXELEM];
	double *pvec;
	double *pneg;
	double *ppos;
	double *yvec;
	double ypos, yneg, yval;
	double pposNorm1;
	double pnegNorm1;
	double sum;
	int i, j;
	int LUwidth;
	double *LUstore;
	double zNormInf;
	double rNormInf;

	if (A.nr != n || A.nc != n)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdLu::condest: %K\n").raise();
	   return -1;
	 }
	if (!(flags & RX_LU_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdLu::condest: %K\n").raise();
	   return -1;
	 }
	if (flags & RX_LU_SINGULAR)
	 { return (HUGE_VAL);
	 }

	LUwidth = LU->width;
	LUstore = LU->store;

#define LUelem(ii,jj)	LUstore[(ii)*LUwidth+(jj)]

	if (n > MAXELEM)
	 { pvec = (double*)malloc(n*3*sizeof(double));
	   pneg = pvec + n;
	   ppos = pneg + n;
	   yvec = ppos + n;
	 }
	else
	 { pvec = pvbuf;
	   pneg = pnbuf;
	   ppos = ppbuf;
	   yvec = yvbuf;
	 }

	for (j=0; j<n; j++)
	 { pvec[j] = 0;
	 }
	for (j=0; j<n; j++)
	 { pposNorm1 = 0;
	   pnegNorm1 = 0;
	   ypos = ( 1 - pvec[j])/LUelem(j,j);
	   for (i=j+1; i<n; i++)
	    { ppos[i] = pvec[i] + ypos*LUelem(j,i);
	      pposNorm1 += ABS(ppos[i]);
	    }
	   yneg = (-1 - pvec[j])/LUelem(j,j);
	   for (i=j+1; i<n; i++)
	    { pneg[i] = pvec[i] + yneg*LUelem(j,i);
	      pnegNorm1 += ABS(pneg[i]);
	    }
	   if (ABS(ypos) + pposNorm1 >= ABS(yneg) + pnegNorm1)
	    { yvec[j] = ypos;
	      for (i=j+1; i<n; i++)
	       { pvec[i] = ppos[i];
	       }
	    }
	   else
	    { yvec[j] = yneg;
	      for (i=j+1; i<n; i++)
	       { pvec[i] = pneg[i];
	       }
	    }
	 }

	rNormInf = 0;
	zNormInf = 0;

	// Solve L^T r = y

	for (i=n-1; i>=0; i--)
	 { sum = 0;
	   for (j=i+1; j<n; j++)
	    { sum += yvec[j]*LUelem(j,i);
	    }
	   yval = yvec[i] - sum;
	   if (ABS(yval) > rNormInf)
	    { rNormInf = ABS(yval);
	    }
	   yvec[i] = yval;
	 }

	// Solve L w = P r

	for (i=0; i<n; i++)
	 { sum = yvec[perm[i]];
	   yvec[perm[i]] = yvec[i];
	   for (j=0; j<i; j++)
	    { sum -= yvec[j]*LUelem(i,j);
	    }
	   yvec[i] = sum;
	 }

	// Solve U z = w

	for (i=n-1; i>=0; i--)
	 { sum = 0;
	   for (j=i+1; j<n; j++)
	    { sum += yvec[j]*LUelem(i,j);
	    }
	   yval = (yvec[i]-sum)/LUelem(i,i);
	   if (ABS(yval) > zNormInf)
	    { zNormInf = ABS(yval);
	    }
	   yvec[i] = yval;
	 }

	// Compute the infinity norm of A

	if (n > MAXELEM)
	 { free (pvec);
	 }

	return (A.normInf() * zNormInf / rNormInf);
}

double RxMatdLu::det() const
{
	double prod;
	int w = LU->width;
	int i;

	if (!(flags & RX_LU_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdLu::det: %K\n").raise();
	   return -1;
	 }
	if (flags & RX_LU_SINGULAR)
	 { return (0.0);
	 }
	prod = 1.0;
	for (i=0; i<n; i++)
	 { prod *= LU->store[i*w+i];
	   if (perm[i] != i)
	    { prod = -prod;
	    }
	 }
	return (prod);
}

RxMatd &RxMatdLu::inv (RxMatd *M) const
{
	if (M->nc != n || M->nr != n)
	 { RxMathError (RxMathError::BadSize,
			"RxMatdLu::inv: %K\n").raise();
	   return *M;
	 }
	if (!(flags & RX_LU_CALCDONE))
	 { RxMathError (RxMathError::InitNeeded,
			"RxMatdLu::inv: %K\n").raise();
	   return *M;
	 }
	M->setIdent();
	solve (M, *M);	
	return *M;
}
