#include <RxMatdSvd.h>
#include <RxTimeval.h>
#include <cdefs.h>
#include <stdlib.h>
#include <stdio.h>
#include <cdefs.h>

#define MAXD 10
#define EPS     (1e-10)

#define ERROR_EXIT() \
	printf ("n=%d m=%d ii=%d jj=%d\n", n, m, ii, jj); \
	A.print ("A"); \
	U.print ("U"); \
	V.print ("V"); \
	sig.print ("sig %8.4f\n"); \
	exit (1);
	
char **argvec;
int doTiming = 0;

main(int argc, char **argv)
{
	int i, j;
	int n, m;
	int ii, jj;
	int mind;
	double sigBuf[100];
	double det;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (ARGUMENT("-timing"))
	    { doTiming = 1;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, 
"Usage: %s [timing]\n", argvec[0]);
	   exit (1);
	 }

	double buf[MAXD*MAXD] = {  
			    1, 2, 6, 4, 5, 6, 7, 0, 0, -1,
			    6, 5, 4, 3, 2, 1, 0, 0, 0, 1,
			    1, 2, 1, 2, 1, 2, 1, 0, 1, -2,
			    1, 1, 1, 9, 1, 1, 1, 0, 9, 0,
			    1, 3, 2, 5, 4, 9, 6, 0, 8, 9,
			    8, 8, 7, 7, 2, 5, 4, 1, 5, -1,
			    4, 3, 5, 6, 7, 5, 4, 1, 6, 10,
			    0, 0, 0, 0, 1, 1, 1, 1, 7, 0, 
			    1, 3, 3, 7, 8, 1, 2, 8, 13, 0,
			   -3,-7, 2, 5, 0, 0, 0,-9, 1, 1 };

	if (doTiming)
	 { 
	   RxMatd M3x3 (3, 3, buf, MAXD);
	   RxMatd M5x5 (5, 5, buf, MAXD);
	   RxMatd M7x7 (7, 7, buf, MAXD);
	   RxMatdSvd svd;

	   TIMEIT (svd.calc (M3x3), 1000);
	   TIMEIT (svd.calc (M5x5), 1000);
	   TIMEIT (svd.calc (M7x7), 1000);
	 }
	else
	 { 
	   for (m=1; m<=MAXD; m++)

	    { for (n=1; n<=MAXD; n++)
	       { for (ii=0; ii<=MAXD-m; ii++)
		  { for (jj=0; jj<=MAXD-n; jj++)
		     { 
		       mind = MIN(m,n);

		       RxMatd A(m, n, &buf[ii*MAXD+jj], MAXD);
		       RxMatd U(m,mind);
		       RxMatd V(n,mind);
		       RxVecd sig(mind);
		       RxMatd S(mind,mind);

		       RxMatd I(n,n);

		       RxMatdSvd svd(A, &sig, &U, &V);
		       S.setDiag (sig.buf());

		       if (!A.equal (U*S*V.T(), EPS))
			{ printf ("U*S*V.T() != A\n");
			  ERROR_EXIT();
			}
		       if (n == m && n <= 3)
			{ if (n == 1)
			   { det = A(0,0);
			   }
			  else if (n == 2)
			   { det = A(0,0)*A(1,1) - A(0,1)*A(1,0);
			   }
			  else if (n == 3)
			   { det = A(0,0)*A(1,1)*A(2,2) + 
				   A(0,1)*A(1,2)*A(2,0) + 
				   A(0,2)*A(1,0)*A(2,1) -
				   A(0,2)*A(1,1)*A(2,0) -
				   A(0,0)*A(1,2)*A(2,1) -
				   A(0,1)*A(1,0)*A(2,2);
			   }
			  if (ABS(det-svd.det()) > EPS)
			   { printf ("det=%g vs. %g\n", svd.det(), det);
			     ERROR_EXIT();
			   }
			}
		       RxMatd Itest(n,n);
		       RxVecd bc(m);
		       RxVecd xc(n);
		       RxVecd br(n);
		       RxVecd xr(m);

		       if (svd.cond() < 1/EPS && m == n)
			{ RxMatd Itest(n,n);

			  Itest.setIdent();
			  svd.inv(&I);
			  if (!(I*A).equal (Itest, EPS))
			   { printf ("A*A.inv() != I\n");
			     (A*A.inv()).print ("A*A.inv()");
			     ERROR_EXIT();
			   }

			  bc.set (buf);
			  svd.solve (&xc, bc);
			  if (!(A*xc).equal (bc, EPS))
			   { printf ("A*xc != bc\n");			  
			     bc.print ("bc%8.4f\n");
			     xc.print ("xc%8.4f\n");
			     (A*xc).print ("A*xc%8.4f\n");
			     ERROR_EXIT();
			   }

			  br.set (buf);
			  svd.Tsolve (&xr, br);
			  if (!(xr*A).equal (br, EPS))
			   { fprintf (stderr, "xr*A != br\n");
			     br.print ("br%8.4f\n");
			     xr.print ("xr%8.4f\n");
			     (xr*A).print ("xr*A%8.4f\n");
			     ERROR_EXIT();
			   }
			}
		     }
		  }
	       }
	    }
	   printf ("\nROUTINES OK\n\n");
	   exit (0);
	 }
}
