#ifndef _RXVECDN2_H_WRAP_
#define _RXVECDN2_H_WRAP_

#include <RxVecd.h>

class RxVecd2: public RxVecd
{
	double vbuf[2];

    public:

	// access

	double &x()
	 { return vbuf[0];
	 }
	const double &x() const
	 { return vbuf[0];
	 }
	double &y()
	 { return vbuf[1];
	 }
	const double &y() const
	 { return vbuf[1];
	 }

	// query 
	
	int length() const
	 { return 2;
	 }
	double& operator()(int i)
	 { return vbuf[i];
	 } 
	double operator()(int i) const
	 { return vbuf[i];
	 } 


	// assignment

	RxVecd2& set (const RxVecd2 &v2) 
	 { x()=v2.x(); y()=v2.y();
	   return (*this);
	 }
	RxVecd2& operator=(const RxVecd2 &v2) 
	 { return (set (v2));
	 }
	RxVecd2& set (double xx, double yy)
	 { x() = xx; y() = yy;
	   return (*this);
	 }
	void get (double *xp, double *yp) const
	 { *xp = x(); *yp = y();
	 }
	RxVecd2& set (const double *dp)
	 { x() = dp[0]; y() = dp[1];
	   return (*this);
	 }
	void get (double *dp) const
	 { dp[0] = x(); dp[1] = y();
	 }

	// constructors	

	RxVecd2 () : RxVecd (2, vbuf) 
	 { 
	 }
	RxVecd2 (double x, double y) : RxVecd (2, vbuf)
	
	 { set (x, y);
	 }
	RxVecd2 (const RxVecd2 &v2) : RxVecd (2, vbuf)
	 { set (v2);
	 }
	RxVecd2 (double *dp) : RxVecd (2, vbuf)
	 { set (dp);
	 }

	// T <- T,T operations

	RxVecd2& add (const RxVecd2 &v1, const RxVecd2 &v2);
	RxVecd2& add (const RxVecd2 &v2);
	RxVecd2& operator+=(const RxVecd2 &v2)
	 { return (add (v2));
	 }
	friend RxVecd2 operator+(const RxVecd2 &v1, const RxVecd2 &v2);
	RxVecd2& addQ (const RxVecd2 &v1, const RxVecd2 &v2)
	 { x() = v1.x()+v2.x(); y() = v1.y()+v2.y();
	   return (*this);
	 }

	RxVecd2& sub (const RxVecd2 &v1, const RxVecd2 &v2);
	RxVecd2& sub (const RxVecd2 &v2);
	RxVecd2& operator-=(const RxVecd2 &v2)
	 { return (sub (v2));
	 }
	friend RxVecd2 operator-(const RxVecd2 &v1, const RxVecd2 &v2);
	RxVecd2& subQ (const RxVecd2 &v1, const RxVecd2 &v2)
	 { x() = v1.x()-v2.x(); y() = v1.y()-v2.y();
	   return (*this);
	 }

	double cross (const RxVecd2 &v1, const RxVecd2 &v2) const
	 { return (v1.x()*v2.y() - v1.y()*v2.x());
	 }
	double cross (const RxVecd2 &v2) const
	 { return (cross(*this,v2));
	 }
	RxVecd2& extrap (const RxVecd2 &v1, double a, const RxVecd2 &v2);
	RxVecd2& extrap (double a, const RxVecd2 &v2);
	RxVecd2& extrapQ (const RxVecd2 &v1, double a, const RxVecd2 &v2)
	 { x() = v1.x()+a*v2.x(); y() = v1.y()+a*v2.y();
	   return (*this);
	 }

	RxVecd2& interp (const RxVecd2 &v1, double a, const RxVecd2 &v2);
	RxVecd2& interp (double a, const RxVecd2 &v2);
	RxVecd2& interpQ (const RxVecd2 &v1, double a, const RxVecd2 &v2)
	 { x() = (1-a)*v1.x()+a*v2.x(); y() = (1-a)*v1.y()+a*v2.y();
	   return (*this);
	 }

	// T <- T,O operations

	RxVecd2& scale (double a, const RxVecd2 &v1);
	RxVecd2& scale (double a);
	RxVecd2& operator*=(double a)
	 { return (scale (a));
	 }
	RxVecd2& operator/=(double a)
	 { return (scale (1/a));
	 }
	friend RxVecd2 operator*(const RxVecd2 &v1, double a);
	friend RxVecd2 operator*(double a, const RxVecd2 &v1);
	friend RxVecd2 operator/(const RxVecd2 &v1, double a);
	RxVecd2& scaleQ (double a, const RxVecd2 &v1)
	 { x() = v1.x()*a; y() = v1.y()*a;
	   return (*this);
	 }

	// O <- T,T operations

	double dot(const RxVecd2 &v2) const
	 { return (x()*v2.x() + y()*v2.y());
	 }
	double dotQ(const RxVecd2 &v2) const
	 { return (x()*v2.x() + y()*v2.y());
	 }
	double dist(const RxVecd2 &v2) const
	 { RxVecd2 diff;
	   diff.sub(*this, v2);
	   return diff.norm2();
	 }

	int operator==(const RxVecd2 &v2) const
	 { return x() == v2.x() && y() == v2.y();
	 }

	// T <- T operations

	RxVecd2 &unit(const RxVecd2 &v1);
	RxVecd2 &unit()
	 { return (unit(*this));
	 }
	RxVecd2 &neg(const RxVecd2 &v1)
	 { x() = -v1.x(); y() = -v1.y();
	   return (*this);
	 }
	RxVecd2 &neg()
	 { return (neg (*this));
	 }
	friend RxVecd2 operator-(const RxVecd2 &v1);

	// T <- O operations

	RxVecd2 &setZero();

	// O <- T operations

	double norm1() const;
	double norm2() const;
	double normInf() const;

	// constants

	static const RxVecd2 ZERO;
	static const RxVecd2 I;
	static const RxVecd2 J;
	static const RxVecd2 Ineg;
	static const RxVecd2 Jneg;
};

inline RxVecd2 operator+(const RxVecd2 &v1, const RxVecd2 &v2)
{
	RxVecd2 vr;
	return (vr.add (v1, v2));
}

inline RxVecd2 operator-(const RxVecd2 &v1, const RxVecd2 &v2)
{
	RxVecd2 vr;
	return (vr.sub (v1, v2));
}

inline RxVecd2 operator*(const RxVecd2 &v1, double a)
{
	RxVecd2 vr;
	return (vr.scale (a, v1));
}

inline RxVecd2 operator*(double a, const RxVecd2 &v1)
{
	RxVecd2 vr;
	return (vr.scale (a, v1));
}

inline RxVecd2 operator/(const RxVecd2 &v1, double a)
{
	RxVecd2 vr;
	return (vr.scale (1/a, v1));
}

inline RxVecd2 operator-(const RxVecd2 &v1)
{
	RxVecd2 vr;
	return (vr.neg (v1));
}

#endif
