#include <RxVecd3.h>
#include <RxTrsf.h>
#include <math.h>
#include <cdefs.h>

const RxVecd3 RxVecd3::ZERO(0,0,0);
const RxVecd3 RxVecd3::I(1,0,0);
const RxVecd3 RxVecd3::J(0,1,0);
const RxVecd3 RxVecd3::K(0,0,1);
const RxVecd3 RxVecd3::Ineg(-1,0,0);
const RxVecd3 RxVecd3::Jneg(0,-1,0);
const RxVecd3 RxVecd3::Kneg(0,0,-1);

#if 0
RxVecd3& RxVecd3::set (const double *dp)
{
	vbuf[0] = dp[0];
	vbuf[1] = dp[1];
	vbuf[2] = dp[2];
	return (*this);
}

void RxVecd3::get (double *dp)
{
	dp[0] = vbuf[0];
	dp[1] = vbuf[1];
	dp[2] = vbuf[2];
}
#endif

RxVecd3& RxVecd3::add(const RxVecd3 &v1, const RxVecd3 &v2)
{
	vbuf[0] = v1.vbuf[0] + v2.vbuf[0];
	vbuf[1] = v1.vbuf[1] + v2.vbuf[1];
	vbuf[2] = v1.vbuf[2] + v2.vbuf[2];
	return (*this);
}

RxVecd3& RxVecd3::add(const RxVecd3 &v2)
{
	vbuf[0] += v2.vbuf[0];
	vbuf[1] += v2.vbuf[1];
	vbuf[2] += v2.vbuf[2];
	return (*this);
}

RxVecd3& RxVecd3::sub(const RxVecd3 &v1, const RxVecd3 &v2)
{
	vbuf[0] = v1.vbuf[0] - v2.vbuf[0];
	vbuf[1] = v1.vbuf[1] - v2.vbuf[1];
	vbuf[2] = v1.vbuf[2] - v2.vbuf[2];
	return (*this);
}

RxVecd3& RxVecd3::sub(const RxVecd3 &v2)
{
	vbuf[0] -= v2.vbuf[0];
	vbuf[1] -= v2.vbuf[1];
	vbuf[2] -= v2.vbuf[2];
	return (*this);
}

RxVecd3& RxVecd3::cross(const RxVecd3& v1, const RxVecd3& v2)
{
	register double xx, yy, zz;

	xx = v1.y() * v2.z() - v1.z() * v2.y();
	yy = v1.z() * v2.x() - v1.x() * v2.z();
	zz = v1.x() * v2.y() - v1.y() * v2.x();
	
	x() = xx;
	y() = yy;
	z() = zz;
	return(*this);
}

RxVecd3& RxVecd3::extrap(const RxVecd3 &v1, double a, const RxVecd3 &v2)
{
	vbuf[0] = v1.vbuf[0] + a*v2.vbuf[0];
	vbuf[1] = v1.vbuf[1] + a*v2.vbuf[1];
	vbuf[2] = v1.vbuf[2] + a*v2.vbuf[2];
	return (*this);
}

RxVecd3& RxVecd3::extrap (double a, const RxVecd3 &v2)
{
	x() += a*v2.x();
	y() += a*v2.y();
	z() += a*v2.z();
	return (*this);
}

RxVecd3& RxVecd3::interp (const RxVecd3 &v1, double a, const RxVecd3 &v2)
{
	x() = (1-a)*v1.x() + a*v2.x();
	y() = (1-a)*v1.y() + a*v2.y();
	z() = (1-a)*v1.z() + a*v2.z();
	return (*this);
}

RxVecd3& RxVecd3::interp (double a, const RxVecd3 &v2)
{
	x() = (1-a)*x() + a*v2.x();
	y() = (1-a)*y() + a*v2.y();
	z() = (1-a)*z() + a*v2.z();
	return (*this);
}

RxVecd3& RxVecd3::scale(double a, const RxVecd3 &v1)
{
	vbuf[0] = v1.vbuf[0] * a;
	vbuf[1] = v1.vbuf[1] * a;
	vbuf[2] = v1.vbuf[2] * a;
	return (*this);
}

RxVecd3& RxVecd3::scale(double a)
{
	vbuf[0] *= a;
	vbuf[1] *= a;
	vbuf[2] *= a;
	return (*this);
}

RxVecd3& RxVecd3::xform (const RxVecd3 &v1, const RxRot& R)
{
	/* Let R be a rotation from frame A to frame B. Then v1 is 
	   converted from frame A to frame B according to 
		      T
		vr = R  v1
	 */
	double v1x, v1y, v1z;

	v1x = v1.x();
	v1y = v1.y();
	v1z = v1.z();

	x() = R.n.x()*v1x + R.n.y()*v1y + R.n.z()*v1z;
	y() = R.o.x()*v1x + R.o.y()*v1y + R.o.z()*v1z;
	z() = R.a.x()*v1x + R.a.y()*v1y + R.a.z()*v1z;	
	return (*this);
}

RxVecd3& RxVecd3::Ixform (const RxVecd3 &v1, const RxRot& R)
{
	/* Let R be a rotation from frame A to frame B. Then v1 is 
	   converted from frame B to frame A according to 

		vr = R  v1
	 */
	double v1x, v1y, v1z;

	v1x = v1.x();
	v1y = v1.y();
	v1z = v1.z();

	x() = R.n.x()*v1x + R.o.x()*v1y + R.a.x()*v1z;
	y() = R.n.y()*v1x + R.o.y()*v1y + R.a.y()*v1z;
	z() = R.n.z()*v1x + R.o.z()*v1y + R.a.z()*v1z;	
	return (*this);
}

RxVecd3& RxVecd3::xform (const RxVecd3 &v1, const RxTrsf& T)
{
	/* Let T be a spatial transform from frame A to frame B. Then v1 is 
	   converted from frame A to frame B according to 
		      -1
		vr = T  v1
	 */
	double v1x, v1y, v1z;

	v1x = v1.x() - T.p.x();
	v1y = v1.y() - T.p.y();
	v1z = v1.z() - T.p.z();

	x() = T.R.n.x()*v1x + T.R.n.y()*v1y + T.R.n.z()*v1z;
	y() = T.R.o.x()*v1x + T.R.o.y()*v1y + T.R.o.z()*v1z;
	z() = T.R.a.x()*v1x + T.R.a.y()*v1y + T.R.a.z()*v1z;	
	return (*this);
}

RxVecd3& RxVecd3::Ixform (const RxVecd3 &v1, const RxTrsf& T)
{
	/* Let T be a spatial transform from frame A to frame B. Then v1 is 
	   converted from frame B to frame A according to 

		vr = T v1
	 */
	double v1x, v1y, v1z;

	v1x = v1.x();
	v1y = v1.y();
	v1z = v1.z();

	x() = T.R.n.x()*v1x + T.R.o.x()*v1y + T.R.a.x()*v1z + T.p.x();
	y() = T.R.n.y()*v1x + T.R.o.y()*v1y + T.R.a.y()*v1z + T.p.y();
	z() = T.R.n.z()*v1x + T.R.o.z()*v1y + T.R.a.z()*v1z + T.p.z();
	return (*this);
}

double RxVecd3::dot(const RxVecd3 &v2) const
{
	return (vbuf[0]*v2.vbuf[0] + vbuf[1]*v2.vbuf[1] + vbuf[2]*v2.vbuf[2]);
}

RxVecd3& RxVecd3::unit(const RxVecd3 &v1)
{
	double mag;
 
	if ((mag = sqrt(v1.vbuf[0]*v1.vbuf[0] + v1.vbuf[1]*v1.vbuf[1] + v1.vbuf[2]*v1.vbuf[2])) > 0)
	 { vbuf[0] = v1.vbuf[0]/mag;
	   vbuf[1] = v1.vbuf[1]/mag;
	   vbuf[2] = v1.vbuf[2]/mag;
	 }
	else
	 { vbuf[0] = vbuf[1] = vbuf[2] = 0;
	 }
	return (*this);
}

RxVecd3& RxVecd3::neg(const RxVecd3 &v1)
{
	vbuf[0] = -v1.vbuf[0];
	vbuf[1] = -v1.vbuf[1];
	vbuf[2] = -v1.vbuf[2];
	return (*this);
}

RxVecd3& RxVecd3::setZero()
{
	vbuf[0] = 0;
	vbuf[1] = 0;
	vbuf[2] = 0;
	return (*this);
}

double RxVecd3::norm1() const
{
	return (ABS(vbuf[0]) + ABS(vbuf[1]) + ABS(vbuf[2]));
}

double RxVecd3::norm2() const
{
	return (sqrt(vbuf[0]*vbuf[0] + vbuf[1]*vbuf[1] + vbuf[2]*vbuf[2]));
}

double RxVecd3::normInf() const
{
	double xabs, yabs, zabs;
 
	xabs = ABS(vbuf[0]);
	yabs = ABS(vbuf[1]);
	zabs = ABS(vbuf[2]);
	if (xabs > yabs)
	 {  return ((xabs > zabs) ? xabs : zabs);
	 }
	else
	 {  return ((yabs > zabs) ? yabs : zabs);
	 }
}
