#include <RxVecd4.h>
#include <math.h>
#include <cdefs.h>

RxVecd4& RxVecd4::set (const double *dp)
{
	v[0] = dp[0];
	v[1] = dp[1];
	v[2] = dp[2];
	v[3] = dp[3];
	return (*this);
}

void RxVecd4::get (double *dp)
{
	dp[0] = v[0];
	dp[1] = v[1];
	dp[2] = v[2];
	dp[3] = v[3];
}

RxVecd4& RxVecd4::add(const RxVecd4 &v1, const RxVecd4 &v2)
{
	v[0] = v1.v[0] + v2.v[0];
	v[1] = v1.v[1] + v2.v[1];
	v[2] = v1.v[2] + v2.v[2];
	v[3] = v1.v[3] + v2.v[3];
	return (*this);
}

RxVecd4& RxVecd4::add(const RxVecd4 &v2)
{
	v[0] += v2.v[0];
	v[1] += v2.v[1];
	v[2] += v2.v[2];
	v[3] += v2.v[3];
	return (*this);
}

RxVecd4& RxVecd4::sub(const RxVecd4 &v1, const RxVecd4 &v2)
{
	v[0] = v1.v[0] - v2.v[0];
	v[1] = v1.v[1] - v2.v[1];
	v[2] = v1.v[2] - v2.v[2];
	v[3] = v1.v[3] - v2.v[3];
	return (*this);
}

RxVecd4& RxVecd4::sub(const RxVecd4 &v2)
{
	v[0] -= v2.v[0];
	v[1] -= v2.v[1];
	v[2] -= v2.v[2];
	v[3] -= v2.v[3];
	return (*this);
}

RxVecd4& RxVecd4::extrap(const RxVecd4 &v1, double a, const RxVecd4 &v2)
{
	v[0] = v1.v[0] + a*v2.v[0];
	v[1] = v1.v[1] + a*v2.v[1];
	v[2] = v1.v[2] + a*v2.v[2];
	v[3] = v1.v[3] + a*v2.v[3];
	return (*this);
}

RxVecd4& RxVecd4::scale(double a, const RxVecd4 &v1)
{
	v[0] = v1.v[0] * a;
	v[1] = v1.v[1] * a;
	v[2] = v1.v[2] * a;
	v[3] = v1.v[3] * a;
	return (*this);
}

RxVecd4& RxVecd4::scale(double a)
{
	v[0] *= a;
	v[1] *= a;
	v[2] *= a;
	v[3] *= a;
	return (*this);
}

double RxVecd4::dot(const RxVecd4 &v2) const
{
	return (v[0]*v2.v[0] + 
		v[1]*v2.v[1] + 
		v[2]*v2.v[2] +
		v[3]*v2.v[3]
	       );
}

RxVecd4& RxVecd4::unit(const RxVecd4 &v1)
{
	double mag;
 
	if ((mag = v1.norm2()) > 0)
	 { v[0] = v1.v[0]/mag;
	   v[1] = v1.v[1]/mag;
	   v[2] = v1.v[2]/mag;
	   v[3] = v1.v[3]/mag;
	 }
	else
	 { setZero();
	 }
	return (*this);
}

RxVecd4& RxVecd4::neg(const RxVecd4 &v1)
{
	v[0] = -v1.v[0];
	v[1] = -v1.v[1];
	v[2] = -v1.v[2];
	v[3] = -v1.v[3];
	return (*this);
}

RxVecd4& RxVecd4::setZero()
{
	v[0] = 0;
	v[1] = 0;
	v[2] = 0;
	v[3] = 0;
	return (*this);
}

double RxVecd4::norm1() const
{
	return (ABS(v[0]) +
		ABS(v[1]) +
		ABS(v[2]) +
		ABS(v[3])
	       );
}

double RxVecd4::norm2() const
{
	return (sqrt(v[0]*v[0] +
		     v[1]*v[1] +
		     v[2]*v[2] +
		     v[3]*v[3]
		    ));
}

double RxVecd4::normInf() const
{
	double tmp;
	double max;
 
	max = ABS(v[0]);
	if ((tmp = ABS(v[1])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[2])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[3])) > max)
	 { max = tmp;
	 }

	return (max);
}



