#include <RxVecd6.h>
#include <math.h>
#include <cdefs.h>

RxVecd6& RxVecd6::set (const double *dp)
{
	v[0] = dp[0];
	v[1] = dp[1];
	v[2] = dp[2];
	v[3] = dp[3];
	v[4] = dp[4];
	v[5] = dp[5];
	return (*this);
}

void RxVecd6::get (double *dp)
{
	dp[0] = v[0];
	dp[1] = v[1];
	dp[2] = v[2];
	dp[3] = v[3];
	dp[4] = v[4];
	dp[5] = v[5];
}

RxVecd6& RxVecd6::add(const RxVecd6 &v1, const RxVecd6 &v2)
{
	v[0] = v1.v[0] + v2.v[0];
	v[1] = v1.v[1] + v2.v[1];
	v[2] = v1.v[2] + v2.v[2];
	v[3] = v1.v[3] + v2.v[3];
	v[4] = v1.v[4] + v2.v[4];
	v[5] = v1.v[5] + v2.v[5];
	return (*this);
}

RxVecd6& RxVecd6::add(const RxVecd6 &v2)
{
	v[0] += v2.v[0];
	v[1] += v2.v[1];
	v[2] += v2.v[2];
	v[3] += v2.v[3];
	v[4] += v2.v[4];
	v[5] += v2.v[5];
	return (*this);
}

RxVecd6& RxVecd6::sub(const RxVecd6 &v1, const RxVecd6 &v2)
{
	v[0] = v1.v[0] - v2.v[0];
	v[1] = v1.v[1] - v2.v[1];
	v[2] = v1.v[2] - v2.v[2];
	v[3] = v1.v[3] - v2.v[3];
	v[4] = v1.v[4] - v2.v[4];
	v[5] = v1.v[5] - v2.v[5];
	return (*this);
}

RxVecd6& RxVecd6::sub(const RxVecd6 &v2)
{
	v[0] -= v2.v[0];
	v[1] -= v2.v[1];
	v[2] -= v2.v[2];
	v[3] -= v2.v[3];
	v[4] -= v2.v[4];
	v[5] -= v2.v[5];
	return (*this);
}

RxVecd6& RxVecd6::extrap(const RxVecd6 &v1, double a, const RxVecd6 &v2)
{
	v[0] = v1.v[0] + a*v2.v[0];
	v[1] = v1.v[1] + a*v2.v[1];
	v[2] = v1.v[2] + a*v2.v[2];
	v[3] = v1.v[3] + a*v2.v[3];
	v[4] = v1.v[4] + a*v2.v[4];
	v[5] = v1.v[5] + a*v2.v[5];
	return (*this);
}

RxVecd6& RxVecd6::scale(double a, const RxVecd6 &v1)
{
	v[0] = v1.v[0] * a;
	v[1] = v1.v[1] * a;
	v[2] = v1.v[2] * a;
	v[3] = v1.v[3] * a;
	v[4] = v1.v[4] * a;
	v[5] = v1.v[5] * a;
	return (*this);
}

RxVecd6& RxVecd6::scale(double a)
{
	v[0] *= a;
	v[1] *= a;
	v[2] *= a;
	v[3] *= a;
	v[4] *= a;
	v[5] *= a;
	return (*this);
}

double RxVecd6::dot(const RxVecd6 &v2) const
{
	return (v[0]*v2.v[0] + v[1]*v2.v[1] + v[2]*v2.v[2] +
		v[3]*v2.v[3] + v[4]*v2.v[4] + v[5]*v2.v[5]);
}

RxVecd6& RxVecd6::unit(const RxVecd6 &v1)
{
	double mag;
 
	if ((mag = v1.norm2()) > 0)
	 { v[0] = v1.v[0]/mag;
	   v[1] = v1.v[1]/mag;
	   v[2] = v1.v[2]/mag;
	   v[3] = v1.v[3]/mag;
	   v[4] = v1.v[4]/mag;
	   v[5] = v1.v[5]/mag;
	 }
	else
	 { setZero();
	 }
	return (*this);
}

RxVecd6& RxVecd6::neg(const RxVecd6 &v1)
{
	v[0] = -v1.v[0];
	v[1] = -v1.v[1];
	v[2] = -v1.v[2];
	v[3] = -v1.v[3];
	v[4] = -v1.v[4];
	v[5] = -v1.v[5];
	return (*this);
}

RxVecd6& RxVecd6::setZero()
{
	v[0] = 0;
	v[1] = 0;
	v[2] = 0;
	v[3] = 0;
	v[4] = 0;
	v[5] = 0;
	return (*this);
}

double RxVecd6::norm1() const
{
	return (ABS(v[0]) + ABS(v[1]) + ABS(v[2]) +
		ABS(v[3]) + ABS(v[4]) + ABS(v[5]));
}

double RxVecd6::norm2() const
{
	return (sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2] +
		     v[3]*v[3] + v[4]*v[4] + v[5]*v[5]));
}

double RxVecd6::normInf() const
{
	double tmp;
	double max;
 
	max = ABS(v[0]);
	if ((tmp = ABS(v[1])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[2])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[3])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[4])) > max)
	 { max = tmp;
	 }
	if ((tmp = ABS(v[5])) > max)
	 { max = tmp;
	 }
	return (max);
}

