#include <RxVecd.h>
#include <RxVecdBaseTest.h>
#include <RxTimeval.h>
#include <assert.h>
#include <cdefs.h>
#include <math.h>
#include <unistd.h>
#include <iostream.h>
#include <fstream.h>

#define EPS	(1e-14)
#define EQ(a,b)	(ABS(a-b) < EPS)

#define LEN	4

double testVals[][LEN] = {
 { 0.0, 0.0, 0.0, 0.0 },
 { 1.0, 2.0, 3.0, 4.0 },
 {-3.0, -2.0, -1.0, 3.0 }};

char *testIOFile = "RxVecdTest.out";

int hook()
{
}

testIO (char *str, char *fmt, int prec, RxVecd *vp)
{
	char buf[1024];
	FILE *fp;
	char *cp;
	int status;
	RxVecd *vr = NULL;
	int doEof, k, i;

	if (vp)
	 { vr = new RxVecd (vp->length());
	 }
	else
	 { vr = new RxVecd (LEN);
	 }
	for (k=0; k<2; k++)
	 { for (doEof=0; doEof<2; doEof++)
	    {
	      if ((fp = fopen (testIOFile, "w")) == NULL)
	       { fprintf (stderr,
"testIO: can't open file %c for writing\n", testIOFile);
		 exit (1);
	       }
	      for (i=0; i<strlen(str); i++)
	       { putc (str[i], fp);
		 if (doEof && i>=strlen(str)/2)
		  { break;
		  }
	       }
	      fclose (fp);
	      switch (k)
	       { case 0:
		  { if ((fp = fopen (testIOFile, "r")) == NULL)
		     { fprintf (stderr,
"testIO: can't open file %c for reading\n", testIOFile);
		       exit (1);
		     }
		    if (doEof) hook();
		    status = vr->scan (fp);
		    fclose (fp);
		    break;
		  }
		 case 1:
		  { ifstream *isp = new ifstream (testIOFile);
		    vr->scan (*isp);
		    if (isp->eof())
		     { status = EOF;
		     }
		    else
		     { status = (isp->good() ? 1 : 0);
		     }
		    delete isp;
		    break;
		  }
	       }
	      if (status == 1)
	       { if (vp == NULL)
		  { fprintf (stderr,
"testIO: expecting error but got vector anyway\n");
		    exit (1);
		  }
		 else
		  { assert (vr->equal(*vp, EPS));
		  }
	       }
	      else if (status == EOF)
	       { if (doEof == 0)
		  { fprintf (stderr, "testIO: unexpected EOF\n");
		    exit (1);
		  }
	       }
	      else if (status == 0)
	       { if (vp != NULL)
		  { fprintf (stderr, "testIO: unexpected error\n");
		    exit (1);
		  }
	       }
	    }
	   if (vp)
	    { switch (k)
	       { case 0:
		  { fp = fopen (testIOFile, "w");
		    vp->print (fp, fmt);
		    fclose (fp);
		    break;
		  }
		 case 1:
		  { ofstream *osp = new ofstream (testIOFile);
		    vp->print (*osp);
		    osp->flush();
		    delete osp;
		    break;
		  }
	       }
	      fp = fopen (testIOFile, "r");
	      cp = buf;
	      while ((*cp = getc(fp)) != EOF)
	       { cp++;
	       }
	      *cp = '\0';
	      if (strcmp (buf, str) != 0)
	       { fprintf (stderr,
"testIO: expecting output\n%s\nbut got instead\n%s\n", str, buf);
		 exit (1);
	       }
	    }
	 }
	unlink (testIOFile);
}

main()
{
	static double vec1[LEN] = { 11.1, 22.2, 33.3, 44.4 };
	static double vec2[LEN] = { 0.0, 0.0, 0.0, 0.0 };
	RxVecd v1(LEN);
	RxVecd v2(LEN);
	RxVecd vd(vec1, LEN);
	int i, j;

	for (i=0; i<LEN; i++)
	 { assert (EQ(vd(i), vec1[i]));
	 }
	RxVecd v3(vd);
	for (i=0; i<LEN; i++)
	 { assert (EQ(v3(i), vec1[i]));
	 }
	assert (v3.length() == LEN);

	assert (v3.equal(vd,EPS));
	assert (!v3.equal(v1,EPS));

	for (i=0;i<LEN;i++)
	 { v1(i) = vec1[i];
	 }
	assert (v1.equal(v3,EPS));

	for (i=0;i<LEN;i++)
	 { v1(i) = i*10+2;
	 }
	assert (!v2.equal(v1,EPS));
	v2 = v1;
	assert (v2.equal(v1,EPS));

	v1.set (vec1);
	for (i=0;i<LEN;i++)
	 { assert (EQ(v1(i), vec1[i]));
	 }
	v1.get (vec2);		
	for (i=0;i<LEN;i++)
	 { assert (EQ(vec1[i], vec2[i]));
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { for (j=i; j<ARRAY_SIZE(testVals); j++)
	    { v1.set(testVals[i]);
	      v2.set(testVals[j]);

	      assert (testAdd (v1+v2, v1, v2, EPS));
	      assert (testAdd (v3.add(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testAdd (v3.add(v2), v1, v2, EPS));
	      v3 = v1;					    
	      assert (testAdd (v3 += v2, v1, v2, EPS));

	      assert (testSub (v1-v2, v1, v2, EPS));
	      assert (testSub (v3.sub(v1,v2), v1, v2, EPS));
	      v3 = v1;
	      assert (testSub (v3.sub(v2), v1, v2, EPS));
	      v3 = v1;					    
	      assert (testSub (v3 -= v2, v1, v2, EPS));

	      assert (testExtrap (v3.extrap(v1,11.1,v2), v1, 11.1, v2, EPS));
	      v3 = v1;
	      assert (testExtrap (v3.extrap(11.1,v2), v1, 11.1, v2, EPS));

	      v3 = v1;
	      assert (testDot (v3.dot(v2), v1, v2, EPS));
	    }
	 }

	for (i=0; i<ARRAY_SIZE(testVals); i++)
	 { v1.set(testVals[i]);

	   assert (testScale (v3.scale(M_PI,v1), M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3.scale(M_PI), M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3 *= M_PI, M_PI, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3 /= M_PI, 1/M_PI, v1, EPS));
	   assert (testScale (M_PI*v1, M_PI, v1, EPS));
	   assert (testScale (v1*M_PI, M_PI, v1, EPS));
	   assert (testScale (v1/M_PI, 1/M_PI, v1, EPS));

	   assert (testUnit (v3.unit(v1), v1, EPS));
	   v3 = v1;
	   assert (testUnit (v3.unit(), v1, EPS));
	   assert (testScale (v3.neg(v1), -1.0, v1, EPS));
	   v3 = v1;
	   assert (testScale (v3.neg(), -1.0, v1, EPS));
	   assert (testScale (-v1, -1.0, v1, EPS));

	   assert (testNorm1 (v1.norm1(), v1, EPS));
	   assert (testNorm2 (v1.norm2(), v1, EPS));
	   assert (testNormInf (v1.normInf(), v1, EPS));
	 }

	for (i=0; i<LEN; i++)
	 { v1(i) = 0;
	 }
	assert ((v2.setZero()).equal(v1, EPS));

	// test out the I/O routines
	 { char buf[256];
	   int status;

	   v1.set(vec1);
	   status = v1.print (buf, "foo %4.1f bar\n");

	   assert (strcmp (buf, "foo [11.1 22.2 33.3 44.4] bar\n") == 0);
	   assert (status == strlen(buf));
	   status = v2.scan (buf+3);
	   assert (status == 1);
	   assert (v1.equal (v2, EPS));

	   status = v1.print (buf, "foo %4.1f,%5.2f bar\n");
	   assert (strcmp (buf, "foo [11.1,22.20,33.30,44.40] bar\n") == 0);
	   assert (status == strlen(buf));

	   status = v1.print (buf, "foo %4.1f,%5.2f bar\n", 12);
	   assert (strcmp (buf, "foo [11.1,2") == 0);
	   assert (status == -1);
	 }

	RxVecd vp(3);
	vp(0) = 1.2;
	vp(1) = 3.3;
	vp(2) = 3.4;
	testIO ("[1.2 3.3 3.4]", "%3.1f", 1, &vp);
	testIO ("[1.2 foo bar]", "", 0, NULL);
	testIO ("[1.2 foo+bar]", "", 0, NULL);

	printf ("\nROUTINES OK\n\n");
	return 0;
}
