%{

/* try /local/mouse/bin/localsym */

/* defines to cover the global yacc symbols */

#include <stdio.h>
#include <stdiox.h>
#include <ctype.h>
#include <RxParseLex.h>
#include <RxParser.h>

#define yyerror(s)	rxParseError (s, (ParseArg*)yyarg, yychar, &yylval)
int yyparse (void *);

int RxParser::doparse (void *arg)
{
	return yyparse (arg);
}

#define YYPARSE_PARAM   yyarg

static void rxParseError (char *str, ParseArg *parg, int yychar, 
			 YYSTYPE *yylvalp) 
{
	if (parg->errp->getMsg() == NULL)
	 { char *tokenStr;

	   switch (yychar)
	    { case BOOLEAN_LIT:
	      case INTEGER_LIT:
	      case DOUBLE_LIT:
	      case STRING_LIT:
	      case REFERENCE:
	      case IDENTIFIER_TOK:
	      case TYPENAME:
	       { tokenStr = yylvalp->s_raw;
		 break;
	       }
	      default:
	       { tokenStr = tokenToStr(yychar);
	       }
	    }
	   if (yychar == EOF_TOK)
	    { parg->eof = 1;
	    } 
	   else
	    { parg->errp->set (RxSysError::ParsingError,
"RxParseObj::scan: %s, near %s\n", str, tokenStr);
	    }
	   parg->errp->lineno() = parg->lineno;
	 }
}

#define DEBUG 1
#define PARSE_ARG ((ParseArg*)yyarg)
#if DEBUG
#define REDUCE_MSG(s) if (PARSE_ARG->debug) printf ("REDUCE %s", s);
#else
#define REDUCE_MSG(s)
#endif

#define SET_ERROR(fmt,line,arg1,arg2) \
	YYARG->errp->set (RxSysError::ParsingError, \
			  fmt, "RxParseObj::scan:", arg1, arg2); \
	YYARG->errp->lineno() = line;

%}
%token INTEGER_LIT DOUBLE_LIT STRING_LIT BOOLEAN_LIT
%token LEFTPAREN RIGHTPAREN LEFTBRACKET RIGHTBRACKET EQUAL 
%token IDENTIFIER_TOK REFERENCE TYPENAME COMMA
%token QUESTION_MARK ASSIGN
%token PLUS MINUS TIMES DIVIDE
%token SEP ERR EOF_TOK
%left PLUS MINUS
%left TIMES DIVIDE
%right ASSIGN
%start object
%%
object		: atomic_expr
		   { PARSE_ARG->node = $1.node;
	             REDUCE_MSG ("object : atomic_expr\n");
		     YYACCEPT;
		   }	
		;
literal		: DOUBLE_LIT
		   { $$.node = new RxParseObj();
		     $$.node->typeFlag = RxParseObj::DOUBLE;
		     $$.node->val.d = $1.u.d;
		     $$.node->lineno = $1.lineno;
		     free ($1.s_raw);
		     REDUCE_MSG ("literal : DOUBLE\n");
		   }
		| INTEGER_LIT
		   { $$.node = new RxParseObj();
		     $$.node->typeFlag = RxParseObj::INTEGER;
		     $$.node->val.i = $1.u.i;
		     $$.node->lineno = $1.lineno;
		     free ($1.s_raw);
		     REDUCE_MSG ("literal : INTEGER\n");
		   }
		| BOOLEAN_LIT
		   { $$.node = new RxParseObj();
		     $$.node->typeFlag = RxParseObj::BOOLEAN;
		     $$.node->val.i = $1.u.i;
		     $$.node->lineno = $1.lineno;
		     REDUCE_MSG ("literal : BOOLEAN\n");
		   }
		| STRING_LIT
		   { $$.node = new RxParseObj();
		     $$.node->typeFlag = RxParseObj::STRING;
		     $$.node->val.s = $1.u.s;
		     $$.node->lineno = $1.lineno;
		     free ($1.s_raw);
		     REDUCE_MSG ("literal : STRING\n");
		   }
		;
list		: TYPENAME LEFTBRACKET listBody RIGHTBRACKET
		   { $$.node = $3.node;
		     $$.node->typeName = $1.u.s;
		     $$.node->lineno = $1.lineno;
		     free ($1.s_raw);
		     REDUCE_MSG ("list : TYPENAME LEFTBRACKET listBody RIGHTBRACKET\n");
		   }
		| LEFTBRACKET listBody RIGHTBRACKET
		   { $$.node = $2.node;
		     $$.node->lineno = $1.lineno;
		     REDUCE_MSG ("list : LEFTBRACKET listBody RIGHTBRACKET\n");
		   }
		;
listBody	:
		   { $$.node = new RxParseObj();
		     $$.node->depth = 1;
		     $$.node->typeFlag = RxParseObj::LIST;
		     $$.node->val.listp = &($$.node->list_);
		     REDUCE_MSG ("listBody :\n");
		   }
		| listBody COMMA listEntry
		   { RxParseNode *linkNode = new RxParseNode($3.node);
		     $$.node->val.listp->addTail (linkNode);
		     $3.node->linkNode = linkNode;
		     $$.node->depth = MAX($$.node->depth, $3.node->depth+1);
		     if ($3.node->typeFlag & RxParseObj::LIST)
		      { $$.node->typeFlag |= RxParseObj::LIST_IN_LIST;
		      }
		     else
		      { $$.node->typeFlag |= $3.node->typeFlag;
		      }
		     REDUCE_MSG ("listBody : listBody COMMA listElem\n");
		   }
		| listBody listEntry
		   { RxParseNode *linkNode = new RxParseNode($2.node);
		     $$.node->val.listp->addTail (linkNode);
		     $2.node->linkNode = linkNode;
		     $$.node->depth = MAX($$.node->depth, $2.node->depth+1);
		     if ($2.node->typeFlag & RxParseObj::LIST)
		      { $$.node->typeFlag |= RxParseObj::LIST_IN_LIST;
		      }
		     else
		      { $$.node->typeFlag |= $2.node->typeFlag;
		      }
		     REDUCE_MSG ("listBody : listBody listElem\n");
		   }
		;
listEntry	: IDENTIFIER_TOK EQUAL listElem
		   { $$.node = $3.node;
		     $$.node->tagName = $1.u.s;
		     REDUCE_MSG ("listEntry : IDENTIFIER_TOK EQUAL listElem\n");
		   }
		| listElem
		   { $$.node = $1.node;
		     REDUCE_MSG ("listEntry : listElem\n");
		   }
		;
listElem	: atomic_expr
		   { $$.node = $1.node;
		     REDUCE_MSG ("listElem : atomic_expr\n");
		   }
		| IDENTIFIER_TOK
		   { $$.node = new RxParseObj();
		     $$.node->typeFlag = RxParseObj::IDENTIFIER;
		     $$.node->val.s = $1.u.s;
		     $$.node->lineno = $1.lineno;
		     REDUCE_MSG ("listElem : IDENTIFIER\n");
		   }
		;
atomic_expr	: LEFTPAREN expr RIGHTPAREN
		   { $$.node = $2.node;
		     REDUCE_MSG ("atomic_expr : LEFTPAREN expr RIGHTPAREN\n");
		   }
		| IDENTIFIER_TOK ASSIGN atomic_expr
		   { RxStrHashTab *htab;
		     $$.node = $3.node;
		     $$.node->lineno = $1.lineno;
		     $$.node->reference();
		     if ((htab = YYARG->cinp->getHashTab()))
		      { RxStrHashNode *hash;
			if ((hash = htab->find($1.u.s)))
			 { delete (RxParseObj*)hash->ptr();
			   hash->setPtr($$.node);
			 }
			else
			 { htab->insert ($1.u.s, $$.node);
			 }
		      }
		     free ($1.u.s);
		     REDUCE_MSG ("atomic_expr : IDENTIFIER_TOK ASSIGN atomic_expr\n");
		   }
		| literal
		   { $$.node = $1.node;
		     REDUCE_MSG ("atomic_expr : literal\n");
		   }
		| list
		   { $$.node = $1.node;
		     REDUCE_MSG ("atomic_expr : list\n");
		   }
		| REFERENCE
		   { RxStrHashTab *htab;
		     REDUCE_MSG ("atomic_expr : REFERENCE\n");
		     if ((htab = YYARG->cinp->getHashTab()))
		      { RxStrHashNode *hash;
			if ((hash = htab->find($1.u.s)))
			 { RxParseObj *ref = (RxParseObj*)hash->ptr();
			   $$.node = new RxParseObj();
			   $$.node->setRef (*ref);
			   $$.node->lineno = $1.lineno;
			 }
			else
			 { 
SET_ERROR ("%s undefined reference @%s\n", $1.lineno, $1.u.s, 0);
			   YYERROR;
			 }
		      }
		     else
		      { 
SET_ERROR ("%s no hash table to lookup reference\n", $1.lineno, 0, 0);
			YYERROR;
		      }
//		     free ($1.u.s);
		     free ($1.s_raw);
		   }
		| MINUS atomic_expr
		   { if (RxParser::unnop ($2.node, '-', yyarg) < 0)
		      { YYERROR;
		      }
		     $$.node = $2.node;
		     REDUCE_MSG ("atomic_expr : MINUS atomic_expr\n");
		   }
		| PLUS atomic_expr
		   { if (RxParser::unnop ($2.node, '+', yyarg) < 0)
		      { YYERROR;
		      }
		     $$.node = $2.node;
		     REDUCE_MSG ("atomic_expr : PLUS atomic_expr\n");
		   }
		| ERR
		   {
		     REDUCE_MSG ("atomic_expr : ERR\n");
SET_ERROR ("%s %s", $1.lineno, $1.u.s, 0);
		     YYERROR;
		   }
		;
expr		: atomic_expr
		   { $$.node = $1.node;
		     REDUCE_MSG ("expr : atomic_expr\n");
		   }
		| expr PLUS expr
		   { if (RxParser::binop ($1.node, $3.node, '+', yyarg) < 0)
		      { YYERROR;
		      }
		     $$.node = $1.node;
		     delete $3.node;
		     REDUCE_MSG ("expr : expr PLUS expr\n");
		   }
		| expr MINUS expr
		   { if (RxParser::binop ($1.node, $3.node, '-', yyarg) < 0)
		      { YYERROR;
		      }
		     $$.node = $1.node;
		     delete $3.node;
		     REDUCE_MSG ("expr : expr MINUS expr\n");
		   }
		| expr TIMES expr
		   { if (RxParser::binop ($1.node, $3.node, '*', yyarg) < 0)
		      { YYERROR;
		      }
		     $$.node = $1.node;
		     delete $3.node;
		     REDUCE_MSG ("expr : expr TIMES expr\n");
		   }
		| expr DIVIDE expr
		   { if (RxParser::binop ($1.node, $3.node, '/', yyarg) < 0)
		      { YYERROR;
		      }
		     $$.node = $1.node;
		     delete $3.node;
		     REDUCE_MSG ("expr : expr DIVIDE expr\n");
		   }
		;
%%

void RxParser::assignReference (RxParseObj *obj, RxParseObj *ref)
{
	obj->typeFlag = ref->typeFlag;
	obj->typeName = (ref->typeName ? strdup(ref->typeName) : NULL);
	obj->val = ref->val;
}

int RxParser::binop (RxParseObj *n1, RxParseObj *n2, int op, void *yyarg)
{
	double val1, val2;

	if (n1->typeFlag == RxParseObj::INTEGER && 
            n2->typeFlag == RxParseObj::INTEGER)
	 { switch (op)
	    { case '+':
	       { n1->val.i = n1->val.i + n2->val.i;
		 break;
	       }
	      case '-':
	       { n1->val.i = n1->val.i - n2->val.i;
		 break;
	       }
	      case '*':
	       { n1->val.i = n1->val.i * n2->val.i;
		 break;
	       }
	      case '/':
	       { if (n2->val.i == 0)
		  { 
SET_ERROR ("%s integer divide by zero\n", n2->lineno, 0, 0);
		    return (-1);
		  }
		 n1->val.i = n1->val.i / n2->val.i;
		 break;
	       }
	    }
	 }
	else if ((n1->typeFlag == RxParseObj::INTEGER ||
		  n1->typeFlag == RxParseObj::DOUBLE) &&
		 (n2->typeFlag == RxParseObj::INTEGER ||
		  n2->typeFlag == RxParseObj::DOUBLE))
	 { val1 = (n1->typeFlag==RxParseObj::INTEGER ? n1->val.i : n1->val.d);
	   val2 = (n2->typeFlag==RxParseObj::INTEGER ? n2->val.i : n2->val.d);
	   switch (op)
	    { case '+':
	       { n1->val.d = val1 + val2;
		 break;
	       }
	      case '-':
	       { n1->val.d = val1 - val2;
		 break;
	       }
	      case '*':
	       { n1->val.d = val1 * val2;
		 break;
	       }
	      case '/':
	       { n1->val.d = val1 / val2;
		 break;
	       }
	    }
	   n1->typeFlag = RxParseObj::DOUBLE;
	 }
	else if (n1->typeFlag == RxParseObj::STRING &&
		 n2->typeFlag == RxParseObj::STRING && op == '+')
	 { char *s1;
	   s1 = n1->val.s;
	   n1->val.s = (char*)malloc(strlen(s1)+strlen(n2->val.s)+1);
	   strcpy (n1->val.s, s1);
	   strcat (n1->val.s, n2->val.s);
	   free (s1);
	 }
	else
	 {
SET_ERROR ("%s invalid types for binary operator '%c'\n", n1->lineno, op, 0);
	   return (-1);
	 }
	return (1);
}

int RxParser::unnop (RxParseObj *n1, int op, void *yyarg)
{
	if (n1->typeFlag != RxParseObj::INTEGER && 
	    n1->typeFlag != RxParseObj::DOUBLE)
	 { 
SET_ERROR ("%s invalid type for unary operator '%c'\n", n1->lineno, op, 0);
	   return (-1);
	 }
	if (n1->typeFlag == RxParseObj::INTEGER)
	 { switch (op)
	    { case '-':
	       { n1->val.i = -n1->val.i;
		 break;
	       }
	    }
	 }
	else
	 { switch (op)
	    { case '-':
	       { n1->val.d = -n1->val.d;
		 break;
	       }
	      case '+':
	       { break;
	       }
	    }
	    
	 }
	return (1);
}

#include "RxParseLex.cc"
