#include <cdefs.h>
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <RxCio.h>
#include <stdlib.h>
#include <RxParseLex.h>
#include <RxParseTokenDefs.h>
#include <stdiox.h>

char **argvec;

YYSTYPE *testTokVals;
int *testToks;
int numTestToks;

void readTestTokens (FILE *fp)
{
	char ibuf[256];
#define MAX_TEST_TOKENS	1024
	int maxToks = MAX_TEST_TOKENS;
	int numToks = 0;
	int token;
	
	testTokVals = (YYSTYPE*)malloc(sizeof(YYSTYPE)*maxToks);
	testToks = (int*)malloc(sizeof(int)*maxToks);
	xscanfBegin (fp);
	do
	 { if (xscanf (fp, "%s", ibuf) != 1)
	    { fprintf (stderr, 
"readTestTokens: can't read token string\n");
	      exit (1);
	    }
	   token = strToToken (ibuf);
	   if (token != EOF_TOK)
	    { if (numToks >= maxToks)
	       { maxToks *= 2;
		 testTokVals = (YYSTYPE*)realloc(testTokVals,
						   sizeof(YYSTYPE)*maxToks);
		 testToks = (int*)realloc(testToks,sizeof(int)*maxToks);
	       }
	      switch (token)
	       { case IDENTIFIER_TOK:
		 case REFERENCE:
		 case TYPENAME:
		  { if (xscanf (fp, "%s", ibuf) != 1)
		     { fprintf (stderr,
"readTestTokens: can't read string value for token\n");
		       exit (1);
		     }
		    testTokVals[numToks].u.s = strdup(ibuf);
		    break;
		  }
		 case DOUBLE_LIT:
		  { if (xscanf (fp, "%lf", &testTokVals[numToks].u.d) != 1)
		     { fprintf (stderr, 
"readTestTokens: can't read double value for token\n");
		       exit (1);
		     }
		    break;
		  }
		 case INTEGER_LIT:
		 case BOOLEAN_LIT:
		  { if (xscanf (fp, "%d", &testTokVals[numToks].u.i) != 1)
		     { fprintf (stderr, 
"readTestTokens: can't read integer value for token\n");
		       exit (1);
		     }
		    break;
		  }
		 case STRING_LIT:
		  { int c0, c;
		    char *cp;
		    do
		     { c = xgetc(fp);
		     }
		    while (isspace(c));
		    c0 = c;
		    cp = ibuf;
		    do
		     { c = xgetc(fp);
		       if (c != c0)
			{ *cp++ = c;
			}
		     }
		    while (c != c0);
		    *cp = '\0';
		    testTokVals[numToks].u.s = strdup(ibuf);
		    break;
		  }
	       }
	      testToks[numToks] = token;
	      numToks++;
	    }
	 }
	while (token != EOF_TOK);
	xscanfEnd(fp);
	numTestToks = numToks;
}

main (int argc, char **argv)
{
	YYSTYPE yylval;
	int token;
	int automatic = 1;
	int k, j;
	int lastLine;
	ParseArg parg;
	RxCinput cin;

	argvec = argv;
	argc--; argv++;
	while (argc > 0)
	 { if (strcmp(*argv, "-i") == 0)
	    { automatic = 0;
	    }
	   else
	    { goto usage;
	    }
	   argc--; argv++;
	   continue;
usage:
	   fprintf (stderr, "Usage: %s [-i]\n", argvec[0]);
	   exit (-1);
	 }

	parg.lineno = 0;
	parg.cinp = &cin;
	if (automatic)
	 { FILE *fp;
	
	   if ((fp = fopen ("RxParseLexTest.data", "r")) == 0)
	    { fprintf (stderr, "Can't open file lextestData\n");
	      exit (-1);
	    }
	   cin.setStream (fp);
//	   yylexinit(fp);

	   /* start by computing the necessary lexical values */

	   readTestTokens(fp);
	   lastLine = 0;
	   j = 0;
	   for (k=0; (token = rxParseLexer(&yylval, &parg)) != EOF_TOK; k++)
	    { /* j counts the tokens from the start of the line */
	      if (lastLine != parg.lineno)
	       { lastLine = parg.lineno;
		 j = 1;
	       }
	      else
	       { j++;
	       }
	      if (k >= numTestToks)
	       { printf ("Too many tokens in test file\n");
		 printf ("k=%d numTestToks=%d\n", k, numTestToks);
		 exit (-1);
	       }
printf ("token=%d\n", token);
	      if (token != testToks[k])
	       { printf ("Token number %d, line %d:\n", j, parg.lineno);
		 printf ("Expecting %s, got %s (%d)\n", 
			  tokenToStr(testToks[k]),
			  tokenToStr(token), token);
		 
		 exit (-1);
	       }
	      switch (token)
	       { 
		 case INTEGER_LIT:
		  { if (yylval.u.i != testTokVals[k].u.i)
		     { printf ("Token number %d, line %d:\n", j, parg.lineno);
		       printf ("INTEGER_LIT: Expecting value %d, got %d\n", 
				testTokVals[k].u.i, yylval.u.i);
		       exit (-1);
		     }
		    break;
		  }
		 case BOOLEAN_LIT:
		  { if (yylval.u.i != testTokVals[k].u.i)
		     { printf ("Token number %d, line %d:\n", j, parg.lineno);
		       printf ("BOOLEAN_LIT: Expecting value %d, got %d\n", 
				testTokVals[k].u.i, yylval.u.i);
		       exit (-1);
		     }
		    break;
		  }
		 case DOUBLE_LIT:
		  { if (yylval.u.d != testTokVals[k].u.d)
		     { printf ("Token number %d, line %d:\n", j, parg.lineno);
		       printf ("DOUBLE_LIT: Expecting value %g, got %g\n", 
				testTokVals[k].u.d, yylval.u.d);
		       exit (-1);
		     }
		    break;
		  }
		 case IDENTIFIER_TOK:
		  { if (strcmp (yylval.u.s, testTokVals[k].u.s) != 0)
		     { printf ("Token number %d, line %d:\n", j, parg.lineno);
		       printf ("IDENTIFIER_TOK: Expecting value %s, got %s\n", 
				testTokVals[k].u.s, yylval.u.s);
		       exit (-1);
		     }
		    break;
		  }
		 case TYPENAME:
		  { if (strcmp (yylval.u.s, testTokVals[k].u.s) != 0)
		     { printf ("Token number %d, line %d:\n", j, parg.lineno);
		       printf ("TYPENAME: Expecting value %s, got %s\n", 
				testTokVals[k].u.s, yylval.u.s);
		       exit (-1);
		     }
		    break;
		  }
		 case REFERENCE:
		  { if (strcmp (yylval.u.s, testTokVals[k].u.s) != 0)
		     { printf ("Token number %d, line %d:\n", j, parg.lineno);
		       printf ("REFERENCE: Expecting value %s, got %s\n", 
				testTokVals[k].u.s, yylval.u.s);
		       exit (-1);
		     }
		    break;
		  }
		 case STRING_LIT:
		  { if (strcmp (yylval.u.s, testTokVals[k].u.s) != 0)
		     { printf ("Token number %d, line %d:\n", j, parg.lineno);
		       printf ("STRING_LIT: Expecting value %s, got %s\n", 
				testTokVals[k].u.s, yylval.u.s);
		       exit (-1);
		     }
		    break;
		  }
	       }
	    }	      
	   if (k != numTestToks)
	    { printf ("Not enough tokens in test file\n");
	      exit (-1);
	    }
	   printf ("\nLexical ROUTINES OK\n\n");
	 }
	else
	 { 
	   cin.setStream (stdin);
	    
	   while ((token = rxParseLexer(&yylval, &parg)) != EOF_TOK)
	    { switch (token)
	       { case INTEGER_LIT:
		  { printf ("INTEGER_LIT %d\n", yylval.u.i);	
		    break;
		  }
		 case DOUBLE_LIT:
		  { printf ("DOUBLE_LIT %g\n", yylval.u.d);	
		    break;
		  }
		 case STRING_LIT:
		  { printf ("STRING_LIT %s\n", yylval.u.s);	
		    break;
		  }
		 case IDENTIFIER_TOK:
		  { printf ("IDENTIFIER_TOK %s\n", yylval.u.s);	
		    break;
		  }
		 default:
		  { printf ("%s\n", tokenToStr(token));	
		    break;
		  }
	       }
	    }
	 }
	printf ("\nROUTINES OK\n\n");
	exit (0);
}

#include <RxParseLex.cc>
