#include <RxParseObj.h>
#include <RxSysError.h>
#include <stdlib.h>
#include <stdiox.h>
#include <RxParseLex.h>
#include <RxParser.h>
#include <assert.h>

static RxParseObj eofObj;
RxParseObj *RxParseObj::eof = &eofObj;

int RxParseObj::allocCnt = 0;

RxParseNode::~RxParseNode()
{
	if (obj_)
	 { delete obj_;
	 }
}

#if 0
void *RxParseObj::operator new (size_t n)
{
	void *p;
	p = malloc(n);
	allocCnt++;
	return p;
}

void RxParseObj::operator delete (void *p, size_t n)
{
	RxParseObj *obj = (RxParseObj*)p;

	if (obj->refcnt > 0)
	 { obj->refcnt--;
	 }
	else
	 { allocCnt--;
	   free (p);
	 }
}
#endif

void RxParseObj::clearType()
{
	if (typeName)
	 { free (typeName);
	   typeName = NULL;
	 }
	if (typeFlag == STRING)
	 { free (val.s);
	 }
	else if (typeFlag & LIST)
	 {
	   assert (val.listp == &list_ || refobj);
	   if (val.listp == &list_)
	    { list_.removeAndDeleteAll();	/* FIX THIS */
	    }
	   else /* refobj exists */
	    { delete refobj;
	      refobj = 0;
	    }
	 }
	depth = 0;
}

void RxParseObj::setRef (RxParseObj &ref)
{
	typeFlag = ref.typeFlag;
	if (ref.typeName)
	 { typeName = strdup(ref.typeName);
	 }
	else
	 { typeName = NULL;
	 }
	depth = ref.depth;

	if (ref.isList())
	 { val.listp = &ref.list_;
	   ref.reference();		// ref the object so it stays around
	   refobj = &ref;		// keep track of what we've ref'd
	 }
	else
	 { val = ref.val;
	 }
}

int RxParseObj::getValue (union Value *valp)
{
	*valp = val;
	return typeFlag;
}

double RxParseObj::getNumberValue() const
{
	if (typeFlag == DOUBLE)
	 { return val.d;
	 }
	else if (typeFlag == INTEGER)
	 { return val.i;
	 }
	else
	 { return 0;
	 }
}

int RxParseObj::getNumberListValues (double *vec, int n, RxParseObj **objp)
{
	RxParseObj *obj;
	int i;

	if (objp && *objp)
	 { obj = *objp;
	 }
	else
	 { obj = firstChild();
	 }
	for (i=0; i<n; i++)
	 { if (!obj || !obj->isNumber())
	    { break;
	    }
	   vec[i] = obj->getNumberValue();
	   obj = nextChild();
	 }
	if (objp)
	 { *objp = obj;
	 }
	return (i == n);
}

RxParseObj *RxParseObj::firstChild()
{
	curNode = val.listp->head();
	if (curNode)
	 { return curNode->obj();
	 }
	else
	 { return NULL;
	 }
}

RxParseObj *RxParseObj::nextChild()
{
	if (curNode)
	 { curNode = curNode->next();
	 }
	if (curNode)
	 { return curNode->obj();
	 }
	else
	 { return NULL;
	 }
}

void RxParseObj::init()
{
	typeFlag = 0;
	tagName = NULL;
	typeName = NULL;
	refptr = NULL;
	refobj = NULL;
	lineno = 1;
	curNode = NULL;
	depth = 0;
	linkNode = NULL;
	list_.setAutoDelete(0);
}

RxParseObj::~RxParseObj()
{
	if (ref.deleteOK())
	 { 
	   if (tagName)
	    { free (tagName);
	    }
	   clearType();
	 }
}

#if 0
void RxParseObj::freeObj (RxParseObj *obj)
{
	if (obj->refcnt == 0)
	 { delete obj;
	 }
}
#endif

void RxParseObj::freeHashedObjs (class RxCinput *cip)
{
	RxStrHashTab *htab = cip->getHashTab();

	if (htab)
	 { RxStrHashNode *hash;
	   for (hash=htab->first(); hash; hash=htab->next())
	    { delete (RxParseObj*)hash->ptr();
	    }
	   htab->clear();
	 }
}

RxParseObj* RxParseObj::scan (class RxCinput *cip, int *linenop)
{
	RxSysError err;
	ParseArg parseArg;
	int status;
	RxParser parser;
	RxError::action act;
	RxParseObj *obj = NULL;

	act = RxSysError::setAction(RxError::POST);
	parseArg.errp = new RxSysError();
	RxSysError::setAction(act);
	parseArg.level = 0;
	if (linenop)
	 { parseArg.lineno = *linenop;
	 }
	else if (cip->type() == RxCinput::IN_FILE_X)
	 { parseArg.lineno = xscanfLine ((FILE*)cip->getStream());
	 }
	else
	 { parseArg.lineno = 1;
	 }
	parseArg.cinp = cip;
	parseArg.debug = 0;
	parseArg.eof = 0;

	status = parser.doparse (&parseArg);
	if (linenop)
	 { *linenop = parseArg.lineno;
	 }
	if (status)
	 { if (parseArg.eof)
	    { obj = eof;
	    }
	   else
	    { parseArg.errp->raise();
	    }
	 }
	else
	 { obj = parseArg.node;
	 }
	delete parseArg.errp;
	return obj;
}

int RxParseObj::print (RxCoutput *co, RxPrintSpec *ps)
{
	RxParseNode *node;

	if (ps)	co->output_str (ps->prefix());
	if (typeFlag & LIST)
	 { co->output_char ('[');
	   for (node=val.listp->head(); node; node=node->next())
	    { node->obj()->print (co, NULL);
	      if (node != val.listp->tail())
	       { co->output_str (", ");
	       }
	    }
	   co->output_char (']');
	 }
	else
	 { switch (typeFlag)
	    { case BOOLEAN:
	       { if (val.i)
		  { co->output_str ("true");
		  }
		 else
		  { co->output_str ("false");
		  }
		 break;
	       }
	      case INTEGER:
	       { co->output_int ("%d", val.i);
		 break;
	       }
	      case DOUBLE:
	       { co->output_double ("%g", val.d);
 		 break;
	       }
	      case STRING:
	       { co->output_char ('"');
		 co->output_str (val.s);
		 co->output_char ('"');
		 break;
	       }
	      case IDENTIFIER:
	       { co->output_str (val.s);
		 break;
	       }
	    }
	 }
	if (ps) co->output_str (ps->suffix());
	return co->nchars();
}
