#include <RxS3RKin.h>
#include <RxMatdSvd.h>
#ifdef HAS_GRF4
#include <grf4.h>
#endif
#include <math.h>
#include <cdefs.h>
#include <stdio.h>
#include <mathx.h>
#include <useful.h>

int noGraphics = 0;
int winw = 600;
int winh = 600;
int nsamps = 200;
RxVecd3 p0;
RxVecd3 p1;
#if 1
double plotRng = (5*M_PI/4);
double plotOff = -(1*M_PI/16); /* -(DTOR*66.385); */
#else
double plotRng = 2*M_PI;
double plotOff = 0;
#endif
char *psFileName = "sing.idr";
double psWidth = 3.25;          /* width of PostScript output */   
double psHeight = 3.25;          /* height of PostScript output */
int angIdx = 2;

#ifdef HAS_GRF4

RxS3RKin s3Rkin;

int onResize (Grf4Event *ev, int)
{
	winw = ev->val1;
	winh = ev->val2;
	grf4Ortho (0, winw, 0, winh, -1.0, 1.0, /*aspectRatio=*/1);
	return (1);
}

doInvK (RxVecd3 *j3p, const RxTrsf T)
{
	int nr;

	nr = s3Rkin.invK (&j3p[0], T, 0, NULL, 0);
	if (nr < 2)
	 { return (0);
	 }
	if (nr == 2)
	 { s3Rkin.invK (&j3p[1], T, 1, NULL, 0);
	   return (2);
	 }
	if (nr == 4)
	 { s3Rkin.invK (&j3p[1], T, 1, NULL, 0);
	   s3Rkin.invK (&j3p[2], T, 2, NULL, 0);
	   s3Rkin.invK (&j3p[3], T, 3, NULL, 0);
	   return (4);
	 }
}

double closestAng (RxVecd3 *j3, int nr, int j, double ang)
{
	double mind;
	int mini;
	double d;
	int i;

	for (i=0; i<nr; i++)
	 { d = ABS(j3[i](j)-ang);
	   if (i==0 || d < mind)
	    { mini = i;
	      mind = d;
	    }
	 }
	return (j3[mini](j));
}

void doDraw()
{
	RxTrsf T;
	RxVecd3 j3[4];
	RxVecd3 j3prev[4];
	int nr;
	int nrprev;

	double len = (p0-p1).norm2();
	double scaleX = winw/(double)nsamps;
	double scaleY = winh/plotRng;
	double ang0, ang1;
	int i, j, k;

p0.print ("p0 %9.4f\n");
p1.print ("p1 %9.4f\n");
	grf4Clear();
	grf4PushTrans();
	grf4Translate (0, winh/2, 0);

	T.setIdent();

	grf4Move2 (0, plotOff*scaleY);
	grf4Draw2 (0, (M_PI/2+plotOff)*scaleY);
	for (i=0; i<nsamps; i++)
	 { T.p = p0 + (p1-p0)*i/(double)nsamps;

	   nr = doInvK (j3, T);
#if 1
	   if (i>0)
	    { if (nr >= nrprev)
	       { for (k=0; k<nr; k++)
		  { ang1 = j3[k](angIdx)+plotOff;
		    ang0 = closestAng (j3prev, nrprev, angIdx, j3[k](angIdx)) 
			 + plotOff;
		    grf4Move2 ((i-1)*scaleX, ang0*scaleY);
		    grf4Draw2 ((i)*scaleX, ang1*scaleY);
		  }
	       }
	      else
	       { for (k=0; k<nrprev; k++)
		  { ang0 = j3prev[k](angIdx)+plotOff;
		    ang1 = closestAng (j3, nr, angIdx, j3prev[k](angIdx)) 
			 + plotOff;
		    grf4Move2 ((i-1)*scaleX, ang0*scaleY);
		    grf4Draw2 ((i)*scaleX, ang1*scaleY);
		  }
	       }
	    }
#else
	   for (k=0; k<nr; k++)
	    { ang0 = rangePi(j3[k](angIdx));
	      grf4Point2 (i*scaleX, (ang0+plotOff)*scaleY);
	    }
#endif
	   nrprev = nr;
	   for (k=0; k<nr; k++) j3prev[k] = j3[k];
	 }
	grf4PopTrans();
	grf4Flush();
}

#endif

doBisect (RxVecd3 *pr, RxVecd3 &p1, RxVecd3 &p2)
{
	RxVecd3 pp1, pp2;
	RxTrsf T;
	RxVecd3 j3;
	int nr;

	pp1 = p1;
	pp2 = p2;

	T.p = pp1;
	nr = s3Rkin.invK (&j3, T, 0, NULL, 0);
	if (nr != 4)
	 { fprintf (stderr, "doBisect: nr=%d for pp1\n", nr);
	   exit (1);
	 }
	T.p = pp2;
	nr = s3Rkin.invK (&j3, T, 0, NULL, 0);
	if (nr != 2)
	 { fprintf (stderr, "doBisect: nr=%d for pp2\n", nr);
	   exit (1);
	 }
	while (ABS(pp1.z()-pp2.z()) > 1e-15 || ABS(pp1.y()-pp2.y()) > 1e-15)
	 { T.p = (pp1+pp2)/2;
/*  printf ("zdiff %g ydiff %g\n", pp2.z()-pp1.z(), pp2.y()-pp1.y()); */
	   nr = s3Rkin.invK (&j3, T, 0, NULL, 0);
	   if (nr == 2)
	    { pp2 = T.p;
	    }
	   else if (nr == 4)
	    { pp1 = T.p;
	    }
	   else
	    { fprintf (stderr, "doBisect: nr=%d\n", nr);
	      exit (1);
	    }
	 }
	*pr = pp1;
}

void psWriteOutput (char *name)
{
	FILE *fp;

	if ((fp = fopen(name, "w")) == NULL)
	 { fprintf (stderr, "Can't open %s\n", psFileName);
	   perror ("fopen");
	   exit (1);
	 }
	printf ("Drawing IDRAW PostScript ... \n");
	grf4PostScriptInit (fp, name, /*flags=*/GRF_IDRAW_PS,
			    /*width=*/72*psWidth, /*height=*/72*psHeight);
	grf4HideLines(1);
	doDraw();	
	grf4HideLines(0);
	grf4PostScriptFinish (fp);
	printf ("Done.\n");
	fclose (fp);
}

main()
{
	RxMatd J(6,6);
	RxMatdSvd svd;
	RxMatd Jr(3,3);
	RxVecd3 j3;
	RxVecd3 p2, ps, po;
	RxVecd3 pt, pb, pl, px;
	RxTrsf T;
	double ang = 0;
	double dy;
	double dz;
	double deg;
	int cnt;

	int nr;

	ps.x() = 0;
	ps.y() = 2.465552;
	ps.z() = 1.99871255;
#if 1
	ps.y() = 2.46555000910061928;
	ps.z() = 1.99871869512318967;
#endif
	/* ang = 66.385 */
	
	T.p = ps;
	nr = s3Rkin.invK (&j3, T, 1, NULL, 0);
	(RTOD*j3).print ("j3: %8.3f\n");

	s3Rkin.Jset (&J, j3);
	J.print ("J = [%10.5f,%10.5f]\n");
	s3Rkin.fwdK (&T, (int*)0, j3);

printf ("A\n");
	Jr = J(0,0,3,3);
	Jr.mul (T.R, Jr);
	T.R.print ("R %9.5f\n");
printf ("B\n");
	Jr.print ("Jr = [%10.5f,%10.5f]\n");
	svd.calc (Jr);
printf ("C\n");
	svd.getSig()->print ("sig %10.4f\n");
printf ("D\n");
	svd.getU()->print ("U   %10.5f\n");

#if 0
	px = ps;
	cnt = 0;	
	while (1)
	 { po = px;
	   po.z() += .1;
	   doBisect (&pt, px, po);
	   po.z() -= .2;
	   doBisect (&pb, px, po);
	   po = pt;
	   po.y() -= .1;
	   doBisect (&pl, pt, po);
	   px = (pt+pb)/2;
	   dz = pt.z() - pb.z();
	   dy = pl.y() - pt.y();
	   po = px;
	   po.y() += 100*dy;
	   po.z() += 100*dz;

	   T.p = po;
	   if (s3Rkin.invK (&j3, T, 0, NULL, 0) == 4)
	    { px = po;
	    }
	   else
	    { doBisect (&px, px, po);
	    }
	   cnt++;

	   if (cnt % 10 == 0)
	    { printf ("%d\n", cnt);
	      pt.print ("pt %20.17f\n");
	      pb.print ("pb %20.17f\n");
	      printf ("diff=%g\n", dz);
	      pl.print ("pl %20.17f\n");
	      printf ("diff=%g\n", dy);
	      px.print ("px %20.17f\n");

	      T.p = px;
	      s3Rkin.invK (&j3, T, 0, NULL, 0);
	      printf ("%9.4f\n", RTOD*j3(2));
	      s3Rkin.invK (&j3, T, 1, NULL, 0);
	      printf ("%9.4f\n", RTOD*j3(2));
	      s3Rkin.invK (&j3, T, 2, NULL, 0);
	      printf ("%9.4f\n", RTOD*j3(2));
	      s3Rkin.invK (&j3, T, 3, NULL, 0);
	      printf ("%9.4f\n", RTOD*j3(2));
	      (RTOD*j3).print ("jnts: %9.4f\n");
	    }
	 }
#endif
	if (!noGraphics)
	 { grf4PlaceWindow (200, 400, winw, winh);
	   grf4Init ("DAO", GRF_WHITE, GRF_BLACK, GRF_2DONLY | GRF_X11);

	   grf4AttachEvent (GRF_RESIZE_EVENT, onResize, 0);

	   grf4Ortho (0, winw, 0, winh, -1.0, 1.0, /*aspectRatio=*/1);
	   grf4SetDrawFxn (doDraw, (void*)0);
	 }
#ifdef HAS_GRF4
	for (deg=109.89; deg<110; deg+= 5)
	 { ang = DTOR*deg;
printf ("ang=%g\n", deg);
	   dy = cos(ang);
	   dz = sin(ang);
	   p0 = ps;
	   p1 = ps; 
	   p0.y() -= dy;
	   p1.y() += dy;
	   p0.z() -= dz;
	   p1.z() += dz;
	   ps.print ("ps %9.4f\n");
	   doDraw ();
	   getchar();
	 }
	if (psFileName)
	 { psWriteOutput (psFileName);
	 }
	if (!noGraphics)
	 { doDraw();
	   grf4Wait (-1);
	 }
#endif

}
