#include <RcCurveX.h>
#include <math.h>
#include <cdefs.h>

CurveSegDef *RcCurveXPoint::segs;
int RcCurveXPoint::numSegs;
int RcCurveXPoint::numRpts;

void RcCurveXPoint::setSegDefs (RxTrsf *Xp, double *slasts, int ns, int nr)
{
	CurveSegDef *sp;
	int i, k;

	if (ns <= 0)
	 { return;
	 }
	if (segs)
	 { delete[] segs;
	 }
	nr = MIN(MAX(1,nr),4);
	segs = new CurveSegDef[ns];
	numSegs = ns;
	numRpts = nr;

	sp = segs;
	for (i=0; i<ns; i++)
	 { sp->slast = slasts[i];

	   sp->va[3] = Xp[4*i+3].p-3*Xp[4*i+2].p+3*Xp[4*i+1].p-  Xp[4*i+0].p;
	   sp->va[2] =             3*Xp[4*i+2].p-6*Xp[4*i+1].p+3*Xp[4*i+0].p;
	   sp->va[1] =                           3*Xp[4*i+1].p-3*Xp[4*i+0].p;
	   sp->va[0] =                                           Xp[4*i+0].p;
	
	   RxRot D;

	   sp->RP[0] = Xp[4*i+0].R;
	   for (k=0; k<nr-1; k++)
	    { D.Imul (Xp[4*i+k].R, Xp[4*i+k+1].R);
	      D.getAxis (&sp->ra[k], &sp->rd[k]);
	      if (k < nr-2)
	       { sp->RP[k+1] = Xp[4*i+k+1].R;
	       }
	    }
	   sp++;
	 }
}

void RcCurveXPoint::calcRot (RxRot *RR, RxVecd3 *axisp, double *distp, 
		     CurveSegDef *sp, double sl)
{
	RxRot RT[5];
	RxRot D;
	RxVecd3 axis;
	double dist;

	switch (numRpts)
	 { case 1:
	    { axisp->setZero();
	      *distp = 0;
	      if (RR) *RR = sp->RP[0];
	      break;
	    }
	   case 2:
	    { *axisp = sp->ra[0];
	      *distp = sp->rd[0]*sl;
	      if (RR)
	       { D.setAxis (*axisp, *distp*sl);
		 RR->mul (sp->RP[0], D);
	       }
	      break;
	    }
	   case 3:
	    {
	      D.setAxis (sp->ra[0], sp->rd[0]*sl);
	      RT[0].mul (sp->RP[0], D);
	      D.setAxis (sp->ra[1], sp->rd[1]*sl);
	      RT[1].mul (sp->RP[1], D);

	      D.Imul (RT[0], RT[1]);
	      D.getAxis (axisp, distp);
	      if (RR)
	       { D.setAxis (*axisp, *distp*sl);
		 RR->mul (RT[0], D);
	       }
	      break;
	    }
	   case 4:
	    { 
	      D.setAxis (sp->ra[0], sp->rd[0]*sl);
	      RT[0].mul (sp->RP[0], D);
	      D.setAxis (sp->ra[1], sp->rd[1]*sl);
	      RT[1].mul (sp->RP[1], D);
	      D.setAxis (sp->ra[2], sp->rd[2]*sl);
	      RT[2].mul (sp->RP[2], D);

	      D.Imul (RT[0], RT[1]);
	      D.getAxis (&axis, &dist);
	      D.setAxis (axis, dist*sl);
	      RT[3].mul (RT[0], D);
	      D.Imul (RT[1], RT[2]);
	      D.getAxis (&axis, &dist);
	      D.setAxis (axis, dist*sl);
	      RT[4].mul (RT[1], D);

	      D.Imul (RT[3], RT[4]);
	      D.getAxis (axisp, distp);
	      if (RR)
	       { D.setAxis (*axisp, *distp*sl);
		 RR->mul (RT[3], D);
	       }
	      break;
	    }
	 }
}

int chook()
{
	return (0);
}

void RcCurveXPoint::set (double s)
{
	CurveSegDef *sp;	/* pointer to segment info */
	RxVecd3 raxis;		/* rotation axis */	
	double rdist;		/* rotation distance (rad) */
	double sfinal;		/* last s value over the whole interval */
	double prevSlast;	/* value of slast for previous segment */
	double slen;		/* local length of s interval */
	double sl;		/* s value local to the segment */
	int k;

	sfinal = segs[numSegs-1].slast;
	if (s < 0)
	 { s = 0;
	 }
	for (k=0; k<numSegs; k++)
	 { if (s < segs[k].slast)
	    { break;
	    }
	 }
	if (k == numSegs)
	 { s = sfinal;
	   k--;
	 }
	seg = k;
	sp = &segs[k];

	if (k == 1)
	 { chook();
	 }

	prevSlast = ((sp == segs) ? 0 : (sp-1)->slast);
	slen = sp->slast-prevSlast;
	sl = (s-prevSlast)/slen;

	X.p = ((sp->va[3]*sl + sp->va[2])*sl + sp->va[1])*sl + sp->va[0];
	calcRot (&X.R, &raxis, &rdist, sp, sl);

	ur.v() = ( (3*sp->va[3]*sl + 2*sp->va[2])*sl + sp->va[1] ) / slen;
	dldsr = ur.v().norm2();
	ur.w() = ( raxis*rdist ) / slen;

	if (s != 0)
	 { if ((sp != segs) && (s == (sp-1)->slast))
	    { sp--;
	      prevSlast = ((sp == segs) ? 0 : (sp-1)->slast);
	      slen = sp->slast-prevSlast;
	      sl = (s-prevSlast)/slen;
	      ul.v() = (3*sp->va[3]*sl + 2*sp->va[2])*sl + sp->va[1];
	      dldsl = ul.v().norm2();
	      calcRot ((RxRot*)0, &raxis, &rdist, sp, sl);
	      ul.w() = ( raxis*rdist ) / slen;
	    }
	   else
	    { ul = ur;
	      dldsl = dldsr;
	    }
	 }
	s0 = s;
}
