#include <RcPathTrajectory.h>
#include <RxPumaKin.h>
#include <math.h>

// Simple example in which the PUMA is driven along a straight line
// trajectory between two points pa and pb, with fixed orientation

RxVecd3 pa;
RxVecd3 pb;
double roll;
double pitch;
double yaw;
int conf;
RxPumaKin *kin;

int fwdKinFxn (RxTrsf *X, const RxVecd &jnts)
{
	kin->fwdK (X, (int*)0, jnts);
	return 1;
}

int invKinFxn (RxVecd *jnts, RxTrsf *X, RxVecd *old, double s)
{
	RxTrsf _X;
	int status;
	int rcode = 0;

	_X.R.setRpy (roll, pitch, yaw);
	_X.p = pa + s*(pb-pa);
		 
	if (X)
	 { *X = _X;
	 }
	if (jnts)
	 { status = kin->invK (jnts, _X, conf, old);
	   if (status & RX_GENKIN_INDETERMINATE)
	    { rcode |= RC_PATHKIN_INDETERMINATE;
	    }
	   if (status & RX_GENKIN_BOUNDARY_LIMIT)
	    { rcode |= RC_PATHKIN_BOUNDARY_LIMIT;
	    }
	 }
	return (rcode);
}

main()
{
	RcPathTrajectory pathTraj(6);
	RxVecd jnts(6);

	// set up the path endpoints and the fixed orientation

	pa.x() = 0;
	pa.y() = 550;
	pa.z() = 200;
	
	pb.x() = 0;
	pb.y() =-550;
	pb.z() = 200;

	roll = M_PI;
	pitch = 0;
	yaw = M_PI;
	
	// set the desired configuration for the PUMA inverse kinematics
	conf = RX_PUMAKIN_CONF_FLIP;

	// allocate a PUMA kinematics data structure
	kin = new RxPumaKin();

	pathTraj.setFwdKinFxn (fwdKinFxn);
	pathTraj.setInvKinFxn (invKinFxn);

	pathTraj.setMaxCartVel (250, 180*DTOR);
	pathTraj.setMaxCartAcc (2500, 500*DTOR);
	pathTraj.setMaxJntVel (180*DTOR, 0);
	pathTraj.setMaxJntAcc (500*DTOR, 0);

	pathTraj.setPathTol (0.1, 0.1*DTOR);
	pathTraj.setCharDist (1000);

	pathTraj.enableOptions (RC_PATHTRAJ_USE_PATH_LEN);

	// generate the path
	pathTraj.generate (0, 1);
	printf ("num knots = %d\n", pathTraj.length());

	// integrate the trajectory
	pathTraj.integrateInit(&jnts);
	for ( ; pathTraj.integrate(&jnts, 0.05); )
	 { printf ("jnts 123: %8.3f %8.3f %8.3f\n",
		    RTOD*jnts(0), RTOD*jnts(1), RTOD*jnts(2));
	 }
}
